;;; -*-  Mode: Lisp; Package: USER; Syntax: Common-Lisp; Base: 10 -*- ;;;;

;  Copyright (C) 1994 by Matt Kaufmann and Computational Logic, Inc.
;  All Rights Reserved.

;  Copying of this file is authorized to those who have read and
;  agreed with the terms in the "Pc-Nqthm-1992 GENERAL PUBLIC SOFTWARE
;  LICENSE" at the beginning of the Pc-Nqthm file "basis-pc.lisp".

;  Pc-Nqthm Version 1992 (for use with Nqthm-1992)

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

(defparameter *pc-nqthm-version-number* "1992")

#|

                 Pc-Nqthm-1992 GENERAL PUBLIC SOFTWARE LICENSE
                           Computational Logic, Inc.
                          1717 West Sixth, Suite 290
                           Austin, Texas 78703-4776

Please  read this license carefully before using the Pc-Nqthm-1992 Software. By
using the Pc-Nqthm-1992 Software, you are agreeing to be bound by the terms  of
this license. If you do not agree to the terms of this license, promptly return
the Pc-Nqthm-1992 Software to the place where you obtained it.

The Pc-Nqthm-1992 Software was developed  by  Computational  Logic,  Inc.(CLI).
You  own  the  disk  or  other  medium  on  which the Pc-Nqthm-1992 Software is
recorded, but CLI retains title to the Pc-Nqthm-1992 Software.  The purposes of
this  license  are  to  identify  the  Pc-Nqthm-1992  Software  and to make the
Pc-Nqthm-1992 Software, including its source  code,  freely  available.    This
license  allows  you  to  use,  copy,  distribute  and modify the Pc-Nqthm-1992
Software, on the condition that you comply with all the  Copying  Policies  set
out below.

COPYING POLICIES

1. You may copy and distribute verbatim copies of the Pc-Nqthm-1992 Software as
you receive it, in any medium, including embedding it  verbatim  in  derivative
works,  provided  that  you  a) conspicuously and appropriately publish on each
copy a valid  copyright  notice  "Copyright  (C)  1994  by  Matt  Kaufmann  and
Computational  Logic,  Inc.  All Rights Reserved.", b) keep intact on all files
the notices that refer to this License Agreement and  to  the  absence  of  any
warranty,  and  c)  give all recipients of the Pc-Nqthm-1992 Software a copy of
this License Agreement along with the program.

2. You may modify your copy or copies of  the  Pc-Nqthm-1992  Software  or  any
portion  of  it,  and  copy and distribute such modifications provided you tell
recipients that what they have is a modification by your  organization  of  the
CLI version of the Pc-Nqthm-1992 Software.

3.  You may incorporate parts of the Pc-Nqthm-1992 Software into other programs
provided that you

    a) acknowledge Computational Logic, Inc. in the program documentation, and

    b) send a copy of a user's manual for the program to

       Software-Request                 or     Software-Request@CLI.COM
       Computational Logic, Inc.
       1717 West Sixth, Suite 290
       Austin, TX 78703-4776

CLI also requests, but does not require, that any improvements or extensions to
the  Pc-Nqthm-1992 Software be returned to one of these addresses, so that they
may be shared with other Pc-Nqthm-1992  users.    The  Pc-Nqthm-1992  Software,
including its source, can be obtained by contacting one of the above addresses.

NO WARRANTY

BECAUSE  THE  Pc-Nqthm-1992  SOFTWARE  IS  LICENSED  FREE OF CHARGE, WE PROVIDE
ABSOLUTELY NO WARRANTY.  THE SOFTWARE IS PROVIDED "AS IS" WITHOUT  WARRANTY  OF
ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE
RISK  AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
Pc-Nqthm-1992 SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST  OF  ALL  NECESSARY
SERVICING, REPAIR OR CORRECTION.

IN  NO  EVENT  WILL MATT KAUFMANN OR COMPUTATIONAL LOGIC, INC. BE LIABLE TO YOU
FOR ANY DAMAGES, ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL, INCIDENTAL OR
CONSEQUENTIAL  DAMAGES  ARISING  OUT  OF  THE  USE  OR  INABILITY  TO  USE  THE
Pc-Nqthm-1992 SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED  INACCURATE  OR  LOSSES  SUSTAINED BY THIRD PARTIES), EVEN IF YOU HAVE
ADVISED US OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM  BY  ANY  OTHER
PARTY.
|#

;      An Interactive Enhancement to the Boyer-Moore Theorem Prover                              

;                                 Matt Kaufmann

;                                 Austin, Texas

;                                  March, 1988

;                         (Version "1992":  December, 1992)



;                          Computational Logic, Inc.
;                         1717 W. 6th St., Suite 290
;                             Austin, Texas 78703

;                               (512) 322-9951

;                              kaufmann@cli.com


; This system runs on top of the released version Nqthm-1992 of the
; Boyer-Moore theorem prover.

; This file contains all the variable declarations and initializations
; and all the macros.  Here's a brief table of contents for this file:

; 1) Record declarations
; 2) Instructions
; 3) Flags
; 4) Other globals
; 5) Macros
; 6) Macro command definitional support

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;			 RECORD DECLARATIONS                     ;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; A state is a list, as follows:

(DEF-NQTHM-RECORD STATE
	   (INSTRUCTION CURRENT-TERM GOVERNORS CURRENT-ADDR-R
			GOAL OTHER-GOALS CUMULATIVE-LEMMAS-USED
			FREWRITE-DISABLED-RULES ABBREVIATIONS
			FREE-VARIABLES)
	   NIL)

; where

; instruction = instruction executed to get to this state
; current-term = subterm of goal "goal-name" at which
;                instruction is executed
; governors = subterms of goal which govern current-term, in reverse
;             order (i.e. topmost governor is last in this list)
; current-addr-r = reverse of current address, i.e. of numeric
;                  sequence leading from goal down to current term
; goal = current goal  (See below for the form of the goals.)
; other-goals = list of the other goals, each consed on to a list
;               (current-term governors current-addr-r) saying where
;               to start up when (re-)considering such goal
; cumulative-lemmas-used = set of all lemmas used thus far (from the
;                          history upon entering the interactive prover),
;                          not necessarily including those in GROUND-ZERO.
;                          Note:  We don't bother storing all the function
;                          names from the original term here, since they
;                          will always be added in the dependents by
;                          PROVE-LEMMA; however, we are careful to update
;                          this field whenever new function symbols are
;                          introduced (e.g., by CLAIM or even by adding
;                          abbreviations)
; abbreviations = list of abbreviations, where an abbreviation is a pair
;                 of the form  (variable . term).  Here,  variable  is an
;                 atom beginning with the character  @ , and  term  is a
;                 term in internal (translated) form.
; free-variables = free variables, a list of variables which
;                  are not yet instantiated.

; Here, each goal has a form given as follows:

(DEF-NQTHM-RECORD GOAL
  (CONC HYPS DEPENDS-ON GOAL-NAME)
  NIL)

; where  conc = conclusion
;        hyps = hypotheses (each CONSed on to a flag (quote a) or (quote h,)
;                           i.e. active or hidden, and negated)
;        depends-on = list of names of goals on which current
;                     goal depends
;        goal-name = name of goal

; NOTE:  The organization here is that we separate out the roles of
; the goal's hypotheses and of the governors, i.e. the terms governing
; the current subterm.  Unlike the former, the latter are never
; dropped (though they may be hidden).

; The main goal is called MAIN, while subgoals of a goal named G are usually
; named (G . n) for some n>0.

;; Design decision: I won't try to keep track of the "real"
;; substitution that results from using "copy".  It shouldn't be hard
;; to COMPUTE such a thing during the proof if one really wants to.
;; Alternatively, I could probably add a new field to FREE-VARIABLES
;; records, without too much trouble, which keeps track of the "real"
;; substitution.

;; The following macros were originally DEFUNs in macro-commands-aux.lisp,
;; which explains why so little use is made of them in the code.

(defmacro INSTRUCTION (&optional state)
  `(ACCESS STATE INSTRUCTION
	   ,(or state '(CAR STATE-STACK))))

(defmacro CURRENT-TERM (&optional state)
  `(ACCESS STATE CURRENT-TERM
	   ,(or state '(CAR STATE-STACK))))

(defmacro GOVERNORS (&optional state)
  `(ACCESS STATE GOVERNORS
	   ,(or state '(CAR STATE-STACK))))

(defmacro CURRENT-ADDR-R (&optional state)
  `(ACCESS STATE CURRENT-ADDR-R
	   ,(or state '(CAR STATE-STACK))))

(defmacro GOAL (&optional state)
  `(ACCESS STATE GOAL
	   ,(or state '(CAR STATE-STACK))))

(defmacro OTHER-GOALS (&optional state)
  `(ACCESS STATE OTHER-GOALS
	   ,(or state '(CAR STATE-STACK))))

(defmacro CUMULATIVE-LEMMAS-USED (&optional state)
  `(ACCESS STATE CUMULATIVE-LEMMAS-USED
	   ,(or state '(CAR STATE-STACK))))

(defmacro FREWRITE-DISABLED-RULES (&optional state)
  `(ACCESS STATE FREWRITE-DISABLED-RULES
	   ,(or state '(CAR STATE-STACK))))

(defmacro ABBREVIATIONS (&optional state)
  `(ACCESS STATE ABBREVIATIONS
	   ,(or state '(CAR STATE-STACK))))

(defmacro FREE-VARIABLES (&optional state)
  `(ACCESS STATE FREE-VARIABLES
	   ,(or state '(CAR STATE-STACK))))

(defmacro conc (&optional state)
  `(access goal conc
	   (access state goal
		   ,(or state '(car state-stack)))))

(defmacro hyps (&optional state)
  `(access goal hyps
	   (access state goal
		   ,(or state '(car state-stack)))))

(defmacro depends-on (&optional state)
  `(access goal depends-on
	   (access state goal
		   ,(or state '(car state-stack)))))

(defmacro goal-name (&optional state)
  `(access goal goal-name
	   (access state goal
		   ,(or state '(car state-stack)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;			     INSTRUCTIONS                        ;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; An instruction may create a new state; it then depends only on the
; existing state (not on earlier ones).  However, the user may
; invoke "meta-instructions" like UNDO which manipulate the state stack
; rather than individual states.  Changes of context (e.g. changing
; the current goal or the current subterm and address) are however
; instructions in the original sense, i.e. they create a new state from
; the existing state.  (We'll call these "state-change" instructions.)
; A third kind of instruction (a "help" instruction, such as HYPS or HELP)
; only prints out information.

; The main function, INTERACTIVELY-REWRITE, simply takes a state stack
; and, after prompting for an instruction, calls itself with a new state-
; stack.  If the instruction is for information only (e.g. HELP or HYPS)
; then no change is made but the appropriate information is reported.
; If the instruction is a meta-instruction, the whole state-stack
; is manipulated with CHANGE-STATE-STACK.  Otherwise, the function
; RETURN-NEW-STATE is invoked to create a new state.  (If it fails
; to create a new state, then it returns NIL.)  An invariant is that
; STATE-STACK is always a nonempty list of states which is correct
; in the sense that each instruction field is a valid instruction
; which creates that state from the previous state (or is the
; "instruction" START for the first state).

; I'd like the system to be easy to extend.  To add a new instruction
; it will be sufficient to add it to INSTRUCTIONS and then to add its
; effect by adding a CHANGE-STATE-WITH, CHANGE-STATE-STACK-WITH, or
; PRINT-HELP-WITH function.  In addition the appropriate HELP and
; HELP-LONG information should be added to "help.lisp" and the name
; should be added in PRINT-HELP-WITH-HELP and PRINT-HELP-WITH-HELP-LONG.
; Users should use the DEFINSTR facility instead, of course.  

; Let's begin with the list of instructions along with a function which
; converts the user's instruction into an instruction-type pair (where
; the type is (quote C) (quote H), or (quote M) according to whether the
; instruction is a state-change, help, or meta- instruction).  That function
; returns NIL for clearly invalid instructions.

; Here is a list of all of the valid "official" instructions along
; with their types, and lists of acceptable numbers of arguments.

(DEFPARAMETER INSTRUCTIONS
  (convert-alist-to-eq-hash-table
   (QUOTE (				; start with state-changing instructions 
	   (DIVE C ())
	   (UP C (0))
	   (NX C (0))
	   (BK C (0))
	   (TOP C (0))
	   (S-PROP C ())
	   (X-DUMB C (0))
	   (CONTRADICT C (1))
	   (PUSH C (0 1))
	   (= C (0 1 2 3))
	   (CLAIM C (1 2))
	   (PROMOTE C (0 1))
	   (DEMOTE C ())
	   (REWRITE C (0 1 2 3))
	   (S C (0 1))
	   (X C (0 1))
	   (CHANGE-GOAL C (0 1 2))
	   (HIDE-HYPS C ())
	   (HIDE-GOVS C ())
	   (SHOW-HYPS C ())
	   (SHOW-GOVS C ())
	   (DROP C ())
	   (GENERALIZE C (1 2))
	   (SPLIT C (0))
	   (SUBV C ())
	   (PROVE C ())
	   (BASH C ())
	   (ENABLE C ())
	   (DISABLE C ())
	   (ENABLE-THEORY C ())
	   (DISABLE-THEORY C ())
	   (BOOKMARK C (1))
	   (COMMENT C ())
	   (INDUCT C (0 1))
	   (USE-LEMMA C (1 2))
	   (ADD-ABBREVIATION C (1 2))
	   (REMOVE-ABBREVIATIONS C ())
	   (COPY C ())
	   (PUT C ())
	  
	   (PP H (0))
	   (P H (0))
	   (PP-TOP H (0))
	   (GOALS H (0 1))
	   (HYPS H (0 2))
	   (SHOW-REWRITES H (0 1))
	   (SHOW-INDUCTIONS H (0))
	   (COMMANDS H (0 1))
	   (SHOW-ABBREVIATIONS H ())
	   (UNDONE-INSTRUCTIONS H (0))
	   (SHOW-FREE-VARIABLES H (0 1))
	   (HELP H ())
	   (HELP-LONG H ())
	  
	   (UNDO M (0 1))
	   (RESTORE M (0))
	   (LISP M (1))
	   (BIND M ())
	   (DO-ALL M ())
	   (PROG2 M (2))
	   (NEGATE M (1))
	   (PROTECT M (1))
	   (SUCCEED M (1))

	   (EXIT NONE (0 1 2 3))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;				FLAGS                            ;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; The first flag keeps track of whether one is currently inside an interactive session
; (but has nothing to do with calls to PROVE-LEMMA).
(DEFPARAMETER IN-PROOF-CHECKER-FLG NIL)

(DEFPARAMETER IN-DEFN-FLG NIL)

; The following, when T, causes the system to query the user when doing a MAKE-LIB on a history
; that has incomplete proofs (i.e. a non-NIL STATE-STACK-LIST).  The default is to save them
; all (though in all cases we only save the top and bottom of each state stack).
(DEFVAR SAVE-STATE-STACK-LIST-QUERY-FLG NIL)

(defparameter blurb-flg t)		;turns off printing by BLURB when set to NIL

; The following is used by setting SHELVED-LEMMAS to the AND of
; the flag and itself, locally in any use.
(DEFVAR SHELVED-LEMMAS-FLG T)

; The following, when reset to NIL, should turn off error-handling in INTERACTIVE-SINGLE-STEP
(DEFVAR NO-DEBUG-FLG T)

; The following is used for printing instructions when inside invocation of a macro command
; or a meta command.
(DEFPARAMETER PRINT-INSTR-FLG NIL)

; The following flag is reset locally to NIL upon exit from the
; interactive prover with a response of "Y" or "YES" to the ensuing prompt.
(DEFPARAMETER REPLAY-FLG T)

; The following is set to NIL for (S LEMMAS) command, and is understood thus by
; FREWRITE-WITH-LEMMAS.
(DEFPARAMETER FNS-ALLOWED-FOR-FAST-REWRITE-FLG T)

; When set to NIL, printing of old macro commands (upon redefinition or removal)
; is suppressed.  Otherwise, they're printed.
(DEFVAR PRINT-OLD-MACRO-DEF-FLG NIL)

;; Used in order to make prover not revert to induct on original goal (as in nthm)
(defparameter otf-flg nil)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;			    OTHER GLOBALS                        ;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; The interactive system's prompt; may be reset by the user
(defvar *pc-prompt* (QUOTE |->: |))

; The following are set in INITIALIZE-FREWRITE

(DEFVAR FREWRITE-RULES)
(DEFVAR ALL-FREWRITE-RULES)
(DEFVAR ALL-FREWRITE-RULE-NAMES)
(DEFVAR ALL-DEFINED-FUNCTION-NAMES)
(DEFVAR FREWRITE-QUICK-OPENERS)
(DEFPARAMETER OLD-CHRONOLOGY NIL)

; The main data structure is a stack of states:

(DEFVAR STATE-STACK)
(DEFVAR OLD-SS)

; We also keep around a list of state-stacks, which is manipulated
; by PROVE-LEMMA and UBT.  It is initialized to NIL, but is not re-initialized
; when (for some reason) one reloads the proof-checker.
;   Notice that STATE-STACK-LIST is an association list which assigns something
; to every interactive PROVE-LEMMA event whose conclusion is T.  However, because
; one has the option of not saving partial proofs in MAKE-LIB, some of these event
; names may be associated with NIL.  It is a hard error though if such a name is
; not in the domain of STATE-STACK-LIST.  The invariant is maintained through the
; following line of code in PROVE-LEMMA, (SETQ FINISH-FLG (NOT (EQUAL TERM TRUE))),
; togther with the fact that MAKE-LIB doesn't change the domain of STATE-STACK-LIST
; when storing it.

(DEFVAR STATE-STACK-LIST NIL)		;used DEFVAR so this would not be re-initialized

; The following global stores the text of the event to submit next.
(DEFVAR EV0)

; More variable declaraions.

; The following should be set before calling EXPAND-FNS, EXPAND-FNCALL, or
; FAST-REWRITE (and its subsidiaries), which is why I use DEFVAR instead of
; DEFPARAMETER here.
(DEFVAR MY-ALL-LEMMAS-USED)

(defvar basic-macro-command-names
  '(TH DV COMM UNDO! REDUCE EQSUB PRUNE EX SHRINK RETAIN REPLAY HYP PROVE+ CLAIM0 CASESPLIT
       QQ BACKCHAIN FORWARD SK* PRINT-GOAL
       PRO R$ BASH+ CLAIM+ =+ S* DEFN ELIM USE
       PRINT SAVE PG TAUT THEN UNDO+ PLAY WRAP))

(defvar undone-instrs nil)		;defvar so it's not changed upon reloading

; The macros SHELVE and UNSHELVE update the SHELVED-LEMMAS.
(DEFVAR SHELVED-LEMMAS NIL)

;the list of macro commands, stored appropriately -- notice that it persists upon reloading
(DEFVAR MACRO-COMMANDS NIL)

(DEFPARAMETER COMMAND-BINDINGS NIL)	;the binding list understood by the BIND meta command

(DEFPARAMETER *PC-MACRO*-PREFIX (QUOTE *PC-MACRO*-))

; Here's the stack depth for restarting the proof-checker automatically.  It's pretty
; arbitrary where it should be set; in fact I'll allow the option of setting it to
; NIL, which means "infinity".  Actually, though, I'll initialize it to 0 because
; both Symbolics Release 7.2 (I think) and AKCL on the Sun-3 have problems with
; my use of UNWIND-PROTECT when there's stack overflow. -- fixed for kcl.

#-kcl
(DEFVAR RESTART-STACK-DEPTH 0)

#+kcl
(DEFVAR RESTART-STACK-DEPTH 8)

; The following two globals pertain to interactive termination proofs.
(defparameter interactive-defn-hint nil)
(defvar start-marker)

; For SAVE and RETRIEVE:
(defvar retrieve-alist nil)

;; used in BASH
(defvar bash-clauses)

;; (DEFVAR HINTS) already appears in basis.lisp, but I want hints to
;; be set even if we're not inside prove-lemma.  I'll arrange that
;; HINTS is set when we call BASH or PROVE from inside the interactive
;; loop.  It will then be safe to set IN-PROVE-LEMMA-FLG.
(DEFparameter HINTS nil)

;; I use defvar below because if *UNTRANSLATE-EVENT-TYPES* is already
;; defined, then I don't want to overwrite this variable.  I'm redefining
;; UNTRANSLATE-EVENT because DEFN can take 5 arguments, but I want to do
;; so in a way that is compatible with DEFN-SK.
(DEFVAR *UNTRANSLATE-EVENT-TYPES*
  '(ADD-SHELL BOOT-STRAP DCL TOGGLE TOGGLE-DEFINED-FUNCTIONS
	      DEFTHEORY SET-STATUS))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;				MACROS                           ;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(DEFMACRO BLURB (BINDINGS &REST X)

; Used all over the place for printing.  Notice that it always returns NIL if it
; doesn't cause an error.

  `(if blurb-flg
       (PROGN (PRINEVAL (PQUOTE (PROGN ,@X))
			(LIST ,@(ITERATE FOR PAIR IN BINDINGS COLLECT
					 (COND ((ATOM PAIR)
						`(CONS (QUOTE ,PAIR) ,PAIR))
					       (T `(CONS (QUOTE ,(CAR PAIR))
							 ,(CADR PAIR))))))
			0 PROVE-FILE)	;last argument used to be NIL
	      NIL)
       nil))

(defmacro neq (x y)
  `(not (eq ,x ,y)))

; The following is omitted since it's probably not Common Lisp.
;(DEFMACRO TRACE-FREWRITE ()

; handy for tracing, since FREWRITE and such have some very large arguments

;  (QUOTE
;    (PROGN
;      (TRACE (FREWRITE :ENTRYPRINT (FIRSTN 4 ARGLIST) :VALUE))
;      (TRACE (FREWRITE-IF :ENTRYPRINT (FIRSTN 6 ARGLIST) :VALUE))
;      (TRACE (FREWRITE-WITH-LEMMAS :ENTRYPRINT (FIRSTN 3 ARGLIST) :VALUE)))))

(DEFUN ALL-ATOMS (FORM)
  (COND
    ((ATOM FORM)
     (LIST FORM))
    ((EQ (CAR FORM) (QUOTE QUOTE))
     NIL)
    (T (ITERATE FOR X IN FORM
	     APPEND (ALL-ATOMS X)))))

(DEFMACRO LET-FIELDS (&REST A)
  (LET ((VARS (ALL-ATOMS A)))
    (LET ((STATE-VARS
	    (INTERSECTION-EQ (CADR (ASSOC-EQ (QUOTE STATE) RECORD-DECLARATIONS)) VARS))
	  (GOAL-VARS
	    (INTERSECTION-EQ (CADR (ASSOC-EQ (QUOTE GOAL) RECORD-DECLARATIONS)) VARS)))
      (COND (GOAL-VARS (SETQ STATE-VARS (ADD-TO-SET (QUOTE GOAL) STATE-VARS))))
      `(LET
	 ,(ITERATE FOR X IN STATE-VARS
		COLLECT (LIST X `(ACCESS STATE ,X STATE)))
	 (LET
	   ,(ITERATE FOR X IN GOAL-VARS
		  COLLECT (LIST X `(ACCESS GOAL ,X GOAL)))
	   ,@A)))))

(DEFMACRO MY-UNWIND-PROTECT (FORM &REST CLEANUP-FORMS)
  ;; Only exceutes the CLEANUP-FORMS if FORM does not complete successfully.
  `(LET (MY-UNWIND-PROTECT-FINISHED-FLG)
	(UNWIND-PROTECT
	 (PROG1 ,FORM (SETQ MY-UNWIND-PROTECT-FINISHED-FLG T))
	 (COND
	  (MY-UNWIND-PROTECT-FINISHED-FLG NIL)
	  (T ,@CLEANUP-FORMS)))))

(defmacro retrieve (&optional VAR)
  ;; to get back the state-stack saved in VAR by e.g. an invocation of SAVE
  `(let ((ss (cdr (assoc-eq
		   (or ',var 'temp-ss)
		   retrieve-alist))))
     (cond
      (ss (setq state-stack ss)
	  (verify))
      (t (blurb ((x (or ',var 'temp-ss))) cr |There| |was| |no| SAVE |into| |the| |variable|
		(!ppr x (quote |.|)) cr)))))

(DEFMACRO SHELVE (&REST NAMES)
  `(ITERATE FOR NAME IN (quote ,NAMES) DO
	    (COND
	     ((OR (GET NAME 'EVENT)
		  (GET NAME 'MAIN-EVENT))
	      (SETQ SHELVED-LEMMAS (ADD-TO-SET-EQ NAME SHELVED-LEMMAS)))
	     (T (BLURB (NAME) CR |***WARNING*** --| (!PPR NAME NIL) |is|
		       |neither| |an| |event| |nor| |a| |satellite| |,|
		       |and| |hence| |is| |not| |being| |shelved| |.| CR)))))

(DEFMACRO UNSHELVE (&REST NAMES)
  `(ITERATE FOR NAME IN (QUOTE ,NAMES) DO
	    (SETQ SHELVED-LEMMAS (REMOVE NAME SHELVED-LEMMAS :TEST #'EQ))))

(DEFMACRO VERIFY (&optional x instrs)
  (COND
   (IN-PROOF-CHECKER-FLG
    (BLURB (X) IT IS ILLEGAL TO
	(COND (X ENTER A NEW INTERACTIVE SESSION)
	      (T REENTER AN INTERACTIVE SESSION))
	WHILE ALREADY INSIDE AN INTERACTIVE SESSION EXCL |(Hint| |:| |if| |you| |are|
	|looking| |at| |Lisp's| |error| |prompt| |,| |then| |you| |probably| |need|
	|to| |abort| |.| |Under| |most| |circumstances| |,| |this| |will| |take| |you|
	|back| |in| |to| |the| |current| |interactive| |session.)| CR))
   ((and instrs
	 (or (atom instrs)
	     (cdr (our-last instrs))))
    (er soft (instrs) |The| |optional| |second| |argument| |to| VERIFY |should|
	|be| |a| |proper| |list| |,| |but| (!ppr instrs nil) |is| |not| |.|))
   (X
    (COND
     ((ERROR1-SET
       (PROGN
	(SETQ X (TRANSLATE-AND-CHK X))
	(COND
	 ((NOT (BOUNDP (QUOTE CHRONOLOGY)))
	  (ER SOFT NIL |The| |theorem-prover's| |database| |must| |be| |initialized|
	      |.| |try| |(BOOT-STRAP)| |or| |invoke| NOTE-LIB |.|))
	 ((EQUAL X TRUE)
	  (ER SOFT NIL |Don't| |be| |silly| EXCL |Why| |would| |anyone| |want|
	      |to| |prove| TRUE |?|)))))
      `(LET ((IN-PROOF-CHECKER-FLG T))
	    (START-PROOF-CHECKER (QUOTE ,X) (quote ,instrs))))
     (T
      (BLURB NIL CR Back to the Lisp reader CR))))
   ((or (not (boundp 'state-stack))
	(NULL STATE-STACK))
    (BLURB NIL CR |There| |is| |no| |interactive| |verification|
	   |to| |re-enter| EXCL |Back| |to| |the| |Lisp| |reader ---| CR)) 
   (T
    `(LET ((IN-PROOF-CHECKER-FLG T))
	  (RESTART-PROOF-CHECKER NIL ',instrs)))))

(defun start-term-for-verify-defn (fnname formals body relation-measure-lst)
  (start-term-for-interactive-prove-termination
   formals
   (CAR RELATION-MEASURE-LST)
   (TERMINATION-MACHINE FNNAME BODY NIL)))

(DEFMACRO VERIFY-DEFN (FNNAME ARGS BODY &OPTIONAL RELATION-MEASURE-LST instrs)
  (COND
   (IN-PROOF-CHECKER-FLG
    (BLURB NIL IT IS ILLEGAL TO ENTER A NEW INTERACTIVE SESSION
	   WHILE ALREADY INSIDE AN INTERACTIVE SESSION EXCL CR))
   ((NOT (BOUNDP (QUOTE CHRONOLOGY)))
    (ER SOFT NIL |The| |theorem-prover's| |database| |must| |be| |initialized|
	|.| |try| |(BOOT-STRAP)| |or| |invoke| NOTE-LIB |.|)) 
   (T
    (LET ((START-MARKER `(START (DEFN ,FNNAME ,ARGS ,BODY ,RELATION-MEASURE-LST))))
	 (MATCH! (CHK-ACCEPTABLE-DEFN FNNAME ARGS BODY RELATION-MEASURE-LST)
		 (LIST FNNAME ARGS BODY RELATION-MEASURE-LST))
	 (LET ((ARITY-ALIST (CONS (CONS fnNAME (LENGTH ARGS)) ARITY-ALIST)))
	      ;; Who knows?  Maybe I need the following setup for the starting term.
	      (DEFN-SETUP (LIST (QUOTE DEFN) FNNAME ARGS BODY RELATION-MEASURE-LST))
	      (COND
	       ((NOT
		 (OR RELATION-MEASURE-LST
		     (SETQ RELATION-MEASURE-LST
			   (GUESS-RELATION-MEASURE-LST ARGS (TERMINATION-MACHINE FNNAME BODY NIL)))))
		(BLURB NIL |There| |is| |no| |heuristically-chosen| |relation-measure| |list| |.|
		       |You| |are| |advised| |to| |specify| |one| |yourself| |as| |a| |fourth|
		       |argument| |to| |VERIFY-DEFN| |.| CR))
	       (T
		`(let ((in-proof-checker-flg t))
		   (start-proof-checker-defn
		    (quote ,(start-term-for-verify-defn fnname args body relation-measure-lst))
		    (quote ,start-marker)
		    (quote ,instrs))))))))))

(DEFMACRO RE-ENTER (FORM &optional instrs)
  (COND
   (IN-PROOF-CHECKER-FLG
    (BLURB (FORM) IT IS ILLEGAL TO
	   ENTER A NEW INTERACTIVE SESSION
	   WHILE ALREADY INSIDE AN INTERACTIVE SESSION EXCL CR))
   ((null form)
    (BLURB (FORM) |It| |is| |illegal| |to| |give| | NIL | |as| |a| |name| |to| RE-ENTER |.|
	   |Probably| |you| |forgot| |the| |argument| |entirely| |and| |your| |Lisp|
	   |interpreted| |(RE-ENTER)| |as| |(RE-ENTER NIL)| |.|)) 
   ((and instrs
	 (or (atom instrs)
	     (cdr (our-last instrs))))
    (er soft (instrs) |The| |optional| |second| |argument| |to| RE-ENTER |should|
	|be| |a| |proper| |list| |,| |but| (!ppr instrs nil) |is| |not| |.|))
   (T
    `(LET ((IN-PROOF-CHECKER-FLG T)) (RESTART-PROOF-CHECKER
				      ,(KWOTE FORM)
				      ',instrs)))))

(defmacro free-variablep (var free)
  `(member-eq ,var ,free))

;    The macro DEFINSTR is to be the one used by most users
; to extend the input language.  BODY should be the form that the
; user wants the instruction expanded to.

;; The following macro requires defn-sk functions, so I omit it.

; (defvar *last-verify-sk-name* nil)

; (defmacro verify-sk (name formula)
;   (let ((vars (formula-free-vars (translate-to-formula formula))))
;     `(progn
;        (defn-sk+ ,name
;          ,(unstable-sort vars (function alphorder))
;          ,formula)
;        (setq *last-verify-sk-name* ',name)
;        (verify ,(cons-term name vars)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;		  MACRO COMMAND DEFINITIONAL SUPPORT             ;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(DEFMACRO DEFINE-MACRO-COMMAND-INTERNAL (NAME ARGS BODY)
  (COND ((EQ NAME 'USER)
	 (BLURB NIL CR |**NOTE**| |:| |Now| |defining| |the| |macro| |command|
		USER |,| |which| |will| |be| |wrapped| |around| |each| |instruction|
		|that| |is| |given| |at| |the| |top-level| |prompt| |.| CR)))
  `(PROGN
    (QUOTE COMPILE)
    (SETF (GET (QUOTE ,NAME) '*PC-MACRO*-ARITY)
	  (MACRO-ARITY (QUOTE ,ARGS)))
    (DEFUN ,(PACK (LIST *PC-MACRO*-PREFIX NAME)) ,ARGS ,BODY)))

(DEFUN CHK-ACCEPTABLE-MACRO-NAME (NAME)
  (COND
   ((get-instr-info name)
    (ER SOFT (NAME) |The| |name| (!PPR NAME NIL) |is| |the| |name| |of| |a|
	|primitive| |instruction| |,| |and| |thus| |may| |not| |be| |used| |as|
	|the| |name| |of| |a| |macro-command| |.|))
   ((NOT (SYMBOLP NAME))
    (ER SOFT (NAME) |The| |"name"| (!PPR NAME NIL) |is| |not| |a| |symbol| |.|))))

(DEFUN PUT-MACRO-HELP (NAME FORM DOCUMENTATION)
  (LET (OLD-DEF)
       (COND ((AND PRINT-OLD-MACRO-DEF-FLG (SETQ OLD-DEF (MACRO-COMMAND-DEF NAME)))
	      (TERPRI NIL)
	      (PRINC "Previous macro command definition:")
	      (TERPRI NIL)
	      (PPRINT (APPEND OLD-DEF (LIST (GET NAME '*PC-MACRO*-DOC))))
	      (TERPRI NIL)))
       (SETF (GET NAME '*PC-MACRO*-DEF) FORM)
       (SETF (GET NAME '*PC-MACRO*-DOC) DOCUMENTATION)
       (COND
	((NOT (MEMBER-EQ NAME MACRO-COMMANDS))
	 (SETQ MACRO-COMMANDS (APPEND MACRO-COMMANDS (LIST NAME)))))))

(DEFUN MACRO-ARITY (ARGS)
  (ITERATE FOR ARG IN ARGS
	   WITH LOWER = 0 AND UPPER = 0 AND LOWER-DONE-FLG
	   DO
	   (COND
	    ((EQ ARG '&REST)
	     (RETURN (CONS LOWER NIL)))
	    ((EQ ARG '&OPTIONAL)
	     (SETQ LOWER-DONE-FLG T))
	    (LOWER-DONE-FLG
	     (SETQ UPPER (1+ UPPER)))
	    (T
	     (SETQ LOWER (1+ LOWER))
	     (SETQ UPPER (1+ UPPER))))
	   FINALLY
	   (RETURN (CONS LOWER UPPER))))

(DEFUN MACRO-COMMAND-DEF (NAME)
  (AND (SYMBOLP NAME)
       (GET NAME '*PC-MACRO*-DEF)))

(DEFMACRO DEFINE-MACRO-COMMAND (NAME ARGS BODY &OPTIONAL DOCUMENTATION)
  `(PROGN
    (QUOTE COMPILE)
    (CHK-ACCEPTABLE-MACRO-NAME (QUOTE ,NAME))
    (PUT-MACRO-HELP
     (QUOTE ,NAME)
     (LIST 'DEFINE-MACRO-COMMAND (QUOTE ,NAME) (QUOTE ,ARGS) (QUOTE ,BODY))
     (OR (QUOTE ,DOCUMENTATION)
	 (LIST '|[ No| '|documentation| '|for| (QUOTE ,NAME) '|]|)))
    (DEFINE-MACRO-COMMAND-INTERNAL ,NAME ,ARGS ,BODY)))

(DEFUN DEFINSTR-BODY-TO-DEFINE-MACRO-COMMAND-BODY-AUX (ARGS BODY)
  (COND
   ((ATOM BODY)
    (IF (MEMBER-EQ BODY ARGS)
	BODY
	(AND BODY (KWOTE BODY))))
   (T
    `(CONS ,(DEFINSTR-BODY-TO-DEFINE-MACRO-COMMAND-BODY-AUX ARGS (CAR BODY))
	   ,(DEFINSTR-BODY-TO-DEFINE-MACRO-COMMAND-BODY-AUX ARGS (CDR BODY))))))

(defun compress-conses (form)
  ;; Turns an s-expression with CONSes into one which is equivalent from
  ;; the point of view of EVAL but which takes advantage of QUOTE notation.
  (let (x y a b)
    (cond
     ((match form (list 'cons x y))
      (setq x (compress-conses x))
      (setq y (compress-conses y))
      (cond
       ((match x (list 'quote a))
	(cond
	 ((null y)
	  (list 'quote (list a)))
	 ((match y (list 'quote b))
	  (list 'quote (cons a b)))
	 (t `(cons ,x ,y))))
       (t `(cons ,x ,y))))
     (t form))))

(defun list*ify (form)
  ;; Turns an s-expression with CONSes into one with LIST*s which is
  ;; equivalent from the point of view of EVAL.  In fact, in some sense all
  ;; CONSes are eliminated in favor of LIST*s from the outside in.
  (let (x y a)
    (cond
     ((match form (list 'cons x y))
      (setq x (list*ify x))
      (setq y (list*ify y)) 
      (cond
       ((null y)
	`(list* ,x nil))
       ((match y (cons 'list* a))
	`(list* ,x ,@a))
       (t `(list* ,x ,y))))
     (t form))))

(defun listify (form)
  ;; Turns an s-expression with LIST*s into one with LISTs which is
  ;; equivalent from the point of view of EVAL.
  (let (x)
    (cond
     ((and (match form (cons 'list* x))
	   (consp x)
	   (null (cdr (our-last x))))
      (cond
       ((eq (car (our-last x)) nil)
	(cons 'list
	      (iterate for tail on x while (cdr tail) collect (listify (car tail)))))
       (t
	(cons 'list*
	      (iterate for y in x collect (listify y))))))
     (t form))))

(defun compress-list-structure (form)
  ;; Turns an s-expression with CONSes into one with LIST*s and LISTs which is
  ;; equivalent from the point of view of EVAL and takes advantage of QUOTE notation.
  (listify
   (list*ify
    (compress-conses form))))

(DEFUN DEFINSTR-BODY-TO-DEFINE-MACRO-COMMAND-BODY (ARGS BODY)
  (COMPRESS-LIST-STRUCTURE
   (DEFINSTR-BODY-TO-DEFINE-MACRO-COMMAND-BODY-AUX ARGS BODY)))
  
(DEFMACRO DEFINSTR (&OPTIONAL NAME ARGS BODY DOCUMENTATION)
  (COND
   ((OR (NULL NAME) (NULL BODY))
    (ER SOFT NIL |The| |first| |and| |third| |arguments| |of| DEFINSTR
	|are| |required| |to| |be| |non-NIL| |.|))
   ((NOT (AND (SYMBOLP NAME)
	      (OR (NULL ARGS)
		  (AND (CONSP ARGS) (NULL (CDR (OUR-LAST ARGS)))))
	      (NO-DUPLICATESP ARGS)
	      BODY))
    (BLURB NIL CR |**SORRY**| |--| DEFINSTR |failed| |to| |create| |a| |new| |macro|
	   |command| |.| |You| |must| |use| DEFINSTR |in| |the| |form| |(DEFINSTR NAME ARGS BODY)|
	   |,| |where| NAME |is| |a| |symbol| |,| ARGS |is| |a| |list| |of| |distinct| |atoms| |,|
	   |and| BODY |is| |some| |structure| |which| |is| |the| |body| |of| |the|
	   |definition| |(see| |user's| |manual| |for| |details)| |.|))
   (T
    `(PROGN
      (QUOTE COMPILE)
      (CHK-ACCEPTABLE-MACRO-NAME (QUOTE ,NAME))
      (ITERATE FOR ARG IN (QUOTE ,ARGS) DO
	       (COND
		((AND (EQUAL (CHAR (STRING ARG) 0) #\&)  ;; was EQ here
		      (NEQ ARG '&OPTIONAL))
		 (ER SOFT (ARG) |The| |argument| (!PPR ARG NIL) |to| DEFINSTR
		     |is| |illegal| |because| |it| |starts| |with| |&| |but| |is|
		     |not| |&OPTIONAL| |.| |You| |may| |wish| |to| |use| |the|
		     DEFINE-MACRO-COMMAND |facility| |instead| |.|))))
      (PUT-MACRO-HELP
       (QUOTE ,NAME)
       (LIST 'DEFINSTR (QUOTE ,NAME) (QUOTE ,ARGS) (QUOTE ,BODY))
       (OR (QUOTE ,DOCUMENTATION) (LIST '|[ No| '|documentation| '|for| (QUOTE ,NAME) '|]|)))
      (DEFINE-MACRO-COMMAND-INTERNAL ,NAME ,ARGS
	,(DEFINSTR-BODY-TO-DEFINE-MACRO-COMMAND-BODY
	   (ITERATE FOR ARG IN ARGS WHEN (NEQ ARG '&OPTIONAL) COLLECT ARG)
	   BODY))))))

(DEFMACRO UNDEFINSTR (NAME)
  `(LET ((OLD-DEF (MACRO-COMMAND-DEF (QUOTE ,NAME))))
	(COND
	 (OLD-DEF
	  (COND (PRINT-OLD-MACRO-DEF-FLG
		 (TERPRI NIL)
		 (PRINC "Removing macro command definition:")
		 (TERPRI NIL)
		 (PPRINT (APPEND OLD-DEF (LIST (GET (QUOTE ,NAME) '*PC-MACRO*-DOC))))
		 (TERPRI NIL)))
	  (REMPROP (QUOTE ,NAME) '*PC-MACRO*-DEF)
	  (REMPROP (QUOTE ,NAME) '*PC-MACRO*-DOC)
	  (REMPROP (QUOTE ,NAME) '*PC-MACRO*-ARITY)
	  (SETQ MACRO-COMMANDS (REMOVE1 (QUOTE ,NAME) MACRO-COMMANDS))
	  (QUOTE ,NAME))
	 (T
	  (BLURB ((NAME (QUOTE ,NAME))) CR |Sorry| |--| |there| |is| |currently| |no| |macro| |command|
		 |named| (!PPR NAME (QUOTE |.|)) CR)))))

(DEFUN CHK-ACCEPTABLE-MACRO-COMMAND (NAME ARGS-LENGTH)
  (AND (>= ARGS-LENGTH (CAR (GET NAME '*PC-MACRO*-ARITY)))
       (OR (NULL (CDR (GET NAME '*PC-MACRO*-ARITY)))
	   (<= ARGS-LENGTH (CDR (GET NAME '*PC-MACRO*-ARITY))))))

(DEFMACRO COMPILE-MACRO-COMMANDS (&REST NAMES)
  `(COMPILE-UNCOMPILED-MACRO-FUNCTIONS	 
    (OR (QUOTE ,NAMES) MACRO-COMMANDS)))