;;; -*-  Mode: Lisp; Package: USER; Syntax: Common-Lisp; Base: 10 -*- ;;;;

;  Copyright (C) 1994 by Matt Kaufmann and Computational Logic, Inc.
;  All Rights Reserved.

;  Copying of this file is authorized to those who have read and
;  agreed with the terms in the "Pc-Nqthm-1992 GENERAL PUBLIC SOFTWARE
;  LICENSE" at the beginning of the Pc-Nqthm file "basis-pc.lisp".

;  Pc-Nqthm Version 1992 (for use with Nqthm-1992)

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

;;************************************************************;;
;;                      HELP MESSAGES                         ;;
;;************************************************************;;

(DEFPARAMETER COPY-SHORT-HELP-MESSAGE
  (QUOTE
   (|#| |(COPY n1 ... nk)| |:| |Copy| |the| |indicated| |active| |hypotheses| |,|
    |renaming| |all| |free| |variables| |in| |the| |copies| |that| |do| |not| |appear| |in|
    |any| |other| |goal| |.| CR
    EXAMPLE |:| |(COPY 3)| CR CR |#|
    OTHER FORM |(see| |HELP-LONG)| |:| COPY |.|)))

(DEFPARAMETER COPY-LONG-HELP-MESSAGE
  (QUOTE
   (|#| |(COPY hyp-number)| |:| |Copy| |the| |indicated| |active| |hypothesis| |,|
    |renaming| |all| |free| |variables| |in| |the| |copy| |that| |do| |not| |appear| |in|
    |the| |conclusion| |or| |in| |any| |other|
    |goal| |.| |A| |message| |prints| |out| |a| |list| |that| |associates| |each|
    |variable| |copied| |with| |the| |corresponding| |new| |variable| |.| CR |#|
    COPY |:| |Same| |as| |(COPY n1 ... nk)| |where| |(n1 ... nk)| |is| |the|
    |list| |of| |all| |active| |hypotheses| |,| |except| |that| |a| |hypothesis|
    |is| |only| |copied| |when| |the| |resulting| |copy| |contains| |at|
    |least| |one| |renamed| |variable| |.| CR |#|
    NOTE |that| |the| |system| |picks| |out| |new| |variable|
    |names| |according| |to| |the| |following|
    |algorithm| |.| |Given| |a| |variable| |v| |,| |a| |variable| |of| |the| |form|
    |w$n| |where| |n| |is| |a| |sequence| |of| |digits| |is| |created| |as|
    |follows| |.| |If| |v| |ends| |in| |$| |,| |then| |w$n| |is| |obtained| |by|
    |suffixing| |the| |character| |0| |to| |v| |.| |If| |v| |ends| |in| |$n| |where| |n|
    |is| |a| |sequence| |of| |digits| |,| |then| |w| |is| |v$m| |where| |m| |is| |the|
    |sequence| |of| |digits| |corresponding| |to| |adding| |1| |to| |the| |number| |represented|
    |by| |n| |.| |Otherwise| |w| |is| |v$0| |.| |Finally| |,| |the| |number| |represented|
    |by| |n| |is| |incremented| |as| |necessary| |so| |that| |w$n| |does| |not| |occur|
    |in| |any| |current| |goal| |and| |is| |not| |equal| |to| |any| |new| |variable| |that|
    |has| |already| |been| |generated| |by| |this| |command| |,| |and| |the| |resulting|
    |w$n| |is| |the| |new| |variable| |associated| |with| |v| |.|)))

(DEFPARAMETER PUT-SHORT-HELP-MESSAGE
  (QUOTE
   (|#| |(PUT (V1 term1) ... (Vn termn))| |:| |Apply| |the| |indicated| |substitution|
    |to| |the| |entire| |proof| |state| |,| |where| |each| |Vi| |is| |a|
    |free| |variable| |in| |the| |current| |proof| |state| |.| CR
    EXAMPLE |:| |(PUT (V$ (PLUS X Y)))|)))

(DEFPARAMETER PUT-LONG-HELP-MESSAGE
  (QUOTE
   (|#| |(PUT (V1 term1) ... (Vn termn))| |:| |Apply| |the| |indicated| |substitution|
    |to| |the| |entire| |proof| |state| |,| |where| |each| |Vi| |is| |a|
    |free| |variable| |in| |the| |current| |proof| |state| |.| CR |#|
    NOTE |:| |the| |effect| |of| |a| |PUT| |on| |the| |free| |variables| |of|
    |the| |state| |is| |that| |all| |of| |the| |variables| |in| |the| |domain| |of|
    |the| |indicated| |substitution| |are| |removed| |from| |the| |list| |of|
    |free| |variables| |,| |but| |then| |all| |variables| |in| |the| |range| |of|
    |the| |substitution| |that| |do| |not| |occur| |in| |its| |domain| |or| |in|
    |any| |goal| |of| |the| |current| |state| |are| |put|
    |onto| |that| |list| |.| |In| |particular| |,| |it| |is| |permissible| |to|
    |(PUT (var var))| |,| |which| |simply| |has| |the| |effect| |of| |removing|
    |var| |from| |the| |list| |of| |free| |variables| |of| |the| |proof| |state| |.|)))

(DEFPARAMETER SHOW-FREE-VARIABLES-SHORT-HELP-MESSAGE
  (QUOTE
   (|#| SHOW-FREE-VARIABLES |:| |Show| |the| |free| |variables| |of|
    |the| |current| |proof| |state| |,| |each| |listed| |together| |with|
    |the| |goals| |in| |which| |it| |appears| |(if| |any)| |.| CR CR |#|
    |(SHOW-FREE-VARIABLES T)| |:| |As| |above| |,| |except| |that| |printing|
    |of| |the| |goal| |information| |is| |suppressed| |.|)))

(DEFPARAMETER SHOW-FREE-VARIABLES-LONG-HELP-MESSAGE
  (QUOTE
   (|#| SHOW-FREE-VARIABLES |:| |Show| |the| |free| |variables| |of|
    |the| |current| |proof| |state| |,| |each| |listed| |together| |with|
    |the| |goals| |in| |which| |it| |appears| |(if| |any)| |.| CR |#|
    |(SHOW-FREE-VARIABLES T)| |:| |As| |above| |,| |except| |that| |printing|
    |of| |the| |goal| |information| |is| |suppressed| |.| |(Actually| |,|
    |any| |argument| |in| |place| |of| |T| |will| |have| |the| |same| |effect.)|)))
   
(DEFPARAMETER DIVE-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(DIVE N)| |:| |For| |N>0,| |move| |to|
	 |the| |Nth| |argument| |.| CR EXAMPLE |:| |If| |the| |current| |subterm|
	 |is| CR |#| (TIMES (PLUS A B) C) |,| CR |then| |after| |(DIVE 1)| |it| |is|
	 CR |#| |(PLUS A B)| |.| CR CR |#|
	 DIVE |:| |Dive| |down| |to| |the| |leftmost| |unproved| |branch| |in| |the|
	 |IF-structure| |(from| |the| |top| |of| |the| |conclusion| |,| |unless|
	 |the| |current| |term| |is| |an| |IF| |term)| |.| CR
	 EXAMPLE |:| |If| |the| |conclusion| |is| CR |#| (IF A (IF B T (G X)) Y) |,|
	 CR |then| DIVE |puts| |the| |current| |subterm| |at| |(G X)| |.| CR CR |#|
	 NOTE |:| N |is| |an| |abbreviation| |for| |(DIVE N)| |,| |so| |we| |could|
	 |have| |typed| |1| |instead| |of| |(DIVE 1)| |in| |the| |first| |example|
	 |above| |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:| |(DIVE N1 ... Nk)| |.|)))

(DEFPARAMETER DIVE-LONG-HELP-MESSAGE
   (QUOTE
     (|#| |(DIVE N)| |or| N |:| |For| |N>0,| |move| |to|
	  |the| |Nth| |argument| |.| CR |#|
	  |(DIVE N1 ... Nk)| |:| |Same| |as| |(DIVE N1)|
	  |followed| |by| |(DIVE N2)| |followed| |by| |...|
	  |,| |but| |all| |wrapped| |into| |one| |instruction|
	  |.| CR |#| 
	  DIVE |:| |Dive| |down| |in| IF |structure|
	  |from| |the| |top| |of| |the| |goal| |until|
	  |locating| |the| |leftmost| |branch| |which| |is| |not| |an|
	  |explicit| |value| |or| FALSE |.| |However| |,|
	  |if| |the| |current| |subterm| |is| |of| |the|
	  |form| |(IF x y z)| |then| |the| DIVE |begins|
	  |from| |the| |current| |subterm| |.| CR |#|
	  NOTE |:| |If| |the| |current| |subterm| |is| |not| |of|
	  |the| |form| |(IF x y z)| |(and| |hence| |DIVing| |is|
	  |attempted| |from| |the| |top| |of| |the| |current|
	  |conclusion)| |,| |and| |if| |every| |branch| |is|
	  |a| |non-FALSE| |explicit| |value| |,| |then| |the|
	  |goal| |is| |considered| |to| |be| |proved| |.|)))

(DEFPARAMETER UP-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| UP |:| |Move| |up| |to| |enclosing| |subterm| |.| CR
	 EXAMPLE |:| |If| |the| |conclusion| |is| CR |#| (G (H X (FOO Y)) Z) CR |and|
	 |the| |current| |subterm| |is| CR |#| (FOO Y) |,| CR |then| |after|
	 |executing| UP |,| |the| |current| |subterm| |will| |be|
	 CR |#| (H X (FOO Y)) |.|)))

(DEFPARAMETER UP-LONG-HELP-MESSAGE
  (QUOTE
    (|#| UP |:| |Move| |up| |to| |enclosing| |subterm| |(if| |not| |already| |at|
	 |the| |top| |of| |the| |conclusion)| |.|)))

(DEFPARAMETER X-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| X |:| |Expand| |function| |call| |at| |current| |subterm| |,|
	 |and| |simplify| |.| CR
	 EXAMPLE |:| |If| |the| |current| |subterm| |is| CR |#|
	 (APPEND (CONS A X) Y) |,| CR
	 |then| |after| |executing| X |the| |current| |subterm| |will| |be|
	 CR |#| (CONS A (APPEND X Y)) |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:| |(X ALL),| |(X LEMMAS),| |(X n).|)))

(DEFPARAMETER X-LONG-HELP-MESSAGE
  (QUOTE
   (|#| X |:| |Expand| |function| |call| |at| |current| |subterm| |,|
    |and| |simplify| |the| |result| |using| |the| |primitive|
    |fast| |rewrite| |rules| |which| |are| |not| |currently|
    |disabled| |.| |(Type| |(HELP-LONG S)| |for| |a| |discussion| |of|
    |primitive| |fast| |rewrite| |rules.)| |This| |only| |has| |an|
    |effect| |if| |the| |current| |subterm| |is| |an| |application| |of|
    |a| |"defined"| |function| |,| |i.e.| |one| |defined| |in| |the|
    BOOT-STRAP |or| |by| |a| DEFN |event| |.| CR |#|
    |(X ALL)| |:| |As| |above| |,| |but| |remove| |the| |"primitive"|
    |restriction| |.| CR |#|
    |(X LEMMAS)| |:| |Same| |as| |(X ALL)| |,| |but| |no| |function|
    |definitions| |are| |expanded| |(except| |for| |the| |top-level| |one)|
    |,| |i.e.| |only| |lemmas| |are| |used| |for| |rewriting| |.| CR |#|
    |(X n)| |(where| |n| |is| |a| |nonnegative| |integer)| |:| |Same| |as|
    |(X LEMMAS)| |,| |except| |that| |the| |fast| |rewriter| |is| |allowed|
    |to| |backchain| |n| |levels| |.| |In| |particular| |,| |(X 0)|
    |is| |equivalent| |to| |(X LEMMAS).| CR |#|)))

(DEFPARAMETER X-DUMB-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| X-DUMB |:| |Expand| |function| |call| |at| |current| |subterm|
	 |without| |simplifying| |the| |result| |.| CR
	 EXAMPLE |:| |If| |the| |current| |subterm| |is| CR |#|
	 (APPEND (CONS A X) Y) |,| CR
	 |then| |after| |executing| X |the| |current| |subterm| |will| |be|
	 CR |#| (IF (LISTP (CONS A X))
		    (CONS (CAR (CONS A X))
			  (APPEND (CDR (CONS A X)) Y))
		    Y))))

(DEFPARAMETER X-DUMB-LONG-HELP-MESSAGE
  (QUOTE
    (|#| X-DUMB |:| |Expand| |function| |call| |at| |current| |subterm|
	 |without| |simplifying| |the| |result| |.| |This| |only| |has| |an|
	 |effect| |if| |the| |current| |subterm| |is| |an| |application| |of|
	 |a| |"defined"| |function| |,| |i.e.| |one| |defined| |in| |the|
	 BOOT-STRAP |or| |by| |a| DEFN |event| |.|)))

(DEFPARAMETER SUBV-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| (SUBV |(v1 t1)| |...| |(vn tn)|) |:|
	 |Do| |parallel| |substitution| |of| |each| |term| | ti | |for|
	 |the| |respective| |variable| | vi | |into| |the| |current| |conclusion|
	 |and| |into| |all| |active| |top-level| |hypotheses| |,|
	 |if| |justified| |by| |the| |current| |active| |hypotheses| |.| CR
	 EXAMPLE |:| |If| |the| |equalities| (EQUAL X (PLUS A B)) |and|
	 (EQUAL (TIMES C D) Y) |are| |among| |the| |active| |hypotheses| |,|
	 |and| |if| |the| |current| |subterm| |is| CR |#| (DIFFERENCE X Y) |,| CR |then|
	 |after| |executing| CR |#| (SUBV (X (PLUS A B)) (Y (TIMES C D))) |,| CR |the|
	 |current| |subterm| |will| |be| CR |#|
	 (DIFFERENCE (PLUS A B) (TIMES C D)) |.| CR CR |#|
	 SUBV |:| |As| |above| |,| |but| |every| |such| |substitution| |will| |be|
	 |performed| |,| |subject| |to| |certain| |restrictions|
	 |(see| |(HELP-LONG SUBV)).| CR
	 EXAMPLE |:| |In| |the| |situation| |described| |above| |,| SUBV |would| |have| |had|
	 |the| |same| |effect| |as| |would| |the| |command|
	 |(SUBV (X (PLUS A B)) (Y (TIMES C D)))| |.|)))

(DEFPARAMETER SUBV-LONG-HELP-MESSAGE
  (QUOTE
    (|#| (SUBV |(v1 t1)| |...| |(vn tn)|) |:|
	 |Do| |parallel| |substitution| |of| |each| |term| | ti | |for|
	 |the| |respective| |variable| | vi | |into| |the|
	 |conclusion| |and| |into| |all| |active| |top-level| |hypotheses| |.|
	 |It| |is| |required| |that| |for| |each| |i| |,| |either| |(EQUAL vi ti)|
	 |or| |(EQUAL ti vi)| |(or| |both)| |is| |among| |the|
	 |active| |hypotheses| |.| |Also| |,| |for| |distinct| |i| |and| |j|
	 |we| |require| |that| |vi| |and| |vj| |are| |distinct| |variables| |.| CR |#|
	 |SUBV| |:| |Same| |as| |above| |,| |except| |that| |we| |impose| |the|
	 |requirement| |that| |the| |variables| |finally| |eliminated| |by|
	 |SUBV| |will| |not| |occur| |in| |the| |resulting| |goal|
	 |(except| |possibly| |in| |hidden| |hypotheses).|
	 |Thus| |,| |the| |substitution|
	 |consists| |of| |all| |pairs| |(v t)| |that| |are| |constructed|
	 |according| |to| |the| |following| |algorithm| |.| |Starting| |with|
	 |the| |last| |active| |hypothesis| |and| |working| |forward| |,|
	 |each| |hypothesis| |is| |considered| |in| |turn| |.| |If| |the|
	 |hypothesis| |currently| |considered| |is| |an| |equation| |,| |where| |at| |least|
	 |one| |side| |of| |the| |equation| |is| |a| |variable| | v | |that| |has| |not|
	 |yet| |been| |placed| |into| |the| |domain| |of| |the| |substitution|
	 |being| |built| |,| |and| |the| |other| |side| |contains| |a| |term| | t |
	 |which| |is| |neither| |that|
	 |variable| |nor| |any| |variable| |placed| |so| |far| |into| |the|
	 |domain| |of| |the| |substitution| |being| |built| |,| THEN |we| |add|
	 |the| |pair| |(v t)| |to| |the| |substitution| |being| |built| |.|
	 BUT |if| |the| |resulting| |substitution| |is| |empty| |then|
	 |the| |command| |fails| |,| |that| |is| |,| |there| |is| |no| |change| |.| CR |#|
	 |ADDITIONAL| |NOTES| |:| |CR| |#|
	 |(1)| |One| |must| |be| |at| |the| |top| |of| |the| |conclusion|
	 |in| |order| |to| |apply| SUBV |or| |(SUBV ...)| |.| |If| |you| |feel| |that|
	 |you| |wish| |you| |could| |apply| |SUBV| |only| |to| |a| |subterm| |,|
	 |try| |using| |the| |`='| |command| |instead| |.| |The| |idea| |of| |SUBV|
	 |is| |global| |elimination| |of| |one| |or| |more| |variables| |in| |a| |goal| |.|
	 CR |#|
	 |(2)| |For| |each| |active| |equality| |hypothesis| |(equal v1 v2)| |,| |if|
	 |(v1 v2)| |or| |(v2 v1)| |is| |used| |in| |the| |substitution| |that| |is|
	 |applied| |,| |then| |that| |hypothesis| |becomes| |hidden| |as| |a| |result|
	 |of| |this| |command| |.| |Moreover| |,| |the| |substitution| |is| NOT
	 |applied| |to| |such| |hypotheses| |.|)))

(DEFPARAMETER S-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| S |:| |Simplify| |the| |current| |subterm| |,| |expanding|
	 |nonrecursive| |function| |calls| |and| |doing| |various| |nice|
	 |things| |.| |(For| |details| |,| |use| |(HELP-LONG S).)| CR
	 EXAMPLE |:| (ZEROP X) |simplifies| |to| CR |#|
	 (IF (EQUAL X 0)
	     T
	     (IF (NUMBERP X) F T))
	 CR |if| |there| |are| |no| |active| |hypotheses| |or| |governors| |,|
	 |but| |in| |the| |presence| |of| |the| |hypothesis|
	 (NOT (EQUAL X 0)) |it| |simplifies| |to| CR |#|
	 (IF (NUMBERP X) F T) |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:| |(S ALL),| |(S LEMMAS),| |(S n),|
	 |(S (<lemma-1> ... <lemma-N>))| |.|)))

(DEFPARAMETER S-LONG-HELP-MESSAGE
  (QUOTE
    (|#| S |:| |Simplify| |the| |current| |subterm| |with| |the| |fast| |rewriter| |,|
	 |expanding| |all| |nonrecursive| |function| |calls| |and| |using|
	 |all| |available| |fast| |rewrite| |rules| |which| |are|
	 |primitive| |.| |("Fast"| |and| |"primitive"| |are| |defined|
	 |below.) | |Of| |course| |,| |disabled| |fast| |rewrite| |rules|
	 |and| |functions| |are| |ignored| |here| |.| CR |#|
	 |(S ALL)| |:| |As| |above| |,| |but| |remove| |the| |"primitive"|
	 |restriction| |.| CR |#|
	 |(S LEMMAS)| |:| |Same| |as| |(S ALL)| |,| |but| |no| |function|
	 |definitions| |are| |expanded| |,| |i.e.| |only| |lemmas| |are|
	 |used| |for| |rewriting| |.| CR |#|
	 |(S n)| |(where| |n| |is| |a| |nonnegative| |integer)| |:| |Same| |as|
	 |(S LEMMAS)| |,| |except| |that| |the| |fast| |rewriter| |is| |allowed|
	 |to| |backchain| |n| |levels| |.| |In| |particular| |,| |(S 0)|
	 |is| |equivalent| |to| |(S LEMMAS).| CR |#|
	 |(S (<lemma-1> ... <lemma-N>))| |:| |Same| |as| |(S LEMMAS)| |,| |except|
	 |that| |only| |the| |fast| |rewrite| |rules| |with| |the| |given| |names| |are|
	 |used| |(whether| |disabled| |or| |not)| |.| CR CR
	 |#| |A| |rewrite| |rule| |is| FAST |if|  |its| |conclusion| |is| |an| |equality| |,|
	 |every| |variable| |in| |a| |hypothesis| |occurs| |in| |the|
	 |left| |side| |of| |the| |conclusion| |,| |and| |the| |rewrite|
	 |rule| |is| |not| |permutative| |.| |Note| |that| |nonrecursive|
	 |definitions| |are| |also| |considered| |to| |be| |fast| |rewrite|
	 |rules| |.| CR
	 |#| |A| |fast| |rewrite| |rule| |is| PRIMITIVE |if| |it|
	 |contains| |no| |function| |symbol| |other| |than| |ones|
	 |which| |are| |added| |by| BOOT-STRAP |and| |have| |no| |definition|
	 |,| |e.g.| CAR |,| CDR |,| CONS |,| |or| |else| |are| |added| |by|
	 |the| |shell| |principle| |.| CR
	 |#| |The| FAST REWRITER |uses| |fast| |rewrite| |rules| |but| |also|
	 |has| |some| |additional| |optimizations| |.| |For| |example| |,| |it|
	 |knows| |that| |(EQUAL term term)| |simplifies| |to| T |for| |every| |term;|
	 |it| |may| |simplify| |a| |term| |to| |one| |which| |is|
	 |propositionally| |equivalent| |rather| |than| |equal| |in| |some|
	 |cases;| |and| |it| |expands| |recursive| |function| |calls| |when| |an|
	 |argument| |which| |gets| |smaller| |on| |each| |recursive| |call| |is|
	 |a| |bottom| |object| |.| |Also| |,| |even| |if| |no| |backchaining| |is|
	 |allowed| |(or| |if| |the| |backchaining| |level| |has| |been| |exhausted)|
	 |,| |some| |small| |attempt| |(quite| |short| |of| |full| |rewriting)|
	 |is| |made| |to| |relieve| |hypotheses| |.| |Finally| |,| |terms| |of| |the|
	 |form| |(FN x1 ... xn)| |in| |which| |each| |xi| |is| |an| |explicit| |value|
	 |term| |are| |evaluated| |(reduced)| |completely| |,| |provided| |that| |(1)|
	 |either| |FN| |is| |not| |DEFNed| |or| |else| |the| |*1*functions| |are|
	 |enabled| |(in| |the| |sense| |of| |the| |nqthm| |event| |TOGGLE-DEFINED-FUNCTIONS),|
	 |and| |(2)| |FN| |is| |enabled| |in| |the| |proof-checker| |and| |so| |is|
	 |every| |function| |in| |the| |body| |of| |its| |definition| |(if|
	 |it| |has| |a| |definition).|)))

(DEFPARAMETER S-PROP-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| S-PROP |:| |Expand| |all| |calls| |of| IMPLIES |,| AND |,| OR |,| |and|
	 NOT |in| |the| |current| |subterm| |,| |and| |then| |push| |all| |calls|
	 |of| IF |to| |the| |top| |of| |the| |current| |subterm| |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:| (S-PROP NIL) |,|
	 |(S-PROP |x1| ... |xn|)| |.|)))

(DEFPARAMETER S-PROP-LONG-HELP-MESSAGE
  (QUOTE
    (|#| S-PROP |:| |Same| |as| |(S-PROP IMPLIES AND OR NOT)| |(see|
	 |below)| |.| |Use| |(S-PROP NIL)| |if| |you| |don't|
	 |want| |those| |boolean| |functions| |expanded| |.| CR |#|
	 |(S-PROP x1 x2 ... xn)| |:| |Push| |the| |IFs| |to| |the|
	 |top| |of| |the| |current| |subterm| |,| |after| |expanding| |certain|
	 |function| |calls| |as| |follows| |.| |If| |xi| |is| |the|
	 |name| |of| |a| |function| |,| |then| |all| |calls| |of| |that|
	 |function| |are| |expanded| |.| |In| |addition| |,| |if| |xj|
	 |is| |of| |the| |form| |(f a1 ... ak)| |,| |then| |each|
	 |occurrence| |of| |that| |term| |is| |expanded| |.|)))

(DEFPARAMETER ENABLE-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(ENABLE <name1> ... <nameK>)|
	 |:| |Enable| |the| |given| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |for| |the| |fast| |rewriter| |.| |(These| |terms| |are| |explained|
	 |with| |(HELP-LONG S).)| CR
	 EXAMPLE USE |:| (ENABLE APPEND ZEROP) |.| CR CR |#|
	 ENABLE |:| |Enable| |all| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |.|)))

(DEFPARAMETER ENABLE-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(ENABLE name1 ... nameK)|
	 |:| |Enable| |the| |given| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |for| |the| |fast| |rewriter| |.| |(These| |terms| |are| |explained|
	 |with| |(HELP-LONG S).)| CR |#|
	 ENABLE |:| |Enable| |all| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |.|)))

(DEFPARAMETER DISABLE-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(DISABLE name1 ... nameK)|
	 |:| |Disable| |the| |given| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |for| |the| |fast| |rewriter| |.| |(These| |terms| |are| |explained|
	 |with| |(HELP-LONG S).)| CR
	 EXAMPLE USE |:| (DISABLE APPEND ZEROP) |.| CR CR |#|
	 DISABLE |:| |Disable| |all| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |.|)))

(DEFPARAMETER DISABLE-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(DISABLE name1 ... nameK)|
	 |:| |Disable| |the| |given| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |for| |the| |fast| |rewriter| |.| |(These| |terms| |are| |explained|
	 |with| |(HELP-LONG S).)| CR |#|
	 DISABLE |:| |Disable| |all| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |.|)))

(DEFPARAMETER ENABLE-THEORY-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(ENABLE-THEORY <name1> ... <nameK>)|
	 |:| |Enable| |the| |rules| |and| |function| |definitions| |from|
	 |the| |given| |theories| |for| |the| |fast| |rewriter| |.| |(These|
	 |terms| |are| |explained|
	 |with| |(HELP-LONG S).)| CR
	 EXAMPLE USE |:| (ENABLE-THEORY LIST-FACTS ARITHMETIC) |.| CR CR |#|
	 NOTE |:| |to| |enable| |all| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |,| |use| ENABLE |.|)))

(DEFPARAMETER ENABLE-THEORY-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(ENABLE-THEORY <name1> ... <nameK>)|
	 |:| |Enable| |the| |rules| |and| |function| |definitions| |from|
	 |the| |given| |theories| |for| |the| |fast| |rewriter| |.| |(These|
	 |terms| |are| |explained|
	 |with| |(HELP-LONG S).)| CR |#|
	 NOTE |:| |to| |enable| |all| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |,| |use| ENABLE |.|)))

(DEFPARAMETER DISABLE-THEORY-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(DISABLE-THEORY <name1> ... <nameK>)|
	 |:| |Disable| |the| |rules| |and| |function| |definitions| |from|
	 |the| |given| |theories| |for| |the| |fast| |rewriter| |.| |(These|
	 |terms| |are| |explained|
	 |with| |(HELP-LONG S).)| CR
	 EXAMPLE USE |:| (DISABLE-THEORY LIST-FACTS ARITHMETIC) |.| CR CR |#|
	 NOTE |:| |to| |disable| |all| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |,| |use| DISABLE |.|)))

(DEFPARAMETER DISABLE-THEORY-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(DISABLE-THEORY <name1> ... <nameK>)|
	 |:| |Disable| |the| |rules| |and| |function| |definitions| |from|
	 |the| |given| |theories| |for| |the| |fast| |rewriter| |.| |(These|
	 |terms| |are| |explained|
	 |with| |(HELP-LONG S).)| CR |#|
	 NOTE |:| |to| |disable| |all| |fast| |rewrite| |rules| |and| |function|
	 |definitions| |,| |use| DISABLE |.|)))

(DEFPARAMETER NX-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| NX |:| |Move| |forward| |one| |argument| |in| |the| |enclosing| |term| |.| CR
	 EXAMPLE |:| |If| |the| |conclusion| |is| CR |#| (G (H X (FOO Y)) Z) CR |and|
	 |the| |current| |subterm| |is| X |,| |then| |after|
	 |executing| NX |,| |the| |current| |subterm| |will| |be|
	 (FOO Y) |.|)))

(DEFPARAMETER NX-LONG-HELP-MESSAGE
  (QUOTE
    (|#| NX |:| |Move| |forward| |one| |argument| |in| |the| |term|
	 |immediately| |enclosing| |the| |current| |subterm| |,| |if| |possible| |.|)))

(DEFPARAMETER BK-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| BK |:| |Move| |backward| |one| |argument| |in| |the| |enclosing| |term| |.| CR
	 EXAMPLE |:| |If| |the| |conclusion| |is| CR |#| (G (H X (FOO Y)) Z) CR |and|
	 |the| |current| |subterm| |is| (FOO Y) |,| |then| |after|
	 |executing| BK |,| |the| |current| |subterm| |will| |be|
	 X |.|)))

(DEFPARAMETER BK-LONG-HELP-MESSAGE
  (QUOTE
    (|#| BK |:| |Move| |backward| |one| |argument| |in| |the| |term|
	 |immediately| |enclosing| |the| |current| |subterm| |,| |if| |possible| |.|)))

(DEFPARAMETER TOP-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| TOP |:| |Move| |to| |the| |top| |of| |the| |current| |conclusion| |.| CR
	 EXAMPLE |:| |If| |the| |conclusion| |is| CR |#| (G (H X (FOO Y)) Z) CR |and|
	 |the| |current| |subterm| |is| (FOO Y) |,| |then| |after|
	 |executing| TOP |,| |the| |current| |subterm| |will| |be|
	 CR |#| (G (H X (FOO Y)) Z) |.|)))

(DEFPARAMETER TOP-LONG-HELP-MESSAGE
  (QUOTE
    (|#| TOP |:| |Move| |to| |the| |top| |of| |the| |current| |conclusion| |.|)))

(DEFPARAMETER DROP-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| DROP |:| |Drop| |all| |the| |top-level| |hypotheses| |.| CR CR |#|
	 |(DROP n1 ... nk)| |:| |Drop| |the| |top-level| |hypotheses|
	 |with| |the| |indicated| |indices| |.| CR
	 EXAMPLE |:| (DROP 2 5) |.| CR CR |#|
	 |(DROP HIDDEN)| |:| |Drop| |the| |hidden| |top-level|
	 |hypotheses| |.|)))

(DEFPARAMETER DROP-LONG-HELP-MESSAGE
  (QUOTE
    (|#| DROP |:| |Drop| |all| |the| |top-level| |hypotheses| |.| CR |#|
	 |(DROP n1 ... nk)| |:| |Drop| |the| |top-level| |hypotheses|
	 |with| |the| |indicated| |indices| |.| CR |#|
	 |(DROP HIDDEN)| |:| |Drop| |the| |hidden| |top-level|
	 |hypotheses| |.|)))

(DEFPARAMETER HIDE-HYPS-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |HIDE-HYPS|
	 |:| |Hide| |all| |of| |the| |hypotheses| |.| CR CR |#|
	 |(HIDE-HYPS ...)| |:| |Hide| |the| |indicated| |hypotheses| |,|
	 |referenced| |by| |number| |.| |To| |obtain| |these| |numbers|
	 |,| |type| HYPS |or| |(HYPS ALL NIL)| |.| CR
	 EXAMPLE |:| (HIDE-HYPS 2 4) |.|)))

(DEFPARAMETER HIDE-HYPS-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |HIDE-HYPS|
	 |:| |Hide| |all| |of| |the| |hypotheses| |.| CR |#|
	 |(HIDE-HYPS ...)| |:| |Hide| |the| |indicated| |hypotheses| |,|
	 |referenced| |by| |number| |.| |To| |obtain| |these| |numbers|
	 |,| |type| HYPS |or| |(HYPS ALL NIL)| |.| CR |#|
	 NOTE |:| |Hypotheses| |of| |(TRUE)| |are| |already| |generally|
	 |treated| |as| |hidden| |by| |the| |other| |commands| |.|)))

(DEFPARAMETER HIDE-GOVS-SHORT-HELP-MESSAGE
  (QUOTE
    (|HIDE-GOVS|
      |:| |Hide| |all| |of| |the| |governors| |.| CR CR |#|
      |(HIDE-GOVS ...)| |:| |Hide| |the| |indicated| |governors| |,|
      |referenced| |by| |number| |.| |To| |obtain| |these| |numbers|
      |,| |type| HYPS |or| |(HYPS NIL ALL)| |.|)))

(DEFPARAMETER HIDE-GOVS-LONG-HELP-MESSAGE
  (QUOTE
    (|HIDE-GOVS|
      |:| |Hide| |all| |of| |the| |governors| |.| CR |#|
      |(HIDE-GOVS ...)| |:| |Hide| |the| |indicated| |governors| |,|
      |referenced| |by| |number| |.| |To| |obtain| |these| |numbers|
      |,| |type| HYPS |or| |(HYPS NIL ALL)| |.|)))

(DEFPARAMETER SHOW-HYPS-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |SHOW-HYPS|
	 |:| |Activate| |all| |of| |the| |top-level| |hypotheses| |.| CR CR |#|
	 |(SHOW-HYPS ...)| |:| |Activate| |the| |indicated| |top-level|
	 |hypotheses| |,| |referenced| |by| |number| |.| |To| |obtain|
	 |these| |numbers| |,| |type| |(HYPS ALL NIL)| |.| CR
	 EXAMPLE |:| (SHOW-HYPS 2 4) |.|)))

(DEFPARAMETER SHOW-HYPS-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |SHOW-HYPS|
	 |:| |Activate| |all| |of| |the| |top-level| |hypotheses| |.| CR |#|
	 |(SHOW-HYPS ...)| |:| |Activate| |the| |indicated| |top-level|
	 |hypotheses| |,| |referenced| |by| |number| |.| |To| |obtain|
	 |these| |numbers| |,| |type| |(HYPS ALL NIL)| |.|)))

(DEFPARAMETER SHOW-GOVS-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |SHOW-GOVS|
	 |:| |Activate| |all| |of| |the| |governors| |.| CR CR |#|
	 |(SHOW-GOVS ...)| |:| |Activate| |the| |indicated| |governors|
	 |,| |referenced| |by| |number| |.| |To| |obtain| |these| |numbers|
	 |,| |type| |(HYPS NIL ALL)| |.| CR
	 EXAMPLE |:| (SHOW-GOVS 2 4) |.|)))

(DEFPARAMETER SHOW-GOVS-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |SHOW-GOVS|
	 |:| |Activate| |all| |of| |the| |governors| |.| CR
	 |(SHOW-GOVS ...)| |:| |Activate| |the| |indicated| |governors|
	 |,| |referenced| |by| |number| |.| |To| |obtain| |these| |numbers|
	 |,| |type| |(HYPS NIL ALL)| |.|)))

(DEFPARAMETER PUSH-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| PUSH |:| |Replace| |the| |current| |subterm| |with| T |(assuming|
	 |it's| |boolean)| |,| |and| |create| |a| |new| |goal| |to| |prove|
	 |that| |term| |under| |the| |active| |hypotheses| |and|
	 |governors| |.| CR
	 EXAMPLE |:| |Suppose| |that| |we| |have| |the|
	 |single| |active| |hypothesis| |and| |single| |active| |governor| CR |#|
	 |H1.|  (LESSP X Y) CR |#|
	 |G1.|  (NOT (ZEROP Z)) |,| CR
	 |and| |conclusion| CR |#|
	 (IF (ZEROP Z)
	     T
	     (LESSP (TIMES X Z) (TIMES Y Z))) CR
	 |with| |current| |subterm| CR |#|
	 (LESSP (TIMES X Z) (TIMES Y Z)) |.| CR
	 |Then| |after| |the| |execution| |of| PUSH |,| |the| |current| |subterm|
	 |is| T |,| |the| |new| |conclusion| |is| CR |#|
	 (IF (ZEROP Z) T T) |,| CR
	 |and| |the| |new| |subgoal| |has| |hypotheses| CR |#|
	 |H1.|  (LESSP X Y) CR |#|
	 |H2.|  (NOT (ZEROP Z)) CR
	 |and| |conclusion| CR |#|
	 (LESSP (TIMES X Z) (TIMES Y Z)) |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:| (PUSH |name|) |.|)))

(DEFPARAMETER PUSH-LONG-HELP-MESSAGE
  (QUOTE
    (|#| PUSH |:| |Replace| |the| |current| |subterm| | x | |with| T |(or| |else|
	 |with| |(IF x x T)|
	 |,| |if| |using| T |is| |not| |known| |to| |be| |sound| |because|
	 | x | |is| |not| |known| |to| |be| |boolean)| |,|
	 |and| |create| |a| |new| |goal| |to| |prove| | x | |under|
	 |the| |active| |hypotheses| |and| |governors| |.|
	 |the| |new| |goal| |has| |a| |name| |of| |the| |form|
	 |(name . n)| |,| |where|
	 | name | |is| |the| |current| |goal's| |name| |and|
	 | n | |is| |a| |positive| |integer| |.| CR |#|
	 (PUSH |name|) |:| |as| |above| |,| |except| |that|
	 |the| |new| |goal| |will| |be| |given| |the| |indicated| |name|
	 |(if| |that| |name| |is| |unused)| |.|)))

(DEFPARAMETER CHANGE-GOAL-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(CHANGE-GOAL <goal-name>)|
	 |:| |Change| |to| |the| |goal| |with| |the|
	 |name| |<goal-name>| |,| |i.e.| |make| |it|
	 |the| |current| |goal| |.| CR
	 EXAMPLE |:| (CHANGE-GOAL (MAIN . 1)) CR CR |#|
	 CHANGE-GOAL |:| |Change| |to| |the| |first| |unproved|
	 |goal| |(in| |the| |list| |shown| |by| |the| |command|
	 |GOALS)|  |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:|
	 |(CHANGE-GOAL <goal-name> T).|)))

(DEFPARAMETER CHANGE-GOAL-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(CHANGE-GOAL <goal-name>)|
	 |:| |Change| |to| |the| |goal| |with| |the|
	 |name| |<goal-name>| |,| |i.e.| |make| |it|
	 |the| |current| |goal| |.| CR |#|
	 CHANGE-GOAL |:| |Change| |to| |the| |first| |unproved|
	 |goal| |(in| |the| |list| |shown| |by| |the| |command|
	 |GOALS)| |.| CR |#|
	 |(CHANGE-GOAL <goal-name> T)| |:| |Same| |as| |(CHANGE-GOAL <goal-name>)|
	 |except| |that| |the| |current| |goal| |will| |become| |the| |last| |goal|
	 |in| |the| |stack| |of| |goals| |(as| |shown| |by| |the| GOALS |command)| |,|
	 |rather| |than| |the| |one| |next| |to| |the| |top| |of| |the| |stack| |.|
	 |Actually| |,| |the| |presence| |of| |any| |second| |argument| |at| |all|
	 |has| |the| |same| |effect| |as| T |.|)))

(DEFPARAMETER GENERALIZE-SHORT-HELP-MESSAGE
  (QUOTE
   (|#| |(GENERALIZE ((term1 V1) ... (termn Vn)))| |:|
    |Replace| |each| |of| |the| |given| |terms| |by| |the|
    |indicated| |corresponding| |new| |variable| |,| |which| |must| |not| |occur|
    |anywhere| |in| |the| |current| |proof| |state| |.| |A| |question| |mark|
    |(?)| |may| |be| |used| |in| |place| |of| |any| |or| |all|
    |variables| |Vi| |;| |however| |,| |the| |variables| |actually|
    |supplied| |must| |be| |distinct| |.| |Appropriate| |hypotheses| |are|
    |added| |(use| HELP-LONG |for| |details).| CR
    EXAMPLE |:| (GENERALIZE (((PLUS X Y) ?))) |.| CR CR |#|
    OTHER FORMS |(see| |HELP-LONG)| |:|
    |(GENERALIZE ((term1 V1) ... (termn Vn)) new-goal-name)| |.|)))

(DEFPARAMETER GENERALIZE-LONG-HELP-MESSAGE
  (QUOTE
   (|#| |(GENERALIZE ((term1 V1) ... (termn Vn)) new-goal-name)| |:|
    |Replace| |each| |of| |the| |given| |terms| |by| |the|
    |indicated| |new| |variable| |(which| |must| |not| |occur|
    |anywhere| |in| |the| |current| |proof| |state| |,| |and| |must|
    |not| |be| |an| |abbreviation| |variable)| |.| CR |#| NOTE |:|
    CR |#| |(1) | |Each| |pair| |(term Vi)| |may| |instead| |be| |the|
    |pair| |(term ?)| |-- i.e.| |the| |`?' | |symbol| |indicates| |that| |a|
    |new| |variable| |is| |to| |be| |generated| |for| |that|
    |term| |.| |However| |,| |the| |actual| |variables| |specified|
    |must| |be| |distinct| |and| |must| |not| |occur| |in| |the|
    |current| |proof| |state| |.| CR |#| |(2) | |The|
    |new-goal-name| |is| |optional| |.| |If| |it| |is| |supplied| |,|
    |then| |it| |may| |not| |be| |the| |name| |of| |an|
    |existing| |goal| |.| CR |#|
    |(3) | |For| |technical| |reasons| |,| |the|
    |indicated| |substitution| |will| |not| |be| |pushed| |inside|
    |explicit| |values| |.| |For| |example| |,| |if| |the| |substitution|
    |((1 x))| |is| |specified| |then| |(PLUS 1 4)| |will| |be|
    |replaced| |by| |(PLUS x 4)| |but| |(ADD1 1)| |,| |i.e.| |2| |,|
    |will| |not| |be| |replaced| |by| |(ADD1 x)| |.| CR |#|
    |(4)| |Generalization| |may| |only| |be| |used| |when| |the|
    |current| |subterm| |is| |the| |entire| |conclusion| |.| CR |#|
    |(5) | |The| |same| |mechanism| |used| |by| |the| |Boyer-Moore| |prover| |for|
    |generalization| |(including| |the| |use| |of| GENERALIZE |lemmas)| |is|
    |used| |here| |to| |add| |additional| |hypotheses| |,| |except| |that|
    |if| |anything| |is| |known| |about| |the| |type| |set| |of| |a| |term|
    |being| |generalized| |away| |,| |then| |that| |information| |is| |also| |added|
    |as| |a| |hypothesis| |.| |(In| |the| |Boyer-Moore| |prover| |,| |this|
    |happens| |only| |if| |the| |type| |set| |is| |a| |singleton| |,| |presumably|
    |to| |avoid| |case| |splits.)| CR |#|
    |(6) | |(Subtle)| |It| |is| |legal| |to| |specify| |new| |generalization|
    |variables| |Vi| |that| |are| |free| |variables| |of| |the| |current| |state| |,|
    |as| |long| |as| |they| |do| |not| |occur| |in| |any| |goal| |of| |the|
    |current| |state| |.| |However| |,| |such| |variables| |will| |not|
    |be| |free| |in| |the| |new| |state| |.| CR |#|
    |(7) | |Finally| |,| |consider| |the| |binary| |relation|
    |on| |the| |free| |variables| |of| |a| |proof| |state| |consisting|
    |of| |all| |pairs| |of| |variables| |that| |occur| |in| |a|
    |common| |goal| |.| |Then| |for| |every| |free| |variable| |V| |that| |occurs|
    |in| |any| |term| |that| |has| |been| |generalized| |away| |(i.e.| |lies|
    |in| |the| |domain| |of| |the| |generalizing| |substitution),| |if| |some|
    |free| |variable| |W| |occurring| |in| |the| |resulting| |generalized| |goal|
    |is| |related| |to| |V| |in| |the| |transitive| |closure| |of| |the|
    |above| |relation| |(in| |the| |new| |proof| |state),| |then| |in| |fact| |V|
    |is| |removed| |from| |the| |list| |of| |free| |variables| |.|
    |This| |is| |necessary| |for| |soundness| |.| |Consider| |,| |as| |a|
    |simple| |example| |,| |generalizing| |away| |the| |term| |(ADD1 X$)| |in| |the|
    |goal| |(LESSP (ADD1 X$) X$)| |,| |where| |X$| |is| |free| |,| |by| |using| |the|
    |command| |(GENERALIZE (((ADD1 X$) Z)))| |.| |A| |subsequent|
    |(PUT (X$ (ADD1 Z)))| |would| |let| |one| |prove| |the|
    |resulting| |goal| EXCL)))

(DEFPARAMETER =-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |=| | :| |Make| |a| |substitution| |for| |the| |current| |subterm| |,|
	 |using| |an| |equality| |which| |is| |explicitly| |among| |the|
	 |current| |active| |hypotheses| |and| |governors| |.| CR
	 EXAMPLE |:| |If| |the| |current| |subterm| |is| (PLUS X Y) |and| |the|
	 |equality| |(EQUAL (TIMES U V) (PLUS X Y))| |is| |an| |active| |hypothesis|
	 |,| |then| |after| |executing| |the| |=| |command| |,| |the| |current|
	 |subterm| |will| |be| (TIMES U V) |(unless| |perhaps| |some| |other|
	 |hypothesis| |equates| (PLUS X Y) |with| |something)| |.| CR CR |#|
	 |(= Exp)| |:| |Replace| |the| |current| |subterm| |by| |Exp| |,| |if| |they|
	 |can| |be| |proved| |equal| |by| |the| |theorem| |prover| |under| |the|
	 |current| |active| |hypotheses| |and| |governors| |.| CR
	 EXAMPLE |:| (= T) |will| |replace| |the| |current| |subterm| |with| T
	 |if| |the| |prover| |can| |prove| |their| |equality| |.| CR CR |#|
	 |(= Exp1 Exp2)| |:| |Replace| |Exp1| |by| |Exp2| |in| |the|
	 |current| |subterm| |,| |if| |they| |can| |be| |proved| |equal| |by|
	 |the| |theorem| |prover| |under| |the| |current| |active| |hypotheses|
	 |and| |governors| |.| CR
	 EXAMPLE |:| |If| |the| |current| |subterm| |is| (LESSP (PLUS X Y) Z) |,|
	 |then| |after| |executing| |the| |command| (= (PLUS X Y) (PLUS Y X)) |,|
	 |the| |new| |current| |subterm| |will| |be| (LESSP (PLUS Y X) Z) |.| CR CR |#|
	 OTHER FORMS |:| |In| |the| |following| |one| |may|
	 |use| |*| |in| |place| |of| |Exp1| |to| |denote|
	 |the| |current| |subterm| |.|
	 CR |(= Exp1 Exp2 (hint1 hint2 ... hintK))| |:| |for| |giving| |hints| |to|
	 |the| |prover| |.| CR
	 |(= Exp1 Exp2 0)| |:| |As| |above| |,| |but| |create| |a|
	 |new| |goal| |to| |prove| |the| |equality| |of| |Exp1| |and| |Exp2|
	 |from| |the| |currently| |active| |hypotheses| |and| |governors| |if|
	 |and| |only| |if| |neither| |this| |equation| |nor| |its|
	 |commuted| |version| |are| |among| |the| |currently| |active|
	 |hypotheses| |or| |governors| |.| CR
	 |(= Exp1 Exp2 ID)| and |(= Exp1 Exp2 IFF)| |:| |See| |(HELP-LONG =).|)))

(DEFPARAMETER =-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |=| | :| |Make| |a| |substitution| |for| |the| |current| |subterm| |,|
	 |using| |an| |equality| |which| |is| |explicitly| |among| |the|
	 |current| |active| |hypotheses| |and| |governors| |.| CR |#|
	 |(= Exp)| |:| |Replace| |the| |current| |subterm| |by| |Exp| |,| |if| |they|
	 |can| |be| |proved| |equal| |by| |the| |theorem| |prover| |under| |the|
	 |current| |active| |hypotheses| |and| |governors| |.| CR |#|
	 |(= Exp1 Exp2)| |:| |Replace| |Exp1| |by| |Exp2| |in| |the|
	 |current| |subterm| |,| |if| |they| |can| |be| |proved| |equal| |by|
	 |the| |theorem| |prover| |under| |the| |current| |active| |hypotheses|
	 |and| |governors| |.| CR |#|
	 |(= Exp1 Exp2 (hint1 hint2 ... hintK))| |:| |As| |above| |,| |but|
	 |with| |the| |indicated| |list| |of| |hints| |given| |to| |the|
	 |theorem-prover| |(with| |the| |same| |syntax| |as| |the| |hints|
	 |to| |PROVE-LEMMA)| |.|
	 NOTE |:| |It| |is| |allowed| |to| |use| |abbreviations| |in| |the|
	 |hints| |to| |the| |prover| |(i.e.| |in| |the| |third| |argument)|
	 |.| CR |#|
	 |(= Exp1 Exp2 0)| |:| |As| |above| |,| |but| |create| |a|
	 |new| |goal| |to| |prove| |(EQUAL Exp1 Exp2)| |from| |the| |currently|
	 |active| |hypotheses| |and| |governors| |if| |and| |only| |if| |neither| |this|
	 |equation| |nor| |its|
	 |commuted| |version| |are| |among| |the| |currently| |active|
	 |hypotheses| |or| |governors| |.| NOTE |:| |Any| |atom| |may| |be|
	 |used| |in| |place| |of| |the| |final| |`0'| |,| except note
	 |that| ID |and| IFF |are| |treated| |specially| |(see below).|
	 CR |#|
	 NOTES |:| CR |#|
	 |(1)| |It| |is| |permissible| |to| |use| |the| |symbol| |*| |for|
	 |Exp1| |to| |denote| |the| |current| |subterm| |.| |This| |convention|
	 |is| |useful| |when| |3| |arguments| |are| |given| |.| CR |#|
	 |(2)| |Only| |the| |0-| |and| |1-| |argument| |versions| |are|
	 |allowed| |when| |substituting| |for| |a| |term| |all| |of| |whose|
	 |function| |symbols| |are| |constructors| |and| |bottom| |objects|
	 |.| CR |#|
	 |(3)| |In| |the| |case| |that| |the| |instruction| |is| |*not*|
	 |of| |the| |form|
	 |(= <...> <...> ID)| |,|
	 |but| |is| |not| |simply| |the| |command| |=| |(with| |no|
	 |arguments)| |,| |then| |the|
	 |prover| |will| |be| |asked| |to|
	 |prove| |a| |term| |of| |the| |form|
	 |(IFF Exp1 Exp2)| |rather| |than| |(EQUAL Exp1 Exp2)| |,|
	 |provided| |that| |the| |following| |conditions| |are| |met| |.|
	 |First| |,| |the| |current| |subterm| |should|
	 |occur| |in| |a| |propositional| |position| |inside| |the| |conclusion|
	 |(e.g.| |in| |the| |first| |argument| |position| |of| |an| IF |expression)| |,|
	 |or| |else| |both| |Exp1| |and| |Exp2| |should| |be| |of| |boolean| |type| |.|
	 |Also| |,| |either| |the| |instruction| |should| |be| |of| |the|
	 |form| |(= <...> <...> IFF)| |,| |or| |else| |not| |both|
	 |Exp1| |and| |Exp2| |may| |be| |of| |boolean| |type| |.|
	 |In| |these| |cases| |,| |if| |a| |new| |subgoal| |is| |created|
	 |then| |it| |will|
	 |have| |a| |conclusion| |of| |the| |form|
	 |(IFF Exp1 Exp2)| |rather| |than| |of| |the| |form|
	 |(EQUAL Exp1 Exp2)| |.|)))

(DEFPARAMETER DEMOTE-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(DEMOTE n1 n2 ... nk)|
	 |:| |Replace| |the| |current| |conclusion| |with| |the| |goal|
	 |of| |proving| |that| |the| |conjunction| |of| |the| |indicated|
	 |hypotheses| |implies| |the| |current| |conclusion| |,| |and|
	 |drop| |these| |hypotheses| |.| |Note| |that| |this| |command|
	 |may| |only| |be| |used| |when| |at| |the| |top| |of| |the|
	 |conclusion| |.| CR CR |#|
	 DEMOTE |:| |As| |above| |,| |but| |demote| |all| |the| |active|
	 |hypotheses| |.| CR
	 EXAMPLE |:| |If| |the| |active| |hypotheses| |are| CR |#| 
	 |H1.|  (LESSP X Y) CR |#|
	 |H2.|  (NOT (ZEROP Z)) CR
	 |and| |the| |conclusion| |and| |current| |subterm| |are| |both| CR |#|
	 (LESSP (TIMES X Z) (TIMES Y Z)) |,| CR
	 |then| |after| |executing| DEMOTE |,| |there| |will| |be| |no|
	 |active| |hypotheses| |and| |the| |conclusion| |will| |be| CR |#|
	 (IMPLIES (AND (LESSP X Y) (NOT (ZEROP Z)))
		  (LESSP (TIMES X Z) (TIMES Y Z))))))

(DEFPARAMETER DEMOTE-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(DEMOTE n1 n2 ... nk)|
	 |:| |Replace| |the| |current| |conclusion| |with| |the| |goal|
	 |of| |proving| |that| |the| |conjunction| |of| |the| |indicated|
	 |hypotheses| |implies| |the| |current| |conclusion| |,| |and|
	 |drop| |these| |hypotheses| |.| |Note| |that| |this| |command|
	 |may| |only| |be| |used| |when| |at| |the| |top| |of| |the|
	 |conclusion| |.| CR |#|
	 DEMOTE |:| |As| |above| |,| |but| |demote| |all| |the| |active|
	 |hypotheses| |.|)))

(DEFPARAMETER PROVE-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |PROVE| |:| |Attempt| |to| |prove| |the| |current| |goal| |.|
	 CR |#| |(PROVE hint1 ... hintk)| |:| |as| |above|
	 |,| |where| |each| |hint| |is| |as| |in| PROVE-LEMMA |and|
	 |the| |theorem| |prover| |is| |to| |use| |these| |hints|
	 |as| |it| |does| |with| PROVE-LEMMA |.| CR
	 EXAMPLE |:| CR |#|
	 (PROVE (DISABLE TIMES APPEND) (INDUCT (PLUS X Y))) |.|)))

(DEFPARAMETER PROVE-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |PROVE| |:| |Attempt| |to| |prove| |the| |current| |goal| |.|
	 |Use| |(= T)| |instead| |if| |you| |are| |not| |at| |the| |top|
	 |of| |the| |conclusion| |.| NOTE |that| |if| |there| |are|
	 |any| |active| |hypotheses| |in| |the| |current| |goal| |,|
	 |then| |what| |is| |actually| |attempted| |is| |a| |proof| |of|
	 |the| |implication| |whose| |antecedent| |is| |the| |conjunction|
	 |of| |the| |active| |hypotheses| |and| |whose| |conclusion| |is|
	 |the| |goal's| |conclusion| |.| CR |#|
	 |(PROVE hint1 ... hintk)| |:| |as| |above|
	 |,| |where| |each| |hint| |is| |as| |in| PROVE-LEMMA |and|
	 |the| |theorem| |prover| |is| |to| |use| |these| |hints|
	 |as| |it| |does| |with| PROVE-LEMMA |.| NOTE |:| |It| |is|
	 |allowed| |to| |use| |abbreviations| |in| |the| |hints| |.|)))

(DEFPARAMETER SPLIT-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| SPLIT |:| |Replace| |the| |current| |goal| |by| |the| |cases| |generated|
	 |from| |its| |propositional| |structure| |.| |This| |command| |can|
	 |only| |be| |used| |at| |the| |top| |of| |the| |conclusion| |.|
	 |Roughly| |speaking| |,| |this| |command| |does|
	 |OR-splitting| |in| |the| |hypotheses| |and| |AND-splitting|
	 |in| |the| |conclusion| |.| CR
	 EXAMPLE |:| |Suppose| |that| |there| |is| |a| |hypothesis|
	 |of| |the| |form| |(IMPLIES (AND P Q) R)| |.| |Then| |a| |subgoal|
	 |would| |be| |generated| |replacing| |this| |hypothesis| |by|
	 |(NOT P)| |,| |and| |similarly| |there| |would| |be| |one| |for|
	 |(NOT Q)| |and| |for| R |.| |Of| |course| |,| |there| |could| |be|
	 |more| |subgoals| |generated| |if| P |,| Q |,| |or| |R| |had|
	 |additional| |propositional| |structure| |or| |if| |other|
	 |hypotheses| |or| |the| |conclusion| |generated| |further|
	 |cases| |.|)))

(DEFPARAMETER SPLIT-LONG-HELP-MESSAGE
  (QUOTE
    (|#| SPLIT |:| |Replace| |the| |current| |goal| |by| |the| |cases| |generated|
	 |from| |its| |propositional| |structure| |.| |More| |precisely| |:|
	 |the| |implication| |formed| |from| |the| |active| |hypotheses| |and|
	 |conclusion| |is| |put| |in| |conjunctive| |normal| |form| |,|
	 |and| |each| |disjunct| |(clause)| |is| |made| |into| |a| |new| |goal|
	 |.| |Duplicate| |hypotheses| |are| |eliminated| |.|
	 |This| |command| |can|
	 |only| |be| |used| |at| |the| |top| |of| |the| |conclusion| |.|
	 |Roughly| |speaking| |,| |this| |command| |does|
	 |OR-splitting| |in| |the| |hypotheses| |and| |AND-splitting|
	 |in| |the| |conclusion| |.| CR |#|
	 NOTE |:| |The| |new| |goals| |will| |all| |have| |their| |hypotheses|
	 |promoted;| |no| |conclusion| |will| |have| |a| |top| |function| |symbol|
	 |of| IMPLIES |.| |Also| |note| |that| |this| |command| |will| |fail| |if|
	 |the| |new| |goal| |is| |sufficiently| |similar| |to| |the| |current|
	 |goal| |,| |i.e.| |if| |it| |has| |the| |same| |number| |of| |hypotheses|
	 |and| |the| |same| |conclusion| |.| |Finally| |,| |note| |that| SPLIT
	 |does| |not| |look| |at| |the| |propositional| |structure| |past| |the| |top|
	 |level| |.| |That| |is| |,| |as| |one| |dives| |into| |a| |term| |one| |stops|
	 |looking| |inside| |subterms| |whose| |top| |function| |symbol| |is| |other|
	 |than| AND |,| OR |,| NOT |,| IMPLIES |,| IF |,| |or| IFF |.| |In| |particular| |,|
	 |one| |may| |wish| |to| |apply| S-PROP |or| |(S-PROP NIL)| |before| |applying|
	 SPLIT |in| |case| |there| |are| IF |expressions| |which| |are| |not| |in|
	 |the| |top-level| |propositional| |structure| |of| |the| |conclusion| |.|)))

(DEFPARAMETER CLAIM-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(CLAIM exp)|
	 |:| |attempt| |to| |prove| | exp | |from| |the| |currently|
	 |active| |hypotheses| |(using| |the| |theorem| |prover)| |,|
	 |and| |if| |this| |can| |be| |done|
	 |,| |then| |add| | exp | |as| |a| |hypothesis| |.| CR
	 EXAMPLE |:| (CLAIM (NUMBERP (PLUS X Y))) |will| |simply| |add|
	 |(NUMBERP (PLUS X Y))| |as| |a| |hypothesis| |in| |the| |current| |goal| |.|
	 |However| |,| |if| |one| |submits| |the| |command| (CLAIM (NUMBERP X))
	 |and| |if| |this| |cannot| |be| |proved| |by| |the| |prover| |(from|
	 |the| |existing| |active| |hypotheses)| |,| |then| |there| |is| |no|
	 |change| |.| CR CR |#|
	 |(CLAIM exp 0)| |:| |Add| | exp | |as| |a| |hypothesis| |,| |and|
	 |create| |a| |new| |subgoal| |with| |the| |negation| |of| | exp |
	 |added| |as| |a| |hypothesis| |.| CR EXAMPLE |:| |(CLAIM (NUMBERP X) 0)|
	 |causes| |a| |case| |split| |on| |(NUMBERP X)| |,| |in| |the| |sense|
	 |that| |(NUMBERP X)| |is| |added| |as| |a| |hypothesis| |of| |the|
	 |current| |goal| |,| |and| |a| |new| |subgoal| |is| |created| |which|
	 |is| |identical| |to| |the| |current| |goal| |except|
	 |that| |(NOT (NUMBERP X))| |is| |added| |as| |a| |hypothesis| |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:| CR
	 |(CLAIM exp (hint1 hint2 ... hintK))| | {to| |give| |hints| |to| |the|
	 |prover}| CR
	 |(CLAIM exp TAUT)| | {to| |replace| |the| |prover| |call| |with| |a|
	 |tautology| |check}| CR)))

(DEFPARAMETER CLAIM-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(CLAIM exp)|
	 |:| |attempt| |to| |prove| | exp | |from| |the| |currently|
	 |active| |hypotheses| |(using| |the| |theorem| |prover)| |,|
	 |and| |if| |this| |can| |be| |done|
	 |,| |then| |add| | exp | |as| |a| |hypothesis| |.| |If| |the|
	 |proof| |fails| |,| |then| |of| |course| |this| |command| |fails| |.| CR
	 NOTE |:| |If| |one| |successfully| |CLAIMs| |a| |term| |(AND x y ...)|
	 |then| |each| |of| |x,| |y,| |...| |will| |be| |added| |as| |a|
	 |separate| |hypothesis| |.| CR |#|
	 |(CLAIM exp (hint1 hint2 ... hintK))|
	 |:| |As| |above| |,| |but| |if| |the| |theorem-prover| |is| |invoked|
	 |then| |it| |will| |use| |the| |indicated| |list| |of| |hints| |(with|
	 |the| |same| |syntax| |as| |the| |hints| |to| |PROVE-LEMMA)| |.| NOTE |:|
	 |It| |is| |allowed| |to| |use| |abbreviations| |in| |the| |hints| |.| CR |#|
	 |(CLAIM exp TAUT)|
	 |:| |As| |above| |,| |but| |no| |proof| |attempt| |is| |made| |other|
	 |than| |tautology| |checking| |.| |However| |,| |in| |this| |case| |,|
	 |the| |negation| |of| |the| |conclusion| |is| |also| |assumed| |for|
	 |the| |purpose| |of| |tautology-checking| |.| CR |#|
	 |(CLAIM exp 0)| |:| |Add| | exp | |as| |a| |hypothesis| |,| |and|
	 |create| |a| |new| |subgoal| |with| |(instead)| |the| |negation| |of| | exp |
	 |added| |as| |a| |hypothesis| |.| |Notice| |that| |no| |proof|
	 |attempt| |or| |tautology| |checking| |is| |made| |whatsoever| |in|
	 |this| |case| |.| |Actually| |,| |any| |atom| |besides| TAUT |will|
	 |behave| |the| |same| |as| | 0 | |here| |.| CR |#|
	 |NOTE| |:| |Any| |variables| |in| | exp | |that| |do| |not| |occur| |anywhere| |else|
	 |in| |the| |current| |proof| |state| |will| |be| |added| |to| |the|
	 |list| |of| |free| |variables| |of| |the| |proof| |state| |.|)))

(DEFPARAMETER PROMOTE-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| PROMOTE |:| |Replace| |(IMPLIES hyps exp)| |with| |simply|
	 | exp| |,| |adding| | hyps | |to| |the| |list| |of| |hypotheses|
	 |(after| |flattening| |its| AND |structure| |,| |if| |it| |is|
	 |an| AND |expression)| |.| CR
	 EXAMPLE |:| |If| |the| |conclusion| |is| (IMPLIES (AND X Y) Z) |,|
	 |then| |after| |execution| |of| PROMOTE |,| |the| |conclusion| |will|
	 |be| Z |and| |the| |terms| X |and| Y |will| |be| |hypotheses| |.|
	 CR CR |#| OTHER FORMS |(see| |HELP-LONG)| |:| CR |#|
	 (PROMOTE T))))

(DEFPARAMETER PROMOTE-LONG-HELP-MESSAGE
  (QUOTE
   (|#| PROMOTE |:| |Replace| |(IMPLIES hyps exp)| |with| |simply|
    | exp| |,| |adding| | hyps | |to| |the| |list| |of| |hypotheses|
    |(after| |flattening| |its| AND |structure| |,| |if| |it| |is|
    |an| AND |expression)| |.| CR |#| NOTE |:| |You| |must| |be| |at| |the|
    |top| |of| |the| |conclusion| |to| |apply| |this| |command| |.| CR |#|
    |(PROMOTE n)| |:| |Similar| |,| |but| |with| |no| |flattening| |of|
    |the| |AND| |structure| |in| |the| |sense| |that| |we| |view| | hyps |
    |as| |a| |conjunction| |of| |at| |most| | n | |hypotheses| |.| |Hence|
    |for| |example| |,| |if| |the| |conclusion| |is| |(IMPLIES (AND X Y Z) W)| |then|
    |(PROMOTE 2)| |will| |result| |in| |two| |new| |hypotheses| |,|
    X |and| |(AND Y Z)| |.| |This| |version| |is| |primarily| |of|
    |interest| |to| |writers| |of| |macro| |commands| |.| NOTE |:| |Use|
    |an| |argument| |of| |0| |if| |there| |is| |no| |desired| |limit| |on|
    |the| |number| |of| |hypotheses| |created| |.| |An| |argument| |of| |1|
    |will| |of| |course| |cause| | hyps | |to| |become| |a| |single| |hypothesis| |.|)))

(DEFPARAMETER REWRITE-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| (REWRITE N) |:| |Replace| |the| |current| |subterm| |with| |a|
	 |new| |term| |by| |using| |the| |Nth| |rewrite| |rule| |,| |as|
	 |displayed| |by| |the| |command| SHOW-REWRITES |.| CR
	 EXAMPLE |:| |Suppose| |that| |the| |current| |subterm| |is|
	 (REVERSE (REVERSE X)) |.| |The| |command| SHOW-REWRITES |might|
	 |result| |in| |the| |following| |being| |printed| |to| |the|
	 |screen| |:| CR |#|
	 |1.| REVERSE-NLISTP CR |#|
	 |New term:| |NIL | CR |#|
	 |Hypotheses:| ((NOT (LISTP (REVERSE X)))) CR |#|
	 |2.| REVERSE-REVERSE CR |#|
	 |New term:| X CR |#|
	 |Hypotheses:| ((PLISTP X)) CR
	 |Then| (REWRITE 2) |results| |in| |a| |new| |current| |subterm| |of|
	 X |together| |with| |a| |new| |subgoal| |of| |proving| (PLISTP X)
	 |under| |the| |current| |goal's| |active| |hypotheses| |.| |(Note|
	 |,| |however| |,| |that| |the| |new| |goal| |would| |not| |be|
	 |created| |if| |(PLISTP X)| |is| |among| |the| |currently|
	 |active| |hypotheses| |or| |governors.)| CR CR |#|
	 NOTE |:| |All| |variables| |introduced| |into| |the| |proof| |state|
	 |by| REWRITE |are| |added| |to| |the| |list| |of| |free|
	 |variables| |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:| CR |#|
	 |One| |may| |specify| |the| |name| |of| |the| |rewrite| |rule|
	 |instead| |of| |the| |number| |.| |One| |may| |also| |specify| |a|
	 |substitution| |for| |the| |free| |variables| |.| |Also| |,|
	 REWRITE |(with| |no| |arguments)|
	 |may| |be| |used| |to| |apply| |the| |first| |hypothesis-free|
	 |rewrite| |rule| |(or| |,| |just| |the| |first| |rule| |if| |all|
	 |have| |hypotheses| |to| |relieve)| |,| |as| |displayed| |by|
	 SHOW-REWRITES |.| |Use| (HELP-LONG REWRITE) |to| |get| |the|
	 |details| |.|)))

(DEFPARAMETER REWRITE-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(REWRITE rule-name substitution optional-lemma-statement)|
	 |:| |Rewrite| |subterm | |to| |new-subterm| |using|
	 |the| |rewrite| |rule| |named| |rule-name| |(or| |numbered|
	 |instead| |--| |see| |"CONVENTIONS"| |below)| |.| |The| |substitution|
	 |argument| |is| |a| |list| |of| |pairs| |of| |the| |form|
	 |(var term)| |,| |where| |var| |is| |a| |free| |variable| |of|
	 |the| |rule| |and| |should| |be| |prefixed| |by| |a| |dollar|
	 |sign| |($)| |,| |as| |displayed| |by| |the| |command| SHOW-REWRITES
	 |.| |(A| |"free variable"| |of| |a| |rule|
	 (IMPLIES |<hyps>| (EQUAL |<left_side>| |<right_side>|)) |is| |a|
	 |variable| |occurring| |in| |the| |rule| |that| |does| |not|
	 |occur| |in| |<left_side>.)|
	 |Free| |variables| |not| |covered| |by| |the| |substitution| |will|
	 |simply| |have| |their| |dollar| |sign| |prefixes| |removed| |,|
	 |with| |a| |warning| |printed| |in| |this| |case| |if| |also|
	 |new| |subgoals| |are| |created| |(to| |relieve| |hypotheses)| |or|
	 |if| |there| |are| |free| |variables| |in| |the| |conclusion| |.|
	 |The| |final| |argument| |is| |simply| |a| |comment| |(see| |also|
	 |below)| |.| CR |#|
	 OPTIONAL ARGUMENTS |:| |Any| |initial| |segment| |of| |the| |three|
	 |arguments| |of| REWRITE |is| |legal| |,| |where| |the|
	 |substitution| |defaults| |to| | NIL | |and| |the|
	 |comment| |is| |omitted| |.| |In| |particular| |,|
	 |the| |atom| REWRITE |means| |that| |any| |rule| |may| |be| |used|
	 |to| |rewrite| |the| |current| |term| |to| |anything| |(but|
	 |see| |Remark| |(1)| |below)| |.|
	 CR |#|
	 NOTE |on| |free| |variables| |of| |the| |proof| |state| |:|
	 |All| |variables| |introduced| |into| |the| |proof| |state|
	 |by| REWRITE |are| |added| |to| |the| |list| |of| |free|
	 |variables| |.| CR |#|
	 CONVENTIONS ON |rule-name| |:| |The| |first| |argument| |may| |be|
	 |either| |a| |positive| |integer| N |,| |indicating| |the| |Nth|
	 |rewrite| |rule| |(as| |shown| |by| |the|
	 |command| |SHOW-REWRITES)| |,|
	 |or| |else| |the| |name| |of| |a| |rewrite| |rule| |.| |The| |system|
	 |will| |store| |the| |command| |using| |the| |name| |of| |the| |rule| |,|
	 |even| |if| |the| |user| |provides| |a| |number| |.| |However| |,| |if|
	 |the| |user| |provides| |a| |number| |and| |there| |are| |two| |or| |more| |rewrite|
	 |rules| |with| |the| |rule's| |name| |that| |match| |the| |current|
	 |subterm| |,| |then| |the| |rule| |that| |is| |actually| |used| |will|
	 |be| |stored| |in| |the| |optional-lemma-statement| |field| |.|
	 CR |#| REMARKS |:| CR |#|
	 |(1)| |Subject| |to| |the| |above| |constraints| |,| |use| |of| |the|
	 |atomic| |command| REWRITE |will| |always| |cause| |an|
	 |attempt| |to| |rewrite| |without|
	 |introducing| |new| |goals| |(to| |relieve| |hypotheses)| |,| |even|
	 |if| |it| |is| |necessary| |to| |automatically| |choose| |additional|
	 |instantiations| |for| |free| |variables| |in| |order| |to|
	 |accomplish| |this| |.| |(No| |additional| |instantiations| |are|
	 |chosen| |,| |however| |,| |unless| ALL |hypotheses| |can|
	 |be| |relieved.)| |By| |the| |way| |,| |the| |algorithm| |for|
	 |"completing"| |substitutions| |is| |rather| |weak| |;| |it's|
	 |not| |much| |more| |than| |trying| |to| |arrange| |all| |such|
	 |hypotheses| |to| |be| |among| |the| |currently| |active|
	 |hypotheses| |and| |governors| |,| |and| |not| |worrying| |about|
	 |instances| |of| |(EQUAL X X)| |.| CR |#|
	 |(2)| |If| |there| |are| |free| |variables| |in| |the|
	 |rewrite| |rule| |that| |have| |not| |been| |specified| |for|
	 |instantiation| |by| |the| |user's| |substitution| |,| |then|
	 |they| |will| |be| |left| |unchanged| |.| |In| |this| |case| |,|
	 |if| |new| |subgoals| |are| |created| |(to| |relieve| |hypotheses)|
	 |or| |if| |there| |is| |at| |least| |one| |such| |uninstantiated|
	 |variable| |in| |the| |conclusion| |,| |then| |an| |appropriate|
	 |warning| |message| |will| |be| |printed| |.| CR |#|
	 |(3)| |If| |the| |current| |subterm| |is| |in| |a| |position| |where|
	 |only| |propositional| |equality| |matters| |,| |the| REWRITE
	 |command| |is| |allowed| |to| |take| |advantage| |of|
	 |this| |situation| |.| CR |#|)))

(DEFPARAMETER INDUCT-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(INDUCT (g v1 ... vn))| |:| |Induct| |as| |in| |the|
	 |corresponding| INDUCT |hint| |given| |to| |the| |theorem|
	 |prover| |,| |creating| |new| |subgoals| |for| |the| |base|
	 |and| |induction| |steps| |.| CR
	 EXAMPLE |:| |If| |the| |current| |conclusion| |is| |of| |the|
	 |form| |(P X)| |and| |there| |are| |no| |hypotheses| |,|
	 |then| |after| (INDUCT (TIMES X Y)) |,| |there| |are| |two|
	 |new| |subgoals| |with| |conclusions| CR |#|
	 (IMPLIES (ZEROP X) (P X)) | and| CR |#|
	 (IMPLIES (AND (NOT (ZEROP X))
		       (P (SUB1 X)))
		  (P X))
	 CR CR |#|
	 INDUCT |:| |Induct| |according| |to| |a| |scheme| |chosen|
	 |by| |the| |theorem| |prover's| |heuristics| |,| |creating|
	 |subgoals| |for| |the| |base| |and| |induction| |steps| |.|)))

(DEFPARAMETER INDUCT-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(INDUCT x)|
	 |:| |The| |argument| |of| INDUCT |can| |take| |any| |of| |three|
	 |forms| |:| CR |#|
	 |If| |the| |argument| |is| |a| TERM |,| |then| |induct|
	 |according| |to| |that| |term| |,| |as| |in| |the|
	 INDUCT |hint| |given| |to| PROVE-LEMMA |.| |The| |result|
	 |will| |be| |the| |creation| |of| |new| |goals| |(for| |the|
	 |resulting| |base| |and| |induction| |steps)| |.| CR |#|
	 |If| |the| |argument| |is| |a| |function| |symbol| |g| |,|
	 |the| |command| |is| |interpreted| |as|
	 |(INDUCT (g v1 ... vn))| |,| |where| |(g v1 ... vn)| |is| |some|
	 |subterm| |of| |the| |conclusion| |whose| |arguments| |are|
	 |distinct| |variable| |symbols| |.| CR |#|
	 |Otherwise| |,| |the| |argument| |should| |be| |a| |positive|
	 |integer| |N| |,| |and| |the| |command| |causes| |new| |goals|
	 |to| |be| |created| |that| |reflect| |the| |Nth|
	 |heuristically-chosen| |induction| |scheme| |(as| |shown|
	 |by| |the| |command| |SHOW-INDUCTIONS)| |.| CR |#|
	 NOTE ALSO |:| CR |#| |(1)| |The| |command| INDUCT |is| |the| |same| |as|
	 |(INDUCT 1)| |,| |and| |may| |be| |used| |quite| |often| |since|
	 |usually| |there| |will| |be| |at| |most| |one|
	 |heuristically-chosen| |induction| |scheme| |.| |(2)| |If| |there|
	 |are| |any| |free| |variables| |of| |the| |state| |that| |occur| |in|
	 |the| |current| |goal| |,| |these| |will| |be| |removed| |from| |the|
	 |free| |variables| |list| |.| |This| |is| |necessary| |for| |soundness| |.|
	 |(3)| |If| |there| |are| |any| |free| |variables| |of| |the| |state| |occurring|
	 |in| |the| |variables| |of| |a| |command| |of| |the| |form|
	 |(INDUCT (g v1 ... vn))| |,| |these| |will| |be| |removed| |from| |the|
	 |free| |variables| |list| |.| |This| |too| |is| |necessary| |for| |soundness| |.|
	 |(4)| |The| |new| |goals| |will| |all| |be| |without| |top-level|
	 |hypotheses| |,| |and| |hence| |some| |of| |them| |will| |have|
	 |conclusions| |which| |have| IMPLIES |as| |the| |top| |function| |symbol| |.|)))

(DEFPARAMETER BOOKMARK-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(BOOKMARK x)| |:| |Makes| |no| |change| |in| |the| |state|
	 |except| |to| |insert| |this| |instruction| |,|
	 |for| |the| |purpose| |of| |possible| |UNDOing|
	 |later| |.| CR
	 EXAMPLE |:| |The| |command| (BOOKMARK HOWDY) |creates| |an|
	 |instruction| |such| |that| |if| |later| |one| |submits|
	 |(UNDO HOWDY)| |,| |then| |the| |state| |stack| |will| |be|
	 |unwound| |back| |to| |(but| |not| |including)| |this| |bookmark| |.|)))

(DEFPARAMETER BOOKMARK-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(BOOKMARK x)| |:| |Makes| |no| |change| |in| |the| |state|
	 |except| |to| |insert| |this| |instruction| |,|
	 |for| |the| |purpose| |of| |possible| |UNDOing|
	 |later| |.| |The| |argument| |may| |not| |be|
	 |a| |number| |.|)))

(DEFPARAMETER COMMENT-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(COMMENT ....)| |:| |A| |no-op| |,| |except| |that| |the| |instruction|
	 |field| |of| |the| |new| |state| |contains| |the| |given| |instruction|
	 |.| CR
	 EXAMPLE |:| |(COMMENT HI THERE HUMAN)| |makes| |this| |instruction|
	 |the| |new| |instruction| |,| |as| |shown| |e.g.| |by| |executing|
	 COMMANDS |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:| CR COMMENT |makes| |a| |comment|
	 |which| |displays| |the| |current| |subterm| |.|)))

(DEFPARAMETER COMMENT-LONG-HELP-MESSAGE
  (QUOTE
    (|#| COMMENT |:| |Makes| |no| |change| |in| |the| |state| |except| |to|
	 |insert| |the| |instruction| |(COMMENT ***CURRENT-TERM*** x)|
	 |,| |where| |x| |is| |the| |current| |term| |(printed| |using|
	 |the| |current| |abbreviations)| |.| |The| |only|
	 |use| |for| |this| |command| |is| |as| |a| |comment| |.|
	 |Note| |that| |the| |current| |term| |is| |displayed| |with|
	 |respect| |to| |the| |current| |abbreviations| |.| CR |#|
	 |(COMMENT ....)| |:| |As| |above| |,| |but| |inserts| |the|
	 |instruction| |literally| |.|)))

(DEFPARAMETER CONTRADICT-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(CONTRADICT n)|
	 |:| |Negate| |the| |current| |conclusion| |and| |make| |it| |the|
	 |n-th| |hypothesis| |,| |while| |negating| |the| |current| |n-th|
	 |hypothesis| |and| |making| |it| |the| |current| |conclusion| |.| CR
	 EXAMPLE |:| |If| |the| |active| |hypotheses| |are| CR |#| 
	 |H1.|  (LESSP X Y) CR |#|
	 |H2.|  (NOT (ZEROP Z)) CR
	 |and| |the| |conclusion| |and| |current| |subterm| |are| |both| CR |#|
	 (LESSP (TIMES X Z) (TIMES Y Z)) |,| CR
	 |then| |after| |executing| (CONTRADICT 2) |,| |the| |hypotheses| |are|
	 CR |#|
	 |H1.|  (LESSP X Y) CR |#|
	 |H2.|  (NOT (LESSP (TIMES X Z) (TIMES Y Z))) CR
	 |and| |the| |conclusion| |is| CR |#|
	 (ZEROP Z) |.|)))

(DEFPARAMETER CONTRADICT-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(CONTRADICT n)|
	 |:| |Negate| |the| |current| |conclusion| |and| |make| |it| |the|
	 |n-th| |hypothesis| |,| |while| |negating| |the| |current| |n-th|
	 |hypothesis| |and| |making| |it| |the| |current| |conclusion| |.|
	 NOTE |that| |this| |command| |may| |only| |be| |invoked| |while|
	 |at| |the| |top| |of| |the| |current| |conclusion| |.|)))

(DEFPARAMETER ADD-ABBREVIATION-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(ADD-ABBREVIATION <var> <exp>)|
	 |:| |Add| |the| |abbreviation| |which| |displays| |<exp>|
	 |as| |<var>| |.| CR |#|
	 NOTE |:| |The| |variable| |<var>| |should| |start| |with|
	 |the| |character| |@| |.| CR
	 EXAMPLE |:| (ADD-ABBREVIATION @V (TIMES X Y)) |causes| |future|
	 |occurrences| |of| (TIMES X Y) |to| |be| |printed| |as| |@V| |.|
	 |Moreover| |,| |user| |input| |can| |henceforth| |contain| |the|
	 |symbol| |@V| |as| |an| |abbreviation| |for| |(TIMES X Y)| |.|
	 CR CR |#|
	 OTHER FORM |(see| |HELP-LONG)| |:| |(ADD-ABBREVIATION <var>).|)))

(DEFPARAMETER ADD-ABBREVIATION-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(ADD-ABBREVIATION <var> <exp>)|
	 |:| |Add| |the| |abbreviation| |which| |displays| |<exp>|
	 |as| |<var>| |.| |Future| |output| |will| |use| |this| |abbreviation|
	 |,| |and| |user| |input| |may| |be| |given| |which| |uses| |this|
	 |abbreviation| |.| CR |#|
	 NOTE |:| |The| |variable| |<var>| |should| |start| |with|
	 |the| |character| |@| |.| CR |#|
	 |(ADD-ABBREVIATION <var>)| |:| |Same| |effect| |as|
	 |(ADD-ABBREVIATION <var> <exp>)| |where| |<exp>| |is| |the|
	 |current| |term| |.|)))

(DEFPARAMETER REMOVE-ABBREVIATIONS-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| REMOVE-ABBREVIATIONS |:| |Remove| |all| |abbreviations| |.| CR CR |#|
	 |(REMOVE-ABBREVIATIONS var1 var2 ... varK)| |:|
	 |Remove| |the| |indicated| |abbreviations| |.| CR
	 EXAMPLE |:| (REMOVE-ABBREVIATIONS @X @Y) |will| |cause| |future|
	 |proof| |states| |to| |be| |unaware| |of| |@X| |and| |@Y| |as|
	 |abbreviations| |(unless| |they| |are| |restored| |by| |future| 
	 |invocations| |of| |ADD-ABBREVIATION)| |.|)))

(DEFPARAMETER REMOVE-ABBREVIATIONS-LONG-HELP-MESSAGE
  (QUOTE
    (|#| REMOVE-ABBREVIATIONS |:| |Remove| |all| |abbreviations| |.| CR |#|
	 |(REMOVE-ABBREVIATIONS var1 var2 ... varK)| |:|
	 |Remove| |the| |indicated| |abbreviations| |,| |i.e.|
	 |no| |longer| |allow| |the| |abbreviations| |which|
	 |result| |in| |the| |indicated| |variables| |.| CR
	 |#| NOTE |:| |Each| |of| |the| |variables| |vi| |should|
	 |start| |with| |the| |character| |@| |.|)))

;; USE-GOAL used to be documented here.

(DEFPARAMETER USE-LEMMA-SHORT-HELP-MESSAGE
  (QUOTE
   (|#| |(USE-LEMMA name ((var1 term1) ... (varK termK)))| |:| |Add|
    |the| |given| |lemma| |or| |axiom| |from| |the| |chronology|
    |to| |the| |list| |of| |hypotheses| |,| |after| |instantiating|
    |it| |according| |to| |the| |given| |substitution| |.| CR
    EXAMPLE |:| |Invocation| |of| (USE-LEMMA TIMES-0 ((X A))) |will|
    |add| |the| |hypothesis| (EQUAL (TIMES A 0) 0) |if| |the| |lemma|
    TIMES-0 |(from| |the| |chronology)| |is| (EQUAL (TIMES X 0) 0)
    |.| CR CR |#|
    USE-LEMMA |:| |Same| |as| |(USE-LEMMA ())| |.| CR CR |#|
    NOTE |:| |All| |variables| |introduced| |into| |the| |proof| |state|
    |by| USE-LEMMA |are| |added| |to| |the| |list| |of| |free|
    |variables| |.|)))

(DEFPARAMETER USE-LEMMA-LONG-HELP-MESSAGE
  (QUOTE
   (|#| |(USE-LEMMA name ((var1 term1) ... (varK termK)))|
    |:| |Add| |the| |given| |lemma| |or| |axiom| |from| |the| |chronology|
    |to| |the| |list| |of| |hypotheses| |,| |after| |instantiating|
    |it| |according| |to| |the| |given| |substitution| |.| CR |#|
    USE-LEMMA |:| |Same| |as| |(USE-LEMMA ())| |.|  CR |#|
    NOTE |on| |free| |variables| |of| |the| |proof| |state| |:|
    |All| |variables| |introduced| |into| |the| |proof| |state|
    |by| USE-LEMMA |are| |added| |to| |the| |list| |of| |free|
    |variables| |.|)))

(DEFPARAMETER EXIT-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(EXIT name lemma-types)| |:| |Exit| |the| |proof| |checker|
	 |,| |printing| |out| |the| |appropriate|
	 |PROVE-LEMMA| |event| |with| |the| |given|
	 |name| |and| |types| |(e.g.| |(REWRITE)| |or|
	 |NIL)| |.| CR 
	 EXAMPLE |:| (EXIT TIMES-ASSOCIATIVITY (REWRITE)) |.| CR CR |#|
	 |(EXIT T)| |:| |Exit| |the| |proof| |checker| |upon| |completion| |of|
	 |an| |interactive| |verification| |of| |termination| |of| |a| |definition| |.|
	 |Thus| |,| |this| |version| |should| |be| |used| |when| |the| |session| |started|
	 |with| |VERIFY-DEFN| |rather| |than| VERIFY |.| CR CR |#|
	 EXIT |:| |Quit| |the| |proof| |checker| |without| |making|
	 |an| |event| |.| |In| |this| |case| |,| |(VERIFY)| |may| |be|
	 |executed| |to| |get| |back| |in| |at| |the| |same| |state| |,| |as|
	 |long| |as| |there| |hasn't| |been| |an| |intervening| |use| |of|
	 |the| |proof-checker| |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:| |(EXIT name lemma-types T)|
	 |and| |(EXIT T T)| |.|)))

(DEFPARAMETER EXIT-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(EXIT name lemma-types)| |:| |Exit| |the| |proof| |checker|
	 |,| |printing| |out| |the| |appropriate|
	 |PROVE-LEMMA| |event| |with| |the| |given|
	 |name| |and| |types| |(e.g.| |(REWRITE)| |or|
	 |NIL)| |.|
	 CR |#|
	 |(EXIT T)| |:| |Exit| |the| |proof| |checker| |upon| |completion| |of|
	 |an| |interactive| |verification| |of| |termination| |of| |a| |definition| |.|
	 |Thus| |,| |this| |version| |may| |be| |used| |when| |the| |session| |started|
	 |with| |VERIFY-DEFN| |rather| |than| VERIFY |.| CR |#|
	 EXIT |:| |Quit| |the| |proof| |checker| |without| |making|
	 |an| |event| |.| |In| |this| |case| |,| |(VERIFY)| |may| |be|
	 |executed| |to| |get| |back| |in| |at| |the| |same| |state| |,| |as|
	 |long| |as| |there| |hasn't| |been| |an| |intervening| |use| |of|
	 |the| |proof-checker| |.| CR |#|
	 NOTES |:| CR |#| |(1)| |The| |first| |form| |,| |(EXIT name lemma-types)| |,| |may| |be| |used|
	 |even| |when| |the| |session| |began| |with| |a| |verification| |of| |termination| |,|
	 |i.e.| |by| |invoking| |VERIFY-DEFN| |.| |In| |that| |case| |the| |session| |will| |be|
	 |as| |though| |it| |was| |started| |with| |(VERIFY <goal>)| |,| |where| |<goal>| |is|
	 |the| |goal| |generated| |by| |the| |termination-proving| |mechanism| |.| |Exception:|
	 |If| |<goal>| |contains| |the| |function| |symbol| |that| |is| |being| |defined| |,|
	 |then| |this| |conversion| |is| |not| |allowed| |.| |Also| |,| |although| |it| |is|
	 |allowed| |to| |have| |an| |occurrence| |of| |'fn| |inside| |an| |EVAL$| |or| |APPLY$|
	 |or| |the| |like| |,| |such| |uses| |may| |in| |obscure| |cases| |cause| |an| |event|
	 |to| |be| |formed| |which| |can| |not| |be| |replayed| |.| CR |#|
	 |(2)| |If| |EXIT| |is| |given| |an| |additional| |argument| |that| |is|
	 |not| |NIL,| |then| |the| |event| |will| |be| |submitted| |immediately| |,| |just|
	 |as| |though| |the| |user| |had| |answered| |Y| |to| |the| |prompt| |.|)))
	
(DEFPARAMETER PP-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| PP |:| |Prettyprint| |current| |subterm| |without| |any| |special|
	 |use| |of| |pretty-printing| |abbreviations| |.| |So| |for|
	 |example| |,| |(AND x y z)| |is| |printed| |as|
	 |(AND x (AND y z));| |compare| |with| P |.|)))
	
(DEFPARAMETER PP-LONG-HELP-MESSAGE
  (QUOTE
    (|#| PP |:| |Prettyprint| |current| |subterm| |without| |any| |special|
	 |use| |of| |pretty-printing| |abbreviations| |.| |So| |for|
	 |example| |,| |(AND x y z)| |is| |printed| |as|
	 |(AND x (AND y z));| |compare| |with| P |.| |Note| |however|
	 |that| |the| |printing|
	 |is| |with| |respect| |to| |the| |current| |abbreviations|
	 |(as| |displayed| |by| |the| |command| |SHOW-ABBREVIATIONS)| |.|)))

(DEFPARAMETER P-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| P |:| |Prettyprint| |current| |subterm| |in| |the| |usual| |manner| |.|
	 |So| |for| |example| |,| |(AND x y z)| |is| |printed|
	 |rather| |than| |(AND x (AND y z));| |compare| |with| PP |.|)))

(DEFPARAMETER P-LONG-HELP-MESSAGE
  (QUOTE
    (|#| P |:| |Prettyprint| |current| |subterm| |in| |the| |usual| |manner| |.|
	 |So| |for| |example| |,| |(AND x y z)| |is| |printed|
	 |rather| |than| |(AND x (AND y z));| |compare| |with| PP |.| |Note|
	 |however| |that| |the|
	 |printing| |is| |with| |respect| |to| |the| |current| |abbreviations|
	 |(as| |displayed| |by| |the| |command| |SHOW-ABBREVIATIONS)| |.|)))

(DEFPARAMETER PP-TOP-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |PP-TOP| |:| |Prettyprint| |entire| |term| |,|| highlighting|
	 |current| |subterm| |.| |Printing| |is| |with| |respect|
	 |to| |the| |same| |conventions| |as| |for| |the| |command|
	 PP |.| CR
	 EXAMPLE |:| |If| |the| |conclusion| |is|
	 (EQUAL (PLUS X X 0) (TIMES X 2)) |and| |the| |current| |subterm|
	 |is| (PLUS X X 0) |,| |then| PP-TOP
	 |will| |cause| |the| |printing| |of|
	 (EQUAL (*** (PLUS X (PLUS X 0)) ***)
		(TIMES X 2)))))

(DEFPARAMETER PP-TOP-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |PP-TOP| |:| |Prettyprint| |entire| |term| |,|| highlighting|
	 |current| |subterm| |.| |Printing| |is| |with| |respect|
	 |to| |the| |same| |conventions| |as| |for| |the| |command|
	 PP |.|)))

(DEFPARAMETER GOALS-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |GOALS|
	 |:| |Print| |the| |name| |of| |each| |unproved|
	 |goal| |,| |current| |goal| |first| |.| CR CR |#|
	 OTHER FORMS |(see| |HELP-LONG)| |:| 
	 |(GOALS ALL)| |,| |(GOALS TREE)| |.|)))

(DEFPARAMETER GOALS-LONG-HELP-MESSAGE
  (QUOTE
   (|#| |GOALS|
    |:| |Print| |the| |name| |of| |each| |unproved|
    |goal| |,| |current| |goal| |first| |.| CR |#|
    |(GOALS ALL)| |:| |Print| |the| |name| |of| |each|
    |goal| |,| |marking| |each| |proved| |one| |with|
    |an| |asterisk| |(*)| |and| |organizing| |the| |output|
    |in| |a| |tree| |form| |to| |show| |dependencies| |.|
    |Note| |that| |the| |current|
    ORDER |of| |goals| |is| |not| |displayed| |in| |this|
    |case| |.| |Each| |goal|
    |is| |followed| |by| |the| |command| |that| |created| |it|
    |together| |with| |the| |position| |(as| |shown| |by| |COMMANDS)|
    |at| |which| |is| |was| |created| |.| CR |#|
    |(GOALS TREE)| |:| |Same| |as| |(GOALS ALL)| |,| |except| 
    |that| |only| |the| |names| |of| |the| |unproved| |goals|
    |(but| |including| |MAIN)| |are| |shown| |.| CR CR |#|
    EXAMPLE |:| |Suppose| |that| |the| |first| |command| |is| INDUCT |,|
    |which| |finishes| |goal| MAIN |but| |creates| |(MAIN . 1)| |and| |(MAIN . 2)|
    |,| |and| |then| |the| |next| |command| |is| BASH |,| |which| |creates|
    |children| |of| |(MAIN . 1)| |named| |((MAIN . 1) . 1)| |and|
    |((MAIN . 1) . 2)| |and| |finishes| |(MAIN . 1)| |.| |Then| GOALS |results|
    |in| |the| |following| |display| |:| CR CR |((MAIN . 1) . 1)| CR
    |((MAIN . 1) . 2)| CR |(MAIN . 2)| CR CR
    |(GOALS ALL)| |results| |in| |the| |following| |display| |.|
    |Notice| |that| |each| |goal| |is| |followed| |by| |the| |number| |of|
    |proof| |steps| |back| |at| |which| |it| |was| |created| |and|
    |the| |command| |that| |created| |it| |.| CR CR
    |*MAIN <<3>> START| CR
    |   *(MAIN . 1) <<2>> INDUCT| CR
    |       ((MAIN . 1) . 1) <<1>> BASH| CR
    |       ((MAIN . 1) . 2) <<1>> BASH| CR
    |    (MAIN . 2) <<2>> INDUCT|)))

(DEFPARAMETER HYPS-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| HYPS |:| |Print| |the| |current| |active| |hypotheses| |,| |both|
	 |top-level| |and| |governors| |.| CR CR |#|
	 (HYPS ALL ALL) |As| |above| |,| |but| |show| |the| |hidden|
	 |hypotheses| |and| |governors| |as| |well| |.| CR
	 EXAMPLE |:| |If| |the| |current| |hypotheses| |are| (EQUAL X1 Y) |and|
	 (EQUAL X2 Y) |(in| |that| |order)| |,| |where| |the| |first| |of|
	 |these| |is| |hidden| |but| |the| |second| |is| |active| |,|
	 |and| |there| |are| |no| |governors| |,| |then| (HYPS ALL ALL)
	 |will| |display| |:| CR CR
	 |*** Top-level hypotheses (with "*" when hidden):| CR
	 |*H1.  (EQUAL X1 Y)| CR
	 |H2.  (EQUAL X2 Y)| CR CR
	 |*** Governors (with "*" when hidden):| CR
	 |There are no governors to display.| CR CR |#|
	 | while| HYPS |will| |simply| |display| |:| CR CR
	 |*** Active top-level hypotheses:| CR
	 |H2.  (EQUAL X2 Y)| CR CR
	 |*** Active governors:| CR
	 |There are no governors to display.| CR CR |#|
	 OTHER FORM |(see| |HELP-LONG)| |:|
	 |(HYPS (H1 H2 ... Hk) (G1 G2 ... Gn))|)))

(DEFPARAMETER HYPS-LONG-HELP-MESSAGE
  (QUOTE
    (|#| HYPS |:| |Print| |the| |current| |active| |hypotheses| |,| |both|
	 |top-level| |and| |governors| |.| CR |#|
	 |(HYPS (H1 H2 ... Hk) (G1 G2 ... Gn))| |:| |As| |just| |above|
	 |,| |but| |only| |show| |the| |hypotheses| |and| |numbers| |as|
	 |shown| |;| |e.g.| |(HYPS (3 4) NIL)| |will| |show|
	 |hypotheses| |3| |and| |4| |(whether| |hidden| |or| |not)|
	 |but| |will| |show| |no| |governors| |.| NOTE |:| |Use| ALL |in|
	 |place| |of| |a| |list| |if| |there| |is| |to| |be| |no|
	 |restriction| |.|)))

(DEFPARAMETER SHOW-REWRITES-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| SHOW-REWRITES
	 |:| |Show| |all| |the| |rewrite| |rules| |which| |the| REWRITE |command|
	 |may| |apply| |to| |the| |current| |subterm| |,|
	 |including| |the| |name| |,| |resulting|
	 |subterm| |,| |and| |hypotheses|
	 |whose| |relieving| |will| |require| |the|
	 |generation| |of| |new| |subgoals| |.| CR
	 EXAMPLE |:| |Suppose| |that| |the| |current| |subterm| |is|
	 (LESSP A (PLUS A 3)) |and| |that| |there| |is| |a| |rewrite|
	 |rule| |for| |transitivity| |of| LESSP |.| |Then| SHOW-REWRITES
	 |might| |display| |the| |following| |:| CR CR
	 |1. LESSP-TRANS| CR
	 |New term: T| CR
	 |Hypotheses: ((LESSP A $Y) (LESSP $Y (PLUS A 3)))| CR CR
	 |The| |dollar| |sign| |indicates| |a| |free| |variable| |,| |i.e.|
	 |a| |variable| |which| |occurs| |in| |the| |hypotheses| |of| |the|
	 |rewrite| |rule| |(or| |,| |in| |rare| |cases| |,| |the| |right|
	 |side| |of| |the| |conclusion| |of| |the| |rule)| |but| |does| |not|
	 |occur| |in| |the| |left| |side| |of| |the| |conclusion| |of| |the|
	 |rule| |.| NOTE |:| |All| |variables| |introduced| |into| |the| |proof| |state|
	 |by| REWRITE |are| |added| |to| |the| |list| |of| |free|
	 |variables| |.| CR CR |#|
	 OTHER FORM |(see| |HELP-LONG)| |:| |(SHOW-REWRITES <argument>)| |.|)))

(DEFPARAMETER SHOW-REWRITES-LONG-HELP-MESSAGE
  (QUOTE
    (|#| SHOW-REWRITES
	 |:| |Show| |all| |the| |rewrite| |rules| |which|
	 |apply| |to| |the| |current| |subterm| |,|
	 |including| |the| |name| |,| |resulting|
	 |subterm| |,| |and| |hypotheses|
	 |whose| |relieving| |will| |require| |the|
	 |generation| |of| |new| |subgoals| |.| CR |#|
	 |(SHOW-REWRITES T)| |:| |as| |above| |,| |but| |only| |an|
	 |abbreviated| |form| |of| |the| |resulting| |subterm|
	 |and| |hypotheses| |will| |be| |displayed| |.| CR |#|
	 |(SHOW-REWRITES n)| |:| |If| |n| |is| |a| |positive| |integer|
	 |or| |a| |name| |,| |this| |will| |cause| |the| |display| |of|
	 |the| |nth| |rewrite| |rule| |(in| |the| |former| |case)| |or|
	 |of| |all| |rewrite| |rules| |with| |name| |n| |(in| |the|
	 |latter| |case)| |,| |in| |the| |same| |form| |as| |given| |by|
	 |the| |command| SHOW-REWRITES |.| CR |#|
	 NOTES |:| CR |#|
	 |(1)| |Free| |variables| |,| |i.e.| |variables|
	 |not| |occurring| |in| |the| |left| |side| |of| |a|
	 |rewrite| |rule| |,| |will| |always| |be| |prefixed|
	 |with| |a| |dollar| |sign| |.| |However| |,| |if|
	 |some| |instantiation| |for| |the| |free| |variables|
	 |causes| |all| |hypotheses| |to| |be| |among| |the|
	 |current| |hypotheses| |,| |then| |this|
	 |instantiation| |will| |be| |made| |unless| |there| |are|
	 |free| |variables| |in| |the| |conclusion| |.|
	 NOTE |that| |all| |variables| |introduced| |into| |the| |proof| |state|
	 |by| REWRITE |are| |added| |to| |the| |list| |of| |free|
	 |variables| |.|  CR |#|
	 |(2)| |Applicable| |rewrite| |rules| |are| |displayed|
	 |in| |reverse| |chronological| |order| |(i.e.| |reverse|
	 |order| |of| |introduction| |into| |the| |chronology| |,|
	 |just| |like| |the| |theorem-prover's| |order| |of|
	 |application| |of| |rewrite| |rules)| |.| CR |#|
	 |(3)| |Rewrite| |rules| |which| |are| |globally| |disabled| |are|
	 |so| |marked| |,| |e.g.| CR |#| |1.| APPEND-ASSOC |(disabled)| CR |#|
	 |New| |term:| |(APPEND X (APPEND Y Z))| CR |#| |Hypotheses:|  |<none>| CR
	 |This| |information| |can| |be| |useful| |in| |case| |one| |intends|
	 |to| |try| |an| |automatic| |proof| |later| |.| |This| |"disabled"|
	 |status| |has| |no| |effect| |on| |the| REWRITE |command| |.|)))

(DEFPARAMETER COMMANDS-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| COMMANDS
	 |:| |Show| |all| |of| |the| |previous| |commands| |,|
	 |in| |reverse| |order| |.| CR
	 EXAMPLE |:| |Here| |is| |some| |possible| |output| |upon|
	 |execution| |of| COMMANDS |:| CR CR
	 |The commands thus far (in reverse order, i.e. last one first) have been:|
	 |1. (CLAIM (LESSP A B))| CR
	 |2. (REWRITE LESSP-TRANS)| CR
	 |3. PROMOTE| CR
	 |4. START| CR CR
	 |Thus| |,| |there| |have| |been| |three| |commands| |so| |far|
	 |,| |with| |the| |CLAIM| |command| |being| |the| |most| |recent|
	 |.| CR CR |#|
	 |(COMMANDS n)| |:| |Show| |the| |last| |n| |previous|
	 |commands| |,| |in| |reverse| |order| |.|)))

(DEFPARAMETER COMMANDS-LONG-HELP-MESSAGE
  (QUOTE
    (|#| COMMANDS
	 |:| |Show| |all| |of| |the| |previous| |commands| |,|
	 |in| |reverse| |order| |.| CR |#|
	 |(COMMANDS n)| |:| |Show| |the| |last| |n| |previous|
	 |commands| |,| |in| |reverse| |order| |.|)))

(DEFPARAMETER SHOW-INDUCTIONS-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| SHOW-INDUCTIONS
	 |:| |Show| |all| |the| |heuristically-chosen| |induction|
	 |schemes| |.|)))

(DEFPARAMETER SHOW-INDUCTIONS-LONG-HELP-MESSAGE
  (QUOTE
    (|#| SHOW-INDUCTIONS
	 |:| |Show| |all| |the| |heuristically-chosen| |induction|
	 |schemes| |.|)))

(DEFPARAMETER SHOW-ABBREVIATIONS-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| SHOW-ABBREVIATIONS
	 |:| |Show| |all| |abbreviations| |.| CR
	 EXAMPLE OUTPUT |:| CR CR
	 |@Y corresponds to| CR |#|
	 |(TIMES B C)| CR CR
	 |@X corresponds to| CR |#|
	 |(PLUS A @Y)| CR |#| | |
	 |i.e. to| CR |#|
	 |(PLUS A (TIMES B C))| CR CR |#|
	 OTHER FORM |(see| |HELP-LONG)| |:|
	 |(SHOW-ABBREVIATIONS var1 var2 ... varK)| |.|)))

(DEFPARAMETER SHOW-ABBREVIATIONS-LONG-HELP-MESSAGE
  (QUOTE
    (|#| SHOW-ABBREVIATIONS
	 |:| |Show| |all| |abbreviations| |.| CR |#|
	 |(SHOW-ABBREVIATIONS var1 var2 ... varK)| |:| |Show|
	 |the| |terms| |abbreviated| |by| |the| |indicated|
	 |variables| |(each| |of| |which| |should| |have| |@| |as|
	 |its| |first| |character)| |.| CR |#|
	 NOTE |:| |The| |abbreviations| |are| |displayed| |in| |the|
	 |reverse| |of| |the| |order| |in| |which| |they| |were|
	 |introduced| |.| |If| |an| |abbreviation| |depends| |on|
	 |other| |abbreviations| |,| |then| |both| |the|
	 |abbreviated| |version| |(without| |using| |the| |given|
	 |abbreviation)| |and| |the| |raw| |version| |of|
	 |the| |term| |abbreviated| |are| |displayed| |.|)))

(DEFPARAMETER HELP-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| HELP |:| |Print| |names| |of| |selected| |instructions| |,| |including|
	 |a| |selected| |set| |of| |predefined| |macro| |commands| |.| NOTE |:|
	 |To| |get| |the| |names| |of| |all| |instructions| |and| |all|
	 |macro| |commands| |,| |use| HELP-LONG |.| CR CR |#|
	 |(HELP instr1 ...)| |:| |Print| |short| |help| |on| |instr1| |...|
	 |,| |generally| |with| |examples| |.|
	 NOTE |:| |Use| HELP-LONG |to| |get| |more| |complete| |information| |,|
	 |but| |without| |examples| |.|)))

(DEFPARAMETER HELP-LONG-HELP-MESSAGE
  (QUOTE
    (|#| HELP |:| |Print| |names| |of| |selected| |instructions| |,| |including|
	 |a| |selected| |set| |of| |predefined| |macro| |commands| |.| NOTE |:|
	 |To| |get| |the| |names| |of| |all| |instructions| |and| |all|
	 |macro| |commands| |,| |use| HELP-LONG |.| CR |#|
	 |(HELP instr1 ...)| |:| |Print| |short| |help| |on| |instr1| |...|
	 |,| |generally| |with| |examples| |.|
	 NOTE |:| |Use| HELP-LONG |to| |get| |more| |complete| |information| |,|
	 |but| |without| |examples| |.| CR |#| |By| |the| |way| |, the| |long|
	 |help| |is| |sometimes| |not| |any| |longer| |than| |the| |(short)|
	 |help| |,| |though| |often| |it| |does| |contain| |more| |details| |about|
	 |unusual| |situations| |and| |certain| |variations| |on| |the| |commands| |.|)))

(DEFPARAMETER HELP-LONG-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| HELP-LONG |:| |Print| |names| |of| |all| |instructions| |,| |including|
	 |all| |macro| |commands| |.| |For| |a| |shortened| |list| |appropriate|
	 |for| |people| |new| |to| |this| |system| |,| |use| HELP |instead| |.| CR CR |#|
	 |(HELP-LONG instr1 ...)| |:| |Print| |help| |on| |instr1| |...| |.|
	 NOTE |:| |Use| HELP |to| |get| |less| |detailed| |information| |,|
	 |but| |with| |examples| |,| |which| |is| |perhaps| |more| |appropriate|
	 |for| |people| |who| |are| |just| |beginning| |to| |use| |this|
	 |system| |.|)))

(DEFPARAMETER HELP-LONG-LONG-HELP-MESSAGE
  (QUOTE
    (|#| HELP-LONG |:| |Print| |names| |of| |all| |instructions| |,| |including|
	 |all| |macro| |commands| |.| |For| |a| |shortened| |list| |appropriate|
	 |for| |people| |new| |to| |this| |system| |,| |use| HELP |instead| |.| CR |#|
	 |(HELP-long instr1 ...)| |:| |Print| |help| |on| |instr1| |...| |.|
	 NOTE |:| |Use| HELP |to| |get| |less| |detailed| |information| |,|
	 |but| |with| |examples| |,| |which| |is| |perhaps| |more| |appropriate|
	 |for| |people| |who| |are| |just| |beginning| |to| |use| |this|
	 |system| |.| CR |#| |By| |the| |way| |, the| |long|
	 |help| |is| |sometimes| |not| |any| |longer| |than| |the| |(short)|
	 |help| |,| |though| |often| |it| |does| |contain| |more| |details| |about|
	 |unusual| |situations| |and| |certain| |variations| |on| |the| |commands| |.|)))

(DEFPARAMETER UNDO-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(UNDO N)|
	 |:| |Undo| |N| |instructions| |.| CR
	 EXAMPLE |:| (UNDO 3) |undoes| |the| |last| |three| |instructions| |.| CR CR |#|
	 UNDO |:| |Undo| |one| |instruction| |.| CR CR |#|
	 OTHER FORM |(see| |HELP-LONG)| |:|
	 |(UNDO x)| |,| |where| |x| |is| |a| |bookmark| |.|)))

(DEFPARAMETER UNDO-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(UNDO N)|
	 |:| |Undo| |N| |instructions| |.| CR
	 |#| |(UNDO x)| |:| |Undo| |back| |up| |to| |(but|
	 |not| |through)| |the| |instruction|
	 |(BOOKMARK x)| |if| |x| |is| |not| |an| |integer| |.| CR
	 |#| UNDO |:| |Undo| |one| |instruction| |.| CR
	 |#| NOTES |:| CR |#|
	 |(1)| |One| |may| |not| |undo| |back| |past| |the| |start|
	 |of| |the| |current| |sesstion| |,| |i.e.| |an| |instruction| |of|
	 |the| |form| |(START <event-name>)| |may| |not| |be| |undone| |,|
	 |except| |by| |the| |no-argument| |instruction| UNDO |.| CR |#|
	 |(2)| |This| |instruction| |"succeeds"| |just| |in| |case| |it|
	 |has| |any| |effect| |.|)))

(DEFPARAMETER RESTORE-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| RESTORE |:| |Restores| |the| |state| |stack| |to| |its|
	 |value| |just| |before| |the| |immediately|
	 |preceding| RESTORE |or| UNDO |,| |if| |any| |.| CR
	 EXAMPLE |:| |After| |successfully| |executing| (UNDO 3) |,|
	 |the| |execution| |of| RESTORE |will| |put| |the| |state-stack|
	 |back| |to| |its| |value| |before| |that| |UNDOing| |,| |provided| |that|
	 |there| |is| |no| |intervening| |invocation| |of| UNDO |or| RESTORE |.|)))

(DEFPARAMETER RESTORE-LONG-HELP-MESSAGE
  (QUOTE
    (|#| RESTORE |:| |Restores| |the| |state| |stack| |to| |its|
	 |value| |just| |before| |the| |immediately|
	 |preceding| RESTORE |or| UNDO |,| |if| |any| |.|  CR |#| |To| |be| |more|
	 |precise| |:| |two| |state| |stacks| |are| |kept| |around| |,|
	 |in| |the| |global| |variables| STATE-STACK |and| OLD-SS |.|
	 RESTORE |switches| |these| |.| UNDO |saves| |the| |current| |value|
	 |of| STATE-STACK |in| OLD-SS |and| |then| |modifies| STATE-STACK |as|
	 |instructed| |by| |the| UNDO |command| |.| CR |#| |This|
	 |command| |"succeeds"| |unless| |such| |restoration|
	 |is| |impossible| |(which| |happens| |if| |no| UNDO
	 |has| |occurred| |during| |the| |session)| |.|)))

(DEFPARAMETER LISP-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(LISP form)| |:| |Evaluate| |the| |given| |Lisp| |form| |.| CR
	 EXAMPLE |:| (LISP (PROGN (PRINT (+ 2 5)) (TERPRI NIL))) |will|
	 |cause| |the| |numeral| |7| |to| |be| |printed| |,| |followed|
	 |by| |a| |newline| |.|)))

(DEFPARAMETER LISP-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(LISP form)| |:| |Evaluate| |the| |given| |Lisp| |form| |.|
	 |This| |command| |"succeeds"| |if| |and| |only|
	 |if| |the| |result| |is| |non-NIL| |.|)))

(DEFPARAMETER BIND-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(BIND bindings instr1 instr2 ... )| |:| |This| |instruction| |is|
	 |only| |of| |interest| |to| |writers| |of| |macro| |commands| |--|
	 |execute| (HELP-LONG BIND) |if| |you| |really| |care| |.|)))

(DEFPARAMETER BIND-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(BIND <bindings> <instr1> <instr2> ... )|
	 |:| |Push| |the| |indicated| |local| |variable| |declarations|
	 |(and| |initial| |values)| |on| |the| |top| |of| |the|
	 |COMMAND-BINDINGS| |stack| |,| |for| |reference| |by| |commands|
	 |of| |the| |form| |(! <instr>)| |.| |The| |instruction|
	 |"succeeds"| |if| |and| |only| |if| |all| |the| |given|
	 |instructions| |"succeed"| |.| |The| |syntax| |of| BIND |is| |the|
	 |same| |as| |that| |of| |Lisp's| LET |,| |i.e.| |each| |member|
	 |of| |<bindings>| |should| |be| |a| |pair| |(<var> <form>)| |or|
	 |simply| |an| |atom| |<var>| |(which| |is| |treated| |like|
	 |(<var> NIL))| |.| CR |#| NOTE |:| |For| |each| |pair|
	 |(var exp)| |in| |the| |given| |bindings| |,| |exp| |is| |evaluated|
	 |in| |the| |result| |of| |extending| |the| |global| |environment|
	 |by| |the| |existing| COMMAND-BINDINGS |,| |and| |the| |result|
	 |is| |bound| |to| |var| |in| |the| |new| COMMAND-BINDINGS |.| |These|
	 |bindings| |are| |done| |in| |parallel| |.| CR
	 EXAMPLE |:| (BIND ((Y 3)) (BIND ((X (+ 2 Y))) (UNDO (! X)) DIVE)) |will|
	 |cause| |the| |execution| |of| (UNDO 5) |followed| |by| |the| |execution|
	 |of| DIVE |(whether| |or| |not| |the| (UNDO 5) |"succeeds")| |.|)))

(DEFPARAMETER DO-ALL-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(DO-ALL instr1 instr2 ...)|
	 |:| |Run| |all| |of| |the| |given| |instructions| |.| CR
	 EXAMPLE |:| (DO-ALL UNDO DIVE (PLAY S S-PROP)) |.|)))

(DEFPARAMETER DO-ALL-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(DO-ALL instr1 instr2 ...)|
	 |:| |Run| |all| |of| |the| |given| |instructions| |,| |"succeeding"| |if|
	 |and| |only| |if| |all| |the| |given| |instructions| |"succeed"| |.|)))

(DEFPARAMETER PROG2-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(PROG2 instr1 instr2)|
	 |:| |This| |instruction| |is|
	 |only| |of| |interest| |to| |writers| |of| |macro| |commands| |--|
	 |execute| (HELP-LONG PROG2) |if| |you| |really| |care| |.|)))

(DEFPARAMETER PROG2-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(PROG2 instr1 instr2)|
	 |:| |Run| |command| |instr1| |.| |If| |it| |"succeeds"| |,| |then|
	 |run| |instr2| |and| |"succeed"| |if| |and| |only| |if|
	 |instr2| |"succeeds"| |.| |If| |instr1| |"fails"| |then| |this|
	 |command| |"fails"| |.|)))

(DEFPARAMETER NEGATE-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(NEGATE instr)|
	 |:| |This| |instruction| |is|
	 |only| |of| |interest| |to| |writers| |of| |macro| |commands| |--|
	 |execute| (HELP-LONG NEGATE) |if| |you| |really| |care| |.|)))

(DEFPARAMETER NEGATE-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(NEGATE instr)|
	 |:| |Run| |the| |command| |instr| |,| |but| |consider| |this|
	 |instruction| |to| |"succeed"| |if| |and| |only| |if| |the|
	 |run| |of| |instr| |"fails"| |.|)))

(DEFPARAMETER PROTECT-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(PROTECT instr)| |:| |Treats| |instr| |as| |an| |atomic| |entity| |,|
	 |in| |the| |sense| |that| |if| |it| |is| |a| |macro| |command| |,|
	 |then| |either| |instr| |"succeeds"| |or| |else| |the| |global| |state|
	 |reverts| |to| |what| |it| |was| |before| |invocation| |of| |the|
	 PROTECT |command| |.| CR
	 EXAMPLE |:| (PROTECT (DO-ALL DIVE PUSH TOP S)) |will| |either| |cause|
	 |execution| |of| |all| |instructions| DIVE |,| PUSH |,| TOP |,| |and|
	 S |(in| |that| |order)| |,| |or| |else| |will| |be| |a| |no-op| |on|
	 |the| |global| |state| |(in| |case| |any| |of| |these| |four| |instructions|
	 |does| |not| |create| |a| |new| |state)| |.|)))

(DEFPARAMETER PROTECT-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(PROTECT instr)|
	 |:| |Run| |the| |command| |instr| |,| |but| |with| |the|
	 |single| |change| |that| |if| |this| |command| |"fails"| |,|
	 |then| |revert| |to| |the| |global| |state| |present|
	 |before| |this| |execution| |.|)))

(DEFPARAMETER SUCCEED-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |(SUCCEED instr)|
	 |:| |This| |instruction| |is|
	 |only| |of| |interest| |to| |writers| |of| |macro| |commands| |--|
	 |execute| (HELP-LONG SUCCEED) |if| |you| |really| |care| |.|)))

(DEFPARAMETER SUCCEED-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |(SUCCEED instr)|
	 |:| |Run| |the| |command| |instr| |,| |and| |consider| |it|
	 |to| |have| |"succeeded"| |even| |if| |it| |actually| |did|
	 |not| |.|)))

(DEFPARAMETER BASH-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| |BASH| |:| |Attempt| |to| |simplify| |the| |current| |goal|
	 |using| |the| |theorem| |prover's| |simplifier| |(which|
	 |includes| |the| |rewriter)| |.|
	 CR |#| |(BASH hint1 ... hintk)| |:| |as| |above|
	 |,| |where| |each| |hint| |is| |as| |in| PROVE-LEMMA |and|
	 |the| |theorem| |prover| |is| |to| |use| |these| |hints|
	 |as| |it| |does| |with| PROVE-LEMMA |.| |Special| |case| |:| |the|
	 |"hint"| |T| |indicates| |that| |the| |entire| |prover| |is|
	 |to| |be| |called| |(except| |for| |induction)| |rather| |than|
	 |just| |the| |simplifier;| |see| |also| REDUCE |.| CR
	 EXAMPLE |:| CR |#|
	 (BASH T (DISABLE TIMES APPEND) (INDUCT (PLUS X Y))) |.|)))

(DEFPARAMETER BASH-LONG-HELP-MESSAGE
  (QUOTE
    (|#| |BASH| |:| |Attempt| |to| |simplify| |the| |current| |goal|
	 |using| |the| |theorem| |prover's| |simplifier| |(which|
	 |includes| |the| |rewriter)| |.|
	 |You| |must| |be| |at| |the| |top|
	 |of| |the| |conclusion| |.| NOTE |that| |if| |there| |are|
	 |any| |active| |hypotheses| |in| |the| |current| |goal| |,|
	 |then| |what| |is| |actually| |simplified| |is| |the| |implication|
	 |whose| |antecedent| |is| |the| |conjunction|
	 |of| |the| |active| |hypotheses| |and| |whose| |conclusion| |is|
	 |the| |goal's| |conclusion| |.| CR |#|
	 |(BASH hint1 ... hintk)| |:| |as| |above|
	 |,| |where| |each| |hint| |is| |as| |in| PROVE-LEMMA |and|
	 |the| |theorem| |prover| |is| |to| |use| |these| |hints|
	 |as| |it| |does| |with| PROVE-LEMMA |.| NOTE |:| |It| |is|
	 |allowed| |to| |use| |abbreviations| |in| |the| |hints| |.| |Also| |,|
	 |a| |special| |case| |is| |that| |the|
	 |"hint"| |T| |indicates| |that| |the| |entire| |prover| |is|
	 |to| |be| |called| |(except| |for| |induction)| |rather| |than|
	 |just| |the| |simplifier;| |see| |also| REDUCE |.| CR |#|
	 NOTE |:| |The| |new| |goals| |will| |all| |be| |without| |top-level|
	 |hypotheses| |,| |and| |hence| |they| |may| |have|
	 |conclusions| |which| |have| IMPLIES |as| |the| |top| |function| |symbol| |.| CR |#|
	 FINAL NOTE |:| |If| |the| |"hint"| T |is| |supplied| |,| |then| |in| |the| |prover|
	 |call| |,| |the| |sometime| |behavior| |of| |the| |prover| |that| |prefers|
	 |to| |induct| |on| |the| |original| |goal| |is| |suppressed| |.| |(Otherwise| |,|
	 |new| |subgoals| |would| |not| |be| |generated| |,| |and| |the| BASH |command|
	 |would| |fail.)|  |If| |you| |really| |want| |this| |behavior| |to| |be| |allowed| |after|
	 |all| |,| |then| |use| |any| |other| |non-NIL| |atom| |as| |a| |"hint"|
	 |in| |place| |of| |T| |.|)))

(DEFPARAMETER UNDONE-INSTRUCTIONS-SHORT-HELP-MESSAGE
  (QUOTE
    (|#| UNDONE-INSTRUCTIONS
	 |:| |Print| |the| |list| |of| |instructions| |which| |were|
	 |given| |to| |,| |but| |not| |run| |by| |,|
	 |the| |last| |invocation| |of| |any|
	 |command| |which| |keeps| |track| |of| |such| |matters| |,|
	 |such| |as| PROVE-LEMMA |.|)))

(DEFPARAMETER UNDONE-INSTRUCTIONS-LONG-HELP-MESSAGE
  (QUOTE
   (|#| UNDONE-INSTRUCTIONS
    |:| |Print| |the| |list| |of| |instructions| |which| |were|
    |given| |to| |,| |but| |not| |run| |by| |,|
    |the| |last| |invocation| |of| |any|
    |command| |which| |keeps| |track| |of| |such| |matters| |,|
    |such| |as| PROVE-LEMMA |.| CR |#|
    |The| |command| |(UNDONE-INSTRUCTIONS)| |may| |be| |run| |from|
    |Lisp| |as| |well| |as| |from| |inside| |the| |proof-checker| |.|)))