;;; -*-  Mode: Lisp; Package: USER; Syntax: Common-Lisp; Base: 10 -*- ;;;;

;  Copyright (C) 1994 by Matt Kaufmann and Computational Logic, Inc.
;  All Rights Reserved.

;  Copying of this file is authorized to those who have read and
;  agreed with the terms in the "Pc-Nqthm-1992 GENERAL PUBLIC SOFTWARE
;  LICENSE" at the beginning of the Pc-Nqthm file "basis-pc.lisp".

;  Pc-Nqthm Version 1992 (for use with Nqthm-1992)

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

(DEFUN MAKE-EQUALITY-INSTR (N HYPS &OPTIONAL SUPPRESS-PRINT-FAILURE-FLG)
  (LET (LHS RHS HYP-FLG)
       (COND
	((NOT (AND (INTEGERP N)
		   (> N 0)))
	 (AND (NOT SUPPRESS-PRINT-FAILURE-FLG)
	      (BLURB (N) |**NO| |CHANGE**| |--| |The| |argument| (!PPR N NIL) |is| |not|
		     |a| |positive| |integer| |.| CR)))
	((> N (LENGTH HYPS))
	 (AND (NOT SUPPRESS-PRINT-FAILURE-FLG)
	      (BLURB (N (LNG (LENGTH (HYPS)))) |**NO| |CHANGE**| |--| |The| |argument|
		     (!PPR N NIL) |is| |greater| |than| |the| |number| |of| |top-level|
		     |hypotheses| |,| (!PPR LNG (QUOTE |.|)) CR)))
	((PROGN (SETQ HYP-FLG (NTH (1- N) HYPS))
		(EQ (CDR HYP-FLG) 'H))
	 (AND (NOT SUPPRESS-PRINT-FAILURE-FLG)
	      (BLURB (N) |**NO| |CHANGE**| |--| |Hypothesis| |number| (!PPR N NIL)
		     |is| |currently| |hidden| |.| CR)))
	((NOT (MATCH (CAR HYP-FLG) (EQUAL LHS RHS)))
	 (AND (NOT SUPPRESS-PRINT-FAILURE-FLG)
	      (BLURB (N) |**NO| |CHANGE**| |--| |Hypothesis| |number| (!PPR N NIL)
		     |is| |not| |of| |the| |form| |(EQUAL t1 t2)| |.| CR)))
	(T
	 `(= ,(untranslate-abb LHS (abbreviations))
	     ,(untranslate-abb RHS (abbreviations))
	     0)))))

; Here are the command added 12/30/87 for the macro command DV.

(defun expand-address-1 (n form)
; Takes a form which is the output of UNTRANSLATE and an argument
; position n for that form, and returns the list of addresses
; required to get to the correspoding argument of the translated
; form.  We assume that form is not a variable or a QUOTEP.
;   For now I'll punt on FOR.
; One more thing -- in the case of CASE and COND, n will be a cons pair
; of numbers rather than a single number.  The first component of the
; pair tells the argument position, while the second is 1 or 2 indicating
; which of the pair to take.  (Special case:  CASE and a first component
; of 1, the second component is ignored.)  I will assume that n is a cons
; pair of positive integers if the function symbol is CASE or COND and a
; single positive integer otherwise.
  (let (temp (fn (car form)))
    (cond 
     ((not (and (> n 0)
		(< n (length form)))) 
      (er soft (form n) (!ppr n nil) is an illegal argument position into
	  (!ppr form (quote |.|))))
     ((setq temp (car-cdrp fn))		;notice n = 1
      ;; the silly iterate form is 'cause I got compiler warnings
      ;; with the more obvious (iterate for x in temp collect 1)
      (iterate with result
	       while temp
	       do (progn (setq temp (cdr temp))
			 (setq result (cons 1 result)))
	       finally (return result)))
     ((AND (MEMBER-EQ fn (QUOTE (AND OR PLUS TIMES)))
	   (< 3 (length form)))
      (if (INT= n (1- (length form)))
	  (iterate for i from 1 to (1- n) collect 2)
	  (append (iterate for i from 1 to (1- n) collect 2)
		  (list 1))))
     ((eq fn (quote list))
      (append (iterate for i from 1 to (1- n) collect 2)
	      (list 1)))
     (t (list n)))))

(defun expand-address-case (pair form)
  (let ((first (car pair)) (second (cdr pair)))
    (or (and (integerp first)
	     (> first 0)
	     (< first (length form))
	     (or (INT= second 0)
		 (INT= second 1)))
	(er soft (form (lst (list first second))) the address
	    (!ppr lst nil) is not valid for the form
	    (!ppr form (quote |.|))))
    (cond 
     ((INT= second 0)
      (cond ((INT= first (1- (length form)))
	     (er soft (form) |it| |does| |not| |make| |sense| |to| |dive|
		 |to| |the| otherwise |position| |of|
		 (!ppr form (quote |.|))))
	    (t (append (iterate for i from 3 to first collect 3)
		       (list 1 2)))))
     (t					; so second = 1
      (append (iterate for i from 3 to first collect 3)
	      (if (INT= first (1- (length form)))
		  nil
		  (list 2)))))))

(defun expand-address-cond (pair form)
  (let ((first (car pair)) (second (cdr pair)))
    (or (and (integerp first)
	     (> first 0)
	     (< first (length form))
	     (or (INT= second 0)
		 (INT= second 1)))
	(er soft (form (lst (list first second))) |the| |address|
	    (!ppr lst nil) |is| |not| |valid| |for| |the| |form|
	    (!ppr form (quote |.|))))
    (cond
     ((INT= second 0)
      (cond ((INT= first (1- (length form)))
	     (er soft (form) |it| |does| |not| |make| |sense| |to| |dive|
		 |to| |the| |"T"| |test| |in| |the| |final| |position| |of|
		 (!ppr form (quote |.|))))
	    (t (append (iterate for i from 2 to first collect 3)
		       (list 1)))))
     (t					; so second = 1
      (append (iterate for i from 2 to first collect 3)
	      (if (INT= first (1- (length form)))
		  nil
		  (list 2)))))))

(defun dv (addr form term)
  (error1-set
   (progn
     (iterate for x in addr
	      when (not (and (integerp x)
			     (>= x 0)))
	      do (er soft (x) |every| |member|  |of| |the| |address|
		     |list| |must| |be| |a| |non-negative| |integer| |,|
		     |but| (!ppr x nil) |is| |not| excl))
     (expand-address addr form term))))

(defun remove-last-zero (lst)
  (iterate for tail on lst
	   while (or (cdr tail) (not (INT= (car tail) 0)))
	   collect (car tail)))

(defun expand-address (addr form term)
  ; Assumes that TERM is the result of translating FORM.  By the way,
  ; the issue of abbreviations will be ignored, in the sense that one
  ; shouldn't be allowed to DV inside an abbreviation.
  (let (temp)
    (cond
     ((or (null addr) (equal addr '(0)))
      nil)
     ((atom form)
      (er soft nil one is not allowed to dive inside an atom excl))
     ((fquotep term)
      addr)
     ((member-eq (car form) '(case cond))
      (cond
       ((and (INT= (car addr) 1)
	     (eq (car form) 'case))
	(cond ((> (length form) 4)
	       (er warning nil |A| |dive| |into| |the| |first| |argument|
		   |of| |a| case |statement| |has| |resulted| |in| |an|
		   |arbitrary| |choice| |of| |the| |first| |copy| |of|
		   |that| |argument| |in| |the| |internal| |representation|
		   |.|)))
	(append (list 1 1)
		(expand-address
		 (cdr addr)
		 (cadr form)
		 (fargn (fargn term 1) 1))))
       ((consp (cdr addr))
	(append
	 (setq temp
	       (if (eq (car form) 'case)
		   (expand-address-case (cons (car addr) (cadr addr)) form)
		   (expand-address-cond (cons (car addr) (cadr addr)) form)))
	 (expand-address
	  (cddr addr)
	  (nth (cadr addr)
	       (nth (car addr) form))
	  (dive* temp term))))
       (t
	(er soft (form) |attempt| |to| |come| |to| |rest| |at|
	    (!ppr form nil) |,| |which| |is| |not| |the| |representation|
	    |of| |a| |term| |.|))))
     (t
      (append
       (setq temp
	     (expand-address-1
	      (car addr)
	      form))
       (expand-address
	(cdr addr)
	(nth (car addr) form)
	(dive* temp term)))))))

(defun dive* (addr term)
  (cond
   ((null addr)
    term)
   ((variablep term)
    (er soft (term) |attempt| |to| |dive| |into| |the| |variable|
	(!ppr term (quote |.|))))
   ((or (INT= (car addr) 0)
	(> (car addr) (length (sargs term))))
    (er soft (term (n (car addr))) (!ppr n nil) |is| |an| |illegal|
	|argument| |position| |for| (!ppr term (quote |.|))))
   (t
    (dive* (cdr addr) (argn term (car addr))))))

(DEFUN MAKE-LOCAL-HINTS (HINTS STATE-STACK)
  (APPEND
   (ITERATE FOR HINT IN HINTS
	    WHEN (NOT (MEMBER-EQ (CAR HINT) (QUOTE (ENABLE DISABLE ENABLE-THEORY DISABLE-THEORY))))
	    COLLECT HINT)
   (NON-TRIVIAL-PAIRS (DISABLE-AND-ENABLE-HINTS
		       HINTS
		       (ITERATE FOR STATE IN (NEW-PART-OF-STATE-STACK STATE-STACK)
				COLLECT (ACCESS STATE INSTRUCTION STATE))))))

(defun new-make-local-hints (hints state-stack)
  (APPEND
   (ITERATE FOR HINT IN HINTS
	    WHEN (NOT (MEMBER-EQ (CAR HINT) (QUOTE (ENABLE DISABLE ENABLE-THEORY DISABLE-THEORY))))
	    COLLECT HINT)
   (ITERATE FOR STATE IN (NEW-PART-OF-STATE-STACK STATE-STACK)
	    for instr =  (ACCESS STATE INSTRUCTION STATE)
	    with disable-hint and enable-hint
	    and disable-theory-hint and enable-theory-hint
	    do
	    (when (eq instr 'enable)
		  (setq instr '(enable-theory t)))
	    (when (eq instr 'disable)
		  (setq instr '(disable-theory t)))
	    (cond
	     ((not (consp instr))
	      nil)
	     ((eq (car instr) 'enable-theory)
	      (setq enable-theory-hint (union-eq (cdr instr) enable-theory-hint)))
	     ((eq (car instr) 'disable-theory)
	      (setq disable-theory-hint (union-eq (cdr instr) disable-theory-hint)))
	     ((eq (car instr) 'enable)
	      (setq enable-hint (union-eq (cdr instr) enable-hint)))
	     ((eq (car instr) 'disable)
	      (setq disable-hint (union-eq (cdr instr) disable-hint)))
	     (t nil))
	    finally
	    (progn
	      (setq enable-theory-hint
		    (append (cdr (assoc-eq 'enable-theory hints))
			    (set-diff-eq enable-theory-hint
					 (cdr (assoc-eq 'disable-theory hints)))))
	      (setq disable-theory-hint
		    (append (cdr (assoc-eq 'disable-theory hints))
			    (set-diff-eq disable-theory-hint
					 (cdr (assoc-eq 'enable-theory hints)))))
	      (setq enable-hint
		    (append (cdr (assoc-eq 'enable hints))
			    (set-diff-eq enable-hint
					 (cdr (assoc-eq 'disable hints)))))
	      (setq disable-hint
		    (append (cdr (assoc-eq 'disable hints))
			    (set-diff-eq disable-hint
					 (cdr (assoc-eq 'enable hints)))))
	      (return
	       (append
		(and enable-theory-hint
		     (list (cons 'enable-theory enable-theory-hint)))
		(and disable-theory-hint
		     (list (cons 'disable-theory disable-theory-hint)))
		(and enable-hint
		     (list (cons 'enable enable-hint)))
		(and disable-hint
		     (list (cons 'disable disable-hint)))))))))

(DEFUN NON-TRIVIAL-PAIRS (LST)
  (ITERATE FOR X IN LST
	   WHEN (CDR X)
	   COLLECT X))

(DEFUN DISABLE-AND-ENABLE-HINTS (HINTS COMMANDS)
  (IF (OR (MEMBER-EQUAL '(ENABLE-THEORY T) HINTS)
	  (MEMBER-EQUAL '(DISABLE-THEORY T) HINTS))
      HINTS
      (ITERATE FOR COMMAND IN COMMANDS
	       WITH DISABLE-RESULT = (SAFECDR (ASSOC-EQ (QUOTE DISABLE) HINTS))
	       AND ENABLE-RESULT = (SAFECDR (ASSOC-EQ (QUOTE ENABLE) HINTS))
	       AND DISABLE-THEORY-RESULT = (SAFECDR (ASSOC-EQ (QUOTE DISABLE-THEORY) HINTS))
	       AND ENABLE-THEORY-RESULT = (SAFECDR (ASSOC-EQ (QUOTE ENABLE-THEORY) HINTS))
	       AND NAMES AND NEW-THEORY-NAMES
	       AND ENABLE-THEORY-NAMES = (FORM-THEORY (SAFECDR (ASSOC-EQ (QUOTE ENABLE-THEORY) HINTS)))
	       AND DISABLE-THEORY-NAMES = (FORM-THEORY (SAFECDR (ASSOC-EQ (QUOTE DISABLE-THEORY) HINTS)))
	       DO	
	       (COND
		((EQ COMMAND 'ENABLE)
		 ;; We need to worry about names which were enabled by the existing ENABLE-THEORY hint
		 ;; constructed so far and also disabled by the existing DISABLE-THEORY hint, since
		 ;; those are supposed to wind up enabled but would wind up disabled if nothing were
		 ;; done (because of the low precedence of (ENABLE-THEORY T)) -- EXCEPT of course in
		 ;; the case that such name was explicitly enabled or disabled in the existing hints,
		 ;; in which case we're OK.
		 (RETURN
		  (LIST (CONS 'DISABLE DISABLE-RESULT)
			(CONS 'ENABLE
			      (APPEND (SET-DIFF-EQ (INTERSECTION-EQ ENABLE-THEORY-NAMES DISABLE-THEORY-NAMES)
						   (APPEND ENABLE-RESULT DISABLE-RESULT))
				      ENABLE-RESULT))
			(CONS 'DISABLE-THEORY DISABLE-THEORY-RESULT)
			'(ENABLE-THEORY T))))
		((EQ COMMAND 'DISABLE)
		 (RETURN
		  (LIST (CONS 'DISABLE DISABLE-RESULT)
			(CONS 'ENABLE ENABLE-RESULT)
			'(DISABLE-THEORY T)
			(CONS 'ENABLE-THEORY ENABLE-THEORY-RESULT))))
		((MATCH COMMAND
			(CONS (QUOTE ENABLE) NAMES))
		 (ITERATE FOR NAME IN NAMES
			  WHEN (AND (NOT (MEMBER-EQ NAME ENABLE-RESULT))
				    (NOT (MEMBER-EQ NAME DISABLE-RESULT))
				    (NOT (MEMBER-EQ NAME ENABLE-THEORY-NAMES))
				    (NOT (MEMBER-EQ NAME DISABLE-THEORY-NAMES)))
			  DO
			  (SETQ ENABLE-RESULT (CONS NAME ENABLE-RESULT))))
		((MATCH COMMAND
			(CONS (QUOTE DISABLE) NAMES))
		 (ITERATE FOR NAME IN NAMES
			  WHEN (AND (NOT (MEMBER-EQ NAME ENABLE-RESULT))
				    (NOT (MEMBER-EQ NAME DISABLE-RESULT))
				    (NOT (MEMBER-EQ NAME ENABLE-THEORY-NAMES))
				    (NOT (MEMBER-EQ NAME DISABLE-THEORY-NAMES)))
			  DO
			  (SETQ DISABLE-RESULT (CONS NAME DISABLE-RESULT))))
		((MATCH COMMAND
			(CONS (QUOTE ENABLE-THEORY) NAMES))
		 (SETQ NAMES
		       (ITERATE FOR NAME IN NAMES
				WHEN (AND (NOT (MEMBER-EQ NAME ENABLE-THEORY-RESULT))
					  (NOT (MEMBER-EQ NAME DISABLE-THEORY-RESULT)))
				COLLECT NAME))
		 (SETQ ENABLE-THEORY-RESULT
		       (APPEND NAMES ENABLE-THEORY-RESULT))
		 (SETQ NEW-THEORY-NAMES (SET-DIFF-EQ (FORM-THEORY NAMES) ENABLE-THEORY-NAMES))
		 (SETQ ENABLE-THEORY-NAMES (APPEND NEW-THEORY-NAMES ENABLE-THEORY-NAMES))
		 (SETQ DISABLE-RESULT
		       ;; We're worried about the names that are now seemingly enabled but shouldn't be,
		       ;; because they were already disabled by a DISABLE-THEORY hint that was already
		       ;; created, and there's no ENABLE or DISABLE hint that already applies (in which
		       ;; case there would be no problem, since such would take precedence over the new
		       ;; part of the ENABLE-THEORY hint anyhow).  There's no corresponding problem in
		       ;; the DISABLE-THEORY case, because that hint already has lowest priority (short
		       ;; of (ENABLE-THEORY T) and (DISABLE-THEORY T)) anyhow.
		       (APPEND (SET-DIFF-EQ
				(INTERSECTION-EQ NEW-THEORY-NAMES DISABLE-THEORY-NAMES)
				(APPEND ENABLE-RESULT DISABLE-RESULT))
			       DISABLE-RESULT)))
		((MATCH COMMAND
			(CONS (QUOTE DISABLE-THEORY) NAMES))
		 (SETQ NAMES
		       (ITERATE FOR NAME IN NAMES
				WHEN (AND (NOT (MEMBER-EQ NAME ENABLE-THEORY-RESULT))
					  (NOT (MEMBER-EQ NAME DISABLE-THEORY-RESULT)))
				COLLECT NAME))
		 (SETQ DISABLE-THEORY-RESULT
		       (APPEND NAMES DISABLE-THEORY-RESULT))
		 (SETQ NEW-THEORY-NAMES (SET-DIFF-EQ (FORM-THEORY NAMES) DISABLE-THEORY-NAMES))
		 (SETQ DISABLE-THEORY-NAMES (APPEND NEW-THEORY-NAMES DISABLE-THEORY-NAMES))))
	       FINALLY
	       (RETURN
		(LIST (CONS 'DISABLE DISABLE-RESULT)
		      (CONS 'ENABLE ENABLE-RESULT)
		      (CONS 'DISABLE-THEORY DISABLE-THEORY-RESULT)
		      (CONS 'ENABLE-THEORY ENABLE-THEORY-RESULT))))))

(defun print-instrs ()
  (ppr (cdr (nreverse (iterate for state in (new-part-of-state-stack state-stack)
			       collect (access state instruction state))))
       nil))

(defun print-remaining-instrs ()
  (ppr (nreverse (iterate for state in old-ss
			  as i from (1+ (length state-stack)) to (length old-ss)
			  collect (access state instruction state)))
       nil))

(DEFUN PLAY-INSTRS (INSTRS)
  ;; Runs the given instructions in the context of the current
  ;; state stack, printing each instruction as it is tried.
  ;; Unlike the function NONINTERACTIVELY-REWRITE-FOR-PROVE-LEMMA,
  ;; we allow any kind of instruction here and we do evaluate
  ;; arguments of instructions.
  ;;   No attempt is made here to control OLD-SS.
  (SETQ UNDONE-INSTRS INSTRS)
  (COND
   ((NULL INSTRS)
    NIL)
   ((INTERACTIVE-SINGLE-STEP (CAR INSTRS))
    (PLAY-INSTRS (CDR INSTRS)))))

(DEFUN TOP-LEVEL-PLAY-INSTRS (INSTRS &OPTIONAL PROVE-LEMMA-FLG)
  ;; Returns T if the whole PLAY succeeds, otherwise NIL.  Stops as soon as it
  ;; hits a failed instruction.  If the optional argument is set, then
  ;; the instructions are run just as they would be run for PROVE-LEMMA.
  ;;   This should only be called when INSTRS is a non-NIL list.
  ;;   No attempt is made here to control OLD-SS.
  (COND
   ((ATOM INSTRS)
    (ER SOFT (INSTRS) |The| |Lisp| |function| RUN-PLAY |expects|
	|a| |non-empty| |list| |of| |instructions| |,| |but|
	(!PPR INSTRS NIL) |is| |not| |.|))
   (T
    (BLURB NIL |Beginning| |play| |of| |instructions| |.| CR)
    (IF PROVE-LEMMA-FLG
	(NONINTERACTIVELY-REWRITE-FOR-PROVE-LEMMA INSTRS)
	(PLAY-INSTRS INSTRS)))))

(DEFUN PLAY-FN (INSTRS PROVE-LEMMA-FLG)
  (LET ((INSTRS INSTRS) (PROVE-LEMMA-FLG PROVE-LEMMA-FLG))
       (COND
	((ATOM INSTRS)
	 (BLURB (INSTRS (INTENT (IF PROVE-LEMMA-FLG 
				    `(PLAY (,INSTRS) ,PROVE-LEMMA-FLG)
				    `(PLAY (,INSTRS)))))
		|**NO| |CHANGE**| |--|
		|The| |command| PLAY |expects|
		|a| |non-empty| |list| |of| |instructions| |,| |but|
		(!PPR INSTRS NIL) |is| |not| |.| |Perhaps| |you| |meant|
		(!PPR INTENT (QUOTE |.|)) CR))
	(T
	 (LET
	  ((SAVED-SS STATE-STACK))
	  (UNWIND-PROTECT
	   (TOP-LEVEL-PLAY-INSTRS INSTRS PROVE-LEMMA-FLG)
	   (BLURB (UNDONE-INSTRS) CR (COND (UNDONE-INSTRS |Attempted|))
		  PLAY |of| |given| |instructions| |is| |complete| |.|) 
	   (COND
	    ((NULL UNDONE-INSTRS)
	     (COND
	      ('PROVE-LEMMA-FLG
	       (BLURB NIL |Every| |instruction| |caused| |the|
		      |creation| |of| |a| |new| |state| |.|))
	      (T
	       (BLURB NIL CR |Every| |instruction| |was| |a| |"success"| |.|))))
	    (T
	     (BLURB ((INSTR (CAR UNDONE-INSTRS)) (FLG 'PROVE-LEMMA-FLG))
		    |However| |,| |the| |following| |instruction|
		    (COND (FLG |failed| |to| |create| |a| |new| |state|)
			  (T |"failed"|))
		    |:| (!PPR INSTR (QUOTE |.|)) CR |#|
		    |Submit| |the| |command| |(UNDONE-INSTRUCTIONS)| |to|
		    |see| |the| |list| |of| |all| |remaining|
		    |instructions| |,| |starting| |with| |that| |one| |.|)))
	   (BLURB NIL CR |#| |Submit| RESTORE |to| |get|
		  |back| |to| |the| |state| |existing| |at| |the|
		  |time| |the| PLAY |was| |begun| |.| CR)
	   (SETQ OLD-SS SAVED-SS))
	  (NULL UNDONE-INSTRS))))))

(DEFUN REPLAY-FN ()
  (LET
   ((SAVED-SS STATE-STACK)
    (INSTRS
     (CDR
      (NREVERSE
       (ITERATE FOR S IN (NEW-PART-OF-STATE-STACK STATE-STACK)
		COLLECT (ACCESS STATE INSTRUCTION S))))))
   (COND
    ((NULL INSTRS)
     (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |no| |instructions|
	    |to| |replay| |.| CR))
    (T
     (SETQ OLD-SS STATE-STACK)
     (SETQ STATE-STACK (NTHCDR (LENGTH INSTRS) STATE-STACK))
     (UNWIND-PROTECT
      (TOP-LEVEL-PLAY-INSTRS INSTRS T)
      (BLURB (UNDONE-INSTRS) CR (COND (UNDONE-INSTRS |Attempted|))
	     REPLAY |of| |given| |instructions| |is| |complete| |.|) 
      (COND
       ((NULL UNDONE-INSTRS)
	(BLURB NIL |Every| |instruction| |caused| |the|
	       |creation| |of| |a| |new| |state| |.|))
       (T
	(BLURB ((INSTR (CAR UNDONE-INSTRS)))
	       |However| |,| |the| |following| |instruction|
	       |failed| |to| |create| |a| |new| |state|
	       |:| (!PPR INSTR (QUOTE |.|)) CR |#|
	       |Submit| |the| |command| |(UNDONE-INSTRUCTIONS)| |to|
	       |see| |the| |list| |of| |all| |remaining|
	       |instructions| |,| |starting| |with| |that| |one| |.|)))
      (BLURB NIL CR |#| |Submit| RESTORE |to| |get|
	     |back| |to| |the| |state| |existing| |at| |the|
	     |time| |the| REPLAY |was| |begun| |.| CR)
      (SETQ OLD-SS SAVED-SS))
     (NULL UNDONE-INSTRS)))))

; check below -- these are, I think, for prove++ and prove+++

(defun all-supporters-of-term (term)
  (make-set!
    (copy-list
     (iterate for fn in (all-fnnames term)
	      append (data-base 'all-supporters fn)))))

(defun all-supporters-of-term-lst (term-list)
  (make-set!
    (iterate for term in term-list
	     nconc (all-supporters-of-term term))))

(defun lemma-names-about (fn)
  (iterate for x in (get fn 'lemmas)
	   collect (access rewrite-rule name x)))

(defun lemma-names-about-list (fn-list)
  (make-set!
    (iterate for fn in fn-list
	     nconc (lemma-names-about fn))))

(defun gross-hint (term-list)
  (let ((fn-list (all-supporters-of-term-lst term-list)))
    (append fn-list
	    (lemma-names-about-list fn-list))))

(defun all-lemmas-event (ev)
  (append (get ev 'lemmas)
	  (iterate for x in (get ev 'satellites)
		   append (all-lemmas-event x))))

(defun all-lemmas ()
  (iterate for ev in chronology
	   append (all-lemmas-event ev)))

(defun find-original-goal (name)
  (iterate for state in (reverse state-stack)
	   with goal
	   when (setq goal (find-goal name (all-goals state)))
	   do (return goal)))

(defun undo-goal-number (gl-name)
  (iterate for s in state-stack
	   as i from 0
	   when
	   (iterate for goal in (all-goals s)
		    always
		    (not (equal (access goal goal-name goal)
				gl-name)))
	   do (return i)))

(defun all-keys (hash-table)
  (setq ans nil)
  (maphash #'(lambda (key val) (setq ans (cons key ans))) hash-table)
  ans)

(DEFUN SORTED-INSTRUCTIONS ()
  (SORT (ALL-KEYS INSTRUCTIONS)
	#'SYMBOL<))

(DEFUN SYMBOL< (X Y)
  (STRING< (STRING X) (STRING Y)))

(defun check-use-hints-for-uninstantiated-vars (hints)
  ;; from what was use-patch.lsp
  (iterate for hint in hints with vars
	   when (setq vars (set-diff-eq (all-vars (formula-of (car hint)))
					(mapcar #'car (cdr hint))))
	   do (er warning (vars hint) |the| (PLURAL? vars |variables| |variable|)
		  (!ppr-list vars nil) (plural? vars |have| |has|)
		  |not| |been| |instantiated| |in|
		  |the| |hint| (!ppr hint (quote |.|)))))

(defun find-bookend (bookend instrs)
  ;; returns 0-based index of matching bookend, if any
  (iterate for instr in instrs
	   as i from 0
	   with x
	   when (and (match instr (bookmark (end x)))
		     (equal x bookend))
	   do (return i)))

(defun compress-instrs (instrs &aux bookend)
  ;; Assumes that we start with the "start instruction".  Replaces every
  ;; (bookmark (begin ...)) with (*compound* ...) and replaces the instructions
  ;; between it and its "matching bookend" with nil.  If its matching bookend
  ;; doesn't exist then it's treated like an ordinary instructions, but a note
  ;; is made by pushing it onto *incomplete-bookends*.
  (cond
   ((atom instrs) nil)
   ((match (car instrs) (bookmark (begin bookend)))
    (let ((n (find-bookend bookend (cdr instrs))))
      (if n
	  (cons (car instrs)
		(nconc (iterate for i from 1 to n
				collect nil)
		       (let ((rest (subseq (cdr instrs) n)))
			 (cons (car rest)
			       (compress-instrs (cdr rest))))))
	(cons `(bookmark (begin ,bookend) ****NOT COMPLETED****)
	      (compress-instrs (cdr instrs))))))
   (t (cons (car instrs) (compress-instrs (cdr instrs))))))

(defun prettyify-instrs (instrs)
  (iterate for x in instrs
	   with a and b
	   collect
	   (if (match x (rewrite a b &))
	       `(rewrite ,a ,b <FORMULA>)
	     x)))

(defun instrs-for-comm (n)
  (nreverse
   (prettyify-instrs
    (compress-instrs
     (nreverse
      (iterate for i from 1 to n
	       as s in state-stack
	       while s
	       collect (access state instruction s)))))))

(DEFUN copyable-hyps (STATE)
  (LET-FIELDS
   (LET* ((OTHER-VARS
	   (UNION-EQ-LIST
	    (ITERATE FOR GL-INFO IN OTHER-GOALS
		     COLLECT (ALL-VARS-GOAL (CAR GL-INFO)))))
	  (COPY-HYPS
	   (ITERATE FOR HYP-FLG IN HYPS
		    AS I FROM 1
		    WHEN (EQ (CDR HYP-FLG) 'A)
		    COLLECT I)) 
	  (FREE-HYP-VARS
	   (INTERSECTION-EQ FREE-VARIABLES
			    (ALL-VARS-LST
			     (ITERATE FOR HYP-FLG IN HYPS
				      AS I FROM 1
				      WHEN (MEMBER-EQUAL I COPY-HYPS)
				      COLLECT (CAR HYP-FLG)))))
	  (STATE-VARS (UNION-EQ (ALL-VARS-GOAL GOAL) OTHER-VARS))
	  (COPY-ALIST (MAKE-COPY-ALIST
		       (SET-DIFF-EQ FREE-HYP-VARS OTHER-VARS)
		       STATE-VARS))) 
	 (ITERATE FOR HYP-FLG IN HYPS
		  WITH COPY-VARS = (MAPCAR #'CAR COPY-ALIST)
		  AS I FROM 1
		  WHEN (AND (MEMBER-EQUAL I COPY-HYPS)
			    (INTERSECTP-EQ COPY-VARS
					   (ALL-VARS (CAR HYP-FLG))))
		  COLLECT I))))

(defun active-indices ()
  (iterate for i from 1 to (length (hyps))
	   as hyp-flg in (hyps)
	   when (eq (cdr hyp-flg) 'a)
	   collect i))

(DEFUN MAKE-NEW-R$-VAR (V OLD-VARS &AUX (VAR-INDEX (PREPARE-FOR-COPY V)))
  (IF (MEMBER-EQ (CAR VAR-INDEX) OLD-VARS)
      (MAKE-NEW-VAR (CAR VAR-INDEX) (CDR VAR-INDEX) OLD-VARS)
      (CAR VAR-INDEX)))

(defun normalize-implies (x &aux hyp hyp2 conc conc2 n)
  (cond
   ((match x (implies hyp conc))
    (match (normalize-implies conc) (cons n conc))
    (cons (1+ n)
	  (if (match conc (implies hyp2 conc2))
	      (fcons-term* 'implies (fcons-term* 'and hyp hyp2) conc2)
	    (fcons-term* 'implies hyp conc))))
   (t (cons 0 x))))

(defun backchain-leaves (x)
  ;; breaks x up into the leaves of its tree when view as being built only
  ;; from backchain connectives
  (if (and (nvariablep x)
	   (not (fquotep x))
	   (eq (ffn-symb x) 'and))
      (cons (fargn x 1)
	    (append (backchain-leaves (fargn x 1))
		    (cons (fargn x 2)
			  (backchain-leaves (fargn x 2)))))
    (list x)))

(defun backchain-level-subst (source target free &aux conc)
  ;; Returns "level" consed onto corresponding unifying substitution, if such a pair
  ;; exists, and NIL otherwise.  Here source corresponds to a hypothesis.
  (cond
   ((unify source target free)
    (cons 0 unify-subst))
   ((match source (implies & conc))
    (let ((level-subst (backchain-level-subst conc target free)))
      (and level-subst
	   (cons (1+ (car level-subst)) (cdr level-subst)))))
   ((and (nvariablep source)
	 (not (fquotep source))
	 (eq (ffn-symb source) 'and))
    (let ((level-subst
	   (iterate for x in (backchain-leaves source)
		    with temp
		    when (and (setq temp (backchain-level-subst x target free))
			      ;; I don't want to nest IMPLIES inside ANDs for backchaining, yet
			      (eq (car temp) 0)) 
		    do (return temp))))
      (and level-subst
	   (cons (1+ (car level-subst)) (cdr level-subst)))))
   (t nil)))

;;;;;;;;; UNIFICATION

(defun unify (term1 term2 free)
  (setq unify-subst nil)
  (and (catch 'unify
	 (list (setq unify-subst (unify1 term1 term2 free))))
       t))

(DEFUN OCCUR-VAR (var term)
  (cond ((variablep term)
	 (eq var term))
	((fquotep term)
	 nil)
	(t (iterate for arg in (fargs term)
		    thereis (occur-var var arg)))))

(defun unify-var (var term)
  ;; assumes var is *free*
  (cond
   ((eq var term) nil)
   ((occur-var var term) (throw 'unify nil))
   (t (list (cons var term)))))

(defun unify1 (term1 term2 free)
  ;; Returns substitution or throws nil to 'unify.
  ;; The resulting substitution has a domain which is
  ;; disjoint from the variables in its range.
  (cond
   ((free-variablep term1 free)
    (unify-var term1 term2))
   ((free-variablep term2 free)
    (unify-var term2 term1))
   ((or (variablep term1)
	(variablep term2)
	(and (quotep term1)
	     (quotep term2)))
    (if (equal term1 term2)
	nil
      (throw 'unify nil)))
   ((eq (fn-symb term1) (fn-symb term2))
    (unify-list (sargs term1) (sargs term2) free))
   (t (throw 'unify nil))))

(defun unify-list (termlist1 termlist2 free &aux unifier)
  ;; assumes these have the same length
  (if (null termlist1)
      nil
    (compose-unifiers
     (setq unifier (unify1 (car termlist1) (car termlist2) free))
     (unify-list (iterate for x in (cdr termlist1)
			  collect (sublis-var unifier x))
		 (iterate for x in (cdr termlist2)
			  collect (sublis-var unifier x))
		 free))))

(defun compose-unifiers (subst1 subst2)
  (nconc (iterate for pair in subst1
		  collect (cons (car pair)
				(sublis-var subst2 (cdr pair))))
	 subst2))

;;;;; End Of Unification

;;; SK stuff:

(defun parity-error (current-addr term quiet-flg)
  (if quiet-flg
      'fail
    (er soft (current-addr term) PARITY |called| |on| |bad| |args| |,|
	(!ppr current-addr nil) |and| (!ppr term (quote |.|)))))

(defun parity (current-addr term &optional quiet-flg)
  ;; Returns T if parity is positive in the sense that
  ;; the whole conclusion is considered positive, else NIL
  ;; However, returns 'fail (or causes an error) if parity can't
  ;; be determined.  Here parity is actually defined in terms
  ;; of monotonicity of the term's value in terms of the subterm's value.
  (cond
   ((null current-addr)
    t)
   ((fquotep term)
    (parity-error current-addr term quiet-flg))
   ((and (eq (ffn-symb term) 'if)
	 (int= (car current-addr) 1))
    (cond
     ((and (equal (fargn term 2) false)
	   (equal (fargn term 3) true))
      nil)
     ((and (equal (fargn term 2) true)
	   (equal (fargn term 3) false))
      t)
     (t
      (parity-error current-addr term quiet-flg))))
   ((member-eq (ffn-symb term) '(and or if))
    (parity (cdr current-addr) (fargn term (car current-addr)) quiet-flg))
   ((eq (ffn-symb term) 'implies)
    (if (int= (car current-addr) 1)
	(not (parity (cdr current-addr) (fargn term 1) quiet-flg))
      (parity (cdr current-addr) (fargn term 2) quiet-flg)))
   ((eq (ffn-symb term) 'not)
    (not (parity (cdr current-addr) (fargn term 1) quiet-flg)))
   (t (parity-error current-addr term quiet-flg))))

(defun all-defn-sk-addresses (term)
  ;; Returns all addresses meeting the spec of defn-sk-address, where the
  ;; outer ones should come after the inner ones.
  (cond
   ((or (variablep term) (fquotep term))
    nil)
   ((eq (car (get (fn-symb term) 'event)) 'defn-sk)
    (cons nil
	  (iterate for arg in (fargs term)
		   as i from 1
		   append (iterate for x in (all-defn-sk-addresses arg)
				   collect (cons i x)))))
   (t (iterate for arg in (fargs term)
	       as i from 1
	       append (iterate for x in (all-defn-sk-addresses arg)
			       collect (cons i x))))))

(defun sk-rule-name (parity term)
  (and (nvariablep term)
       (eq (car (get (fn-symb term) 'event)) 'defn-sk)
       (pack (list (ffn-symb term)
		   (if parity '-suff '-necc)))))

(defun bad-bookend-position ()
  (let ((instrs (nreverse (iterate for s in state-stack
				   collect (access state instruction s)))))
    (iterate for tail on instrs
	     as i from 1
	     with x
	     when (and (match (car tail) (bookmark (begin x)))
		       (not (find-bookend x (cdr tail))))
	     do (return (1+ (- (length state-stack) i))))))

(defun last-bookend-start-position ()
  (iterate for s in state-stack
	   with x and y and flg
	   as i from 1
	   when (cond
		 (flg
		  (and (match (access state instruction s)
			      (bookmark (begin y)))
		       (equal x y)))
		 (t
		  (setq flg (match (access state instruction s)
				   (bookmark (end x))))
		  nil))
	   do (return i)))

(defun unexpand-if (term hyps-so-far &aux x y)
  ;; utility for macro command PROMOTE-IF
  ;; returns a pair (hyps . conc)
  (cond ((match term (if x y (quote '*1*true)))
	 (if (quotep y)
	     (if (equal y false)
		 (unexpand-if (dumb-negate-lit x) hyps-so-far)
	       (unexpand-if true hyps-so-far))
	   (unexpand-if y (cons x hyps-so-far))))
	((match term (if x (quote '*1*true) y))
	 (if (quotep y)
	     (if (equal y false)
		 (unexpand-if x hyps-so-far)
	       (unexpand-if true hyps-so-far))
	   (unexpand-if y (cons (dumb-negate-lit x) hyps-so-far))))
	(t (cons (nreverse hyps-so-far)
		 (if (match term (if x (quote '*1*true) (quote '*1*false)))
		     x
		   (if (match term (if x (quote '*1*false) (quote '*1*true)))
		       (dumb-negate-lit x)
		     term))))))

(defvar *constrain-to-do* nil)

(defun constrain-helper-name ()
  (declare (special *constrain-to-do*))
  (pack (list (car *constrain-to-do*) '- 'help)))

(defmacro verify-constrain (name types body witnesses
				 &rest hints &aux new-body)
  (declare (ignore hints)
	   (special *constrain-to-do*))
  (setq *constrain-to-do* `(,name ,types ,body ,witnesses))
  (setq new-body (sublis-fn (iterate for pair in witnesses
				     collect (cons (car pair) (cadr pair)))
			    body))
  (blurb (new-body (exit-comm `(exit ,(constrain-helper-name) nil)))
	 cr |Entering| |the| |proof-checker| |to| |prove| CR CR |#|
	 (!ppr new-body (quote |.|)) CR CR |You| |should|
	 |exit| |this| |interactive| |proof| |with| EXIT-CONSTRAIN
	 |or| |with| (!ppr exit-comm (quote |.|)) |After| |that| |you|
	 |may| |submit| |the| |original| |desired| CONSTRAIN |event|
	 |by| |submitting| |the| |command| |(FINISH-CONSTRAIN)| |to| |Lisp| |.| CR)
  `(verify ,new-body))

(defun finish-constrain-form ()
  (declare (special *constrain-to-do*))
  (let (name types body witnesses)
    (match! *constrain-to-do* (list name types body witnesses))
    `(constrain ,name ,types ,body ,witnesses
		((disable-theory t)
		 (use (,(constrain-helper-name)))))))

(defmacro finish-constrain ()
  (let ((form (finish-constrain-form)))
    (blurb (form) cr (!ppr form nil) cr)
    form))

;;;;; The following are for the SHRINK macro command

(defun all-goal-names (state)
  (iterate for goal in (all-goals state)
	   collect (access goal goal-name goal)))

(defun goal-creation-info (goal-name)
  ;; returns (cons state i) where i is the number of states from
  ;; the top (starting at 1), corresponding to the first time
  ;; goal-name appears.
  (iterate for state in (reverse state-stack)
	   as i from 1
	   when (member-equal goal-name (all-goal-names state))
	   do (return (cons state (1+ (- (length state-stack) i))))))

(defun state-instr-name (state)
  (instr-name (access state instruction state)))

(defun remove-initial-change-goals (ss)
  (iterate for tail on ss
	   until (not (eq (state-instr-name (car tail)) 'change-goal))
	   finally (return tail)))

(defun shrink-change-goal (ss &aux first-instr)
  ;; returns a new ss with the second state not being a change-goal.
  (if (consp ss)
      (if (and (eq (instr-name (setq first-instr
				     (access state instruction (car ss))))
		   'change-goal)
	       (eq (state-instr-name (cadr ss)) 'change-goal))
	  (let ((rest-ss (remove-initial-change-goals (cddr ss)))
		(current-goal-name (access goal goal-name (access state goal (car ss)))))
	    (if (equal current-goal-name
		       (access goal goal-name (access state goal (car rest-ss))))
		;; then there's no need to change goals at all!
		(cons (car rest-ss)
		      (shrink-change-goal (cdr rest-ss)))
	      (cons (if (and (not (eq first-instr 'change-goal))
			     ;; a raw CHANGE-GOAL should be converted in this case, I think
			     (equal (access goal goal-name
					    (access state goal
						    (let ((blurb-flg nil))
						      (change-state-with-change-goal
						       first-instr (car rest-ss)))))
				    current-goal-name))
			(car ss)
		      (make-new-record (quote state)
				       (list (cons (quote instruction)
						   (list 'change-goal current-goal-name)))
				       (car ss)))
		    (shrink-change-goal rest-ss))))
	(cons (car ss)
	      (shrink-change-goal (cdr ss))))
    ss))

(defun shrink-bookends (ss &aux mark1 mark2)
  (if (consp ss)
      (if (and (match (access state instruction (car ss))
		      (bookmark (end mark2)))
	       (match (access state instruction (cadr ss))
		      (bookmark (begin mark1)))
	       (equal mark1 mark2))
	  (shrink-bookends (cddr ss))
	(cons (car ss)
	      (shrink-bookends (cdr ss))))
    ss))

(defun movement-instruction-p (instr)
  (and (member-eq (instr-name instr) '(top dive up nx bk))
       (not (eq instr 'dive))))

(defun shrink-top (ss)
  (if (consp ss)
      (if (eq (access state instruction (car ss)) 'top)
	  (iterate for tail on ss
		   when (not (movement-instruction-p
			      (access state instruction (car tail))))
		   do (return
		       (if (access state current-addr-r (car tail))
			   (cons (car ss)
				 (cons (car tail)
				       (shrink-top (cdr tail))))
			 (cons (car tail)
			       (shrink-top (cdr tail))))))
	(cons (car ss)
	      (shrink-top (cdr ss))))
    ss))

(defun shrink-instrs-1 (ss)
  ;; Combines e.g. (dive 1) and (dive 2) to (dive 1 2).
  ;; Cancels all movement in front of TOP.
  ;; Removes adjacent matching begin and end bookmarks.
  ;; Cancels consecutive change-goal commands so that only
  ;; the last in the sequence counts, unless it's to the
  ;; original goal in which case it's dropped too.
  ;; Could do more, e.g. with abbreviations, but forget it.
  (setq ss (shrink-change-goal ss))
  (setq ss (shrink-bookends ss))
  (setq ss (shrink-top ss))
  (setq ss (shrink-movements ss)))

(defun shrink-instrs (ss)
  ;; assumes ss is non-nil 
  (iterate with past-ss 
	   until (= (length past-ss) (length ss))
	   do
	   (setq past-ss ss)
	   (setq ss (shrink-instrs-1 ss))
	   finally (progn (setq old-ss past-ss)
			  (return ss))))

(defun shrink-fn (&optional no-complain)
  (let* ((new-ss (shrink-instrs state-stack))
	 (loss (- (length state-stack) (length new-ss))))
    (if (int= loss 0)
	(if no-complain
	    nil
	  (blurb nil |**NO| |CHANGE**| |--| |No| |way| |was| |found| |to|
		 |decrease| |the| |length| |of| |the| |state| |stack| |.| cr))
      (progn (setq old-ss state-stack)
	     (setq state-stack new-ss)
	     (blurb (loss) |The| |state| |stack| |has| |been| |shrunk| |and|
		    |is| |now| (!ppr loss nil) |shorter| |than| |it| |was| |.|
		    |Submit| |the| |command| |RESTORE| |to| |undo| |this| |effect| |.| cr)
	     t))))

(defun shrink-movements (ss)
  ;; The optional argument only matters if the first instruction in ss
  ;; is a movement instruction.
  ;;   Returns a new ss.
  (if (consp ss)
      (if (and (movement-instruction-p (access state instruction (car ss)))
	       (movement-instruction-p (access state instruction (cadr ss))))
	  (let* ((rest
		  (iterate for tail on ss
			   until (not (movement-instruction-p
				       (access state instruction (car tail))))
			   finally (return tail)))
		 (previous-addr-r
		  (access state current-addr-r (car rest)))
		 (s (car ss))
		 (s-addr-r (access state current-addr-r s)))
	    (cond
	     ((equal s-addr-r previous-addr-r)
	      (shrink-movements rest))
	     ((null s-addr-r)
	      (cons
	       (if (eq (access state instruction s) 'top) ;an optimization to save consing
		   s
		 (make-new-record
		  (quote state)
		  (list (cons (quote instruction) 'top))
		  s))
	       (shrink-movements rest)))
	     ((equal s-addr-r (cdr previous-addr-r))
	      ;; notice that s-addr-r is non-nil hence so is previous-addr-r
	      (cons (make-new-record
		     (quote state)
		     (list (cons (quote instruction) 'up))
		     s)
		    (shrink-movements rest)))
	     ((and previous-addr-r
		   ;; and s-addr-r is already non-nil here
		   (equal (cdr s-addr-r) (cdr previous-addr-r))
		   (= (car s-addr-r) (1+ (car previous-addr-r))))
	      (cons (make-new-record
		     (quote state)
		     (list (cons (quote instruction) 'nx))
		     s)
		    (shrink-movements rest)))
	     ((and previous-addr-r
		   ;; and s-addr-r is already non-nil here
		   (equal (cdr s-addr-r) (cdr previous-addr-r))
		   (= (car s-addr-r) (1- (car previous-addr-r))))
	      (cons (make-new-record
		     (quote state)
		     (list (cons (quote instruction) 'bk))
		     s)
		    (shrink-movements rest)))
	     ((and (< (length previous-addr-r) (length s-addr-r))
		   (equal (nthcdr (- (length s-addr-r) (length previous-addr-r))
				  s-addr-r)
			  previous-addr-r))
	      (cons (make-new-record
		     (quote state)
		     (list (cons (quote instruction)
				 (cons 'dive
				       (subseq (reverse s-addr-r) (length previous-addr-r)))))
		     s)
		    (shrink-movements rest)))
	     (t (cons (make-new-record
		       (quote state)
		       (list (cons (quote instruction) (cons 'dive (reverse s-addr-r))))
		       s)
		      (if previous-addr-r
			  (cons (make-new-record
				 (quote state)
				 (list (cons (quote instruction) 'top))
				 s)
				(shrink-movements rest))
			(shrink-movements rest))))))
	(cons (car ss)
	      (shrink-movements (cdr ss))))
    ss))

;;;;; end of SHRINK support

;;;;; The following are in support of the PRUNE macro command

(defun new-bash-formulas (old-state new-state)
  ;; Returns the formulas used to create the new goals in new-state
  ;; after running a bash command on old-state, presumably in the
  ;; correct order for change-state-with-bash-cheat.
  ;;   The idea is to see how many new goals there are.  If there are
  ;; none, return nil.  Otherwise, the current goal provides the first
  ;; formula, and the first (n-1) other goals' current terms provide
  ;; the others (in that order, as seen by looking at the code for
  ;; update-other-goals, which in turn is called by
  ;; change-state-with-bash-cheat), where n is the number of new goals.
  (let ((n (- (length (access state other-goals new-state))
	      (length (access state other-goals old-state)))))
    (if (= n 0)
	nil
      (cons (conc new-state)
	    (iterate for i from 1 to (1- n)
		     as gl-info in (other-goals new-state)
		     collect (cadr gl-info))))))

(defun instr-goal-other-list (state-list)
  (iterate for tail on state-list
	   while (cdr tail)
	   for current-goal-name = (goal-name (car tail))
	   for instr = (instruction (cadr tail))
	   collect
	   (cond
	    ((eq (instr-name instr) 'bash)
	     (list instr current-goal-name
		   (cumulative-lemmas-used (cadr tail))
		   (new-bash-formulas (car tail) (cadr tail))))
	    ((eq instr 'change-goal)
	     (list `(change-goal ,(goal-name (cadr tail))) current-goal-name))
	    (t (list instr current-goal-name
		     (cumulative-lemmas-used (cadr tail)))))))

(defvar *prove1-cheat-flg* nil)

(DEFUN PROVE1-cheat (TERM return-clauses-flg)
  (declare (ignore term)
	   (special *prove1-cheat-flg*))
  (BLURB NIL CR |*****| |Now| |cheating| |(!!!)| |by| |NOT| |entering|
	 |the| |theorem| |prover| |*****| |:| CR)
  (UNWIND-PROTECT
   (LET (NEW-PROVE-TIME NEW-IO-TIME PROVE-ANS)
	(LET ((PROVE-TIME 0)
	      (IO-TIME 0))
	     (setq *prove1-cheat-flg* t)
	     (SETQ PROVE-ANS t)		;here's the cheat
	     (when return-clauses-flg
		   (setq bash-clauses nil))
	     (SETQ NEW-PROVE-TIME PROVE-TIME)
	     (SETQ NEW-IO-TIME IO-TIME))
	(SETQ PROVE-TIME (+ PROVE-TIME NEW-PROVE-TIME))
	(SETQ IO-TIME (+ IO-TIME NEW-IO-TIME))
	PROVE-ANS)
   (PROGN (ITERATE FOR X IN HINT-VARIABLE-ALIST
		   DO (SET (CADR X) (CADDDR X)))
	  (SETQ LOCAL-DISABLEDP-HASH-FLAG NIL))))

(defvar *state-commands*
  ;; put will be handled specially
  '(enable disable enable-theory disable-theory
	   bookmark comment add-abbreviation remove-abbreviations))

(defvar *bad-goal-names*)

(defun prune-fn (&aux (n (- (length old-ss) (length state-stack))))
  ;; assumes that we've just completed (UNDO n)
  (if (= n 1)
      (blurb nil cr |No| |further| |activity| |for| |pruning| |,| |since|
	     UNDO |only| |undid| |a| |single| |instruction| |.| cr)
    (if (cdr state-stack)
	(let ((old-length (length state-stack))
	      (undo-state (nth (1- n) old-ss)))
	  (replay-instr-goal-other-list
	   (instr-goal-other-list (reverse (subseq old-ss 0 (1- n))))
	   (list (if (eq (state-instr-name undo-state) 'change-goal)
		     (goal-name (nth (- n 2) old-ss))
		   (goal-name undo-state))))
	  (let ((k (- (length state-stack) old-length)))
	    (if (= k 0)
		(blurb ((comm (list 'undo n))) cr |No| |commands| |were| |executed| |after|
		       |executing| (!ppr comm (quote |.|)) cr)
	      (blurb (n k) cr
		     |Completed| |playing| (!ppr k nil)
		     (plural? k |commands| |command|) |after|
		     |undoing| (!ppr n nil) |commands| |.| cr))))
      (blurb nil cr |No| |further| |activity| |for| |pruning| |,| |since|
	     UNDO |took| |us| |back| |to| |the| |start| |.| cr)))
  (blurb nil |Give| |the| |command| RESTORE |to| |undo| |this| PRUNE |command| |.| cr))

(defun replay-instr-goal-other-list
  (instr-goal-other-list *bad-goal-names*)
  (declare (special *bad-goal-names*))
  (let ((*bad-goal-names* *bad-goal-names*)
	(instr-goal-other-list instr-goal-other-list)
	(prove1-def (symbol-function 'prove1)))
    (unwind-protect
	(progn
	  (setf (symbol-function 'prove1)
		(symbol-function 'prove1-cheat))
	  (replay-instr-goal-other-list-rec instr-goal-other-list))
      (setf (symbol-function 'prove1) prove1-def))))

(defvar *new-cumulative-lemmas-used*)
(defvar *new-formulas*)

(DEFUN CHANGE-STATE-WITH-bash-cheat (INSTR STATE)
  ;; implicit parameters also of *new-cumulative-lemmas-used* and *new-formulas*
  (declare (special *new-cumulative-lemmas-used* *new-formulas*))
  (setq instr
	(if (atom instr)
	    'bash
	  (cons 'bash (cdr instr))))
  (blurb (instr) | | (!ppr instr nil) cr) ;since return-new-state-print isn't being called
  (LET-FIELDS
   (BLURB NIL CR |***| |Now| |cheating| |(!!!)| |by| |NOT| |entering|
	  |the| |prover's| |rewriter - simplifier| |***| CR)
   (LET ((NEW-GOAL-NAMES
	  (NEW-GOAL-NAMES GOAL-NAME OTHER-GOALS
			  (LENGTH *NEW-FORMULAS*))))
	(MAKE-NEW-RECORD
	 (QUOTE STATE)
	 (LIST
	  (CONS (QUOTE INSTRUCTION) INSTR)
	  (CONS (QUOTE CURRENT-TERM) TRUE)
	  (CONS (QUOTE GOAL)
		(MAKE-NEW-RECORD
		 (QUOTE GOAL)
		 (LIST (CONS (QUOTE CONC) TRUE)
		       (CONS (QUOTE DEPENDS-ON)
			     (APPEND NEW-GOAL-NAMES DEPENDS-ON)))
		 GOAL))
	  (CONS (QUOTE OTHER-GOALS)
		(if *new-formulas*
		    (UPDATE-OTHER-GOALS
		     OTHER-GOALS
		     GOAL-NAME
		     (ITERATE FOR X IN *NEW-FORMULAS*
			      COLLECT (CONS X NIL))
		     NEW-GOAL-NAMES)
		  other-goals))
	  (cons (quote cumulative-lemmas-used)
		*new-cumulative-lemmas-used*))
	 STATE))))

(defun list-if-not-nil (x)
  (and x (list x)))

(defun return-new-state-print (instr state)
  (blurb (instr) | | (!ppr instr nil) cr)
  (return-new-state instr state))

(defun return-new-states-instr-with-goal 
  (instr-with-goal state &aux
		   instr-name
		   (instr (car instr-with-goal))
		   (goal-name (cadr instr-with-goal))) 
  ;; Returns a stack of new states to be appended to the top of state-stack.
  ;; An atomic result is interpreted as the empty stack.
  (declare (special *prove1-cheat-flg* *state-commands* *bad-goal-names*
		    *new-cumulative-lemmas-used* *new-formulas*))
  (setq instr-name (instr-name instr))
  (cond
   ((member-eq instr-name *state-commands*)
    (list-if-not-nil (return-new-state-print instr state)))
   ((eq instr-name 'put)
    (let ((new-args
	   (iterate for pair in (cdr instr)
		    when (member-eq (car pair)
				    (access state free-variables state))
		    collect pair)))
      (if new-args
	  (list-if-not-nil (return-new-state-print (cons 'put new-args) state))
	t))) 
   ((eq instr-name 'change-goal)
    ;; note that we presumably made the instruction non-atomic in the preprocessing
    (let ((new-goal-name (cadr instr)))
      (if (or (member-equal new-goal-name *bad-goal-names*)
	      (iterate for gl-info in (other-goals state)
		       never (equal new-goal-name (access goal goal-name (car gl-info)))))
	  t
	(list-if-not-nil (return-new-state-print instr state)))))
   ((member-equal goal-name *bad-goal-names*)
    t)
   ((eq instr-name 'copy)
    (setq *bad-goal-names*
	  (cons goal-name *bad-goal-names*))
    t)
   ((not (equal (goal-name state) goal-name))
    (if (iterate for gl-info in (other-goals state)
		 never (equal goal-name (access goal goal-name (car gl-info))))
	t
      (let ((new-state (return-new-state-print `(change-goal ,goal-name) state)))
	(if new-state
	    (append (return-new-states-instr-with-goal instr-with-goal new-state)
		    (list new-state))
	  (er hard ((ins `(change-goal ,goal-name))) |Problem| |with|
	      RETURN-NEW-STATES-INSTR-WITH-GOAL
	      |on| |instruction| (!ppr ins nil))))))
   ((eq instr-name 'bash)
    (let ((*new-cumulative-lemmas-used* (caddr instr-with-goal))
	  (*new-formulas* (cadddr instr-with-goal)))
      (list-if-not-nil
       ;; I have to use return-new-state so that rotating out a true goal
       ;; will take place, etc.
       (return-new-state (if (atom instr)
			     'bash-cheat
			   (cons 'bash-cheat (cdr instr)))
			 state))))
   (t
    (setq *prove1-cheat-flg* nil)
    (let ((new-state (return-new-state-print instr state)))
      (and new-state
	   (list (if *prove1-cheat-flg*
		     (make-new-record
		      (quote state)
		      (list (cons 'cumulative-lemmas-used (caddr instr-with-goal)))
		      new-state)
		   new-state)))))))

(defun replay-instr-goal-other-list-rec (instr-goal-other-list)
  ;; Runs the instructions in question, except that COPY is never run,
  ;; nor is a CHANGE-GOAL to a bad-goal-name, nor is any instruction
  ;; outside of *state-commands* which operates on a bad goal (except for
  ;; PUT, which gets special handling to make it legal).
  ;;   Patterned after NONINTERACTIVELY-REWRITE-FOR-PROVE-LEMMA, though
  ;; we ignore UNDONE-INSTRS here.
  ;;   Notice that *bad-goal-names* is a special variable here, which may
  ;; be updated by return-new-states-instr-with-goal.

  (let (new-states)
       (cond
	((null instr-goal-other-list)
	 nil)
	(t 
	 (setq new-states nil)
	 (cond
	  ((consp (setq new-states
			(return-new-states-instr-with-goal
			 (car instr-goal-other-list) (car state-stack))))
	   (setq state-stack (append new-states state-stack))))
	 (cond 
	  (new-states
	   (replay-instr-goal-other-list-rec (cdr instr-goal-other-list)))
	  (t
	   (er hard ((instr (caar instr-goal-other-list))) cr
	       |Failed| |to| |create| |a| |new| |state| |with|
	       |the| |command| (!ppr instr (quote |.|)) |please| |report|
	       |this| |problem| |to| |Matt| |Kaufmann| |,| |kaufmann@cli.com| |.| cr)))))))

;; for forward$
(defun find-forward-chain-hyp$ ()
  (let ((rev-hyps (reverse (hyps))))
    (iterate for hyp-flg in rev-hyps
	     with x and free = (free-variables)
	     as i from 1
	     when (and (eq (cdr hyp-flg) 'a)
		       (match (car hyp-flg) (implies x &))
		       (iterate for y in rev-hyps
				thereis
				(and (eq (cdr y) 'a)
				     (unify x (car y) free))))
	     do (return (cons i unify-subst)))))

;; for generalize-skolems
(defun skolem-terms-list (termlist free-vars)
  (scrunch (iterate for term in termlist
		    append (skolem-terms term free-vars))))

(defun skolem-terms (term free-vars &aux event-name)
  (cond
   ((variablep term) nil)
   ((fquotep term) nil)
   ((and (setq event-name (get (ffn-symb term) 'main-event))
	 (eq (car (get event-name 'event)) 'defn-sk)
	 (not (intersectp (all-vars term) free-vars)))
    (list term))
   (t
    (skolem-terms-list (fargs term) free-vars))))
