;;; -*-  Mode: Lisp; Package: USER; Syntax: Common-Lisp; Base: 10 -*- ;;;;

;  Copyright (C) 1994 by Matt Kaufmann and Computational Logic, Inc.
;  All Rights Reserved.

;  Copying of this file is authorized to those who have read and
;  agreed with the terms in the "Pc-Nqthm-1992 GENERAL PUBLIC SOFTWARE
;  LICENSE" at the beginning of the Pc-Nqthm file "basis-pc.lisp".

;  Pc-Nqthm Version 1992 (for use with Nqthm-1992)

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

; The evaluation of instructions proceeds as follows.  If the instruction
; is a positive integer N then the instruction (DIVE N) is returned.
; Otherwise, the instruction is first modified by replacing all top-level
; arguments of the form  (! arg)  with the result of evaluating
; arg  in the environment currently provided by
; COMMAND-BINDINGS (on top of the global environment).  The resulting
; instruction is returned, unless the command is a macro command, in
; which case the formals are replaced by these values in the body of
; the macro command and the process is (continually) repeated.  Note
; however that if the "instruction" to be returned is illegal then NIL
; is returned (with an appropriate message printed).

; On the legality of instructions:  An instruction may be clearly legal,
; iffy, or clearly illegal.  An "iffy" instruction is one which is
; a list consisting only of an atom, where that atom is a legal instruction.
; In fact, from now on I will allow the submission of "iffy" instructions,
; with the understanding that they should be converted to atoms before
; being submitted to any functions corresponding to individual commands
; (i.e. the CHANGE-STATE-WITH, PRINT-HELP-WITH, and CHANGE-STATE-STACK-WITH
; functions).

; I'll organize this file roughly as follows.  Part 1:  Commands of general interest.
; Part 2:  Commands of some interest.  Part 3:  Commands of less general interest.

; PART 1:  Commands of general interest.

(definstr andsplit ()
  (wrap andsplit
	(bind ((current-address (reverse (current-addr-r))))
	      (if (match (current-term) (and & &))
		  (do-all
		    (succeed (repeat andsplit1))
		    top
		    (if current-address
			(! (cons 'dive current-address))
			skip)
		    s-prop)
		  (prog2 (lisp (blurb nil cr not an and expression excl cr))
			 fail))))
  (ANDSPLIT |:| |if| |the| |current| |subterm| |is| |of| |the| |form|
	    |(AND x y ...)| |,| |then| |a| |subgoal| |is|
	    |created| |for| |each| |conjunct| |,| |leaving| |one| |with| |T| |(true)|
	    |at| |the| |end| |.|))

(DEFINSTR ANDSPLIT+ NIL
  (WRAP ANDSPLIT+
	(BIND ((CURRENT-ADDRESS (REVERSE (CURRENT-ADDR-R))))
	      (IF (MATCH (CURRENT-TERM) (AND & &))
		  (DO-ALL (SUCCEED (REPEAT ANDSPLIT1+))
			  TOP
			  (IF CURRENT-ADDRESS
			      (! (CONS 'DIVE CURRENT-ADDRESS))
			      SKIP)
			  S-PROP)
		  (PROG2 (LISP (BLURB NIL CR NOT AN AND
				      EXPRESSION EXCL CR))
			 FAIL))))
  (|As| |with| |the| ANDSPLIT |command| |,| |if| |the| |current| |subterm| |is|
   |of| |the| |form| |(AND x y ...)| |,| |then| |a| |subgoal|
   |is| |created| |for| |each| |conjunct| |,| |and| |the| |current| |goal|
   |becomes| |proved| |.| |However| |,| |unlike| ANDSPLIT |,| |hypotheses|
   |are| |copied| |before| |each| |subgoal| |is| |created| |,| |whenever| |the|
   COPY |command| |can| |succeed| |.| |This| |way| |there| |is| |less| |chance|
   |that| |one| |will| |find| |,| |later| |in| |a| |proof| |,| |that| |a| |variable| |to|
   |be| |instantiated| |occurs| |in| |a| |goal| |other| |than| |the|
   |current| |one| |.|))

(definstr axiom (name types)
  (print
   (list 'add-axiom
	 'name
	 'types
	 (untranslate
	  (make-implication
	   (assumptions (hyps))
	   (conc)))))
  (|(AXIOM NAME TYPES)| |:| |prints| |an| ADD-AXIOM |form| |corresponding| |to| |the|
   |current| |goal| |,| |of| |name| NAME |and| |types| TYPES |(e.g.| |(REWRITE))| |.|))

(define-macro-command backchain (&optional n)
  (cond
   ((current-addr-r)
    (blurb nil CR |Please| |only| |use| |backchain| |from| |the| |top|
	   |of| |the| |conclusion| EXCL CR)
    'fail)
   (t
    (let* ((current-term (current-term)) (hyps (hyps)) (free-variables (free-variables))
	   temp
	   (index-level-subst
	    (if n
		(if (and (integerp n) (< 0 n) (<= n (length hyps))
			 (eq (cdr (nth (1- n) hyps)) 'a))
		    (or (and (setq temp
				   (backchain-level-subst
				    (car (nth (1- n) hyps)) current-term free-variables))
			     (cons n temp))
			(blurb (n) |***NO| |CHANGE***| |--| |Hypothesis| |number|
			       (!ppr n nil) |is| |not| |usable|
			       |for| |backchaining| |.| CR))
		  (blurb (n) |***NO| |CHANGE***| |--| |The| |argument| |to|
			 BACKCHAIN |must| |be| |the| |index| |of| |an| |active| |top-level|
			 |hypothesis| |,| |but| (!PPR n nil) |is| |not| |.| CR))
	      (or (iterate for hyp-flg in (reverse hyps) ;;; was hyps
			   as i downfrom (length hyps)
			   when (and (eq (cdr hyp-flg) 'a)
				     (setq temp
					   (backchain-level-subst
					    (car hyp-flg) current-term free-variables)))
			   do (return (cons i temp)))
		  (blurb nil |***NO| |CHANGE***| |--| |No| |hypothesis| |is| |available|
			 |for| |backchaining| |.| CR)))))
      (if index-level-subst
	  `(wrap (backchain ,(car index-level-subst)
			    ,(cadr index-level-subst))
		 (prog2 (if (cddr ',index-level-subst)
			    (put ,@(iterate for pair in (cddr index-level-subst)
					    collect (list (car pair) (cdr pair))))
			  skip) 
			(backchain-no-unify ,(car index-level-subst)
					    ,(cadr index-level-subst))))
	'fail))))
  (|(BACKCHAIN n)| |:| |The| |idea| |is| |that| |if| |hypothesis| |number|
   |n| |is| |active| |and| |of| |the|
   |form| |(IMPLIES P Q)| |,| |where| |Q| |is| |the| |current| |conclusion| |,| |this|
   |results| |in| |proving| |the| |current| |goal| |but| |leaving| |one| |at| |a| |subgoal|
   |to| |prove| |P| |under| |the| |current| |hypotheses| |.| |However| |,|
   BACKCHAIN |is| |more| |powerful| |than| |that| |,| |in| |four| |ways| |.|
   CR |#| |First| |of| |all| |,| |it| |suffices| |that| Q |unify| |with| |the|
   |current| |conclusion| |,| |i.e.| |,| |it| |suffices| |that| |there| |be| |a| PUT |command|
   |that| |will| |make| Q |and| |the| |current| |conclusion| |identical| |.|
   CR |#| |Second| |,| |the| |indicated| |hypothesis| |can| |be| |of| |the| |form|
   |(IMPLIES P1 (IMPLIES P2 (... (IMPLIES Pk Q))))| |,| |for| |any| |non-negative| |k| |.|
   |In| |particular| |,| |it| |can| |simply| |be| Q |.|
   CR |#| |Third| |,| |the| |conditions| |above| |hold| |as| |well| |if| |Q| |is| |a| |conjunction|
   |of| |which| |the| |current| |conclusion| |is| |one| |of| |the| |conjuncts| |(in|
   |the| |general| |sense| |that| |if| |Q| |is| |(AND Q1 Q2)| |,| |then| |everything|
   |above| |holds| |if| |it| |holds| |for| |either| |Q1| |or| |Q2| |in| |place| |of| |Q)| |.|
   CR |#| |Finally| |,| |n| |need| |not| |be| |supplied| |,| |i.e.| |it| |is| |permissible|
   |to| |submit| |the| |command| BACKCHAIN |with| |no| |arguments| |.| |In| |that| |case| |,|
   |an| |attempt| |will| |be| |made| |to| |find| |a| |suitable| |n| |,| |by| |looking|
   |through| |the| |hypotheses| |in| |reverse| |order| |until| |a| |suitable| |one| |is| |found| |.|
   CR |#| NOTE |:| |one| |must| |be| |at|
   |the| |top| |of| |the| |current| |goal| |to| |run| |this| |command| |.| |Also|
   |note| |that| |"bookmarks"| |will| |be| |placed| |around| |the| |instructions|
   |generated| |:| |the| |first| |command| |generated| |will| |be|
   |(BOOKMARK (BEGIN (BACKCHAIN n k)))| |,| |where| |n| |is| |the| |index| |of| |the|
   |hypothesis| |used| |and| |k| |is| |as| |shown| |above| |,| |and| |the| |final|
   |command| |will| |be| |(BOOKMARK (END (BACKCHAIN n k)))| |.|))

(DEFINE-MACRO-COMMAND BASH+ (&REST HINTS)
  (PROGN (CHK-ACCEPTABLE-HINTS-ABB HINTS (ABBREVIATIONS))
	 (LET ((HNTS (MAKE-LOCAL-HINTS HINTS STATE-STACK)))
	      (IF HNTS `(BASH ,@HNTS) 'BASH)))
  (|Call| |the| |theorem| |prover| |in| |the| |sense| |of| |the| BASH
   |command| |,| |but| |with| |hints|
   |that| |match| |the| |local| |disable-enable| |environment| |in|
   |the| |current| |interactive| |session| |,| |as| |updated|
   |(optionally)| |by| |the| |HINTS| |argument| |.| |Notice|
   |that| |this| |command| |has| |the| |same| |syntax|
   |as| BASH |,| |e.g.| |(BASH (DISABLE PLUS TIMES))| |.| |This| |command|
   |"succeeds"| |if| |and| |only| |if| |the| |corresponding| BASH
   |command| |"succeeds"| |,| |i.e.| |the| |proof| |completes| |.|))

(definstr casesplit (exp)
  (claim exp 0)
  (|Add| |exp| |as| |a| |hypothesis| |,| |and| |create|
   |a| |new| |subgoal| |with| |the| |negation| |of| |exp|
   |added| |as| |a| |hypothesis| |.| |(Same| |as| |(CLAIM exp 0).)|))

(define-macro-command cg (&optional goal-name) 
  (if goal-name
      `(change-goal ,goal-name t)
    (if (other-goals) 
	(let ((x (find-unproved-goal-and-info (other-goals))))
	  (if x
	      `(change-goal ,(access goal goal-name (car x)) t) 
	    (progn (BLURB NIL |**NO| |CHANGE**| |--| |All| |other| |goals| |have|
			  |true| |conclusions| |.| CR)
		   'fail)))
      (progn (BLURB NIL |**NO| |CHANGE**| |--| |there| |are| |no| |other| |goals| |.| CR)
	     'fail)))
  (|Same| |as| CHANGE-GOAL |,| |except| |that| |instead| |of| |an| |optional|
   |second| |argument| |,| |the| |current| |goal| |will| ALWAYS |be| |rotated|
   |to| |the| |bottom| |of| |the| |goal| |stack| |when| |this| |command|
   |succeeds| |.|))

(define-macro-command claim0 (exp)
  (let ((new-hyps-length (1+ (length (hyps)))))
    `(wrap claim0
	   (do-all
	    (claim ,exp 0)
	    change-goal
	    top?
	    (contradict ,new-hyps-length)
	    (hide-hyps ,new-hyps-length)
	    change-goal)))
  (|(CLAIM0 <exp>)| |is| |similar| |to| |(CLAIM <exp> 0)| |,| |in| |that| |it| |adds|
   |<exp>| |to| |the| |current| |active| |hypotheses| |.| |However| |,| |the| |new|
   |subgoal| |has| |<exp>| |as| |its| |conclusion| |and| |the| |currently| |active|
   |hypotheses| |as| |its| |active| |hypotheses| |.|))

(DEFINSTR CLAIM+ (EXP &OPTIONAL HINTS)
  (IF (AND (QUOTE HINTS)
	   (ATOM (QUOTE HINTS)))
      (CLAIM EXP HINTS)
      (DO-ALL
       (LISP (PROGN (CHK-ACCEPTABLE-HINTS-ABB (QUOTE HINTS) (ABBREVIATIONS)) T))
       (BIND ((HNTS (MAKE-LOCAL-HINTS (QUOTE HINTS) STATE-STACK)))
	     (IF HNTS
		 (CLAIM EXP (! HNTS))
		 (CLAIM EXP)))))
  (|Same| |as| |using| |the| |command| CLAIM |with| |one| |or| |two| |arguments|
   |(where| |the| |second| |argument| |is| |not| |an| |atom)|
   |,| |except| |that| |the| |if| |the| |theorem| |prover| |is| |called| |,|
   |then| |it| |is| |called| |with|
   |hints| |that| |match| |the| |local| |disable-enable| |environment|
   |in| |the| |current| |interactive| |session| |,| |as| |updated|
   |(optionally)| |by| |the| |HINTS| |argument| |.|))

(DEFINE-MACRO-COMMAND COMM (&OPTIONAL N)
  `(LISP
    (LET ((N (QUOTE ,N)) OVERFLOW-FLG LNG)
	 (COND
	  ((NULL N)
	   (SETQ N (LENGTH STATE-STACK)))
	  ((AND (INTEGERP N)
		(> N 0))
	   (SETQ N
		 (IF (SETQ OVERFLOW-FLG (> N (SETQ LNG (LENGTH STATE-STACK))))
		     LNG
		     N)))
	  (T
	   (BLURB NIL |The| |argument| |to| COMM |must| |be|
		  |a| |positive| |integer| |.| CR)))
	 (COND
	  (N
	   (COND
	    ((NULL (QUOTE ,N))
	     (BLURB NIL |The| |top-level| |commands| |thus| |far| |(in| |reverse|
		    |order| |,| |i.e.| |last| |one| |first)| |have| |been| |:| CR))
	    (T
	     (BLURB (N)
		    (COND
		     ((= N 1)
		      |The| |previous| |command| |was|)
		     (T
		      |In| |reverse| |order|
		      |(i.e.| |last| |one| |first)| |:|))
		    |:| CR)))
	   (ITERATE FOR instr IN (instrs-for-comm n) AS I FROM 1
		    UNTIL (> I N) 
		    DO
		    (when instr
		      (princ i)
		      (princ '|. |)
		      (princ instr)
		      (terpri nil))
		    FINALLY
		    (when OVERFLOW-FLG
			  (BLURB NIL |There| |are| |no| |more| |commands| |.| CR)))))))
  (|Use| COMM |in| |place| |of| COMMANDS |when| |you| |don't| |want| |to| |see|
   |the| |constituent| |commands| |inside| |of| |compound| |macro| |commands| |,|
   |i.e.| |commands| |that| |put| |"bookmarks"| |on| |the| |end| |.|))

(define-macro-command define-macro-command (&rest args)
  `(lisp (define-macro-command ,@args))
  (|Same| |as| |calling| DEFINE-MACRO-COMMAND |from| |Lisp;| |see|
   |the| |manual| |.|))

(definstr definstr (name args body &optional documentation)
  (lisp (definstr name args body documentation))
  (|Same| |as| |calling| DEFINSTR |from| |Lisp;| |see|
   |the| |manual| |.|))

(definstr defn (&optional flg)
  (orelse
   (then split
	 (if (quote flg) bash bash-protect))
   (lisp (blurb ((names (iterate for goal in (all-goals (car state-stack))
				 when (not (equal (access goal conc goal) true))
				 collect (access goal goal-name goal))))
		CR |Goals| |remaining| |to| |prove| |:|
		(!PPR-LIST NAMES (QUOTE |.|)) CR)))
  (DEFN |is| |good| |to| |use| |just| |after| |entering| |an| |interactive|
    |termination| |proof| |with| |VERIFY-DEFN| |.| |The| |effect| |is| |to|
    |try| |to| |prove| |each| |of| |the| |generated| |goals| |(without| |induction)|
    |so| |that| |when| |this| |command| |completes| |,| |one| |is| |left| |with|
    |each| |of| |the| |UNSIMPLIFIED| |goals| |that| |were| |not| |successfully|
    |proved| |.| |However| |,| |if| |one| |submits| |(DEFN T)| |then| |one| |is| |left|
    |with| |the| SIMPLIFIED |goals| |.|))

(define-macro-command dv (&rest addr)
  (cond
   ((null addr)
    (blurb nil |**NO| |CHANGE**| |--| DV |takes| |one| |or| |more| |arguments| |.| CR)
    'fail)
   (t
    (let ((address-set
	   (dv addr (untranslate-abb (current-term)
				     (abbreviations))
	       (current-term))))
      (if address-set
	  `(dive ,@(car address-set))
	'fail))))
  (|Same| |as| DIVE |,| |except| |that| DV |is| |preferred| |when|
   |the| |address| |list| |refers| |to| |the| |"nice"| |print|
   |representation| |of| |the| |term| |,| |i.e.| |that| |given|
   |by| |the| |command| P |(rather| |than| PP |,| |say)| |.|
   CR |#| |For| |example| |,| |to| |dive| |to| |the| |second|
   |argument| |of| |(LIST X Y Z)| |you| |can| |either| |(DV 2)|
   |or| |(DIVE 2 1)| |,| |since| |(LIST X Y Z)| |abbreviates|
   |(CONS X (CONS Y Z))| |.| CR |#| NOTE (minor) |:| |This|
   |command| |does| |not| |dive| |correctly| |into| |terms| |with|
   |function| |symbol| FOR |.|))

(DEFINSTR ELIM (VAR TERM DESTRUCTORS &OPTIONAL COMPLETION)
  (WRAP ELIM (ELIM0 VAR TERM DESTRUCTORS COMPLETION))
  (|Replace| VAR |by| TERM |and| |then| |eliminate| |the| DESTRUCTORS
	     |by| |generalization| |.| |If| |the| |equality| |of| VAR |with| TERM
	     |creates| |a| |subgoal| |,| |then| |this| |subgoal|
	     |will| |try| |to| |be| |established| |by| |the| |fast| |rewriter| |(i.e.|
	     |with| |the| S |command)| |or| |else| |a| |new| |subgoal| |will|
	     |remain| |to| |be| |proved| |.| |If| |the| |optional| |argument|
	     COMPLETION |is| |supplied| |,| |then| |that| |command| |will| |be|
	     |run| |on| |the| |subgoal| |rather| |than| ESTABLISH |(which|
	     |causes| |the| |above| |simplification)| |.|
	     |This| |command| |"succeeds"| |as| |long| |as| |the| |generalization|
	     |"succeeds"| |.| |Use| ELIM0 |instead| |if| |you| |don't|
	     |want| |bookmarks| |.| CR |#| |Sample| |use| |:|
	     |(ELIM X (ADD1 (SUB1 X)) ((SUB1 X)))| |replaces| X |by| (ADD1 Z) |for|
	     |some| |variable| Z |.| CR |#| NOTE |:| |The| |elements| |of| |the|
	     |list| DESTRUCTORS |may| |actually| |be| |either| |terms| |or|
	     |term-variable| |doublets| |.| |Example| |:|
	     |(ELIM X (ADD1 (SUB1 X)) (((SUB1 X) A))).| CR |#|
	     |Obscure| |point| |:| |the| |destructor| |terms| |should| |not| |be|
	     |variables| |or| |explicit| |values| |,| |or| |else| |we| |make|
	     |no| |guarantees| |.|))

(DEFINSTR =+ (OLD NEW &OPTIONAL HINTS)
  (DO-ALL
   (LISP (PROGN (CHK-ACCEPTABLE-HINTS-ABB (QUOTE HINTS) (ABBREVIATIONS)) T))
   (BIND ((HNTS (MAKE-LOCAL-HINTS (QUOTE HINTS) STATE-STACK)))
	 (IF HNTS
	     (= OLD NEW (! HNTS))
	     (= OLD NEW))))
  (|Same| |as| |using| |the| |command| | = | |with| |two| |or| |three| |arguments|
   |(where| |the| |third| |argument| |is| |not| |an| |atom)|
   |,| |except| |that| |the| |theorem| |prover| |is| |called| |with|
   |hints| |that| |match| |the| |local| |disable-enable| |environment|
   |in| |the| |current| |interactive| |session| |,| |as| |updated|
   |(optionally)| |by| |the| |HINTS| |argument| |.|))

(definstr EQSUB (N &OPTIONAL DROP-HYP-FLG)
  (wrap (eqsub n)
	(BIND ((INS (MAKE-EQUALITY-INSTR
		       'N
		       (HYPS)
		       NIL)))
	      (IF INS
		  (bind
		   ((small-hidden (iterate for i from 1 to (1- n)
					   as hyp-flg in (hyps)
					   when (eq (cdr hyp-flg) 'h)
					   collect i))
		    (large-hidden (iterate for i from (1+ n) to (length (hyps))
					   as hyp-flg in (nthcdr n (hyps))
					   when (eq (cdr hyp-flg) 'h)
					   collect (if (not 'drop-hyp-flg)
						       i
						     (1- i))))
		    (demote1 (iterate for i from 1 to (1- n) collect i))
		    (demote2 (iterate for i from (1+ n) to (length (hyps))
				      collect (1+ (- i n)))))
		   (do-strict
		    (top?
		     (if (or small-hidden large-hidden) show-hyps skip)
		     (demote? (! demote1))
		     (demote? (! demote2))
		     (do-all (! INS))
		     (if demote2 (promote (! (length demote2))) skip)
		     ;; now put the eqsub hypothesis back in the proper position
		     (if (not 'drop-hyp-flg) (claim (! (caar (hyps))) taut) skip)
		     (drop 1)
		     (if demote1 (promote (! (length demote1))) skip)
		     (do-all (! (if small-hidden (cons 'hide-hyps small-hidden) 'skip)))
		     (do-all (! (if large-hidden (cons 'hide-hyps large-hidden) 'skip))))))
		  FAIL)))
  (|(EQSUB N)| |:| |If| |the| |Nth| |hypothesis| |is| |active| |and| |of| |the|
   |form| |(EQUAL t1 t2)| |,| |the| |command| |(= t1 t2 0)| |is| |generated| |,|
   |along| |with| |appropriate| |promoting| |and| |demoting| |,| |to|
   |replace| |all| |occurrences| |of| |t1| |by| |t2| |in| |the| |hypotheses| |and|
   |the| |entire| |conclusion| |(not| |just| |the| |current| |term| |--|
   |use| |(HELP =)| |for| |details)| |.| |The| |equality| |is|
   |then| |dropped| |if| DROP-HYPS-FLG |is| |non-NIL| |.| |See| |also| EQSUB-R |.|))

(definstr EQSUB-r (N &OPTIONAL DROP-HYP-FLG)
  (wrap (eqsub-r n)
	(BIND ((INS (MAKE-EQUALITY-INSTR
		       'N
		       (HYPS)
		       NIL)))
	      (IF INS
		  (bind
		   ((small-hidden (iterate for i from 1 to (1- n)
					   as hyp-flg in (hyps)
					   when (eq (cdr hyp-flg) 'h)
					   collect i))
		    (large-hidden (iterate for i from (1+ n) to (length (hyps))
					   as hyp-flg in (nthcdr n (hyps))
					   when (eq (cdr hyp-flg) 'h)
					   collect (if (not 'drop-hyp-flg)
						       i
						     (1- i))))
		    (demote1 (iterate for i from 1 to (1- n) collect i))
		    (demote2 (iterate for i from (1+ n) to (length (hyps))
				      collect (1+ (- i n)))))
		   (do-strict
		    (top?
		     (if (or small-hidden large-hidden) show-hyps skip)
		     (demote? (! demote1))
		     (demote? (! demote2))
		     (do-all (! `(= ,(CADDR INS) ,(CADR INS) 0)))
		     (if demote2 (promote (! (length demote2))) skip)
		     ;; now put the eqsub-r hypothesis back in the proper position
		     (if (not 'drop-hyp-flg) (claim (! (caar (hyps))) taut) skip)
		     (drop 1)
		     (if demote1 (promote (! (length demote1))) skip)
		     (do-all (! (if small-hidden (cons 'hide-hyps small-hidden) 'skip)))
		     (do-all (! (if large-hidden (cons 'hide-hyps large-hidden) 'skip))))))
		  FAIL)))
  (|Same| |as| EQSUB |except| |that| |one| |does| |a| |right-to-left| |substitution| |.|
   |That| |is| |,| |for| |the| |command| |(EQSUB-R N)| |,|
   |if| |the| |Nth| |hypothesis| |is| |active| |and| |of| |the|
   |form| |(EQUAL t1 t2)| |,| |the| |command| |(= t2 t1 0)| |is| |generated| |,|
   |along| |with| |appropriate| |promoting| |and| |demoting| |,| |to|
   |replace| |all| |occurrences| |of| |t2| |by| |t1| |in| |the| |hypotheses| |and|
   |the| |entire| |conclusion| |(not| |just| |the| |current| |term| |--|
   |use| |(HELP =)| |for| |details)| |.| |The| |equality| |is|
   |then| |dropped| |if| DROP-HYPS-FLG |is| |non-NIL| |.|))

(define-macro-command ex (&rest args)
  (progn (shrink-fn t)
	 `(exit ,@args))
  (|(EX ...)| |:| |Same| |as| |(EXIT ....)| |except| |that| SHRINK
   |is| |run| |first| |to| |clean| |up| |the| |state| |stack| |.|))

(definstr exit-constrain ()
  (exit (! (constrain-helper-name)) nil)
  (|This| |command| |is| |part| |of| |a| |mechanism| |for| |interactive| |proofs|
   |for| CONSTRAIN |events| |.| |The| |whole| |thing| |works| |as| |follows| |.|
   |Suppose| |you| |have| |a| CONSTRAIN |event| |that| |you| |wish| |to| |submit|
   |interactively| |,| |say| CR |#| (CONSTRAIN |<name>| |<types>| |<body>| |<witness-alist>|) |.| CR
   |If| |you| |submit| |the| |form| CR |#| (VERIFY-CONSTRAIN |<name>| |<types>| |<body>|) CR
   |to| |Lisp| |,| |you| |will| |find| |yourself| |in| |the| |proof-checker| |with|
   |the| |goal| |of| |proving| |the| |result| |of| |functionally| |substituting|
   |<witness-alist>| |into| |<body>| |.| |When| |you| |have| |completed| |this|
   |proof| |(even| |if| |there| |are| |intervening| |interactive| |proofs| |performed|
   |by| |employing| |the| SAVE |and| RETRIEVE |mechanism)| |,| |give| |the|
   |command| EXIT-CONSTRAIN |and| |you| |will| |exit| |with| |the| |event| |named|
   |<name>-help| |(which| |will| NOT |be| |a| |rewrite| |rule)| |.| |Then| |if|
   |you| |submit| FINISH-CONSTRAIN |to| |Lisp| |,| |the| |appropriate| |CONSTRAIN|
   |event| |will| |be| |printed| |out| |and| |executed| |.| CR CR EXAMPLE |:| |Consider|
   |the| |form| CR CR |#| (verify-constrain fn-comm (rewrite) (equal (fn x y) (fn y x)) ((fn plus)))
   CR CR
   |After| |submitting| |the| |PROVE| |command| |in| |the| |proof-checker| |you| |may|
   |exit| |by| |giving| |the| |command| EXIT-CONSTRAIN |,| |which| |(if| |you| |type|
   Y |to| |the| |prompt)| |will| |create| |the| |event| CR
   (PROVE-LEMMA FN-COMM-HELP NIL (EQUAL (PLUS X Y) (PLUS Y X)) ((INSTRUCTIONS PROVE))) CR
   |Then| |if| |you| |submit| |(FINISH-CONSTRAIN)| |to| |Lisp| |you'll| |get| |the| |event| CR
   (CONSTRAIN FN-COMM (REWRITE)
	      (EQUAL (FN X Y) (FN Y X))
	      ((FN PLUS))
	      ((DISABLE-THEORY T) (USE (FN-COMM-HELP))))))

(DEFINSTR FLATTEN-HYPS ()
  (DO-PROTECT (DEMOTE PROMOTE))
  (|Flattens| |the| |and-structure| |of| |the| |hypotheses| |,| |so| |that|
	      |any| |hypothesis| |of| |the| |form| |(AND A B ...)| |is|
	      |replaced| |by| |separate| |hypotheses| A |,| B |,| |...| |,| |where|
	      |each| |of| |these| |is| |in| |turn| |then| |flattened| |.|))

(define-macro-command forward$ ()
  (let ((n-subst (find-forward-chain-hyp$)))
    (if n-subst
	(if (cdr n-subst)
	    `(do-all (put ,@(iterate for pair in (cdr n-subst)
				     collect (list (car pair)
						   (untranslate-abb (cdr pair) (abbreviations)))))
		     (forward ,(car n-subst)))
	  `(forward ,(car n-subst)))
      'fail))
  (|Like| |forward| |,| |except| |no| |argument| |is| |required| |and| |instead| |,| |the|
   |system| |tries| |to| |find| |(starting| |with| |the| |last| |hypothesis)| |an| |active|
   |hypothesis| |of| |the| |form| |(IMPLIES p q)| |such| |that| |p| |unifies| |with| |some|
   |active| |hypothesis| |.| |If| |the| |system| |finds| |such| |hypotheses| |then| |it|
   |calls| |the| |FORWARD| |command| |appropriately| |after| |doing| |any| |necessary|
   |unification| |.|))

(definstr forward* (hypn &optional r)
  (succeed
   (qq (wrap (! (if 'r '(forward* hypn r) '(forward* hypn)))
	     (progress (forward*-aux hypn r)))))
  (|Essentially| R |repetitions| |of| FORWARD |if| R |is| |supplied| |,|
   |else| |repeat| FORWARD |until| |failure| |.|))

(definstr forward*-aux (hypn r)
  ;; r repetitions of forward
  (if (and r (not (and (integerp r) (>= r 0))))
      (lisp (BLURB NIL CR |***NO| |CHANGE***| |--| |the| |second| |argument| |of|
		   FORWARD* |must| |be| |a| |nonnegative| |integer| |.| CR))
    (if (equal r 0)
	skip
      (prog2 (forward hypn)
	     (forward*-aux (! (length (hyps))) (! (and r (1- r))))))))

(definstr forward (n)
  (wrap (forward n)
	(bind ((hyp-flg (nth (1- n) (hyps))))
	      (if (null hyp-flg)
		  (lisp (BLURB NIL CR |***NO| |CHANGE***| |--| |not|
			       |the| |number| |of| |a| |current| |hypothesis| excl cr))
		(if (eq (cdr hyp-flg) 'h)
		    (lisp (BLURB NIL CR |***NO| |CHANGE***| |--| |the|
				 |indicated| |hypothesis| |is| |currently| |hidden|
				 EXCL CR))
		  (bind ((hyp (car hyp-flg))
			 (current-address (reverse (current-addr-r)))
			 (current-deps (access goal depends-on (goal))))
			(if (match hyp (implies & &))
			    (do-all
			     (if current-address top skip)
			     (demote n)
			     (dive 1 1)
			     (if (not (equal (current-term) true))
				 (do-all push
					 (succeed (protect (do-strict (change-goal s-prop-and-finish)))))
			       skip)
			     up
			     (then (= * (! (fargn hyp 2)) t) (succeed s-prop-and-finish))
			     up
			     promote
			     (if current-address
				 (! (cons 'dive current-address))
			       skip)
			     (succeed
			      (lisp
			       (iterate for new-goal
					in (set-diff (access goal depends-on (goal))
						     current-deps)
					with all-goals = (all-goals (car state-stack))
					and unproved-new-goals
					when (not (equal (access goal conc
								 (find-goal new-goal
									    all-goals))
							 true))
					collect new-goal into unproved-new-goals
					finally
					(when unproved-new-goals
					      (let ((blurb-flg t))
						(blurb (unproved-new-goals) CR NOTE |:| |New|
						       |unproved|
						       (plural? unproved-new-goals
								|goals| |goal|)
						       |created| |by|
						       FORWARD
						       (plural? unproved-new-goals
								|are| |is|)
						       (!PPR-LIST UNPROVED-NEW-GOALS
								  (QUOTE |.|)) CR)))))))
			  (prog2 (lisp (blurb (n) cr |***NO| |CHANGE***| |--|
					      |hypothesis| |number| (!PPR N NIL)
					      |is| |not| |an| |implication| excl))
				 fail)))))))
  (|(FORWARD n)| |:| |If| |hypothesis| |number| |n| |is| |active| |and| |of| |the| |form|
   |(IMPLIES P Q)| |,| |this| |has| |the| |effect| |of| |replacing| |this| |hypothesis| |by|
   |simply| |Q| |(and| |moving| |it| |to| |the| |end| |of| |the| |hypothesis| |list| |,|
   |incidentally)| |,| |and| |creating| |a| |new| |subgoal| |to| |prove| |P| |under| |the|
   |other| |active| |hypotheses| |.| |See| |also| |FORWARD*| |and| |FORWARD!| |.|))

(definstr forward! (n &optional comm demand-success-flg)
  (if demand-success-flg
      (protect (then (forward n)
		     comm))
    (then (forward n)
	  comm))
  (|Same| |as| FORWARD |,| |except| |that| |the| |command| COMM |(default: PROVE+)|
   |is| |run| |on| |each| |newly-created| |subgoal| |not| |proved| |by| FORWARD |.|
   |If| |demand-success-flg| |is| |supplied| |and| |non-NIL| |then| |there| |will| |be|
   |no| |net| |effect| |unless| |the| |command| |COMM| |"succeeds"| |on| |each| |newly|
   |created| |subgoal| |which| |isn't| |proved| |by| FORWARD |.| |(In| |any| |case| |,|
   |this| |is| |the| |criterion| |for| |"success".)|))

(definstr gen ()
  (generalize (! (iterate for abb in (abbreviations)
		       with vars =
			 (all-vars-lst
			   (cons (conc)
				 (iterate for hyp-flg in (hyps) collect (car hyp-flg))))
		       with temp
		       collect
			 (list (car abb)
			       (if (member-eq (setq temp
					       (PACK (CDR (COERCE (STRING (CAR ABB)) 'LIST))))
					 vars)
				   '?
				   temp)))))
  (GEN |:| |Replace| |each| |abbreviation| |@v| |by| |a| |new| |variable| |.| |In| |fact| |,|
       |that| |variable| |will| |be| |the| |corresponding| |one| |(i.e.| |the| |one| |after|
       |the| |@| |sign)| |if| |that| |is| |a| |variable| |not| |already| |occurring| |in|
       |the| |current| |goal| |.|))

(define-macro-command generalize-skolems ()
  (let* ((goal (goal))
	 (skolem-terms
	  (union-equal
	   (skolem-terms (access goal conc goal) (free-variables))
	   (skolem-terms-list
	    (mapcar #'car (access goal hyps goal))
	    (free-variables)))))
    (if skolem-terms
	`(generalize ,(iterate for term in skolem-terms
			       collect (list term '?)))
      'fail))
  (|Generalizes| |away| |all| |terms| |whose| |function| |symbol| |was| |introduced| |as| |a|
   |Skolem| |function| |by| |a| |DEFN-SK| |event| |.| |This| |is| |probably| |a| |useful|
   |thing| |to| |do| |just| |before| |entering| |the| |theorem| |prover| |if| |you| |want|
   |it| |to| |be| |able| |to| |take| |maximal| |advantage| |of| |metafunctions| |.|))

(DEFINSTR HYP (N INSTR)
  (ORELSE
    (HYP0 N INSTR)
    (PROG2 (PRINT "HYP failed")
	   FAIL))
  (|Applies| |instruction| |INSTR| |to| |hypothesis| |number| |N| |.| |If|
	     |this| |"fails"| |,| |then| |there| |is| |no| |change| |made| |to|
	     |the| |state| |stack| |,| |the| |command| |"fails"| |,| |and| |a|
	     |message| |to| |that| |effect| |is| |printed;| |otherwise|
	     |it| |"succeeds"| |.|))

(definstr iff ()
  (if (and (nvariablep (current-term))
	   (not (fquotep (current-term)))
	   (eq (ffn-symb (current-term)) 'equal)
	   (let (type-alist)
	     (and (our-boolean (fargn (current-term) 1))
		  (our-boolean (fargn (current-term) 2)))))
      (wrap iff
	    (do-all (= * (! (fcons-term* 'iff
	                      (untranslate-abb (fargn (current-term) 1)
					       (abbreviations))
			      (untranslate-abb (fargn (current-term) 2)
					       (abbreviations))))
		       0)
		    change-goal
		    (if (hyps) drop skip)
		    (generalize
		     (! (list (list (fargn (fargn (current-term) 1) 1) '?)
			      (list (fargn (fargn (current-term) 1) 2) '?))))
		    prove))
      (prog2 (lisp (blurb nil cr wrong form for iff excl cr))
	     fail))
  (IFF |:| |Replace| |the| |current| |subterm| |,| |if| |of| |the| |form| |(EQUAL x y)|
       |where| |x| |and| |y| |are| |boolean| |,| |by| |(IFF x y)| |.| |This| |command| |is|
       |handy| |for| |taking| |a| |goal| |which| |is| |an| |equation| |between| |two| |boolean|
       |expressions| |and| |replacing| |it| |by| |the| IFF |form| |,| |which| |then| |splits|
       |(via| |the| SPLIT |command)| |into| |cases| |.|))

(define-macro-command ifsplit ()
  (let (x y z)
    (if (match (current-term) (if x y z))
	`(wrap ifsplit
	       (qq (do-all (then-return (claim ,(untranslate x) 0)
					;; in the new subgoal x is false
					(then (= * ,(untranslate z) 0)
					      (do-all (demote (! (length (hyps))))
						      drop?
						      (check-proved (s-prop implies not)))))
			   ;; in the current goal x is true
			   (then (= * ,(untranslate y) 0)
				 (do-all (demote (! (length (hyps))))
					 drop?
					 (check-proved (s-prop implies not)))))))
      `(lisp (blurb nil |**NO| |CHANGE**| |--| |The| |current| |term| |is| |not|
		    |an| IF |term| excl cr))))
  (|Assuming| |that| |the| |current| |term| |is| |an| |IF| |expression| |(IF x y z)| |,|
   |this| |command| |causes| |a| |casesplit| |on| |the| |test| |of| |that|
   |expression| |(i.e., x).| |The| |result| |is| |that| |the| |current| |term| |is| |replaced| |by|
   |y| |in| |the| |current| |goal| |,| |which| |has| |a| |new| |hypothesis| |of| |x| |,|
   |and| |a| |subgoal| |is| |created| |in| |which| |the| |current| |term| |is| |replaced| |by| |z|
   |and| |(NOT x)| |is| |added| |as| |a| |new| |hypothesis| |.|))

(DEFINSTR IFTEST-FALSE (&OPTIONAL CMD)
  ;; One might want to replace the final S by S-PROP below.
  (IF (MATCH (CURRENT-TERM) (IF & & &))
      (ORELSE (PROTECT (DO-STRICT
			((DIVE 1)
			 (IF (QUOTE CMD)
			     (DO-ALL CMD
				     (LISP (EQUAL (CURRENT-TERM) FALSE)))
			     (= F))
			 UP
			 S)))
	      (LISP (BLURB NIL CR |***NO| |CHANGE***| |--| |The|
			   IFTEST-FALSE |command| |didn't| |"succeed"| |.| CR)))
      (LISP (BLURB NIL CR |You| |must| |have| |an|
		   IF |expression| |to| |apply| |the|
		   IFTEST-FALSE |macro| |.| CR)))
  (|This| |command| |is| |used| |to| |first| |prove| |the| |negation| |of|
   |the| |test| |of| |an| |IF| |expression| |and| |then| |simplify|
   |the| |result| |(if| |the| |test| |can| |indeed|
   |be| |proved| |false)| |.| |The| |optional| |argument| |is| |a|
   |command| |to| |use| |in| |place| |of| |the| |default| |,|
   |which| |is| |a| |call| |to| |the| |theorem| |prover| |by|
   |way| |of| |the| |(= F)| |command| |.|))

(DEFINSTR IFTEST-TRUE (&OPTIONAL CMD)
  ;; One might want to replace the final S by S-PROP below.
  (IF (MATCH (CURRENT-TERM) (IF & & &))
      (ORELSE (PROTECT (DO-STRICT
			((DIVE 1)
			 (IF (QUOTE CMD)
			     (DO-ALL CMD
				     (LISP (AND (QUOTEP (CURRENT-TERM))
						(NOT (EQUAL (CURRENT-TERM) FALSE)))))
			     (= T))
			 UP
			 S)))
	      (LISP (BLURB NIL CR |***NO| |CHANGE***| |--| |The|
			   IFTEST-TRUE |command| |didn't| |"succeed"| |.| CR)))
      (LISP (BLURB NIL CR |You| |must| |have| |an|
		   IF |expression| |to| |apply| |the|
		   IFTEST-TRUE |macro| |.| CR)))
  (|This| |command| |is| |used| |to| |first| |prove| |the| |test|
   |of| |an| |IF| |expression| |and| |then| |simplify|
   |the| |result| |(if| |the| |test| |can| |indeed|
   |be| |proved)| |.| |The| |optional| |argument| |is| |a|
   |command| |to| |use| |in| |place| |of| |the| |default| |,|
   |which| |is| |a| |call| |to| |the| |theorem| |prover| |by|
   |way| |of| |the| |(= T)| |command| |.|))

(DEFINSTR INDUCT-PROMOTE (&OPTIONAL ARG)
  (WRAP INDUCT-PROMOTE
	(THEN (IF 'ARG (INDUCT ARG) INDUCT)
	      (PROG2 (LISP (PROGN (BLURB ((gl-name (goal-name)))
					 |Promoting| |on| |goal| (!ppr gl-name nil) |...| CR)
				  T))
		     (QQ PROMOTE*))))
  (|Runs| |the| INDUCT |command| |,| |and| |if| |that| |"succeeds"|
	  |then| |this| |command| |"succeeds"| |too| |,| |and| |PROMOTEs|
	  |all| |hypotheses| |on| |all| |of| |the| |subgoals| |generated| |.|))

(DEFINSTR INTRO-APPLY$ ()
  (BIND ((CTERM (CURRENT-TERM)))
	(BIND
	  ((NEWTERM
	     (TRANSLATE
	       (LIST 'APPLY$
		     (LIST 'QUOTE (FN-SYMB CTERM))
		     (CONS 'LIST (SARGS CTERM))))))
	  (= * (! NEWTERM)
	     (! (SET-DIFF
		  (MAKE-SET
		    (LIST (CONS 'HANDS-OFF
				(CONS (FARGN NEWTERM 2)
				      (CONS CTERM
					    (ASSUMPTIONS (HYPS)))))))
		  (LIST NEWTERM))))))
  (|Replaces| |the| |current| |subterm| |(fn x y ...)| |by| |(APPLY$ 'fn (LIST x y ...))|
	      |if| |possible| |.|))

(definstr lemma (name types)
  (print
   (list 'prove-lemma
	 'name
	 'types
	 (untranslate
	  (make-implication
	   (assumptions (hyps))
	   (conc)))))
  (|(LEMMA NAME TYPES)| |:| |prints| |a| PROVE-LEMMA |form| |corresponding| |to| |the|
   |current| |goal| |,| |of| |name| NAME |and| |types| TYPES |(e.g.| |(REWRITE))| |.|))

(definstr man (&optional ins)
  (if 'ins (help ins) help)
  (|Synonymous| |with| HELP |.|))

(DEFINE-MACRO-COMMAND new-BASH+ (&REST HINTS)
  (PROGN (CHK-ACCEPTABLE-HINTS-ABB HINTS (ABBREVIATIONS))
	 (LET ((HNTS (new-MAKE-LOCAL-HINTS HINTS STATE-STACK)))
	      (IF HNTS `(BASH ,@HNTS) 'BASH)))
  (|Call| |the| |theorem| |prover| |in| |the| |sense| |of| |the| BASH
   |command| |,| |but| |with| |hints|
   |that| |match| |the| |local| |disable-enable| |environment| |in|
   |the| |current| |interactive| |session| |,| |as| |updated|
   |(optionally)| |by| |the| |HINTS| |argument| |.| |Notice|
   |that| |this| |command| |has| |the| |same| |syntax|
   |as| BASH |,| |e.g.| |(BASH (DISABLE PLUS TIMES))| |.| |This| |command|
   |"succeeds"| |if| |and| |only| |if| |the| |corresponding| BASH
   |command| |"succeeds"| |,| |i.e.| |the| |proof| |completes| |.|
   |Unlike| |BASH+| |,| |we| |do| |not| |care| |about| |the| |order|
   |of| |the| |ENABLE| |and| |DISABLE| |hints| |that| |have| |been| |given|
   |to| |the| |proof-checker;| |we| |simply| |throw| |them| |all| |together| |.|
   |In| |this| |"translation"| |,| |we| |view| |the| |atom| |ENABLE| |as|
   |the| |same| |as| |(ENABLE-THEORY T)| |and| |similarly| |we| |view| |the|
   |atom| |DISABLE| |as| |(DISABLE-THEORY T)| |;| |thus| |it|
   |will| |be| |easy| |to| |obtain| |illegal| |hints| |this| |way| |.|
   |Other| |than| |that| |,| |the| |algorithm| |for| |forming| |the|
   |hint| |is| |simply| |that| |all| |ENABLE| |,| DISABLE |,| |ENABLE-THEORY| |,|
   |and| DISABLE-THEORY |hints| |are| |gathered| |up| |from| |the| |state| |stack| |,|
   |and| |appropriately| |appended| |to| |the| |explicitly| |given| |hints| |.| |Here|
   |"appropriately"| |implies| |that| |explicitly| |enabled| |names| |are|
   |removed| |from| |the| |list| |of| |disabled| |names| |obtained| |from| |the|
   |state| |stack| |,| |and| |similarly| |for| |the| |other| |cases| |.|))

(definstr noise ()
  (do-all (setq blurb-flg t)
	  (setq io-fn 'io1))
  (|Undoes| |the| |effect| |of| |the| |command| |Q;| |see| |documentation|
   |for| |Q| |.|))

(define-macro-command orchoose (&optional n)
  (cond
   ((current-addr-r)
    (blurb nil CR |Please| |only| |use| ORCHOOSE |from| |the| |top|
	   |of| |the| |conclusion| EXCL CR)
    'fail)
   ((and n (or (not (integerp n)) (<= n 0)))
    (blurb nil cr ORCHOOSE |expects| |a| |positive| |integer|
	   |argument| |(optional)| |.| CR)
    'fail)
   (t
    `(wrap ,(if n `(orchoose ,n) 'orchoose)
	   (orelse (qq (orchoose-aux ,(or n 2)))
		   (lisp (if (match (conc) (or & &))
			     (blurb ((k n)) cr |***NO| |CHANGE***| |--|
				    |The| OR |structure| |of| |the|
				    |conclusion| |does| |not| |support| |an| |argument|
				    |of| (!ppr k (quote |.|)) cr)
			   (blurb nil cr |***NO| |CHANGE***| |--|
				  |The| |conclusion| |is| |not| |an| OR
				  |term| EXCL cr)))))))
  (|Selects| |the| |Nth| |disjunct| |of| |the| |conclusion| |,| |which|
   |should| |be| |a| |disjunction| |(default| |is| |N=2)| |,| |for| |proof| |.|))

(define-macro-command orsplit (n)
  (if (not (and (integerp n)
		(> n 0)
		(<= n (length (hyps)))))
      (progn (blurb (n) cr Sorry -- bad arg cr)
	     'skip)
    (let (arg1 arg2 (lng (length (hyps))))
      (if (match (car (nth (1- n) (hyps)))
		 (or arg1 arg2))
	  `(wrap (orsplit ,n)
	    (qq (do-strict ((claim (not ,arg1) 0)
			    (claim (not ,arg2) 0)
			    (contradict ,n)
			    (retain ,(1+ lng) ,(+ 2 lng))
			    taut
			    (drop ,n ,(1+ lng))
			    change-goal
			    (drop ,n)))))
	(progn (blurb (n) cr |Sorry|
		      |but| |hypthesis| (!ppr n nil) |is| |not|
		      |an| OR |term| excl cr)
	       'skip))))
  (|Assuming| |that| |the| |nth| |hypothesis| |is| |of| |the| |form| |(OR x y)| |,|
   |creates| |two| |subgoals| |--| |one| |with| |x| |as| |an| |added| |hypothesis| |and|
   |the| |other| |with| |y| |,| |with| |that| |OR| |hypothesis| |dropped| |--|
   |and| |tautologically| |proves| |the| |resulting| |current| |goal| |.|))

(definstr pg (ins)
  (bind ((old-goal-names
	  (iterate for gl in (all-goals (car state-stack)) 
		   collect (access goal goal-name gl))))
	(prog1
	 ins
	 (if (member-eq (instr-name 'ins) '(play restore replay))
	     skip
	   (bind ((new-goals
		   (iterate for gl in (all-goals (car state-stack))
			    when (not (member (access goal goal-name gl)
					      old-goal-names))
			    collect gl)))
		 (if new-goals
		     (lisp
		      (iterate for x in new-goals
			       initially
			       (blurb (new-goals) cr |The| |following| |new|
				      (plural? new-goals |goals have| |goal has|)
				      |been| |created| |:| cr cr)
			       do
			       (blurb ((term (untranslate-abb
					      (make-implication
					       (assumptions (access goal hyps x))
					       (access goal conc x))
					      (abbreviations)))
				       (name (access goal goal-name x)))
				      (!ppr name (quote |:|))
				      (!ppr term nil) cr)))
		   skip)))))
  (|Runs| |the| |given| |instruction| |,| |"succeeding"| |if| |and|
	  |only| |if| |the| |given| |instruction| |"succeeds"| |,| |and|
	  |then| |prints| |out| |all| |the| |new| |goals| |that| |are|
	  |created| |in| |the| |course| |of| |execution| |of| |this|
	  |instruction| |.| |However| |,| |if| |the| |instruction|
	  |is| RESTORE |,| PLAY |,| |or| REPLAY |,| |then| |this|
	  |printing| |of| |goals| |is| |suppressed| |.|))

(define-macro-command pgbash (&rest args)
  `(pg (bash ,@args))
  (|Same| |as| BASH |but| |also| |prints| |out| |all|
   |the| |newly| |created| |subgoals| |.|))

(DEFINSTR PLAY (INSTRS &OPTIONAL PROVE-LEMMA-FLG)
  (LISP (PLAY-FN 'INSTRS 'PROVE-LEMMA-FLG))
  (|(PLAY (instr1 instr2 ...) &OPTIONAL PROVE-LEMMA-FLG)|
   |:| |Play| |the| |given| |instructions|
   |from| |the| |current| |state| |,| |stopping| |if| |any| |fails|
   |and| |otherwise| |noting| |if| |all| |instructions| |run|
   |successfully| |(in| |which| |case| |the| |PLAY| |"succeeds")| |.|
   NOTE |:| |If| |not| |all| |of| |the| |instructions| |"succeed"| |,|
   |then| |the| |command| |(UNDONE-INSTRUCTIONS)| |will| |print| |a| |list|
   |of| |all| |the| |instructions| |which| |did| |not| |run| |(starting|
   |with| |the| |one| |that| |"failed")| |.|
   CR |#| |If| |the| |optional| |argument| |is| |anything| |except|
   | NIL| |,| |then| |only| CHANGE |commands| |(i.e.| |primitive|
   |commands| |as| |opposed| |to| HELP |,| META |,| |or| MACRO |commands)|
   |will| |be| |allowed| |,| |which| |is| |the| |same| |restriction|
   |applied| |to| |the| INSTRUCTIONS |hint| |for| PROVE-LEMMA |.| CR |#|
   |After| |invoking| PLAY |,| |the| RESTORE |command| |will| |put|
   |one| |back| |to| |the| |state| |existing| |just| |when| |the|
   PLAY |command| |was| |invoked| |.|))

(definstr play-rest ()
  (play (! (cdr undone-instrs)))
  (|Play| |the| |instructions| |shown| |by| |the| |command| UNDONE-INSTRUCTIONS |,|
   |except| |for| |the| |first| |one| |.|))

(definstr ppe (x)
  (lisp (ppe x))
  (|Same| |as| |calling| PPE |in| |Lisp| |,| |e.g.| |(PPE 'APPEND-ASSOC)| |.|))

(DEFINSTR PRINT (X) (LISP (PROGN (PPR X NIL) (TERPRI NIL) T))
	   (|Prettyprints| |the| |result| |of| |evaluating| |its| |argument| |in|
	    |Lisp| |,| |and| |always| |"succeeds"| |.|))

(definstr print-all-goals (&optional name)
  (bind ((gl-name (if 'name 'name 'main)))
	(lisp (iterate for gl in (all-goals (car state-stack))
		       when (not (equal (access goal conc gl) true))
		       do (progn
			    (print (access goal goal-name gl))
			    (terpri nil)
			    (ppr (untranslate
				  (make-implication
				   (assumptions
				    (access goal hyps gl))
				   (access goal conc gl)))
				  nil)
			    (terpri nil)))))
  (|Print| |each| |goal| |as| |a| |Boyer-Moore| |term| |(i.e.|
	   |as| |an| |implication| |if| |there| |are| |hypotheses)|
	   |,| |together| |with| |its| |name| |.|))

(DEFINSTR PRINT-ALL-HELP ()
  (DO-ALL (LISP (BLURB NIL CR |***BUILT-IN| |COMMANDS***| |(SHORT)| HELP |:| CR CR))
	  (PRINT-ALL-PRIMITIVE-HELP)
	  (LISP (BLURB NIL CR |***MACRO-COMMANDS***| |(SHORT)| HELP |:| CR CR))
	  (PRINT-ALL-MACRO-HELP))
  (|Same| |as| |(HELP <instr1> <instr2> ...)| |,| |where| |(<instr1> <instr2> ...)|
   |is| |a| |list| |of| |all| |built-in| |commands| |(in| |alphabetic| |order)|
   |followed| |by| |a| |list| |of| |all| |macro| |commands| |.|))

(DEFINSTR PRINT-ALL-HELP-LONG ()
  (DO-ALL (LISP (BLURB NIL CR |***BUILT-IN| |COMMANDS***| |(LONG)| HELP |:| CR CR))
	  (PRINT-ALL-PRIMITIVE-HELP-LONG)
	  (LISP (BLURB NIL CR |***MACRO-COMMANDS***| |(LONG)| HELP |:| CR CR))
	  (PRINT-ALL-MACRO-HELP-LONG))
  (|Same| |as| |(HELP-LONG <instr1> <instr2> ...)| |,| |where| |(<instr1> <instr2> ...)|
   |is| |a| |list| |of| |all| |built-in| |commands| |(in| |alphabetic| |order)|
   |followed| |by| |a| |list| |of| |all| |macro| |commands| |.|))

(DEFINE-MACRO-COMMAND PRINT-ALL-MACRO-HELP ()
  `(HELP ,@MACRO-COMMANDS)
  (|Same| |as| |(HELP <instr1> <instr2> ...)| |,| |where| |(<instr1> <instr2> ...)|
   |is| |a| |list| |of| |all| |macro| |commands| |.|))

(DEFINE-MACRO-COMMAND PRINT-ALL-MACRO-HELP-LONG ()
  `(HELP-LONG ,@MACRO-COMMANDS)
  (|Same| |as| |(HELP-LONG <instr1> <instr2> ...)| |,| |where| |(<instr1> <instr2> ...)|
   |is| |a| |list| |of| |all| |macro| |commands| |.|))

(DEFINE-MACRO-COMMAND PRINT-ALL-PRIMITIVE-HELP ()
  `(HELP ,@(SORTED-INSTRUCTIONS))
  (|Same| |as| |(HELP <instr1> <instr2> ...)| |,| |where| |(<instr1> <instr2> ...)|
   |is| |a| |list| |of| |all| |built-in| |commands| |.|))

(DEFINE-MACRO-COMMAND PRINT-ALL-PRIMITIVE-HELP-LONG ()
  `(HELP-LONG ,@(SORTED-INSTRUCTIONS))
  (|Same| |as| |(HELP-LONG <instr1> <instr2> ...)| |,| |where| |(<instr1> <instr2> ...)|
   |is| |a| |list| |of| |all| |built-in| |commands| |.|))

(define-macro-command print-goal (&optional given-name)
  (let ((goal (find-original-goal
	       (or given-name (goal-name)))))
    (cond
     (goal
      (blurb ((hyps (iterate for x in (assumptions (access goal hyps goal))
			     collect (untranslate x)))
	      (conc (untranslate (access goal conc goal)))
	      (name (or given-name (goal-name))))
	     |Original| |hypotheses| |of| (!PPR NAME (QUOTE |:|)) CR
	     (COND (HYPS (!PPR HYPS NIL)) (T |<none>|)) CR
	     |Original| |conclusion| |of| (!PPR NAME (QUOTE |:|)) CR
	     (!PPR CONC NIL) CR))
     (t (blurb (given-name) there is no goal named (!ppr given-name nil) excl cr)))
    'skip)
  (|(PRINT-GOAL <name>)| |:| |Print| |the| ORIGINAL |hypotheses| |and| |conclusion|
   |of| |the| |goal| |named| |<name>| |.| CR PRINT-GOAL |:| |Print| |the| ORIGINAL
   |hypotheses| |and| |conclusion| |of| |the| |current| |goal| |.|))

(definstr print-instrs ()
  (lisp (progn (print-instrs) (blurb nil cr)))
  (|PrettyPrints| |a| |list| |of| |all| |instructions| |in| |the|
   |current| |interactive| |session| |.|))

(definstr print-remaining-instrs ()
  (lisp (progn (print-remaining-instrs) (blurb nil cr)))
  (|Assuming| |that| |one| |has| |just| |done| |an| UNDO |of| |some| |number|
   |of| |instructions| |,| |this| |command| |causes| |the| |list| |of| |undone|
   |instructions| |to| |be| |printed| |out| |in| |an| |order| |which| |allows|
   |them| |to| |be| |PLAYed| |to| |get| |back| |to| |the| |state| |present|
   |just| |before| |the| UNDO |.|))

(definstr pro ()
  (orelse (qq (wrap pro
		    (orelse promote* 
			    promote-if)))
	  (lisp (blurb nil cr |**NO| |CHANGE**| |--| PRO |failed| |.| cr)))
  (|Does| |a| PROMOTE* |or| PROMOTE-IF |.|))

(definstr promote* ()
  (progress (repeat promote))
  (PROMOTE* |:| |Repeat| |promote| |as| |many| |times| |as| |possible| |.|))

(define-macro-command promote-if ()
  (let (h c)
    (cond
     ((current-addr-r)
      (blurb nil cr |**NO| |CHANGE**| |--| |You| |must| |be|
	     |at| |the| |top| |of| |the| |conclusion| |to| PROMOTE-IF |.| CR)
      'fail)
     (t
      (match! (unexpand-if (current-term) nil) (cons h c))
      (if h
	  `(do-strict
	    ((claim (not (implies ,(dumb-conjoin h) ,c)) 0)
	     change-goal
	     (demote ,(1+ (length (hyps))))
	     s-prop-and-finish
	     (contradict ,(1+ (length (hyps))))
	     (drop ,(1+ (length (hyps))))
	     promote*))
	(progn (blurb nil cr |**NO| |CHANGE**| |--| |You| |must| |be|
		      |looking| |at| |a| |term| |of| |the| |form| |(IF X Y T)|
		      |to| PROMOTE-IF |.| CR)
	       'fail)))))
  (|Just| |like| PROMOTE |except| |that| |it| |works| AFTER |you've| |destroyed|
   |the| |IMPLIES| |by,| |for| |example,| |running| |the| S |command| |.|))

(DEFINE-MACRO-COMMAND PROVE+ (&REST HINTS)
  (PROGN (CHK-ACCEPTABLE-HINTS-ABB HINTS (ABBREVIATIONS))
	 (LET ((HNTS (MAKE-LOCAL-HINTS HINTS STATE-STACK)))
	      (IF HNTS `(PROVE ,@HNTS) 'PROVE)))
  (|Call| |the| |theorem| |prover| |with| |hints|
	  |that| |match| |the| |local| |disable-enable| |environment| |in|
	  |the| |current| |interactive| |session| |,| |as| |updated|
	  |(optionally)| |by| |the| |HINTS| |argument| |.| |That| |is| |,| |call| |the|
	  |theorem| |prover| |in| |an| |environment| |which| |is| |as|
	  |though| |the| |current| |interactive| |session's| |modifications|
	  |to| |the| |enable-disable| |environment| |were| |made| |to|
	  |the| |global| |data-base| |.| |Notice|
	  |that| |this| |command| |has| |the| |same| |syntax|
	  |as| PROVE |,| |e.g.| |(PROVE+ (DISABLE PLUS TIMES))| |.| |This| |command|
	  |"succeeds"| |if| |and| |only| |if| |the| |corresponding| PROVE
	  |command| |"succeeds"| |,| |i.e.| |the| |proof| |completes| |.|))

(definstr prove++ ()
  (bind ((hint `(enable ,@(all-supporters-of-term-lst
			   (cons (conc) (assumptions (hyps)))))))
	(prove (! hint)))
  (|Like| PROVE |,| |but| |with| |a| |big| |hint| |generated| |to| |enable|
   |lots| |of| |stuff| |.|))

(definstr prove+++ ()
  (bind ((hint `(enable ,@(gross-hint
			    (cons (conc) (assumptions (hyps)))))))
	(prove (! hint)))
  (|Like| |PROVE++| |,| |only| |more| |so| |.|))

(definstr prune (n)
  ;; Here n is the UNDO number of the instruction we want to undo,
  ;; preserving as much of the proof after that as is reasonably possible.
  (prog2 (progress (undo n))
	 prune-raw)
  (|(PRUNE x)| |:| |Similar| |to| |(UNDO x)| |,| |except| |that| |some|
   |attempt| |is| |made| |to| |save| |that| |part| |of| |the| |proof| |which|
   |doesn't| |"depend| |on"| |the| |command| |that| |was| |undone| |.|
   RESTORE |can| |be| |run| |after| PRUNE |to| |undo| |its| |effect| |.| CR |#|
   |Here| |is| |a| |more| |complete| |description| |of| PRUNE |.| |The| |idea|
   |is| |that| |as| |many| |commands|
   |as| |are| |"meaningfully| |possible"| |are| |run| |after| |the| |UNDOing| |.|
   |Here| |is| |the| |idea| |.| |Suppose| |that| |some| |command| |comes| |later|
   |in| |the| |proof| |than| |the| |one| |being| |undone| |,| |and| |that| |this|
   |later| |command| |is| |on| |a| |goal| |whose| |existence| |and| |nature| |are|
   |logically| |unaffected| |by| |this| |UNDOing| |.| |Then| |we| |would| |like| |that|
   |later| |command| |to| |be| |run| |after| |the| |UNDOing| |.| |More| |precisely| |,|
   |we| |run| |all| |of| |the| |undone| |commands| |which| |either| |apply| |at| |the|
   |state| |level| |(such as ENABLE)| |or| |apply| |on| |a| |goal| |which| |is| |present|
   |at| |the| |time| |of| |replay| |,| |but| |which| |is| |not| |the| |current| |goal|
   |at| |the| |last| |state| |undone| |.| |Still| |more| |precisely;| |we| |do| |not|
   |bother| |to| |run| CHANGE-GOAL |commands| |which| |will| |fail;| |we| |insert|
   |extra| |CHANGE-GOAL| |commands| |as| |necessary;| |we| |modify| PUT |commands|
   |as| |necessary| |so| |that| |they| |do| |not| |apply| |to| |non-existent| |free|
   |variables;| |and| |we| |omit| COPY |commands| |and| |mark| |their| |target| |goals|
   |as| |"bad"| |so| |that| |no| |further| |commands| |are| |applied| |to| |such| |goals| |.|
   |We| |also| |cheat| |by| |assuming| |that| |all| |calls| |to| |the| |prover|
   |will| |succeed| |,| |hence| |actually| |bypassing| |all| |calls| |to| |the| |prover| |.|
   |One| |final| |minor| |point| |:| |if| |the| |last| |command| |undone| |is| |a|
   CHANGE-GOAL |command| |,| |we| |assume| |that| |the| |goal| |to| |be| |marked| |as|
   |"bad"| |is| |the| |goal| |that| |existed| |immediately| AFTER |that| CHANGE-GOAL |command|
   |,| |rather| |than| |the| |current| |goal| |at| |the| |time| |that| |command| |was|
   |about| |to| |be| |executed| |.|))

(definstr push-to-end (&optional name)
  (bind ((last-name (iterate for goal
			     in (reverse (all-goals (car state-stack)))
			     when (not (equal (access goal conc goal) true))
			     do (return (access goal goal-name goal))))
	 (current-name (if 'name 'name (goal-name))))
	(if (not (iterate for gl in (all-goals (car state-stack))
			  thereis
			  (and (equal current-name (access goal goal-name gl))
			       (not (equal (access goal conc gl) true)))))
	    (lisp (progn (blurb (current-name)
				|**NO| |CHANGE**| |--| |There| |is|
				|no| |unproved| |goal| |named|
				(!PPR current-name (quote |.|)) CR)
			 nil))
	    (if (equal last-name current-name)
		skip
		(prog2 (change-goal (! last-name)) (push-to-end
						    (! current-name))))))
  (|Keep| |changing| |goals| |to| |the| |last| |unproved| |goal|
	  |on| |the| |stack| |until| |that's| |the| |goal| |whose|
	  |name| |is| |given| |as| |an| |argument| |(where| |the|
	  |default| |is| |the| |current| |goal's| |name)| |.|))

(define-macro-command q (ins)
  (let ((saved-blurb-flg blurb-flg)
	(saved-io-fn io-fn))
    `(do-all (succeed (setq blurb-flg nil))
	     (succeed (setq io-fn 'no-io))
	     ,ins
	     (succeed (lisp (progn
			      (setq blurb-flg ',saved-blurb-flg)
			      (setq io-fn ',saved-io-fn)
			      (ppr (instruction) nil)
			      (terpri nil))))))
  (|Turns| |off| |the| |printing| |of| |messages| |by| |the|
   |proof-checker| |and| |theorem-prover| |and| |then|
   |runs| |the| |instruction| |and| |finally| |turns|
   |printing| |back| |on| |and| |prints| |the| |last|
   |instruction| |.| |Note| |:| |if| |this| |does| |not|
   |complete| |then| |run| |the| |macro| |command|
   NOISE |.|))

(define-macro-command qq (ins)
  (let ((saved-blurb-flg blurb-flg)
	(saved-io-fn io-fn))
    `(do-all (succeed (setq blurb-flg nil))
	     (succeed (setq io-fn 'no-io))
	     ,ins
	     (succeed (lisp (progn
			      (setq blurb-flg ',saved-blurb-flg)
			      (setq io-fn ',saved-io-fn))))))
  (|Same| |as| |Q| |except| |that| |the| |last| |instruction| |is|
   |not| |printed| |.|))

(define-macro-command r (&rest args) `(rewrite ,@args)
  (|Same| |as| REWRITE |.|))

(define-macro-command r$ (&rest args)
  (let ((fvs (free-variables)))
    `(protect
      (prog2 (rewrite ,@args)
	     (bind ((new-fvs (free-variables))
		    (extra-fvs (set-diff-eq (free-variables) ',fvs)))
		   (if extra-fvs
		       (do-all
			undo 
			(rewrite ,(car args)
				 (! (nconc (iterate for v in extra-fvs
						    collect (list (pack (list '$ v))
								  (make-new-r$-var v new-fvs)))
					   ',(cadr args)))
				 ,@(if (cddr args) (list (caddr args)) nil)))
		     skip)))))
  (|Same| |as| REWRITE |,| |except| |that| |new| |free| |variables| |are| |suffixed|
   |with| |$| |signs| |when| |they| |don't| |already| |end| |in| |$| |signs| |or|
   |$| |signs| |followed| |by| |indices| |.|))

(define-macro-command reduce (&rest args)
  (list* 'bash t args)
  (|Same| |as| (BASH T . args) |.| |This| |is| |generally| |preferred|
   |to| |calling| BASH |(without| |the| |T| |argument)| |,| |since|
   |the| |entire| |prover| |except| |induction| |is| |called| |,|
   |rather| |than| |just| |the| |simplifier| |--| |and| |also| |,|
   |there| |is| |IO| |.|))

(define-macro-command reduce+ (&rest hints)
  (PROGN (CHK-ACCEPTABLE-HINTS-ABB HINTS (ABBREVIATIONS))
	 (LET ((HNTS (MAKE-LOCAL-HINTS HINTS STATE-STACK)))
	      (IF HNTS `(BASH T ,@HNTS) '(BASH T))))
  (|Same| |as| (BASH+ T . args) |.| |See| |the| |documentation|
   |for| |BASH+| |and| |for| |REDUCE;| |this| |command| |gives| |the|
   |effects| |of| |both| |.|))

;; In response to a request by David Goldschlag.
(define-macro-command reduce-induct (&rest hints)
  `(then (reduce ,@hints)
	 (then induct (reduce ,@hints)))
  (|Example| |:| (REDUCE-INDUCT (EXPAND (APPEND X Y))) CR |#|
   |Call| |the| |prover| |with| |the| |indicated| |hints| |,| |just| |like|
   |the| REDUCE |command| |.| |The|
   INDUCT |command| |will| |be| |given| |to| |each| |subgoal| |,| |and| |the| |same|
   |call| |to| |the| |prover| |will| |be| |made| |on| |each| |subgoal|
   |resulting| |from| |each| INDUCT |command| |.|))

(DEFINSTR REPLAY ()
  (LISP (REPLAY-FN))
  (REPLAY |:| |Undo| |back| |to| |the| |starting| |point|
	  |of| |the| |current| |session| |and|
	  |replay| |the| |instructions| |.| |This| |command| |"succeeds"| |if|
	  |the| |proof| |state| |(i.e.| |the| |STATE-STACK)|
	  |gets| |back| |to| |where| |it| |is|
	  |currently| |.|
	  NOTE |:| |If| |not| |all| |of| |the| |instructions| |"succeed"| |,|
	  |then| |the| |command| |(UNDONE-INSTRUCTIONS)| |will| |print| |a| |list|
	  |of| |all| |the| |instructions| |which| |did| |not| |run| |(starting|
	  |with| |the| |one| |that| |"failed")| |.| CR |#|
	  NOTE |that| |this| |instruction| |saves| |the|
	  |current| STATE-STACK |,| |so| |that| RESTORE |may| |be| |used|
	  |to| |get| |back| |to| |where| |one| |was| |before| |executing|
	  |this| |command| |.|))

(define-macro-command retain (&rest args)
  `(drop ,@(set-diff
	    (iterate with tail = (hyps)
		     while tail
		     as i from 1
		     collect (progn (setq tail (cdr tail)) i))
	    ;; (iterate for x in (hyps)
	    ;;     as i from 1 collect i)
	    args))
  (DROP |all| |hypotheses| EXCEPT |those| |with| |numbers| |in|
	|the| |list| ARGS |.|))

(define-macro-command retrieve (&optional var)
  (progn
    var
    (blurb nil |About| |to| EXIT |--| |you| |should| |then|
	   |submit| |this| RETRIEVE |command| |to| |Lisp| |.| CR)
    'exit)
  (|(RETRIEVE var)| |:| |Exit| |the| |interactive| |environment| |with|
   |a| |suggestion| |that| |one| |submit| |this| |command| |to| |Lisp| |.|
   |Such| |a| |submission| |will| |result| |in| |a| |restoration| |to|
   |the| |state| |stored| |the| |last| |time| |that| |(SAVE var)| |was|
   |executed| |in| |an| |interactive| |session| |.|))

(DEFINSTR S-ALL ()
  (PROGRESS
    (BIND ((NAME (GOAL-NAME)))
	  (DO-ALL
	    TOP?
	    S
	    (! `(DO-ALL
		  ,@(ITERATE FOR I FROM 1 TO (LENGTH (HYPS))
			  COLLECT `(DO-ON-GOAL (SH ,I) ,NAME)))))))
  (|Simplify| |the| |conclusion| |(after| |moving| |to| |the| |top|
	      |if| |necessary)| |,| |and| |then| |simplify| |all| |the| |hypotheses| |,|
	      |stopping| |if| |the| |goal| |is| |proved| |at| |any| |point| |.|
	      |This| |command| |"succeeds"| |if| |any| |change| |is| |made|
	      |in| |the| |state-stack| |(in| |the| |sense| |of| |PROGRESS)| |.|))

(DEFINSTR S* ()
  (ORELSE
    (PROGRESS (PROG2-CHECK-UNPROVED S (ALTERNATE S-PROP S)))
    (PROGRESS (PROG2-CHECK-UNPROVED S-PROP (ALTERNATE S S-PROP))))
  (|Alternately| |applies| S |and| S-PROP |until| |either| |the| |goal| |is|
		 |proved| |or| |there| |is| |no| |change| |.|))

(define-macro-command save (&optional var)
  `(lisp (let ((var (or ',var 'temp-ss)))
	   (if (assoc-eq var retrieve-alist)
	       (setf (cdr (assoc-eq var retrieve-alist))
		     state-stack)
	     (setq retrieve-alist
		   (cons (cons var state-stack)
			 retrieve-alist)))))
  (|(SAVE var)| |:| |Saves| |the| |current| |state| |stack|
   |by| |"associating"| |it| |with| |the|
   |given| |variable;| |the| |default| |for| |simply| SAVE |is| TEMP-SS |.|
   |Submit| |the| |command| |(RETRIEVE VAR)| |(or| |,| |simply|
   |(RETRIEVE)| |for| |the| |case| |TEMP-SS)| |to| |Lisp| |to| |get|
   |back| |to| |this| |state| |.|))

(definstr set-time-limit (&optional n)
  (lisp (set-time-limit 'n))
  (|Set| |the| |time| |limit| |(forever| |,| |until| |it| |is| |changed|
   |again)| |for| |calls| |to| |the| |prover| |,| |including| |those| |from|
   BASH |,| PROVE |,| CLAIM |,| |and| |from| DEFN |,| PROVE-LEMMA |,| |and|
   |so| |on| |.|))

(definstr setq (x y)
  (lisp (setq x y))
  (|Same| |as| |calling| SETQ |in| |Lisp| |,| |e.g.| |(SETQ X 5)| |.|))

(DEFINSTR SH (N)
  (HYP N S)
  (|Simplifies| |(in| |the| |sense| |of| |the| |S| |command)| |the| |hypothesis|
   |with| |the| |given| |number| |.|))

(define-macro-command shelve (&rest names)
  `(lisp (shelve ,@names))
  (|(SHELVE <name@-[1]> ... <name@-[n]>)| |:| |Make| |the| |given|
   |rewrite| |rules| |invisible| |from| |the| |point| |of| |view|
   |of| |the| |SHOW-REWRITE| |and| |REWRITE| |commands| |,| |except|
   |that| |REWRITE| |will| |still| |work| |if| |one| |gives| |the|
   |name| |of| |the| |given| |rewrite| |rule| |(not| |a| |number)|
   |as| |an| |argument| |.| |The| |current| |list| |of| |shelved|
   |lemmas| |can| |be| |seen| |using| |(PRINT SHELVED-LEMMAS)| |.| |The|
   |entire| |shelving| |feature| |can| |be| |switched| |from| |on| |to|
   |off| |and| |vice-versa| |using| |(TOGGLE-SHELVED-LEMMAS)| |.|))

(define-macro-command shelved-lemmas ()
  '(print shelved-lemmas)
  (|Submit| |(HELP SHELVE)| |for| |an| |explanation| |of| |this| |command| |.|))

(definstr shrink (&optional no-complain)
  (lisp (shrink-fn no-complain))
  (|The| SHRINK |command| |attempts| |to| |shrink| |the| |state|
   |stack| |in| |several| |ways| |.| CR |#| |(1)|
   |Movement| |commands| UP |,| NX |,| BK |,| |and| |(DIVE ...)|
   |are| |omitted| |when| |they| |occur| |immediately| |before| TOP |.|
   CR |#| |(2)| |Each| |remaining| |sequence| |of| |consecutive| |movement|
   |commands| |is| |combined| |into| |a| |sequence| |of| |at| |most| |two|
   |movement| |commands| |.| CR |#| |(3)| |Consecutive| CHANGE-GOAL
   |commands| |are| |combined| |.| CR |#| |(4)| |Consecutive| |matching| |bookends|
   |(BOOKMARK (BEGIN <expr>))| |,| |(BOOKMARK (END <expr>))| |are| |cancelled| |.|))

(define-macro-command sk () 
  (let ((parity (parity (reverse (current-addr-r)) (conc) t)) rule-name)
    (if (equal parity 'fail)
	'(lisp (blurb nil cr |**SK| |FAILED**| |--| |parity| |cannot| |be|
		      |determined| excl cr))
      (if (setq rule-name (sk-rule-name parity (current-term)))
	  (if parity
	      `(wrap sk
		     (prog2 (backchain-lemma$ ,rule-name)
			    sk-aux))
	    `(wrap sk
		   (prog2 (forwardchain-lemma$ ,rule-name)
			  sk-aux)))
	'(lisp (blurb nil cr |**SK| |FAILED**| |--| |no| |appropriate|
		      |rule-name| excl cr)))))
  (|Assuming| |that| |the| |current| |term| |is| |a| |call| |of| |a|
   DEFN-SK |function| FOO |,| |and| |that| |there| |is| |a| |lemma|
   |called| FOO-SUFF |or| FOO-NECC |(according| |to| |parity| |determined|
   |by| |position| |of| |the| |current| |term)| |,| |we| |want| |to|
   |replace| |the| |current| |term| |by| |the| |appropriate| |Skolem|
   |axiom| |stored| |in| FOO-SUFF |or| FOO-NECC |.|))

(define-macro-command sk* (&optional noisy)
  (let ((indices (iterate for hyp-flg in (hyps)
			  as i from 1
			  when (eq (cdr hyp-flg) 'a)
			  collect i)))
    (let ((comm
	   `(wrap sk*
		  (prog2 top?
			 (progress
			  (do-all sk-conc
				  ,@(iterate for i in indices
					     collect
					     `(protect (do-strict
							((contradict ,i)
							 sk-conc
							 (contradict ,i)))))))))))
      (if noisy comm `(qq ,comm))))
  (|Applies| |SK| |to| |everything| |possible| |in| |the| |current| |goal|
   |(including| |subterms| |of| |active| |hypotheses).|  |With| |a| |non-NIL| |optional|
   |argument| |it| |prints| |out| |what| |it| |is| |doing| |.|))

(definstr sl () (s lemmas)
  (|Same| |as| (S LEMMAS) |.|))

(define-macro-command sp (&rest args) `(s-prop ,@args)
  (|Same| |as| S-PROP |.|))

(define-macro-command sr (&rest args) `(show-rewrites ,@args)
  (|Same| |as| SHOW-REWRITES |.|))

;; omitted, since it requires *last-verify-sk-name* (see basis-pc)
; (define-macro-command suff ()
;   `(r$ ,(pack (list *last-verify-sk-name* '-suff))))

(DEFINSTR TAUT ()
  (ORELSE
    (PROTECT
      (PROG2
	(IF (ASSUMPTIONS (HYPS))
	    (IF (CURRENT-ADDR-R)
		(PROG2 TOP DEMOTE)
		DEMOTE)
	    SKIP)
	(SET-SUCCESS ((NAME (GOAL-NAME)))
		     ((NEW-NAME (GOAL-NAME)))
		     (OR (NOT (EQUAL NAME NEW-NAME))
			 (EQUAL (CONC) TRUE))
		     (S-PROP AND OR NOT IMPLIES FIX ZEROP IFF NLISTP))))
    (PROG2 (LISP (BLURB NIL |**NO| |CHANGE**| |--| TAUT |failed| |.| CR))
	   FAIL))
  (TAUT |:| |Replace| |the| |current| |conclusion| |by| TRUE |if| |it|
	|follows| |tautologically| |from| |the| |current| |hypotheses|
	|(upon| |expansion| |of| |functions| |from| |the| |list| |(AND|
	|OR| |NOT| |IMPLIES| |FIX| |ZEROP| |IFF| |NLISTP))| |.|))

(DEFINSTR TH ()
  VIEW
  (|Same| |as| VIEW |.|))

(define-macro-command then (instr &optional completion)
  `(then-aux ,instr
	     ,(if (null completion)
		  'prove+
		  completion))
  (|If| |the| |given| |instruction| INSTR |"succeeds"| |,| |then| |the|
   |instruction| COMPLETION |is| |run| |on| |each| |of| |the| |new| |subgoals|
   |of| |the| |original| |goal| |(if| |any)| |.| |If| |no| COMPLETION |is|
   |given| |,| |then| PROVE+ |is| |used| |as| |the| COMPLETION |.| |This|
   |command| |"succeeds"| |if| INSTR |"succeeds"| |and| |each| |resulting|
   |invocation| |of| COMPLETION |"succeeds"| |.|))

(define-macro-command then-return (&rest args)
  (let ((x (goal-name)))
    `(do-all (then ,@args)
	     (maybe-change-goal ,x)))
  (|Same| |as| THEN |except| |that| |at| |the| |end| |,|
   |it| |is| |ensured| |that| |one| |is| |at| |the| |goal|
   |where| |one| |began| |.|))

(define-macro-command time (ins n)
  (let ((old-time-limit (time-limit)))
    (if (equal old-time-limit n)
	ins
      `(prog2
	(lisp
	 (and
	  (error1-set (set-time-limit (quote ,n)))
	  (progn (blurb ((tm ,n) (old-t ,old-time-limit)) cr |Now| |changing| |time| |limit| |from|
			(!ppr old-t nil) |to| (!ppr tm (quote |.|)) cr)
		 t)))
	(prog1
	    ,ins
	  (lisp
	   (progn
	     (blurb ((old-t ,old-time-limit)) cr |Now| |changing|
		    |time| |limit| |back| |to| (!ppr old-t (quote |.|)) cr)
	     (set-time-limit ,old-time-limit)))))))
  (|Runs| |the| |given| |instruction| |INS| |with| |the| |prover's| |time|
   |limit| |set| |to| N |.| |After| |completing| INS |,| |the| |time| |limit|
   |is| |set| |back| |to| |what| |it| |was| |(assuming| |that| |it| |was|
   |different| |from| |N)| |.| |Note| |that| |if| |the| |run| |of| INS |is|
   |aborted| |,| |the| |user| |may| |wish| |to| |run| |(SET-TIME-LIMIT <old limit>)|
   |in| |order| |to| |restore| |the| |previous| |time| |.| |This| |instruction| |"succeeds"|
   |if| |and| |only| |if| INS |is| |run| |"successfully"| |.|))

(define-macro-command toggle-shelved-lemmas ()
  '(lisp (toggle-shelved-lemmas))
  (|Submit| |(HELP SHELVE)| |for| |an| |explanation| |of| |this| |command| |.|))

(definstr undefinstr (name)
  (lisp (undefinstr name))
  (|Same| |as| |calling| UNDEFINSTR |from| |Lisp;| |see|
   |the| |manual| |.|))

(define-macro-command undo! (&optional mark)
  (if mark
      `(orelse (undo+ (begin ,mark))
	       (do-all
		(lisp (progn (blurb ((x '(begin (,mark |...|))))
				    |#| BUT |--| |Now| |trying| |to| |find| |bookmark|
				    |of| |the| |form| (!ppr x nil) |....| cr)
			     t))
		(bind ((y (iterate for s in state-stack
				   with x
				   when (and (match (instruction s)
						    (bookmark (begin x)))
					     (consp x)
					     (eq (car x) ',mark))
				   do (return x))))
		      (if y
			  (undo+ (! `(begin ,y)))
			(lisp (blurb nil cr bookmark not found cr))))))
    (let ((i (bad-bookend-position)))
      (cond
       (i
	(blurb (i) |Undoing| |incomplete| |macro| |command| |at| |position| (!ppr i nil) |.| CR)
	`(undo ,i))
       ((setq i (last-bookend-start-position))
	(blurb (i) |Undoing| |completed| |macro| |command| |at| |position| (!ppr i nil) |.| CR)
	`(undo ,i))
       (t
	(blurb nil |**NO| |CHANGE**| |--| |No|
	       BEGIN |bookmark| |found| |.| CR)
	'fail))))
  (|UNDO!| |:|
   |Undoes| |the| |most| |recent| |"top-level"| |macro| |command| |that| |was| |wrapped|
   |inside| |bookmarks| |,| |where| |"top-level"| |means| |that| |no| |other| |wrapped|
   |macro| |command| |contains| |this| |one| |.| |Here| |we| |include| |all| |macro|
   |commands| |,| |whether| |they| |were| |carried| |to| |completion| |or| |not| |.|
   CR |#|
   |(UNDO! <symbol>)| |:| |Handy| |for| |undoing| |macro| |commands| |.|
   |Virtually| |any| |macro| |command| |named| |<symbol>| |which| |was| |defined| |using|
   WRAP |will| |be| |subject| |to| |(UNDO! <symbol>);| |try| |it| excl 
   |To| |be| |grotesquely| |detailed| |:| |this| |is| |actually|
   |the| |same| |as| |(UNDO+ (BEGIN <symbol>))|
   EXCEPT |that| |if| |(UNDO (BEGIN <symbol>))| |fails| |,|
   |then| |for| |the| |most| |recent| |instruction| |of| |the| |form|
   |(BOOKMARK (BEGIN (<symbol> ...)))| |,|
   |an| |UNDO| |back| |through| |that| |bookmark| |is| |executed| |.|))

(DEFINSTR UNDO+ (X)
  (BIND
    ((N (ITERATE FOR S IN STATE-STACK
	      AS I FROM 1
	      WHEN (EQUAL (ACCESS STATE INSTRUCTION S)
			  (LIST 'BOOKMARK (QUOTE X)))
		DO (RETURN I))))
    (IF N
	(UNDO (! N))
	(PROG2
	  (LISP (BLURB ((B (QUOTE X))) |**NO| |CHANGE**| |--| |The| |bookmark|
		       (!PPR B NIL) |was| |not| |found| |.| CR))
	  FAIL)))
  (|(UNDO+ x)| |undoes| |back| |to| |and| |including| |the| |most| |recent|
	       |instruction| |(BOOKMARK x)| |,| |if| |there| |is| |one| |,| |and|
	       |"succeeds"| |in| |case| |it| |finds| |such| |an| |instruction| |.|
	       |Notice| |that| |in| |this| |case| |,| |(UNDO x)| |only| |goes| |back|
	       |to| |,| |but| |not| |including| |,| |such| |a| |bookmark| |.|))

(definstr undo-goal (&optional gl-name)
  (bind ((name (if 'gl-name 'gl-name (goal-name))))
	(bind ((n (undo-goal-number name)))
	      (cond
	       (((= n 0)
		 (lisp (blurb nil cr no goal with that name cr))
		 fail)
		(t
		 (succeed
		  (lisp (blurb ((k n)) |  | |(UNDO| (!ppr k nil) |)| cr)))
		 (undo (! n)))))))
  (|Undoes| |to| |just| |before| |the| |creation| |of| |the| |goal|
   |with| |the| |given| |name| |.|))

(define-macro-command unify (&optional n)
  (let* ((current-term (current-term)) (hyps (hyps)) (free-variables (free-variables))
	 (index
	  (if n
	      (if (and (integerp n) (< 0 n) (<= n (length hyps))
		       (eq (cdr (nth (1- n) hyps)) 'a))
		  (or (and (unify current-term (car (nth (1- n) hyps)) free-variables)
			   n)
		      (blurb () CR |***NO| |CHANGE***| |--| |The| indicated |Hypothesis|
			     |does| |not| |unify| |with| |the| |current| |subterm| |.| CR))
		(blurb () CR |***NO| |CHANGE***| |--| |The| |argument| |to|
		       UNIFY |must| |be| |the| |index| |of| |an| |active| |top-level|
		       |hypothesis| |.| CR))
	    (iterate for hyp-flg in hyps
		     as i from 1
		     when (and (eq (cdr hyp-flg) 'a)
			       (unify current-term (car hyp-flg) free-variables))
		     do (return i)))))
    (if index
	`(put ,@(iterate for pair in unify-subst
			 collect (list (car pair) (cdr pair))))
      (progn (blurb nil CR |***NO| |CHANGE***| |--| |No| |unifier| |found| |.| CR)
	     'fail)))
  (UNIFY |:| |Creates| |a| PUT |command| |that| |unifies| |some| |active| |top-level|
	 |hypothesis| |with| |the| |current| |subterm| |,| |if| |possible;|
	 |fails| |otherwise| |.| |(UNIFY n)| |specifies| |that| |one| |should|
	 |only| |try| |this| |with| |the| |nth| |hypothesis| |.| CR |#|
	 |See| |also| BACKCHAIN |.|))

(define-macro-command unshelve (&rest names)
  `(lisp (unshelve ,@names))
  (|(UNSHELVE <name@-[1]> ... <name@-[n]>)| |:| |Eliminate| |the| |given|
   |rewrite| |rules| |from| |the| |list| |of| |shelved| |lemmas| |,| |i.e.|
   |undo| |the| |effect| |of| |the| |command| |SHELVE| |(see|
   |(HELP SHELVE)| |for| |details)| |.| |The| |current| |list| |of| |shelved|
   |lemmas| |can| |be| |seen| |using| |(PRINT SHELVED-LEMMAS)| |.| |The|
   |entire| |shelving| |feature| |can| |be| |switched| |from| |on| |to|
   |off| |and| |vice-versa| |using| |(TOGGLE-SHELVED-LEMMAS)| |.|))

(define-macro-command use (&rest hints)
  (cond
   ((not (error1-set (chk-acceptable-hints
		      (list (cons 'use
				  (sublis (abbreviations) hints))))))
    '(prog2 (lisp (blurb nil |**NO| |CHANGE**| |--| use failed excl cr))
	    fail))
   (t
    (check-use-hints-for-uninstantiated-vars hints)
    `(orelse (protect (do-strict
		       (,@(iterate for hint in hints
				   collect
				   (append (list 'use-lemma (car hint))
					   (if (cdr hint)
					       (list (cdr hint))
					     nil))))))
	     (prog2 (lisp (blurb nil |**NO| |CHANGE**| |--| use failed excl cr))
		    fail))))
  (|Like| use-lemma |,| |but| |allowing| |multiple| |hints| |and| |with|
   |a| |syntax| |similar| |to| |the| |boyer-moore| USE |hint| |.|
   |Notice| |that| |if| |any| |of| |the| use-lemma |commands| |thus| |generated|
   |actually| |fail| |,| |then| |the| |whole| |thing| |fails| |and| |the| |original| |state|
   |is| |restored| |.| cr EXAMPLE |:|
   (USE (APPEND-ASSOC (X A) (Y B)) (TIMES-0))))

(define-macro-command use-hyps (command hyp-numbers)
  (let* ((lng (length (hyps)))
	 (bad-hyps (iterate for x in hyp-numbers
			    when (not (and (integerp x)
					   (< 0 x)
					   (<= x lng)))
			    collect x)))
    (if bad-hyps
	(progn (blurb (bad-hyps lng) nil |**NO| |CHANGE**| |--| |There| |are|
		      (!ppr lng nil) |hypotheses| |,| |and| |hence| |the| |following|
		      |hypothesis| (!plural? bad-hyps (progn |numbers| |are|) (|number| |is|))
		      |unacceptable| |:| (!ppr-list bad-hyps (quote |.| CR)))
	       'fail)
      (let ((hyps-to-hide
	     (iterate for hyp-flg in (hyps)
		      as i from 1
		      when (and (eq (cdr hyp-flg) 'a)
				(not (member i hyp-numbers)))
		      collect i))
	    (hyps-to-show
	     (iterate for hyp-flg in (hyps)
		      as i from 1
		      when (and (eq (cdr hyp-flg) 'h)
				(member i hyp-numbers))
		      collect i))
	    (original-goal-name (goal-name)))
	`(do-all ,(if hyps-to-hide `(hide-hyps ,@hyps-to-hide) 'skip)
		 ,(if hyps-to-show `(show-hyps ,@hyps-to-show) 'skip)
		 ,command
		 (maybe-change-goal-if-unproved ,original-goal-name)
		 ,(if hyps-to-hide `(show-hyps ,@hyps-to-hide) 'skip)
		 ,(if hyps-to-show `(hide-hyps ,@hyps-to-show) 'skip)))))
  (|Runs| |the| |instruction| COMMAND |after| |massaging| |the| |goal| |so|
   |that| HYP-NUMBERS |are| |exactly| |the| |active| |hypothesis| |numbers| |.|
   |This| |massaging| |is| |then| |undone| |after| COMMAND |has| |been| |run| |.|))

(define-macro-command use-lemma$ (lemma &optional subst)
  (if (change-state-with-use-lemma-chk `(use-lemma ,lemma ,subst) (abbreviations))
      `(use-lemma ,lemma
		  ,(nconc
		    (iterate for v in (set-diff-eq (all-vars (formula-of lemma)) (mapcar #'car subst))
			     with fvs = (free-variables)
			     collect (list v (make-new-r$-var v fvs)))
		    subst))
    'fail)
  (|Like| USE-LEMMA |,| |except| |that| |variables| |in| |the| |lemma|
   |which| |are| |not| |explicitly| |substituted| |for| |are| |brought| |in|
   |as| |free| |variables| |in| |the| |proof| |state| |.|))

(DEFINSTR VIEW NIL
  (DO-ALL
    HYPS
    (LISP (BLURB NIL CR |The| |current| |subterm| |is| |:| CR))
    P)
  (|Print| |the| HYPS |and| |the| |current| |subterm| |.|))

(DEFINSTR VIEW-TOP NIL
  (DO-ALL
    (LISP (BLURB NIL CR))
    HYPS
    (LISP (BLURB NIL CR |The| |conclusion| |(with| |current| |subterm|
		 |highlighted)| |is| |:| CR))
    PP-TOP)
  (|Print| |the| HYPS |and| |the| |conclusion| |,| |highlighting| |the| |current| |subterm| |.|))

(definstr w (instr)
  (bind ((n (length state-stack))
	 (ss state-stack))
	(prog2 (succeed instr)
	       (if (eq ss state-stack)
		   skip
		   (comment (! (- (length state-stack) n)) instr))))
  (|Puts| |a| |comment| |at| |the| |end| |of| |the| |run| |of| INSTR |,| |showing|
   |how| |many| |primitive| |instructions| |were| |run| |in| |the| |course| |of|
   |running| INSTR |(thus| |,| |not| |including| |the| |comment| |in| |the| |tally)| |.|))

(definstr warn-when-losing-hidden-hyps (instr)
  (bind ((depends-on (depends-on)) (name (goal-name)))
	(if (iterate for hyp-flg in (hyps)
		     thereis (eq (cdr hyp-flg) 'h))
	    (do-all instr
		    (if (not (equal depends-on
				    (access goal depends-on
					    (find-goal name (all-goals (car state-stack))))))
			(lisp
			 (blurb (name) cr |**WARNING**| |--| |You| |have| |created| |at|
				|least| |new| |goal| |in| |a| |situation| |where| |there| |were|
				|hidden| |hypotheses| |in| |the| |starting| |goal| |,|
				(!PPR NAME (QUOTE |.|)) |Just| |letting| |you| |know| |in|
				|case| |you| |intended| |to| |use| |some| |of| |those| |hypotheses|
				|in| |the| |new| |goals| |,| |since| |they| |won't| |be|
				|available| |there| |.| cr))
		      skip))
	  instr))
  (|Prints| |a| |message| |any| |time| |that| |one| |creates| |at| |least| |one| |new| |subgoal|
   |of| |the| |current| |goal| |after| |execution| |of| INSTR |,| |in| |the| |case| |that| |at|
   |least| |one| |hypothesis| |of| |the| |current| |goal| |is| |hidden| |(and| |hence| |is|
   |lost| |in| |the| |subgoals)| |.|))

(DEFINSTR WRAP (NAME INSTR)
  (PROTECT
   (ORELSE
    (DO-PROTECT
     ((BOOKMARK (BEGIN NAME))
      INSTR
      (BOOKMARK (END NAME))))
    (PROG2
     (LISP (BLURB ((X (QUOTE NAME))) CR
		  |(WRAP| (!PPR X (QUOTE |)|)) |failed| |.| CR))
     FAIL)))
  (|The| |main| |difference| |between| |(WRAP instr)| |and| |instr| |is|
   |that| |in| |case| |of| |"success"| |,| |the| |former| |will|
   |insert| |two| |bookmarks| |:|
   |(BOOKMARK (BEGIN name))| |before| |the| |first| |new| |primitive| |command|
   |put| |on| |the| |state| |stack| |,| |and| |(BOOKMARK (END name))| |after|
   |the| |last| |.| |"Success"| |is| |the| |same| |for| |both| |,| |however|
   |.| |One| |other| |difference| |is| |that| |if| |instr| |does| |not|
   |"succeed"| |,| |then| |all| |state| |is| |restored| |to| |what| |it|
   |was| |before| |running| |this| |command| |(in| |the| |sense| |of|
   |PROTECT)| |.| |Use| |WRAP+| |instead| |if| |you| |don't| |want|
   |this| |restoration| |to| |take| |place| |in| |case| |of| |failure| |.|))

(DEFINSTR WRAP+ (NAME INSTR)
  (do-all
   (BOOKMARK (BEGIN NAME))
   INSTR
   (BOOKMARK (END NAME)))
  (|The| |main| |difference| |between| |(WRAP+ instr)| |and| |instr| |is|
	 |that| |in| |case| |of| |"success"| |,| |the| |former| |will|
	 |insert| |two| |bookmarks| |:|
	 |(BOOKMARK (BEGIN name))| |before| |the| |first| |new| |primitive| |command|
	 |put| |on| |the| |state| |stack| |,| |and| |(BOOKMARK (END name))| |after|
	 |the| |last| |.| |"Success"| |is| |the| |same| |for| |both| |,| |however|
	 |.| |Unlike| WRAP |,| |there| |is| |no| |unwinding| |in| |case|
	 |of| |failure| |.|))

(definstr xd () x-dumb
  (|Same| |as| X-DUMB |.|))

(DEFINSTR XH (N)
  (DO-ALL (HYP N X-DUMB)
	  FLATTEN-HYPS)
  (|Apply| |the| X-DUMB |command| |to| |hypothesis| |N| |and| |flatten| |the|
	   |and-structure| |of| |the| |resulting| |hypotheses| |.| |This|
	   |command| |should| |probably| |always| |"succeed"| |as| |long| |as|
	   |there| |is| |a| |hypothesis| |number| |N| |and| |the| |X-DUMB|
	   |command| |applies| |to| |it| |.|))

;;; PART 2:  Commands of some interest.

(definstr backchain-no-unify (index level)
  (do-all (succeed (lisp (blurb ((c '(forward* index level))) |Forward|
				|chaining| |:| (!ppr c nil) cr)))
	  (forward* index level)
	  check-possible-conjunction)
  (|Like| |BACKCHAIN| |,| |and| |subsidiary| |to| |it| |,| |except| |that| |it| |requires|
   |the| |index| |of| |the| |hypothesis| |as| |well| |as| |the| |backchaining| |level|
   |(i.e.| |how| |many| |IMPLIES| |need| |to| |be| |stripped| |off| |in| |the|
   |hypothesis)| |to| |be| |given| |explicitly| |.|))

(definstr back-goal ()
  (bind ((n (iterate for s in (cdr state-stack)
		     as i from 0
		     with current-goal-name = (goal-name)
		     when (not (equal (access goal goal-name
					      (access state goal s))
				      current-goal-name))
		     do (return i))))
	(if (null n)
	    (prog2 (lisp (blurb nil cr |**NO| |CHANGE**| |--|
				|The| |current| |goal| |has| |always|
				|been| |the| |current| |goal| |.|))
		   fail)
	  (if (= n 0)
	      (prog2 (lisp (blurb nil cr |**NO| |CHANGE**| |--|
				  |The| |current| |goal| |has| |just|
				  |become| |the| |current| |goal| |.|))
		     fail)
	    (undo (! n)))))
  (|Continually| |UNDO| |until| |this| |would| |cause| |a| |different| |goal|
   |to| |be| |the| |current| |goal| |,| |and| |stop| |there| |.| |That| |is| |,|
   |this| |command| |backs| |up| |to| |the| |last| |time| |work| |was| |begun|
   |on| |the| |current| |goal| |,| |so| |that| |this| |goal| |is| |still| |the|
   |current| |goal| |but| |one| |more| UNDO |would| |put| |one| |at| |a|
   |different| |goal| |.|))

(DEFINSTR CASE (EXP)
  (BIND
    ((TERMABB (UNTRANSLATE-ABB (CURRENT-TERM) (ABBREVIATIONS))))
    (PROG2
      (= *
	 (! (LIST 'IF (QUOTE EXP) TERMABB TERMABB))
	 0)
      (DIVE 2)))
  (|(CASE exp)| |:| |Case| |on| | exp | |at| |current| |subterm| |,|
		|leaving| |you| |in| |TRUE| |branch| |.| |That| |is| |,|
		|the| |current| |subterm| | u | |is| |replaced| |by|
		|(IF exp u u)| |,| |and| |the| |new| |current| |subterm|
		|is| |the| |second| |argument| |of| |that| |IF|
		|expression| |.|))

(definstr check-proved (instr)
  (protect
   (bind ((name (goal-name)))
	 (prog2
	  instr 
	  (lisp (equal (access goal conc
			       (find-goal name
					  (all-goals (car state-stack))))
		       true)))))
  (|The| |given| |instruction| |must| |"succeed"| |and| |the|
	 |current| |goal| |must| |be| |proved| |at| |the| |end|
	 |of| |this| |run| |.| |If| |so| |,| |then| |this|
	 |"succeeds"| |;| |otherwise| |there| |is| |no| |effect|
	 |and| |this| |"fails"| |.|))

(definstr check-proved-verbose (instr)
  (orelse (check-proved instr)
	  (prog2 (lisp (blurb ((x 'instr)) cr |**NO| |CHANGE**| |--|
			      (!ppr x nil) |failed| |to| |"succeed"|
			      |and| |prove| |the| |current| |goal| |.| cr))
		 fail)))

(definstr cond (pairs)
  (if (null 'pairs)
      fail
      (if (! (caar 'pairs))
	  (do-strict (! (cdar 'pairs)))
	  (cond (! (cdr 'pairs)))))
  (|Bears| |the| |same| |relation| |to| IF |as| |the| |Lisp| COND
	   |bears| |to| |the| Lisp IF |,| |except| |that| |there| |is|
	   |the| |usual| |extra| |layer| |of| |parentheses| |so| |that|
	   |it| |takes| |just| |one| |argument| |.|))

(DEFINSTR DO-ON-GOAL (INSTR NAME)
  (ORELSE (LISP (OR (NOT (EQUAL (GOAL-NAME) (QUOTE NAME)))
		    (EQUAL (CONC) TRUE)))
	  INSTR)
  (|Use| |this| |command| |if| |you| |consider| |yourself| |done| |(with|
	 |"success")| |when| |the| |goal| |named| NAME |is| |not| |the| |current|
	 |goal| |,| |but| |otherwise| |you| |want| |to| |run| INSTR |.|))

(DEFINSTR DO-STRICT (INSTRS)
  (IF (NULL (QUOTE INSTRS))
      SKIP
      (PROG2 (! (CAR (QUOTE INSTRS)))
	     (DO-STRICT (! (CDR (QUOTE INSTRS))))))
  (|Runs| |the| |given| |sequence| INSTRS |of| |instructions| |,| |halting|
	  |early| |once| |an| |instruction| |"fails"| |.|))

(definstr do-strict-until-proved (instr-list)
  (do-strict-until-proved-aux instr-list (! (goal-name))))

(define-macro-command =hyps (newterm &rest hyp-numbers)
  (let ((wrap-info `(=hyps ,newterm ,@hyp-numbers))
	(comm (if (and hyp-numbers
		       (not (integerp (car hyp-numbers))))
		  (pop hyp-numbers)
		'prove+)))
    `(wrap ,wrap-info
	   (use-hyps
	    (then (= * ,newterm 0)
		  (succeed ,comm))
	    ,hyp-numbers)))
  (|#| |(=HYPS newterm N1 ... Nk):| |Replace| |the| |current| |term| |with| |newterm| |,|
   |and| |(if necessary)| |create| |a| |subgoal| |stating| |the| |equality| |of| |the|
   |current| |term| |with| |newterm| |and| |prove| |it| |using| |only| |hypotheses|
   |N1, ...,| |Nk| |.| |The| |command| |PROVE+| |is| |used| |on| |the| |subgoal| |.| |However| |,|
   |the| |form| |(=HYPS newterm command N1 ... Nk)| |is| |also| |allowed|
   |(where| |command| |is| |non-numeric);| |command| |is|
   |then| |used| |to| |prove| |the| |subgoal| |instead| |of| |PROVE+| |.| CR
   |#| |Minor| |point| |:| |Notice| |that| |governors| |will| |appear| |as|
   |hypotheses| |in| |the| |subgoal| |.|))

(DEFINSTR FAIL ()
  (LISP NIL)
  (|Does| |nothing| |,| |but| |"fails"| |.|))

(DEFINSTR IF (TEST INSTR1 INSTR2)
  (BIND ((INSTR
	  (IF TEST
	      (QUOTE INSTR1)
	      (QUOTE INSTR2))))
	(DO-ALL (! INSTR)))
  (|Evaluate| |the| TEST |in| |Lisp|
   |.| |Then| |execute| INSTR1 |if| |the| |result| |is| |not| | NIL |
   |and| INSTR2 |otherwise| |.| CR |#| OBSCURE NOTES |:| CR |#| |(1)|
   |In| |fact| |,| |when| |the| TEST |is| |evaluated| |,| |this| |evaluation|
   |takes| |place| |in| |the| |environment| |obtained| |from| COMMAND-BINDINGS |,|
   |which| |is| |an| |association| |list| |set| |by| |the| BIND |command| |.|
   |Thus| |,| |for| |example| |,| |in| |something| |like|
   |(BIND ((X 3)) (IF <test> <instr1> <instr2>))| |,| |the| |test|
   |is| |evaluated| |in| |an| |environment| |where| X |is| |bound| |to| |3| |.|
   CR |#| |(2) ||Note| |that| |this| |is| |not| |"lazy"| |,|
   |i.e.| |something| |like| |(IF A (! B) (! C))| |will| |cause| |B| |and| |C|
   |to| |be| |evaluated| |,| |in| |fact| |before| A |is| |.|))

(definstr mapcar-ins (ins lst)
  (if (null 'lst)
      skip
      (do-all (! (list 'ins (car 'lst))) (mapcar-ins ins (! (cdr 'lst)))))
  (|Maps| |the| |given| |instruction| |,| INS |,| |over| |the| |list|
	  LST |,| |in| |the| |sense| |that| INS |should| |take| |one|
	  |argument| |and| |each| |member| |of| LST |successively| |becomes|
	  |that| |argument| |.| |This| |instruction| |"succeeds"| |if|
	  |and| |only| |if| |every| |such| |invocation| |of| INS
	  |"succeeds"| |.|))

(DEFINSTR MAYBE-CHANGE-GOAL (NAME)
  (IF (EQUAL (QUOTE NAME) (GOAL-NAME))
      SKIP
      (CHANGE-GOAL NAME))
  (|Change| |to| |the| |goal| |with| |the| |indicated| |name| |,| |unless| |we|
	    |are| |already| |at| |that| |goal| |.|))

(define-macro-command maybe-change-goal-if-unproved (goal-name)
  (if (and (not (equal (goal-name) goal-name))
	   (not (equal (access goal conc
			       (find-goal goal-name
					  (all-goals (car state-stack))))
		       true)))
      `(change-goal ,goal-name)
    'skip)
  (|Changes| |to| |the| |indicated| |goal| |if| |it| |is| |not| |the|
   |current| |goal| |and| |is| |not| |already| |proved| |.|))

(define-macro-command orchoose-aux (n)
  (if (> n 2)
      `(prog2 (orchoose1 2) (orchoose-aux ,(1- n)))
    (if (= n 2)
	'(orchoose1 2)
      '(orchoose1 1))))

(define-macro-command orchoose1 (n)
  (let (x y)
    (cond
     ((match (current-term) (or x y))
      (when (equal n 1)
	    (our-swap x y))
      `(do-all (claim (not ,(untranslate-abb x (abbreviations))) 0)
	       change-goal s-prop
	       (then (= * ,(untranslate-abb y (abbreviations)) 0)
		     (succeed (do-all demote-last s-prop)))))
     (t 'fail))))

(DEFINSTR ORELSE (INSTR1 INSTR2)
  (NEGATE (PROG2 (NEGATE INSTR1) (NEGATE INSTR2)))
  (|Runs| |the| |instruction| INSTR1 |.| |If| INSTR1 |"succeeds"| |,| |then|
	  |the| |command| |"succeeds"| |.| |Otherwise| |the|
	  |instruction| INSTR2 |is| |run| |,| |and| |the| |command| |"succeeds"|
	  |if| |and| |only| |if| INSTR2 |"succeeds"| |.| CR |#| |Note| |:|
	  |you| |may| |wish| |to| |use| |(ORELSE (PROTECT INSTR1) INSTR2)|
	  |in| |order| |to| |negate| |the| |effect| |of| INSTR1 |in| |case|
	  |it| |"fails"| |.|))

(definstr prog1 (ins1 ins2)
  (do-all ins1 (succeed ins2))
  (|Runs| |the| |two| |given| |instructions| |,| |and| |"succeeds"|
	  |if| |and| |only| |if| |the| |first| |instruction|
	  |"succeeds"| |.|))

(DEFINSTR PROGRESS (INSTR)
  (SET-SUCCESS ((SS0 STATE-STACK)) ((SS1 STATE-STACK)) (NEQ SS0 SS1)
	       (SUCCEED INSTR))
  (|Runs| INSTR |,| |and| |"succeeds"| |if| |and| |only| |if| |some| |change|
	  |is| |made| |in| |the| |state-stack| |during| |the| |course| |of|
	  |running| INSTR |(i.e.,| |the| |before-and-after| |state| |stacks|
	  |are| |NEQ)| |.|))

(definstr prune-raw ()
  (lisp (prune-fn))
  (|Auxiliary| |to| |PRUNE| |,| |this| |is| |the| |command| |which| |actually|
   |causes| |the| |replay| |which| |follows| |an| UNDO |.| |The| |important| |thing| |,|
   |then| |,| |about| |this| |command| |is| |that| |it| |should| |be| |executed|
   |in| |an| |environment| |in| |which| |STATE-STACK| |can| |be| |seen| |as| |arising|
   |from| |OLD-SS| |by| |way| |of| |an| UNDO |command| |.|))

(DEFINSTR REPEAT (INSTR)
  (PROG2 INSTR
	 (if (and (equal (conc) true)
		  (iterate for gl-info in (other-goals)
			   always (equal (access goal conc (car gl-info)) true)))
	     skip
	   (REPEAT INSTR)))
  (|Runs| INSTR |,| |stopping| |the| |first| |time| |it| |"fails"| |(and|
   |then| |fails| |unless| |all| |goals| |have| |been| |proved| |.|))

(definstr repeat-safe (instr)
  (succeed (repeat instr))
  (|Same| |as| REPEAT |,| |except| |that| |it| |always| |succeeds| |.|))

(definstr rewrite-second-and-finish ()
  (check-proved
   (do-strict
    ((DIVE 2)
     (progress (MAYBE-REWRITE nil))
     UP
     (S-PROP NIL))))
  (|"Succeeds"| |if| |and| |only| |if| |the| |second| |argument|
		|unconditionally| |rewrites| |to| | T | |and| |then|
		|the| |original| |term| |simplifies| |(with| |the| |use|
		|of| |(S-PROP NIL))| |so| |that| |the|
		|current| |goal| |is| |proved| |.| |Otherwise|
		|there| |is| |no| |change| |.|))

(definstr rr (rule-name)
  (wrap rr
	(bind ((rule (iterate for lemma in (all-lemmas)
			      when (eq (access rewrite-rule name lemma) 'rule-name)
			      do (return lemma))))
	      (if (null rule)
		  (prog2 (lisp (blurb nil no rewrite rule of that
				      name excl cr))
			 fail)
		(bind ((concl (access rewrite-rule concl rule)))
		      (if (not (match concl (equal & &)))
			  (lisp (blurb nil the given rule is not an equality cr))
			(bind ((rhs (fargn concl 2)))
			      (if (one-way-unify rhs (current-term))
				  (bind ((newterm
					  (sublis-var unify-subst
						      (fargn concl 1))))
					(then
					 (= * (! (untranslate-abb newterm (abbreviations))) 0)
					 (do-strict
					  ((dive 2)
					   (rewrite rule-name)
					   up
					   s-prop))))
				(lisp (blurb nil did not unify cr)))))))))
  (|Like| REWRITE |,| |but| |takes| |exactly| |one| |argument| |(the| |name| |of|
   |the| |rule)| |and| |rewrites| |in| |reverse| |,| |i.e.| |right| |to| |left| |.|
   |This| |is| |a| |lot| |less| |robust| |than| |the| |real| |REWRITE| |command| |--|
   |if| |you| |use| |it| |and| |find| |it| |inadequate| |,| |please| |send| |a| |note|
   |to| |Matt| |Kaufmann| |at| |kaufmann@cli.com| |.|))

(definstr s-lemmas-and-finish ()
  (check-proved (s lemmas))
  (|"Succeeds"| |if| |and| |only| |if| |S-PROP| |results| |in|
		|the| |current| |goal| |being| |proved| |.| |Otherwise|
		|there| |is| |no| |change| |.|))

(define-macro-command s-prop-and-finish (&optional print-flg)
  (let ((current-term (current-term)))
    (if print-flg
	`(orelse (check-proved s-prop)
		 (do-all (lisp (blurb ((current-term ',current-term))
				      CR S-PROP-AND-FINISH |failed| |to| |prove|
				      (!PPR CURRENT-TERM '|.|) CR))
			 fail))
      '(check-proved s-prop)))
  (|"Succeeds"| |if| |and| |only| |if| |S-PROP| |results| |in|
   |the| |current| |goal| |being| |proved| |.| |Otherwise|
   |there| |is| |no| |change| |.|))

(define-macro-command s-prop-nil-and-finish (&optional print-flg)
  (let ((current-term (current-term)))
    (if print-flg
	`(orelse (check-proved (s-prop nil))
		 (do-all (lisp (blurb ((current-term ',current-term))
				      CR S-PROP-NIL-AND-FINISH |failed| |to| |prove|
				      (!PPR CURRENT-TERM '|.|) CR))
			 fail))
      '(check-proved (s-prop nil))))
  (|"Succeeds"| |if| |and| |only| |if| |(S-PROP NIL)| |results| |in|
   |the| |current| |goal| |being| |proved| |.| |Otherwise|
   |there| |is| |no| |change| |.|))

(definstr s-shell (&optional constructors)
  (bind ((constrs (if 'constructors 'constructors
		      (iterate for x in shell-alist
			       collect (car x))))
	 (disabled-rules (frewrite-disabled-rules)))
	disable
	(do-all
	 (! (cons 'enable
		  (iterate for x in constrs
			   with dis = (frewrite-disabled-rules)
			   append
			   (iterate for y in (get x 'satellites)
				    when (member-eq y dis)
				    collect y)))))
	s
	enable
	(do-all
	 (! (cons 'disable disabled-rules))))
  (|Takes| |a| |list| |of| |shell| |constructors| |and| |has|
	   |the| |effect| |of| |applying| |the| S |simplification|
	   |command| |in| |an| |environment| |where| |the| |only|
	   |fast| |rewrite| |rules| |are| |those| |which| |are|
	   |satellites| |of| |the| |given| |constructors| |.|
	   |The| |default| |,| |i.e.| |the| |atomic| |command|
	   S-SHELL |,| |is| |equivalent| |to| |giving| |all| |shell|
	   |constructors| |,| |including| |those| |in| |the|
	   GROUND-ZERO |theory| |.|))

(DEFINSTR SET-SUCCESS (PRE-BINDINGS POST-BINDINGS CONDITION INSTR)
  (BIND PRE-BINDINGS
	(DO-ALL INSTR
		(BIND POST-BINDINGS
		      (LISP CONDITION))))
  (|Runs| |the| |instruction| INSTR |,| |and| |"succeeds"| |when| |the|
	  |following| |two| |conditions| |are| |both| |met| |.| |First| |,|
	  INSTR |must| |"succeed"| |.| |Second| |,| |if| |variable| |bindings|
	  |are| |made| |according| |to| PRE-BINDINGS |just| |before| INSTR |is|
	  |run| |and| |then| |also| |according| |to| POST-BINDINGS |just| |after|
	  INSTR |is| |run| |,| |the| CONDITION |should| |evaluate| |to| |non-NIL|
	  |in| |the| |extension| |of| COMMAND-BINDINGS |according| |to| |these|
	  |bindings| |.| CR |#| NOTE |:| |If| |"success"| |of| INSTR |is| |not|
	  |a| |concern| |,| |replace| |it| |with| |(SUCCEED INSTR)| |.|))

(DEFINSTR SKIP ()
  DO-ALL (|Does| |nothing| |,| |but| |"succeeds"| |.|))

(definstr thenL (ins ins-lst)
  (BIND
   ((ORIG-DEP-NAMES (ACCESS GOAL DEPENDS-ON (GOAL)))
    (GOAL-NAME (GOAL-NAME)))
   (PROG2 ins
	  (BIND
	   ((NEW-DEP-NAMES (SET-DIFF (ACCESS GOAL DEPENDS-ON
					     (FIND-GOAL GOAL-NAME
							(ALL-GOALS (CAR STATE-STACK))))
				     ORIG-DEP-NAMES)))
	   (BIND ((INSTR-TO-RUN
		   (CONS 'DO-ALL
			 (ITERATE FOR NAME IN NEW-DEP-NAMES
				  as x in 'ins-lst
				  APPEND
				  (LIST (LIST 'MAYBE-CHANGE-GOAL NAME)
					x)))))
		 (DO-ALL (! INSTR-TO-RUN))))))
  (|As| |in| |LCF| |,| |this| |should| |apply| INS |to| |the| |given| |goal| |and|
   |then| |respective| |members| |of| |INS-LST| |to| |the| |subgoals| |generated| |by| INS |.|
   |This| |may| |be| |in| |a| |pretty| |raw| |state| |.|))

(definstr tree (ins)
  (wrap tree (tree0 ins))
  (|Applies| |the| |given| |instruction| |and| |then| |recursively|
   |applies| |it| |to| |each| |of| |the| |arguments|
   |of| |the| |current| |term| |.| |It| |succeeds| |if|
   |the| |resulting| |global| |state| |i.e.| |the| |state|
   |stack)| |is| |not| |identical| |(in| |the| |sense| |of|
   |EQ)| |to| |that| |existing| |when| |this| |command|
   |is| |invoked| |.| |Use| TREE0 |if| |you| |don't| |want|
   |to| WRAP |this| |(see| HELP |for| |WRAP)| |.|))

;;; PART 3:  Commands of less general interest.

(DEFINSTR ALTERNATE (INSTR1 INSTR2)
  (PROG2-CHECK-UNPROVED
    INSTR1
    (ALTERNATE INSTR2 INSTR1)))

(definstr andsplit1 ()
  (if (match (current-term) (and & &))
      (do-all
       (dive 1)
       push
       nx)
    (prog2 push fail))
  (|Auxiliary| |to| ANDSPLIT |.|))

(DEFINE-MACRO-COMMAND ANDSPLIT1+ NIL
  (IF (MATCH (CURRENT-TERM) (AND & &))
      (LET ((COPY-HYPS (COPYABLE-HYPS (CAR STATE-STACK)))
	    (N (LENGTH (HYPS))))
	   (COND
	    (COPY-HYPS
	     `(DO-ALL COPY 
		      (DIVE 1) PUSH NX
		      CHANGE-GOAL
		      (DROP ,@COPY-HYPS)
		      CHANGE-GOAL
		      (DROP ,@(ITERATE FOR I FROM 1 TO (LENGTH COPY-HYPS)
				       COLLECT (+ N I)))))
	    (T '(DO-ALL (DIVE 1) PUSH NX))))
      '(PROG2 PUSH FAIL))
  (|Auxiliary| |to| ANDSPLIT+ |.|))

(define-macro-command backchain-lemma$ (lemma)
  (let (p q)
    (cond
     ((and (symbolp lemma)
	   (match (formula-of lemma) (implies p q))
	   (one-way-unify q (current-term)))
      (let ((subst (iterate for pair in unify-subst
			    collect (list (car pair) (cdr pair)))))
	`(use-lemma ,lemma
		    ,(nconc
		      (iterate for v in (set-diff-eq (all-vars p) (mapcar #'car subst))
			       with fvs = (free-variables)
			       collect (list v (make-new-r$-var v fvs)))
		      subst))))
     (t
      (blurb (lemma) |**NO| |CHANGE**| |--| |There| |is| |no| |lemma| |named|
	     (!ppr lemma nil) |which| |backchains| |against| |the| |current|
	     |subterm| excl cr)
      'fail)))
  (|(Used| |in| |the| |command| |SK.)| |Creates| |a| USE-LEMMA
   |command| |in| |which| |the| |conclusion| |of| |the| |lemma| |is|
   |unified| |with| |the| |current| |subterm| |,| |and| |the| |variables|
   |in| |the| |hypothesis| |which| |are| |not| |bound| |by| |the| |unifying|
   |substitution| |are| |brought| |into| |the| |proof| |state| |as| |free|
   |variables| |.| |Thus| |,| |this| |command| |assumes| |that|
   LEMMA |is| |the| |name| |of| |a|
   |formula| |of| |the| |form| |(implies p q)| |,| |where| |q| |matches| |the|
   |current| |term| |.| |We| |want| |new| |variables| |for| |the| |free|
   |variables| |in| |p| |.|))

(define-macro-command bash-protect (&optional hints)
  (let ((depends-on (access goal depends-on (goal)))
	(goal-name (access goal goal-name (goal))))
    `(orelse
      (protect
       (prog2 (bash ,@hints)
	      (lisp (equal (quote ,depends-on)
			   (access goal depends-on
				   (find-goal (quote ,goal-name) (all-goals (car state-stack))))))))
      (lisp (blurb ((goal-name (quote ,goal-name))) CR BASH |failed| |on|
		   (!PPR GOAL-NAME (QUOTE |.|)) CR))))
  (|Auxiliary| |to| DEFN |macro| |command| |.|))

(define-macro-command check-possible-conjunction ()
  (let ((conc (conc)))
    (if (and (nvariablep conc)
	     (eq (fn-symb conc) 'and))
	'(then (do-all (dive 1)
		       push
		       nx
		       push
		       top
		       s-prop)
	       check-possible-conjunction)
      '(s nil)))
  (|Should| |prove| |the| |current| |goal| |if| |one| |is| |at| |the| |top| |of|
   |the| |goal| |and| |the| |goal| |is| |either| |a| |conjunction| |all| |of|
   |whose| |conjuncts| |are| |among| |the| |active| |top-level| |hypotheses|
   |(recursively| |,| |i.e.| |including| |nested| |conjuncts),| |or| |else| |if|
   |not| |a| |conjunction| |then| |the| |current| |goal| |is| |one| |of|
   |the| |active| |top-level| |hypotheses| |.|))

(DEFINSTR CONTRADICT-DUMB (N)
  (PROG2
   (CONTRADICT N)
   (IF (MATCH (CURRENT-TERM) (NOT ?))
       SKIP
       (IF (LET (TYPE-ALIST) (OUR-BOOLEAN (CURRENT-TERM)))
	   (! `(= * (NOT (NOT ,(CURRENT-TERM))) T))
	   (LISP (BLURB NIL |Sorry| |--| |can't| |use|
			CONTRADICT-DUMB |when| |the| |hypothesis|
			|in| |question| |is| |the| |negation| |of|
			|a| |term| |which| |is| |not| |known| |to|
			|be| |boolean| |.| CR)))))
  (|Like| CONTRADICT |,| |but| |if| |the| |hypothesis| |is| |of| |the| |form|
   (NOT H) |,| |then| |it| |appears| |in| |the| |conclusion| |as|
   |(NOT (NOT H))| |rather| |than| |as| H |.|))

(DEFINSTR CONTRADICT? (N)
  (IF (MEMBER-EQUAL (CURRENT-TERM) (LIST (TRANSLATE '(NOT F)) (TRANSLATE '(NOT (NOT T)))))
      S-PROP
      (CONTRADICT N)))

(define-macro-command demote? (x)
  (if (null x)
      'skip
    (cons 'demote x)))

(define-macro-command demote-last ()
  (if (hyps)
      `(demote ,(length (hyps)))
    (progn (blurb nil cr nothing to demote excl cr)
	   'fail)))

(define-macro-command dive? (addr)
  (if addr
      `(dive ,@addr)
    'skip))

(DEFINSTR DO-PROTECT (INSTRS)
  (IF (NULL (QUOTE INSTRS))
      SKIP
      (PROG2 (PROTECT (! (CAR (QUOTE INSTRS))))
	     (DO-PROTECT (! (CDR (QUOTE INSTRS))))))
  (|Runs| |the| |given| |sequence| INSTRS |of| |instructions| |,| |halting|
	  |early| |once| |an| |instruction| |"fails"| |,| |and| |(in| |that| |case)|
	  |acts| |as| |though| |that| |failed| |instruction| |was| |not| |even|
	  |run| |.|))

(definstr do-strict-until-proved-aux (instr-list name)
  (if (null 'instr-list)
      skip
      (if (and (not (equal 'name (goal-name)))
	       (cdr state-stack)
	       (equal 'name
		      (access goal goal-name
			      (access state goal (cadr state-stack))))
	       (equal (access goal conc
			      (find-goal 'name (all-goals (car state-stack))))
		      true))
	  fail
	  (PROG2 (! (CAR 'INSTR-list))
		 (DO-STRICT-until-proved-aux (! (CDR 'INSTR-list)) name)))))

(definstr drop? ()
  (if (hyps)
      drop
    skip))

(DEFINSTR ELIM0 (VAR TERM DESTRUCTORS &OPTIONAL COMPLETION)
  (if (or (atom 'destructors)
	  (cdr (our-last 'destructors))
	  (iterate for x in 'destructors thereis (atom x)))
      (lisp (blurb ((x 'destructors)) |**NO| |CHANGE**| |--|
		   |the| |list| |of| |destructors| (!ppr x nil) |must|
		   |be| |a| |non-empty| |proper| |list| |of| |non-atoms| |.| CR))
    (BIND ((NAME (GOAL-NAME))
	   (LNG (LENGTH (HYPS)))
	   (NEWVARS
	    (SET-DIFF ELIM-VARIABLE-NAMES
		      (ALL-VARS-LST (CONS (CONC) (ASSUMPTIONS (HYPS)))))))
	  (DO-ALL TOP?
		  (THEN (CLAIM (EQUAL VAR TERM) 0)
			(SUCCEED (IF (QUOTE COMPLETION)
				     COMPLETION
				     ESTABLISH)))
		  (MAYBE-CHANGE-GOAL (! NAME))
		  TOP?
		  (GENERALIZE (! (ITERATE FOR x IN (QUOTE DESTRUCTORS)
					  for y = (car x)
					  AS ELIM-VAR IN NEWVARS
					  COLLECT (if (atom y)
						      ;; then x is just a term
						      (LIST x ELIM-VAR)
						    x))))
		  (SUCCEED (SUBV (! (FARGS (CAR (NTH LNG (HYPS)))))))
		  (DROP (! (1+ LNG))))))
  (|Same| |as| ELIM |,| |but| |without| |inserting| |bookmarks| |.|))

(DEFINSTR ESTABLISH ()
  (DO-ALL (CONTRADICT (! (LENGTH (HYPS))))
	  S)
  (|Simplify| |the| |negation| |of| |the| |last| |hypothesis| |to| TRUE
	      |(using| |the| S |command).|))

(define-macro-command forwardchain-lemma$ (lemma)
  (let (p q)
    (cond
     ((and (symbolp lemma)
	   (match (formula-of lemma) (implies (not q) (not p)))
	   (one-way-unify p (current-term)))
      (let ((subst (iterate for pair in unify-subst
			    collect (list (car pair) (cdr pair)))))
	`(use-lemma ,lemma
		    ,(nconc
		      (iterate for v in (set-diff-eq (all-vars q) (mapcar #'car subst))
			       with fvs = (free-variables)
			       collect (list v (make-new-r$-var v fvs)))
		      subst))))
     (t
      (blurb (lemma) |**NO| |CHANGE**| |--| |There| |is| |no| |lemma| |named|
	     (!ppr lemma nil) |which| |forward-chains| |against| |the| |current|
	     |subterm| excl cr)
      'fail)))
  (|Similar| |to| BACKCHAIN-LEMMA$ |(see| |documentation)| |,| |except|
   |that| |the| |current| |term| |should| |match| |the| |negation| |of|
   |the| |conclusion| |of| |the| |lemma| |.|))

(DEFINSTR HYP0 (N INSTR)
  (BIND ((ADDR (REVERSE (CURRENT-ADDR-R))))
	(PROTECT
	 (DO-STRICT
	  (TOP?
	   (CONTRADICT-DUMB N)
	   (DIVE 1)
	   (DO-ALL (! (SUBST (QUOTE (DO-ALL TOP 1)) (QUOTE TOP) (QUOTE INSTR))))
	   TOP
	   (CONTRADICT? N)
	   (IF ADDR (! (CONS 'DIVE ADDR)) SKIP))))))

(define-macro-command hyp1-is-conc ()
  `(do-strict
    ((generalize ((,(untranslate-abb (conc) (abbreviations)) ?)))
     (s-prop-nil-and-finish))))

(DEFINSTR PROG2-CHECK-UNPROVED (INSTR1 INSTR2)
  (PROG2 (SET-SUCCESS ((S0 (CAR STATE-STACK))) ((S1 (CAR STATE-STACK)))
		      (AND (EQUAL (ACCESS GOAL GOAL-NAME (ACCESS STATE GOAL S0))
				  (ACCESS GOAL GOAL-NAME (ACCESS STATE GOAL S1)))
			   (NOT (EQUAL (ACCESS GOAL CONC (ACCESS STATE GOAL S1))
				       TRUE)))
		      INSTR1)
	 INSTR2))

;; The following four macro commands are for SK*
(define-macro-command pibm-aux (flg)
  ;; flg is t for implies, nil otherwise
  (let (h1 c1 h2 c2)
    (match! (conc)
	    (implies (list & h1 h2) (list & c1 c2)))
    `(protect
      (do-strict
       ((dive 1 1)
	(add-abbreviation @h1)
	nx
	(add-abbreviation @h2)
	top
	(dive 2 1)
	(add-abbreviation @c1)
	nx
	(add-abbreviation @c2)
	top
	;; now claim negation of what's true
	,(cond
	  ((equal h1 c1)
	   '(claim (not (implies @h2 @c2)) 0))
	  ((not (equal h2 c2))
	   `(lisp (er hard nil cr PIBM-AND-OR |:| |Unexpected| |situation| |where|
			 |neither| |argument| |of| |the| |antecedent| |equals|
			 |corresponding| |argument| |of| |the| |consequent| |.| cr)))
	  ;; so (equal h2 c2)
	  (flg;; so we have IMPLIES
	   '(claim (not (implies @c1 @h1)) 0))
	  (t;; so we have AND or OR
	   '(claim (not (implies @h1 @c1)) 0)))
	change-goal
	;; now we're in a goal with an old (redundant) hyp and a tautology 
	demote-last
	drop?;; probably drop should be ok actually
	(generalize ((@h1 ?) (@h2 ?) (@c1 ?) (@c2 ?)))
	s-prop-and-finish
	(remove-abbreviations @h1 @h2 @c1 @c2)
	;; so, back to the original goal
	(contradict 2)
	(drop 2))))))

(define-macro-command pibm-if ()
  ;; flg is t for implies, nil otherwise
  (let (h1 c1 h2 c2 test1 test2)
    (match! (conc)
	    (implies (if test1 h1 h2) (if test2 c1 c2)))
    (if (not (equal test1 test2))
	(cond
	 ((and (equal h1 true) (equal c1 true)
	       (equal h2 false) (equal c2 false))
	  'pibm-if-t-case)
	 ((and (equal h1 false) (equal c1 false)
	       (equal h2 true) (equal c2 true))
	  'pibm-if-f-case)
	 (t
	  (er soft ((test1 test1) (test2 test2))
	      |Monotonicity| |proof| |failed| |because| |the|
	      |tests| (!PPR test1 nil) |and| (!PPR test2 nil)
	      |of| |the| IF |expressions| |differed| |.|)))
      `(protect
	(do-strict
	 ((dive 1 1)
	  (add-abbreviation @test1)
	  up up
	  (dive 1 2)
	  (add-abbreviation @h1)
	  nx
	  (add-abbreviation @h2)
	  top
	  (dive 2 2)
	  (add-abbreviation @c1)
	  nx
	  (add-abbreviation @c2)
	  top
	  ;; now claim negation of what's true
	  ,(cond
	    ((equal h1 c1)
	     '(claim (not (implies @h2 @c2)) 0))
	    ((not (equal h2 c2))
	     `(lisp (er hard nil cr PIBM-IF |:| |Unexpected| |situation| |where|
			|neither| |argument| |of| |the| |antecedent| |equals|
			|corresponding| |argument| |of| |the| |consequent| |.| cr)))
	    ;; so (equal h2 c2)
	    (t
	     '(claim (not (implies @h1 @c1)) 0)))
	  change-goal
	  ;; now we're in a goal with an old (redundant) hyp and a tautology 
	  demote-last
	  drop?;; probably drop should be ok actually
	  (generalize ((@h1 ?) (@h2 ?) (@c1 ?) (@c2 ?) (@test1 ?)))
	  (s-prop-and-finish t)
	  (remove-abbreviations @h1 @h2 @c1 @c2 @test1)
	  ;; so, back to the original goal
	  (contradict 2)
	  (drop 2)))))))

(define-macro-command pibm-if-f-case ()
  ;; for (implies (if a f t) (if b f t))
  `(protect
    (do-strict
     ((dive 1 1)
      (add-abbreviation @rhs)
      up nx 1
      (add-abbreviation @lhs)
      up up
      (= * (implies @lhs @rhs) 0)
      change-goal
      (generalize ((@lhs ?) (@rhs ?)))
      (remove-abbreviations @lhs @rhs)
      s-prop-and-finish))))

(define-macro-command pibm-if-t-case ()
  ;; for (implies (if a t f) (if b t f))
  `(protect
    (do-strict
     ((dive 1 1)
      (add-abbreviation @lhs)
      up nx 1
      (add-abbreviation @rhs)
      up up
      (= * (implies @lhs @rhs) 0)
      change-goal
      (generalize ((@lhs ?) (@rhs ?)))
      (remove-abbreviations @lhs @rhs)
      s-prop-and-finish))))

(define-macro-command pibm-not ()
  `(protect
    (do-strict
     ((dive 1 1)
      (add-abbreviation @h1)
      top
      (dive 2 1)
      (add-abbreviation @c1)
      top
      (claim (not (implies @c1 @h1)) 0)
      change-goal
      ;; now we're in a goal with an old (redundant) hyp and a tautology 
      demote-last
      drop?;; probably drop should be ok actually
      (generalize ((@h1 ?) (@c1 ?)))
      s-prop-and-finish
      (remove-abbreviations @h1 @c1)
      ;; so, back to the original goal
      (contradict 2)
      (drop 2)))))

(define-macro-command prove-implication-by-monotonicity ()
  (let ((conc (conc)) conc1 conc2)
    (cond
     ((equal (caar (hyps)) conc)
      'hyp1-is-conc)
     ((not (match conc (implies conc1 conc2)))
      `(lisp (blurb ((x ',conc)) PROVE-IMPLICATION-BY-MONOTONICITY
		    |encountered| |a| |non-implication| |as|
		    |the| |conclusion| |:| (!ppr x (quote EXCL)) CR)))
     ((equal (caar (hyps))
	     (fcons-term* 'implies
			  (fcons-term* 'not conc2)
			  (fcons-term* 'not conc1)))
      '(do-all (dive 1)
	       (add-abbreviation @conc1)
	       nx
	       (add-abbreviation @conc2)
	       top
	       (generalize ((@conc1 ?) (@conc2 ?)))
	       (remove-abbreviations @conc1 @conc2)
	       demote
	       s-prop-and-finish))
     (t '(do-strict
	  (prove-implication-by-monotonicity-single-step
	   prove-implication-by-monotonicity)))))
  (|Assumes| |that| |we| |are| |at| |the| |top| |of| |the| |conclusion| |and|
   |that| |the| |goal| |is| |of| |the| |form| |(IMPLIES A B) => (IMPLIES C D)| |.|
   |If| |the| |conclusion| |equals| |the| |hypothesis| |,| |then| |we're|
   |done| |.| |Otherwise| |we| |may| |assume| |C| |and| |D| |both| |have| |the|
   |same| |propositional| |function| |symbol| |and| |that| |only| |one| |of| |their|
   |respective| |args| |differs| |,| |and| |split| |accordingly| |.|
   |The| |result| |is| |a| |new| |goal| |of| |the| |form|
   |(IMPLIES A B) => (IMPLIES C' D');| |then| |we| |iterate| |till| |the|
   |goal| |is| |proved| |because| |the| |conclusion| |and| |hypothesis| |are|
   |equal| |or| |are| |contrapositives| |.|))

(define-macro-command prove-implication-by-monotonicity-single-step ()
  (let ((conc (conc)) conc1 conc2)
    (match! conc (implies conc1 conc2))
    (cond
     ((or (variablep conc1)
	  (variablep conc2)
	  (fquotep conc1)
	  (fquotep conc2)
	  (not (eq (ffn-symb conc1) (ffn-symb conc2))))
      `(lisp (blurb ((x ',conc)) PROVE-IMPLICATION-BY-MONOTONICITY-SINGLE-STEP
		    |encountered| |a| |conclusion| |which| |,| |although|
		    |an| |implication| |,| |has| |different| |function| |symbols|
		    |in| |the| |antecedent| |and| |consequent| |:|
		    (!ppr x (quote EXCL)) CR)))
     (t
      (case (ffn-symb conc1)
	    ((and or) '(pibm-aux nil))
	    (not 'pibm-not)
	    (implies '(pibm-aux t))
	    (if '(pibm-if))
	    (otherwise
	     `(lisp (blurb ((x ',conc)) PROVE-IMPLICATION-BY-MONOTONICITY-SINGLE-STEP
			   |encountered| |a| |conclusion| |which| |,| |although|
			   |an| |implication| |with| |the| |same| |function| |symbol|
			   |in| |the| |antecedent| |and| |consequent| |,|
			   |is| |not| |among| |the| |symbols| AND |,| OR |,|
			   NOT |,| |and| IMPLIES |:|
			   (!ppr x (quote EXCL)) CR)))))))
  (|Auxiliary| |to| PROVE-IMPLICATION-BY-MONOTONICITY |.|))

(define-macro-command require-new-state-else-undo (instr)
  (let ((n (length state-stack)))
    `(do-all ,instr
	     (succeed (if (= ,n (length state-stack))
			  undo
			skip))))
  (|Runs| |instr| |,| |and| |"succeeds"| |exactly| |when| |instr| |does| |.|
   |However| |,| |if| |there| |is| |not| |at| |least| |one| |new| |state|
   |created| |(more| |precisely| |,| |if| |the| |state-stack| |is| |not| |changed|
   |in| |length)| |,| |then| |an| UNDO |is| |performed| |.|))

(definstr retain-last ()
  (do-all demote-last drop? promote))

(define-macro-command run-instr-on-goals (completion goal-names)
  (cons 'do-all
	(iterate for name in goal-names
		 collect
		 `(if (equal ',name (goal-name))
		      ,completion
		    (prog2
		     (change-goal ,name)
		     (if (equal (conc) true)
			 (succeed undo)
		       (require-new-state-else-undo ,completion)))))))

(define-macro-command sk-aux ()
  (let ((current-addr (reverse (current-addr-r)))
	(conc (conc))
	(last-hyp (car (car (our-last (hyps)))))
	new newconc (last-position (length (hyps))))
    (or (match last-hyp (implies (not new) (not &)))
	(match! last-hyp (implies new &)))
    (setq newconc
	  (untranslate-abb (deposit-term conc current-addr new) (abbreviations)))
    `(do-strict
      (top?
       (claim (not ,newconc)
	      0)
       change-goal
       demote-last
       retain-last
       prove-implication-by-monotonicity
       (contradict ,(1+ last-position))
       (drop ,(1+ last-position))
       (drop ,last-position))))
  (|Finishes| |the| |job| |that| SK |started| |.| |It| |assumes| |that| |we| |are|
   |trying| |to| |prove| |(H & (q -> p)) => C(p)| |or|
   |(H & ((not p) -> (not q))) => C(p)| |(depending| |on|
   |the| |parity| |of| |p| |inside| |C)| |and| |tries| |to| |leave| |us| |with| |the|
   |goal| |of| |proving| |H => C(q)| |instead| |,| |leaving| |us| |at| |the| |top|
   |of| |the| |conclusion| |.|))

(define-macro-command sk-conc ()
  ;; Assumes we're at the top of the conclusion, and completes in that
  ;; same "configuration".
  (let ((addresses (all-defn-sk-addresses (conc))))
    (if addresses
	`(wrap sk-conc
	       (progress
		(do-all ,@(iterate for addr in addresses
				   collect `(protect (do-all
						      (dive? ,addr)
						      sk))))))
      'fail))
  (|Auxiliary| |to| |sk*| |.|))

(define-macro-command then-aux (instr completion)
  (let ((orig-dep-names (access goal depends-on (goal)))
	(goal-name (goal-name)))
    `(do-all ,instr
	     (run-instr-on-goals
	      ,completion
	      (! (set-diff (access goal depends-on
				   (find-goal ',goal-name
					      (all-goals (car state-stack))))
			   ',orig-dep-names)))))
  (|Exactly| |the| |same| |as| THEN |,| |but| |where| COMPLETION |is|
   |not| |optional| |.|))

(DEFINSTR TOP? ()
  (IF (CURRENT-ADDR-R)
      TOP
      SKIP)
  (|Move| |to| |the| |top| |of| |the| |conclusion| |if| |possible| |.|
	  |This| |command| |always| |"succeeds"| |.|))

(definstr tree-args (n ins)
  (if (> n (length (fargs (current-term))))
      skip
      (do-all
       (protect
	(do-strict
	 ((dive n)
	  (tree0 ins)
	  up)))
       (tree-args (! (1+ n)) ins))))

(definstr tree-args-search (n ins)
  (if (> n (length (fargs (current-term))))
      fail
      (orelse
       (protect
	(do-strict
	 ((dive n)
	  (tree0-search ins)
	  up)))
       (tree-args-search (! (1+ n)) ins)))
  (|Auxiliary| |to| |TREE0-SEARCH| |.|))

(definstr tree0 (ins)
  (progress
   (do-all ins
	   (if (or (variablep (current-term))
		   (fquotep (current-term))
		   (null (fargs (current-term))))
	       skip
	       (tree-args 1 ins))))
  (|Applies| |the| |given| |instruction| |and| |then| |recursively|
	     |applies| |it| |to| |each| |of| |the| |arguments|
	     |of| |the| |current| |term| |.| |It| |succeeds| |if|
	     |the| |resulting| |global| |state| |i.e.| |the| |state|
	     |stack)| |is| |not| |identical| |(in| |the| |sense| |of|
	     |EQ)| |to| |that| |existing| |when| |this| |command|
	     |is| |invoked| |.|))

(definstr tree0-search (ins)
  (orelse ins
	  (if (or (variablep (current-term))
		  (fquotep (current-term))
		  (null (fargs (current-term))))
	      fail
	      (tree-args-search 1 ins)))
  (|"Succeeds"| |if| |at| |least| |one| |outside-in| |invocation|
		|of| |ins| |succeeds| |.| |Halts| |once| |such| |a|
		|success| |occurs| |.|))
