#|

 Copyright (C) 1994 by Computational Logic, Inc.  All Rights Reserved.

 This script is hereby placed in the public domain, and therefore unlimited
 editing and redistribution is permitted.

 NO WARRANTY

 Computational Logic, Inc. PROVIDES ABSOLUTELY NO WARRANTY.  THE EVENT SCRIPT
 IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
 INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND
 PERFORMANCE OF THE SCRIPT IS WITH YOU.  SHOULD THE SCRIPT PROVE DEFECTIVE, YOU
 ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT WILL Computational Logic, Inc. BE LIABLE TO YOU FOR ANY DAMAGES,
 ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SCRIPT (INCLUDING BUT
 NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES
 SUSTAINED BY THIRD PARTIES), EVEN IF YOU HAVE ADVISED US OF THE POSSIBILITY OF
 SUCH DAMAGES, OR FOR ANY CLAIM BY ANY OTHER PARTY.
|#

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

;; This version of bm-trace runs on top of the trace package that
;; comes with akcl (version 1.456, for example, but later versions
;; should work too).  There's special handling of si::arglist to make
;; this possible.  Note that we cause an error if this is loaded in
;; any other lisp besides akcl:

(EVAL-WHEN (LOAD EVAL COMPILE)
	   #-akcl
	   (error "Unable to load the file bm-trace.lisp, because this facility is
only supported in akcl"))

;; e.g. (bm-trace (append :entry (list (length (car arglist)) (length (cadr arglist)))))

;; -- then when the equivalent of (append '(1 2) '(3 4 5 6)) is called, we should see

;;   1> <<APPEND>> '(2 4)
;;;  1> <<APPEND>> '(1 2 3 4 5 6)

#| Example
(bm-trace (append :cond (numberp (caar arglist)) :entry (car arglist) :exit (car value))
	  numberp)
Another example:
(bm-trace (execute1 :entry (car arglist) :exit (access 'instr (car (get-ss value)))))
|#

(defparameter *mods-alist*
  (if (boundp '*mods-alist*)		;so it can work with Nqthm-1987
      (if (not (assoc-eq 'bm-trace *mods-alist*))
	  (cons (cons 'bm-trace
		      "This is a facility for tracing Nqthm functions
inside R-LOOP.  It accepts roughly the same kinds of trace
specifications as does akcl, as explained in the file
\"bm-trace.doc\".  Type TRACE inside R-LOOP to use this
facility, and UNTRACE to stop tracing.")
		*mods-alist*)
	*mods-alist*)
    nil))

(defmacro bm-trace (&rest r)
  `(let ((lst 
	  (if (null (quote ,r))
	      (iterate for x in system::*trace-list*
		       with s and prefix = (string prefix-for-functions) and def
		       when (and (equal (subseq (setq s (string (car x))) 0 (min 3 (length s)))
					prefix)
				 (and (fboundp (car x))
				 (or (compiled-function-p (setq def (symbol-function (car x))))
				     (and (consp def)
					  (consp (nth 3 def))
					  (eq (car (nth 3 def)) 'sys::trace-call)
					; (LAMBDA-BLOCK block-name lambda-list (TRACE-CALL ... ))
					  ))))
		       collect (car x))
	    (mapcan #'bm-trace-one ',r))))
     (iterate for x in lst collect (intern (subseq (string x) 3)))))

(defmacro bm-untrace (&rest r)
  `(let ((lst (if (null ',r)
		  (untrace ,@(iterate for x in (bm-trace)
				      collect (pack (list prefix-for-functions x))))
		(untrace ,@(iterate for x in r collect (pack (list prefix-for-functions x)))))))
     (iterate for x in lst collect (intern (subseq (string x) 3)))))

(defun bm-trace-one (spec)
  (let ((temp (error1-set (make-bm-spec spec))))
    (if temp
	(system::trace-one (car temp))
      (progn (format t "~&Tracing of ~S aborted.~&~%" spec)
	     nil))))

(defun make-bm-symbol (x)
  (pack (list '<< x '>>)))

(defun update-key (key value list)
  (if (consp list)
      (if (eq (car list) key)
	  (cons key (cons value (cddr list)))
	(cons (car list) (update-key key value (cdr list))))
    list))

(defun remove-key (key list)
  ;; To be safe this removes all occurrences of key
  (if (consp list)
      (if (eq (car list) key)
	  (remove-key key (cddr list))
	(cons (car list) (remove-key key (cdr list))))
    list))

(defun make-bm-exit-form (symbol form)
  ;; takes a form that may contain ARGLIST and VALUE and returns the form that should
  ;; actually be the argument fed to :entry in the trace spec
  `(list (quote ,(make-bm-symbol symbol))
	 (untranslate
	  (reduce-term
	   (quote ,(translate form))
	   (list (cons 'arglist si::arglist)
		 (cons 'value (car values)))))))

(defun make-bm-entry-form (symbol form)
  ;; takes a form that may contain ARGLIST and returns the form that should
  ;; actually be the argument fed to :entry in the trace spec
  `(list (quote ,(make-bm-symbol symbol))
	 (untranslate
	  (reduce-term
	   (quote ,(translate form))
	   (list (cons 'arglist si::arglist))))))

(defun make-spec-cond-form (term)
  `(not (equal (reduce-term
		(quote ,term)
		(list (cons 'arglist si::arglist)))
	       false)))  

(defun make-bm-spec (spec)
  (cond
   ((symbolp spec)
    (make-bm-spec (list spec)))
   ((consp spec)
    (let ((name (car spec)))
      (setq spec (cons (pack (list prefix-for-functions name)) (cdr spec)))
      (when (member-eq :cond spec)
	    (setq spec
		  (update-key :cond
			      (make-spec-cond-form (translate (cadr (member-eq :cond spec))))
			      spec)))
      (when (not (member-eq :entry spec))
	    (setq spec (append spec '(:entry arglist))))
      (when (not (member-eq :exit spec))
	    (setq spec (append spec '(:exit value))))
      (when (member-eq :declarations spec)
	    (format *standard-output*
		    "~&:DECLARATIONS are not allowed with BM-TRACE, and hence are being ignored.~&")
	    (setq spec (remove-key :declarations spec)))
      (when (member-eq :entrycond spec)
	    (format *standard-output*
		    "~&:ENTRYCOND is not allowed with BM-TRACE, and hence is being ignored.~&")
	    (setq spec (remove-key :entrycond spec)))
      (when (member-eq :exitcond spec)
	    (format *standard-output*
		    "~&:EXITCOND is not allowed with BM-TRACE, and hence is being ignored.~&")
	    (setq spec (remove-key :exitcond spec)))
      (setq spec
	    (update-key :entry
			(make-bm-entry-form name (cadr (member-eq :entry spec)))
			spec))
      (setq spec
	    (update-key :exit
			(make-bm-exit-form name (cadr (member-eq :exit spec)))
			spec))))
   (t (er soft (spec) |The| |spec| (!PPR spec nil) |must| |be| |a| |symbol| |or| |a| |list|))))

;;; Let's tie this in to R-LOOP

(format t "~&NOTE:  Now redefining R-LOOP1 so that TRACE and UNTRACE have new ~%~
           meanings.  Use OLD-TRACE and OLD-UNTRACE to get the old commands.")

(defun whitespacep (char)
  (member-equal char (list #\tab #\space #\page #\newline #\linefeed #\return)))

(defun eat-whitespace ()
  (iterate while (and (listen) (or (equal (peek-char) #\)) (whitespacep (peek-char))))
	   do (read-char)))

(defun read-forms ()
  (iterate while (listen) collect
	   (read-eating-trailing-whitespace)))

(defun read-eating-trailing-whitespace ()
  (prog1 (read nil)
    (eat-whitespace)))

(defvar bm-trace-loop-first-time-flg t)

(defun print-bm-trace-loop-message ()
  (prineval (pquote (progn cr |NOTE:| |This| |is| |your| |first| |time| |using| |the| |Boyer-Moore|
			   |function| |tracing| |facility| |from| |inside| |R-LOOP| |.| |Keep| |in|
			   |mind| |that| |when| |you| |specify| |functions| |for| |tracing| |or|
			   |untracing| |,| |you| |should| |NOT| |surround| |this| |list| |with|
			   |parentheses| |.| |When| |using| |TRACE| |,| |you| |are| |allowed| |to|
			   |put| |all| |your| |function| |specs| |on| |the| |line| |with| |TRACE|
			   |or| |else| |hit| |<return>| |and| |respond| |to| |the| |prompt| |.|
			   |Please| |report| |any| |bugs| |or| |suggestions| |for| |improvements|
			   |to| |Matt| |Kaufmann| |.| CR))
	    nil 0 prove-file))

(defun bm-trace-loop (&aux temp acc)
  (when bm-trace-loop-first-time-flg
	(print-bm-trace-loop-message)
	(setq bm-trace-loop-first-time-flg nil))
  (cond
   ((setq temp
	  (progn (eat-whitespace)
		 (read-forms)))
    (print (eval `(bm-trace ,@temp)))
    (fresh-line *standard-output*))
   (t
    (when (listen) (read-line))
    (format t "~&~%Function spec(s), please (or <return> for the list of bm-traced functions):~%  ")
    (cond
     ((equal (peek-char) #\Newline)
      (format t "~%Traced Boyer-Moore functions:~%  ~S~%" (bm-trace))
      (terpri nil))
     (t
      (loop
       (setq acc
	     (append (print (eval `(bm-trace ,@(read-forms))))
		     acc))
       (when (listen) (read-line))
       (format t "~&~%Function spec(s), please (or <return> to finish):~%  ")
       (when (equal (peek-char) #\Newline)
	     (format t "~%Terminating bm-trace session.  Newly traced Boyer-Moore functions:~%  ~S.~%"
		     acc)
	     (return nil))))))))
   
(defun bm-untrace-with-prompt ()
  (when bm-trace-loop-first-time-flg
	(print-bm-trace-loop-message)
	(setq bm-trace-loop-first-time-flg nil))
  (let (temp)
     (when (listen) (read-line))
     (format t "~&~%Functions to untrace (or <return> for all): ")
     (setq temp
	   (eval `(bm-untrace ,@(read-from-string (concatenate 'string "(" (read-line) ")")))))
     (if temp
	 (print temp)
       (princ "No Boyer-Moore functions are currently traced."))
     (terpri nil)))

(eval-when (load eval)
	   (when (and bm-trace-loop-first-time-flg
		      r-loop-trace-mode)
		 (format t "~%  NOTE:  Turning off Nqthm's general-purpose tracing.~%")
		 (setq r-loop-trace-mode nil)))

(DEFUN R-LOOP1 (FORM)

;   This is the body of the iteration in R-LOOP.  FORM has been
;   read.  We process it and print the answer.  Returning T
;   means continue iterating.  Returning NIL means quit.

  (LET (ANS VAR)
    (CASE FORM
          (? (PRINC "Permitted R-LOOP commands:
    term              evaluate term and prettyprint result
    (SETQ var term)   evaluate term and let var be result
    TRACE             prompt for BM-TRACE spec(s) (or <return> for the list of traced functions)
    UNTRACE           prompt for functions to give to BM-UNTRACE
    OLD-TRACE         trace mode:  partial
    FULL-TRACE        trace mode:  full
    OLD-UNTRACE       trace mode:  off
    ABBREV            enter abbreviated output mode
    UNABBREV          exit abbreviated output mode
    ?                 print this message
    OK                exit R-LOOP" NIL)
             (TERPRI NIL)
             T)
          (OK (PRINC "Exiting R-LOOP." NIL)
              NIL)
          (TRACE (BM-TRACE-LOOP)
                 T)
          (UNTRACE (BM-UNTRACE-WITH-PROMPT)
                   T)
	  (old-TRACE (SETQ R-LOOP-TRACE-MODE (QUOTE PARTIAL))
                 (PRINC "Trace Mode:  Partial" NIL)
                 (TERPRI NIL)
                 T)
          (FULL-TRACE (SETQ R-LOOP-TRACE-MODE (QUOTE FULL))
                      (PRINC "Trace Mode:  Full" NIL)
                      (TERPRI NIL)
                      T)
          (old-UNTRACE (SETQ R-LOOP-TRACE-MODE NIL)
                   (PRINC "Trace Mode: Off" NIL)
                   (TERPRI NIL)
                   T)
          (ABBREV (SETQ R-LOOP-UNTRANSLATE-MODE T)
                  (PRINC "Abbreviated Output Mode:  On" NIL)
                  (TERPRI NIL)
                  T)
          (UNABBREV (SETQ R-LOOP-UNTRANSLATE-MODE NIL)
                    (PRINC "Abbreviated Output Mode: Off" NIL)
                    (TERPRI NIL)
                    T)
          (OTHERWISE
           (SETQ VAR NIL)
           (MATCH FORM (SETQ VAR FORM))
           (COND ((ERROR1-SET (SETQ FORM (TRANSLATE-AND-CHK FORM)))
                  (SETQ ANS (COND (R-LOOP-TRACE-MODE
                                   (REDUCE-TERM-TRACE FORM R-ALIST))
                                  (T (PROG1 (REDUCE-TERM FORM R-ALIST)
                                            (ISPACES 1 NIL)))))
                  (COND ((AND VAR (NOT (EQ ANS (QUOTE *1*FAILED))))
                         (COND ((SETQ TEMP-TEMP (ASSOC-EQ VAR R-ALIST))
                                (RPLACD TEMP-TEMP (CADR ANS)))
                               (T (SETQ R-ALIST
                                        (CONS (CONS VAR (CADR ANS))
                                              R-ALIST))))))
                  (PPRIND (COND
                           ((EQ ANS (QUOTE *1*FAILED))
                            (QUOTE (NOT REDUCIBLE)))
                           (R-LOOP-UNTRANSLATE-MODE (UNTRANSLATE ANS))
                           (T ANS))
                          1 0 NIL)
                  (TERPRI NIL)
                  T)
                 (T T))))))

;;; Finally, a patch so that I don't have to depend on akcl to handle
;;; untracing and redefinition properly.

(DEFUN KILL-DEFINITION (ATM)
  (REMPROP ATM (QUOTE SEXPR))
  (when (member-eq atm (trace))
	(eval `(untrace ,atm)))
  (FMAKUNBOUND ATM)
  (SETQ LIB-ATOMS-WITH-DEFS (REMOVE ATM LIB-ATOMS-WITH-DEFS)))
