#|

 Copyright (C) 1994 by Computational Logic, Inc.  All Rights Reserved.

 This script is hereby placed in the public domain, and therefore unlimited
 editing and redistribution is permitted.

 NO WARRANTY

 Computational Logic, Inc. PROVIDES ABSOLUTELY NO WARRANTY.  THE EVENT SCRIPT
 IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
 INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND
 PERFORMANCE OF THE SCRIPT IS WITH YOU.  SHOULD THE SCRIPT PROVE DEFECTIVE, YOU
 ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT WILL Computational Logic, Inc. BE LIABLE TO YOU FOR ANY DAMAGES,
 ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SCRIPT (INCLUDING BUT
 NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES
 SUSTAINED BY THIRD PARTIES), EVEN IF YOU HAVE ADVISED US OF THE POSSIBILITY OF
 SUCH DAMAGES, OR FOR ANY CLAIM BY ANY OTHER PARTY.
|#

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

;; This file speeds up the handling of IDENTITY by about a factor of 2.
;; The idea is simply to save the results computed by
;; CHK-ACCEPTABLE-REWRITE-LEMMA so that ADD-REWRITE-LEMMA doesn't have
;; to recompute them all.

;; an `alist' of the form (chronology . alist), except nil initially.
(defvar identity-alist nil)

(DEFUN REMOVE-IDENTITY (TERM WARN-FLG)

; This function returns the result of replacing zero or more calls of IDENTITY
; from TERM by provably equivalent subterms.

; This feature was designed and originally implemented by Matt Kaufmann in
; response to complaints from several users that Nqthm did not provide a
; mechanism by which large constants could be usefully abbreviated.

  (let (entry)

  (COND
   ((OR (VARIABLEP TERM)
        (FQUOTEP TERM))
    TERM)
   ((AND (EQ (FFN-SYMB TERM) 'IDENTITY)
	 (setq entry (find-identity-entry (fargn term 1))))
    entry)
   ((AND (EQ (FFN-SYMB TERM) 'IDENTITY)
         (NULL (ALL-VARS (FARGN TERM 1))))
    (LET* ((ARG (FARGN TERM 1))
           (VAL (REDUCE-TERM ARG NIL)))
          (COND
           ((EQ VAL (QUOTE *1*FAILED))
            (IF WARN-FLG
                (ER WARNING (TERM) |Unable|
                    |to| |evaluate| |the| |term| (!TERM TERM (QUOTE |.|))))
            TERM)
           (T
	    (add-identity-entry arg val)
            VAL))))
   ((FNNAMEP 'IDENTITY TERM)
    (CONS-TERM (FFN-SYMB TERM)
               (ITERATE FOR ARG IN (FARGS TERM)
                        COLLECT (REMOVE-IDENTITY ARG WARN-FLG))))
   (T TERM))))

(defun some-tail-eq (short long)
  (iterate for tail on long
	   thereis (eq short tail)))

(defun make-identity-alist-current ()
  (if (not (and identity-alist
		(some-tail-eq (car identity-alist) chronology)))
      (setq identity-alist (cons chronology nil))))

(defun add-identity-entry (arg val)
  (make-identity-alist-current)
  (setq identity-alist
	(list* (car identity-alist) (cons arg val) (cdr identity-alist))))

(defun find-identity-entry (arg)
  (make-identity-alist-current)
  (cdr (assoc-equal arg (cdr identity-alist))))

(DEFUN CHK-ACCEPTABLE-REWRITE-LEMMA (NAME TYPE TERM)

;   This function is FUNCALLed and therefore may not be made a MACRO.

  TYPE
; The following is certainly sound.  It's a heuristic device for
; avoiding the situation where an out-of-date "long" chronology
; is heading identity-alist.
  (setq identity-alist nil)
  (LET ((TYPE-COUNT 0)
        (COMPOUND-COUNT 0)
        (LINEAR-COUNT 0)
        (REPLACEMENT-COUNT 0))
       (ITERATE
        FOR X IN (UNPRETTYIFY (REMOVE-IDENTITY TERM T))
        WITH (TOP-FNNAME-VAR REWRITE-RULE LHS ALL-VARS-HYPS ALL-VARS-CONCL
                             MAX-TERMS LST HYPS CONCL)
        DO
        (PROGN
         (SETQ HYPS (CAR X))
         (SETQ CONCL (CDR X))
         (SETQ TOP-FNNAME-VAR (TOP-FNNAME CONCL))
         (COND ((ACCEPTABLE-TYPE-PRESCRIPTION-LEMMAP HYPS CONCL)
                (INCF TYPE-COUNT)
                T)
               ((ACCEPTABLE-COMPOUND-RECOGNIZER-LEMMAP HYPS CONCL)
                (INCF COMPOUND-COUNT)
                T)
               ((NULL TOP-FNNAME-VAR)
                (ER SOFT (NAME) (!PPR NAME NIL) |is| |an| |unacceptable|
                    REWRITE |lemma| |because| |it| |rewrites| |a| |variable|
                    |.|))
               ((EQ TOP-FNNAME-VAR (QUOTE IF))
                (ER SOFT (NAME) (!PPR NAME NIL) |is| |an| |unacceptable|
                    REWRITE |lemma| |because| |it| |rewrites| |an|
                    |IF-expression| |.|))
               ((FQUOTEP CONCL) NIL)
               ((AND (NOT NO-BUILT-IN-ARITH-FLG)
                     (OR (MATCH CONCL (NOT (LESSP & &)))
                         (MATCH CONCL (LESSP & &))))
                (SETQ LST (EXTERNAL-LINEARIZE CONCL T))
                (COND ((OR (NOT (AND LST (NULL (CDR LST))))
                           (NOT (AND (CAR LST) (NULL (CDAR LST)))))
                       (ER HARD NIL LINEARIZE |returned| |a| |list| |of| |more|
                           |than| |one| |thing| |,| |even| |though| |called|
                           |on| |a| LESSP |atom| EXCL)))
                (SETQ ALL-VARS-HYPS (ALL-VARS-LST HYPS))
                (SETQ ALL-VARS-CONCL (ALL-VARS CONCL))
                (SETQ MAX-TERMS
                      (ITERATE FOR PAIR IN (ACCESS POLY ALIST (CAR (CAR LST)))
                               WHEN
                               (AND (NVARIABLEP (CAR PAIR))
                                    (SUBSETP-EQ ALL-VARS-CONCL
                                                (UNION-EQ (ALL-VARS (CAR PAIR))
                                                          ALL-VARS-HYPS))
                                    (ITERATE FOR PAIR2 IN
                                             (ACCESS POLY ALIST
                                                     (CAR (CAR LST)))
                                             WHEN (NOT (EQ PAIR2 PAIR))
                                             NEVER
                                             (AND (< (FORM-COUNT (CAR PAIR))
                                                     (FORM-COUNT (CAR PAIR2)))
                                                  (SUBBAGP
                                                   (ALL-VARS-BAG (CAR PAIR))
                                                   (ALL-VARS-BAG
                                                    (CAR PAIR2))))))
                               COLLECT (CAR PAIR)))
                (COND ((NULL MAX-TERMS)
                       (ER SOFT (NAME) (!PPR NAME NIL) |is| |an|
                           |unacceptable| REWRITE |lemma| |because| |the|
                           |atom| |of| |its| |conclusion| |is| |a| LESSP |and|
                           |it| |cannot| |be| |handled| |by| |our| |linear|
                           |arithmetic| |package| |.| |to| |be| |acceptable|
                           |,| |at| |least| |one| |nonvariable| |addend| |of|
                           |the| |conclusion| |must| |satisfy| |two|
                           |properties.| |First| |,| |it| |must|
                           |contain| |all| |the| |variables| |of| |the| |lemma|
                           |that| |are| |not| |in| |the| |hypotheses| |.|
                           |Second| |,| |it| |must| |not| |be| |the| |case|
                           |that| |under| |every| |substitution| |,| |the|
                           |term| |is| |smaller| |than| |another| |addend| |of|
                           |the| |conclusion.| |.|)))
                (ITERATE FOR X IN MAX-TERMS
                         WHEN (NON-RECURSIVE-DEFNP (FFN-SYMB X))
                         DO (ER WARNING (NAME X (FN (FFN-SYMB X)))
                                |Note| |that| |the| |linear| |lemma|
                                (!PPR NAME NIL) |is| |being| |stored| |under|
                                |the| |term| (!TERM X (QUOTE |,|)) |which| |is|
                                |unusual| |because| (!PPR FN NIL) |is| |a|
                                |nonrecursive| |function| |symbol| |.|))
                (ITERATE
                 FOR X IN MAX-TERMS WHEN (NOT (SUBSETP-EQ ALL-VARS-HYPS
                                                          (ALL-VARS X)))
                 DO
                 (ER WARNING
                     (X NAME
                        (VARS (SET-DIFF ALL-VARS-HYPS (ALL-VARS X)))
                        (LST (ITERATE FOR HYP IN HYPS
                                      WITH VARS = (SET-DIFF ALL-VARS-HYPS
                                                            (ALL-VARS X))
                                      WHEN (INTERSECTP VARS (ALL-VARS HYP))
                                      COLLECT
                                      (PROGN (SETQ VARS
                                                   (SET-DIFF VARS
                                                             (ALL-VARS HYP)))
                                             HYP))))
                     |When| |the| |linear| |lemma| (!PPR NAME NIL) |is|
                     |stored| |under| (!TERM X NIL) |it|
                     |contains| |the| |free|
                     (PLURAL? VARS |variables| |variable|)
                     (!LIST VARS) |which| |will| |be| |chosen| |by|
                     |instantiating| |the|
                     (PLURAL? LST |hypotheses| |hypothesis|)
                     (!TERM-LIST LST (QUOTE |.|))))
                (INCF LINEAR-COUNT (LENGTH MAX-TERMS))
                T)
               (T (SETQ REWRITE-RULE (CREATE-REWRITE-RULE NAME HYPS CONCL NIL))
                  (SETQ ALL-VARS-HYPS (ALL-VARS-LST HYPS))
                  (SETQ ALL-VARS-CONCL (ALL-VARS (COND ((MATCH CONCL
                                                               (EQUAL LHS &))
                                                        LHS)
                                                       (T CONCL))))
                  (COND ((NON-RECURSIVE-DEFNP (TOP-FNNAME CONCL))
                         (ER WARNING (NAME (FN (TOP-FNNAME CONCL)))
                             |Note| |that| |the| |rewrite| |rule|
                             (!PPR NAME NIL) |will| |be| |stored| |so| |as|
                             |to| |apply| |only| |to| |terms| |with| |the|
                             |nonrecursive| |function| |symbol|
                             (!PPR FN (QUOTE |.|)))))
                  (COND ((NOT (SUBSETP-EQ ALL-VARS-HYPS ALL-VARS-CONCL))
                         (ER WARNING
                             (NAME
                              (VARS (SET-DIFF ALL-VARS-HYPS ALL-VARS-CONCL))
                              (LST  (ITERATE FOR HYP IN HYPS
                                             WITH VARS =
                                             (SET-DIFF ALL-VARS-HYPS
                                                       ALL-VARS-CONCL)
                                             WHEN (INTERSECTP VARS
                                                              (ALL-VARS HYP))
                                             COLLECT
                                             (PROGN
                                              (SETQ VARS
                                                    (SET-DIFF VARS
                                                              (ALL-VARS HYP)))
                                              HYP))))
                             |Note| |that| (!PPR NAME NIL) |contains| |the|
                             |free| (PLURAL? VARS |variables| |variable|)
                             (!LIST VARS) |which| |will| |be| |chosen| |by|
                             |instantiating| |the|
                             (PLURAL? LST |hypotheses| |hypothesis|)
                             (!TERM-LIST LST (QUOTE |.|))))
                        ((AND (ATTEMPT-TO-REWRITE-RECOGNIZER CONCL) HYPS)
                         (ER WARNING (NAME) (!PPR NAME NIL) |will| |slow|
                             |down| |the| |theorem-prover| |because| |it|
                             |will| |cause| |backward| |chaining| |on| |every|
                             |instance| |of| |a| |primitive| |type|
                             |expression| |.|)))
                  (ITERATE FOR OLD-RULE IN (GET (TOP-FNNAME CONCL)
                                                (QUOTE LEMMAS))
                           UNLESS (OR (DISABLEDP
                                       (ACCESS REWRITE-RULE NAME OLD-RULE))
                                      (META-LEMMAP OLD-RULE))
                           DO (COND ((SUBSUMES-REWRITE-RULE OLD-RULE
                                                            REWRITE-RULE)
                                     (ER WARNING
                                         (NAME (OLDNAME (ACCESS REWRITE-RULE
                                                                NAME
                                                                OLD-RULE)))
                                         |the| |previously| |added| |lemma| |,|
                                         (!PPR OLDNAME (QUOTE |,|)) |could|
                                         |be| |applied| |whenever| |the|
                                         |newly| |proposed| (!PPR NAME NIL)
                                         |could| EXCL))
                                    ((SUBSUMES-REWRITE-RULE REWRITE-RULE
                                                            OLD-RULE)
                                     (ER WARNING
                                         (NAME (OLDNAME (ACCESS REWRITE-RULE
                                                                NAME
                                                                OLD-RULE)))
                                         |the| |newly| |proposed| |lemma| |,|
                                         (!PPR NAME (QUOTE |,|))
                                         |could| |be| |applied| |whenever|
                                         |the| |previously| |added| |lemma|
                                         (!PPR OLDNAME NIL) |could| |.| CR
                                         CR))))
                  (INCF REPLACEMENT-COUNT)
                  T))))
       (IF (AND (NOT IN-ADD-SHELL0)
                (NOT IN-BOOT-STRAP-FLG)
                (NOT (AND (= REPLACEMENT-COUNT 1)
                          (= TYPE-COUNT 0)
                          (= COMPOUND-COUNT 0)
                          (= LINEAR-COUNT 0))))
           (ER WARNING
               (NAME REPLACEMENT-COUNT TYPE-COUNT COMPOUND-COUNT LINEAR-COUNT)
               |Note| |that| |the| |proposed| |lemma|
               (!PPR NAME NIL) |is| |to| |be| |stored| |as|
               (@ TYPE-COUNT) |type| |prescription|
               (PLURAL? TYPE-COUNT |rules| |rule|) |,|
               (@ COMPOUND-COUNT) |compound| |recognizer|
               (PLURAL? COMPOUND-COUNT |rules| |rule|) |,|
               (@ LINEAR-COUNT) |linear|
               (PLURAL? LINEAR-COUNT |rules| |rule|) |,| |and|
               (@ REPLACEMENT-COUNT) |replacement|
               (PLURAL? REPLACEMENT-COUNT |rules| |rule|) |.|))))
