#|

 Copyright (C) 1994 by Computational Logic, Inc.  All Rights Reserved.

 This script is hereby placed in the public domain, and therefore unlimited
 editing and redistribution is permitted.

 NO WARRANTY

 Computational Logic, Inc. PROVIDES ABSOLUTELY NO WARRANTY.  THE EVENT SCRIPT
 IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
 INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND
 PERFORMANCE OF THE SCRIPT IS WITH YOU.  SHOULD THE SCRIPT PROVE DEFECTIVE, YOU
 ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT WILL Computational Logic, Inc. BE LIABLE TO YOU FOR ANY DAMAGES,
 ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SCRIPT (INCLUDING BUT
 NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES
 SUSTAINED BY THIRD PARTIES), EVEN IF YOU HAVE ADVISED US OF THE POSSIBILITY OF
 SUCH DAMAGES, OR FOR ANY CLAIM BY ANY OTHER PARTY.
|#

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

(defparameter *mods-alist*
  (if (boundp '*mods-alist*)		;so it can work with Nqthm-1987
      (if (not (assoc-eq 'pretty-print-tight-event *mods-alist*))
	  (cons (cons 'pretty-print-tight-event
		      "A facility for improving performance by
replacing events by versions that have explicit hints.
See \"ppte.doc\" for documentation.")
		*mods-alist*)
	*mods-alist*)
    nil))

(defun make-enable-hint-for-ppte (supps definite-disables)
  (iterate for x in
	   (iterate for name in supps
		    nconc
		    (let ((ev (get name 'event)))
		      (cond
		       ((not ev)
			(er hard (name) MAKE-ENABLE-HINT-FOR-PPTE |found| |a|
			    |supporter| (!ppr name nil) |that| |has| |no| |non-NIL|
			    |event| |property| EXCL |Please| |send| |an| |email|
			    |message| |with| |a| |transcript| |of| |what| |happended|
			    |to| |Matt| |Kaufmann| |,| |kaufmann@cli.com|))
		       ((eq (car ev) 'defn)
			(list name (pack (list prefix-for-functions name))))
		       ((eq (car ev) 'add-shell)
			;; no need to enable the name itself
			(get name 'satellites))
		       (t (list name)))))
	   when (not (member-eq x definite-disables))
	   collect x))

(defun modify-hints-for-ppte (hints supps)
  (iterate for x in hints
	   with temp
	   when (not (member-eq (car x) '(enable disable enable-theory disable-theory)))
	   collect x into temp
	   finally
	   (return (append
		    (list* '(disable-theory t)
			   '(enable-theory ground-zero)
			   (let ((x (make-enable-hint-for-ppte
				     supps
				     (cdr (assoc-eq 'disable hints)))))
			     (and x (list (cons 'enable x)))))
		    temp))))

(defun ppte (name &optional stream)
  (let ((ev (get name 'event)))
    (cond
     ((and ev
	   (eq (car ev) 'prove-lemma)
	   (not (assoc-eq 'instructions (cadddr (cdr ev)))))
      (setq ev `(prove-lemma ,name ,(caddr ev) ,(cadddr ev)
			     ,(modify-hints-for-ppte (cadddr (cdr ev))
					    (data-base 'immediate-supporters name)))))
     ((and ev (eq (car ev) 'constrain))
      (setq ev `(constrain ,name ,(caddr ev) ,(cadddr ev) ,(cadddr (cdr ev))
			   ,(modify-hints-for-ppte (cadddr (cddr ev))
					  (data-base 'immediate-supporters name)))))
     ((and ev (eq (car ev) 'functionally-instantiate))
      (setq ev `(functionally-instantiate
		 ,name ,(caddr ev) ,(cadddr ev) ,(cadddr (cdr ev)) ,(cadddr (cddr ev))
		 ,(modify-hints-for-ppte (cadddr (cdddr ev))
				(data-base 'immediate-supporters name))))))
    (ppr (untranslate-event ev) stream)))

(defun dump-tight-to-file (outfile)
  (with-open-file (str outfile :direction :output)
		  (iterate for x in (reverse chronology)
			   do
			   (ppte x str)
			   (terpri str)
			   (terpri str))))
