#|

 Copyright (C) 1994 by Computational Logic, Inc.  All Rights Reserved.

 This script is hereby placed in the public domain, and therefore unlimited
 editing and redistribution is permitted.

 NO WARRANTY

 Computational Logic, Inc. PROVIDES ABSOLUTELY NO WARRANTY.  THE EVENT SCRIPT
 IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED,
 INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND
 PERFORMANCE OF THE SCRIPT IS WITH YOU.  SHOULD THE SCRIPT PROVE DEFECTIVE, YOU
 ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT WILL Computational Logic, Inc. BE LIABLE TO YOU FOR ANY DAMAGES,
 ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SCRIPT (INCLUDING BUT
 NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES
 SUSTAINED BY THIRD PARTIES), EVEN IF YOU HAVE ADVISED US OF THE POSSIBILITY OF
 SUCH DAMAGES, OR FOR ANY CLAIM BY ANY OTHER PARTY.
|#

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

(defparameter *mods-alist*
  (if (boundp '*mods-alist*)		;so it can work with Nqthm-1987
      (if (not (assoc-eq 'process-events *mods-alist*))
	  (cons (cons 'process-events
		      "Some random event-processing utilities.  See
\"process-events.doc\" for documentation.")
		*mods-alist*)
	*mods-alist*)
    nil))

(defun safe-cadr (x)
  (and (consp x)
       (consp (cdr x))
       (cadr x)))

(DEFUN POPU NIL
  ;; taken literally from Nqthm code, but included here
  ;; because this function no longer exists in Nqthm-1992.
  (SETQ UNDONE-EVENTS (CAR UNDONE-EVENTS-STACK))
  (SETQ UNDONE-EVENTS-STACK (CDR UNDONE-EVENTS-STACK))
  UNDONE-EVENTS)

(defmacro redo-defn (name &rest args)
  (let ((dependents (and (symbolp name)
			 (data-base 'all-dependents name))))
    `(progn (ubt ,name)
	    (popu)
	    (do-events
	     (cons '(defn ,name ,@args)
		   (iterate for x in (cdr undone-events)
			    with deps = ',dependents
			    collect
			    (if (and (eq (car x) 'prove-lemma)
				     (not (member (cadr x) deps)))
				(list 'add-axiom (cadr x) (caddr x) (cadddr x))
			      x)))))))

(defmacro redo-defn+ (defn-event &rest other-events)
  (let (name args)
    (match! defn-event (cons 'defn (cons name args)))
    (let ((dependents (and (symbolp name)
			   (data-base 'all-dependents name))))
      `(progn (ubt ,name)
	      (popu)
	      (do-events
	       (append ',other-events
		       (cons '(defn ,name ,@args)
			     (iterate for x in (cdr undone-events)
				      with deps = ',dependents
				      collect
				      (if (and (eq (car x) 'prove-lemma)
					       (not (member (cadr x) deps)))
					  (list 'add-axiom (cadr x) (caddr x) (cadddr x))
					x)))))))))

(defun library-filename (filename)
  (let ((pathname (pathname filename)))
    (namestring (make-pathname :host (pathname-host pathname)
			       :device (pathname-device pathname)
			       :directory (pathname-directory pathname)
			       :name (pathname-name pathname)))))

(defun make-lib-conditional (filename save? compile?)
; Do a MAKE-LIB and compile the resulting .lisp file depending on
; the value of the flags SAVE? and COMPILE?
  (cond (save? (make-lib filename compile?))))

(defun do-files (infiles &optional (save nil) (compile nil))
  ;; This does a sequence of files, stopping the first time a failed
  ;; event is encountered. If the flag SAVE is set, a library is saved
  ;; at the end of each file. If the flag SAVE is set and the flag COMPILE
  ;; is set, then the saved lisp file is compiled.
  (if (every #'probe-file infiles)
      (iterate for infile in infiles
	       do (cond ((do-file infile)
			 (make-lib-conditional (library-filename infile) save compile))
			(t (make-lib-conditional (library-filename infile) save compile)
			   (return nil)))
	       finally (return t))
      (format t "~%*** File ~a not found. ***"
	      (some (function (lambda (x) (and (not (probe-file x)) x)))
		    infiles))))

(DEFUN DO-FILE (INFILE &optional start-name)

;; patched so that one can specify the event at which one wants to begin

;   This function executes each of the event commands in the file INFILE. 
;   The events are top level forms in the file.  It prints
;   each event form to PROVE-FILE and then executes it, accumulating the total
;   event times and printing the event names to the terminal if the output is
;   going elsewhere.  It aborts if some event causes an error or fails.  It
;   prints the system configuration and the accumulated times at the end of
;   PROVE-FILE.  It returns T if all events succeeded and NIL if some failed.

  (WITH-OPEN-FILE (INSTREAM (EXTEND-FILE-NAME INFILE NIL)
                            :DIRECTION :INPUT
                            :IF-DOES-NOT-EXIST :ERROR)
    (LET (ANS FORM)
      (PROG NIL

	    (when start-name
		  (loop  (setq form (READ INSTREAM NIL A-VERY-RARE-CONS))
			 (COND ((EQ FORM A-VERY-RARE-CONS)
				(RETURN T)))
			 (when (eq (safe-cadr form) start-name)
			       (go run-form))))
            LOOP 
            (SETQ FORM (READ INSTREAM NIL A-VERY-RARE-CONS))
	    run-form
	    
            (COND ((EQ FORM A-VERY-RARE-CONS)
                   (RETURN T)))
        
;   Print out the event form to PROVE-FILE and, if PROVE-FILE is not the
;   terminal, print the name to the terminal

            (ITERPRIN 1 PROVE-FILE)
            (IPRINC EVENT-SEPARATOR-STRING PROVE-FILE)
            (ITERPRIN 2 PROVE-FILE)
            (PPRIND FORM 0 0 PROVE-FILE)
            (ITERPRI PROVE-FILE)
            (COND ((NOT (EQ PROVE-FILE NIL))
                   (IPRINC (safe-CADR FORM) NIL)))

;   Evaluate the event form

            (SETQ ANS (ERROR1-SET (EVAL FORM)))
            (COND ((NULL ANS)
                 
;   A SOFT ERROR1 occurred during the evaluation.  Perhaps we should
;   let the user edit the form, but we have no standard editor in the
;   system.
                 
                   (RETURN NIL)))

;   Recover the actual value from the CONS produced by the ERROR1-SET
;   protection

            (SETQ ANS (CAR ANS))
                 
;   Print the answer to PROVE-FILE and, if PROVE-FILE is not the terminal,
;   print a comma or the failure message, as appropriate, to the terminal
;   to indicate completion of the event.

            (ITERPRI PROVE-FILE)
            (IPRINC ANS PROVE-FILE)
            (COND ((NOT (EQ PROVE-FILE NIL))
                   (COND ((EQ ANS NIL)
                          (ITERPRI NIL)
                          (IPRINC FAILURE-MSG NIL)
                          (ITERPRI NIL))
                         (T (IPRINC (QUOTE |,|) NIL)
                            (COND ((< (OUR-LINEL NIL NIL)
                                      (IPOSITION NIL NIL NIL))
                                   (ITERPRI NIL)))))))

;   Exit if the command failed.

            (COND ((EQ ANS NIL) (RETURN NIL)))

;   Otherwise, continue looping.

            (GO LOOP)))))

(DEFUN RESUME-FILE (INFILE)

;   This function executes each of the event commands in the file INFILE
;   following the first event which is not already evaluated in the current
;   chronology. The events are top level forms in the file.  It prints
;   each event form to PROVE-FILE and then executes it, accumulating the total
;   event times and printing the event names to the terminal if the output is
;   going elsewhere.  It aborts if some event causes an error or fails.  It
;   prints the system configuration and the accumulated times at the end of
;   PROVE-FILE. 
;   If all evaluated events are successful, then RESUME-FILE returns the number
;   of evaluated events. (Resuming a file in which all events are already in
;   the chronology causes RESUME-FILE to return 0.)
;   If some evaluated event fails, then RESUME-FILE returns NIL.

  ;; Note that disables and enables are skipped (since they don't have
  ;; event names visible).

  (WITH-OPEN-FILE (INSTREAM (EXTEND-FILE-NAME INFILE NIL)
                            :DIRECTION :INPUT
                            :IF-DOES-NOT-EXIST :ERROR)
    (LET (ANS FORM resume-processing number-of-successes)
      (PROG NIL
	    (setq number-of-successes 0)
            LOOP 
            (SETQ FORM (READ INSTREAM NIL A-VERY-RARE-CONS))
            (COND ((EQ FORM A-VERY-RARE-CONS)
                   (RETURN number-of-successes)))

;   Evaluate the event only if it or a previous event is not already in the chronology.

	    (cond ((or resume-processing 
		       (setq resume-processing
			     (and (consp form)
				  (not (eq (car form) 'enable))
				  (not (eq (car form) 'disable))
				  (consp (cdr form))
				  ;; the following line was added 12/23 to prevent core dumps!
				  (symbolp (cadr form))
				  (not (get (cadr form) 'event)))))
		   
;   Print out the event form to PROVE-FILE and, if PROVE-FILE is not the
;   terminal, print the name to the terminal

		   (ITERPRIN 1 PROVE-FILE)
		   (IPRINC EVENT-SEPARATOR-STRING PROVE-FILE)
		   (ITERPRIN 2 PROVE-FILE)
		   (PPRIND FORM 0 0 PROVE-FILE)
		   (ITERPRI PROVE-FILE)
		   (COND ((NOT (EQ PROVE-FILE NIL))
			  (IPRINC (safe-CADR FORM) NIL)))

;   Evaluate the event form

		   (SETQ ANS (ERROR1-SET (EVAL FORM)))
		   (COND ((NULL ANS)
                 
;   A SOFT ERROR1 occurred during the evaluation.  Perhaps we should
;   let the user edit the form, but we have no standard editor in the
;   system.
                 
			  (RETURN NIL)))

;   Recover the actual value from the CONS produced by the ERROR1-SET
;   protection

		   (SETQ ANS (CAR ANS))
                 
;   Print the answer to PROVE-FILE and, if PROVE-FILE is not the terminal,
;   print a comma or the failure message, as appropriate, to the terminal
;   to indicate completion of the event.

		   (ITERPRI PROVE-FILE)
		   (IPRINC ANS PROVE-FILE)
		   (COND ((NOT (EQ PROVE-FILE NIL))
			  (COND ((EQ ANS NIL)
				 (ITERPRI NIL)
				 (IPRINC FAILURE-MSG NIL)
				 (ITERPRI NIL))
				(T (IPRINC (QUOTE |,|) NIL)
				   (COND ((< (OUR-LINEL NIL NIL)
					     (IPOSITION NIL NIL NIL))
					  (ITERPRI NIL)))))))

					;   Exit if the command failed.

		   (COND ((EQ ANS NIL) (RETURN NIL)))
		   (setq number-of-successes (+ 1 number-of-successes))
		   ))

;   Otherwise, continue looping.

            (GO LOOP)))))

(DEFUN RESUME-FILES (INFILES &OPTIONAL (SAVE NIL) (COMPILE NIL))
;
; Resume a list of files. That is, find the first event in the first file that is
; not a member of chronology and begin evaluating there. If that file is successfully
; processed, apply DO-FILE to the remaining files. 
;
; In the following code, FLAG has one of these values:
;
;   a natural number: RESUME-FILE was applied to the file, and all evaluated events
;                     were successful
;   T               : DO-FILE was applied to the file, and all events were successful
;   otherwise       : either RESUME-FILE or DO-FILE was applied to the file, and
;                     some event failed.

  (IF (EVERY #'PROBE-FILE INFILES)
      (ITERATE FOR INFILE IN INFILES
	       WITH EVENTS-RESUMED
	       DO (LET ((FLAG (IF EVENTS-RESUMED (DO-FILE INFILE) (RESUME-FILE INFILE))))
		    (COND ((NUMBERP FLAG)			   
			   (UNLESS (ZEROP FLAG)
				   (SETQ EVENTS-RESUMED T)
				   (MAKE-LIB-CONDITIONAL (LIBRARY-FILENAME INFILE) SAVE COMPILE)))
			  ((EQUAL FLAG T)
			   (MAKE-LIB-CONDITIONAL (LIBRARY-FILENAME INFILE) SAVE COMPILE))
			  (T (MAKE-LIB-CONDITIONAL (LIBRARY-FILENAME INFILE) SAVE COMPILE)
			     (RETURN NIL))))
	       FINALLY (RETURN T))
      (FORMAT T "~%*** File ~a not found. ***"
	      (SOME (FUNCTION (LAMBDA (X) (AND (NOT (PROBE-FILE X)) X)))
		    INFILES))))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                    THE BOOK HACK                    ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; The order of libraries doesn't matter too much to me.  But if it's not
;; too much trouble, it would be nice if as many event lists as possible
;; could be virtually order-independent, i.e. independent except for
;; duplication of definitions.  Here's a reasonably concrete proposal.
;; Suppose we allow ourselves to replace any number of occurrences of
;; DEFN in all our event lists by a macro DEFN? defined as follows.
;;   Notice that logically speaking, the effects of successfully
;; executing PROVE-LEMMA and FUNCTIONALLY-INSTANTIATE are the same.

(defun when-event-exists (criterion event-type name args)
  ;; here criterion uses the designated symbol EVENT 
  `(let ((event (get ',name 'event)))
    (if event
	(if ,criterion
	    (progn (prineval
		    (pquote (progn cr |#| |The| |event|
				   (!ppr n nil) |already| |exists|
				   |in| |a| |compatible| |form| |,|
				   |and| |hence| |we| |will| |skip|
				   |it| |here| |.| cr cr
				   |[No| |time| |triple| |needed.]| cr))
		    (list (cons 'n ',name))
		    0 prove-file)
		   ',name)
	  (er soft ((n ',name) (ev (untranslate-event event)))
	      |The| |event| |named| (!ppr n nil) |already|
	      |exists| |--| |in| |fact| |it's| cr cr |#| (!ppr ev nil) cr cr |--|
	      |and| |this| |differs| |from| |what| |you| |submitted| excl))
      (,event-type ,name ,@args))))

(defun set-equal-to-old (new old)
  ;; Assumes old is already proper with no duplicates.  Notice that we can't
  ;; use subsetp-eq or no-duplicatesp, because metalemma types aren't atoms.
  (and (or (null new)
	   (and (consp new)
		(null (cdr (our-last new)))))
       (= (length new) (length old))
       (subsetp-equal old new)
       (subsetp-equal new old)))

(defmacro prove-lemma? (name types term &optional hints)
  (when-event-exists
   `(and (member-eq (car event) '(prove-lemma functionally-instantiate
					      add-axiom))
	 (set-equal-to-old ',types  (caddr event))
	 (equal (cadddr event) (translate ',term)))
   'prove-lemma
   name
   (if hints (list types term hints) (list types term))))
	 
(defmacro lemma? (name types term &optional hints)
  `(prove-lemma? ,name ,types ,term ,(bevier-mode hints)))

(defmacro defn? (name args body &optional relation-measure-lst)
  (when-event-exists
   `(and (eq (car event) 'defn)
	 (equal (caddr event) ',args)
	 (equal (cadddr event) (translate ',body)))
   'defn
   name
   (if relation-measure-lst
       (list args body relation-measure-lst)
     (list args body))))

(defmacro constrain? (name types term witness-alist &optional hints)
  (when-event-exists
   `(and (eq (car event) 'constrain)
	 (set-equal-to-old ',types  (caddr event))
	 (equal (cadddr event) (translate ',term)))
   'constrain
   name
   (if hints
       (list types term witness-alist hints)
     (list types term witness-alist))))

(defmacro functionally-instantiate? (name types term old-name fs &optional hints)
  (when-event-exists
   `(and (member-eq (car event) '(prove-lemma functionally-instantiate
					      add-axiom))
	 (set-equal-to-old ',types  (caddr event))
	 (equal (cadddr event) (translate ',term)))
   'functionally-instantiate
   name
   (if hints
       (list types term old-name fs hints)
     (list types term old-name fs))))

(defmacro add-axiom? (name types term)
  (when-event-exists
   `(and (member-eq (car event) '(prove-lemma functionally-instantiate
					      add-axiom))
	 (set-equal-to-old ',types  (caddr event))
	 (equal (cadddr event) (translate ',term)))
   'add-axiom
   name
   (list types term)))

(defmacro add-shell? (name btm-fn-symb recognizer destructor-tuples)
  (when-event-exists
   `(equal event
	   '(add-shell ,name ,btm-fn-symb ,recognizer ,destructor-tuples))
   'add-shell
   name
   (list btm-fn-symb recognizer destructor-tuples)))

(defmacro dcl? (name args)
  (when-event-exists
   `(equal event
	   '(dcl ,name ,args))
   'dcl
   name
   (list args)))

(defmacro deftheory? (name eventnames)
  ;; courtesy of Bill Bevier
  (when-event-exists
   `(and (eq (car event) 'deftheory)
	 (set-equal-to-old ',eventnames (caddr event)))
   'deftheory
   name
   (list eventnames)))

(defun maybe-do-event (form)
  ;; Puts a question mark on the front.
  (if (and (consp form)
	   (member-eq (car form)
		      '(prove-lemma lemma defn constrain
				     functionally-instantiate add-axiom
				     add-shell dcl deftheory)))
      (cons (pack (list (car form) '?))
	    (cdr form))
    form))

(DEFUN DO?-FILE (INFILE &optional start-name)
  ;; just like do-file, but turns everything into question marks

  (WITH-OPEN-FILE (INSTREAM (EXTEND-FILE-NAME INFILE NIL)
                            :DIRECTION :INPUT
                            :IF-DOES-NOT-EXIST :ERROR)
    (LET (ANS FORM)
      (PROG NIL

	    (when start-name
		  (loop  (setq form (READ INSTREAM NIL A-VERY-RARE-CONS))
			 (COND ((EQ FORM A-VERY-RARE-CONS)
				(RETURN T)))
			 (when (eq (safe-cadr form) start-name)
			       (go run-form))))
            LOOP 
            (SETQ FORM (READ INSTREAM NIL A-VERY-RARE-CONS))
	    run-form
	    
            (COND ((EQ FORM A-VERY-RARE-CONS)
                   (RETURN T)))

	    ;; Here's the change for DO?-FILE -- put the ? in
	    (setq form (maybe-do-event form))
        
;   Print out the event form to PROVE-FILE and, if PROVE-FILE is not the
;   terminal, print the name to the terminal

            (ITERPRIN 1 PROVE-FILE)
            (IPRINC EVENT-SEPARATOR-STRING PROVE-FILE)
            (ITERPRIN 2 PROVE-FILE)
            (PPRIND FORM 0 0 PROVE-FILE)
            (ITERPRI PROVE-FILE)
            (COND ((NOT (EQ PROVE-FILE NIL))
                   (IPRINC (safe-CADR FORM) NIL)))

;   Evaluate the event form

            (SETQ ANS (ERROR1-SET (EVAL FORM)))
            (COND ((NULL ANS)
                 
;   A SOFT ERROR1 occurred during the evaluation.  Perhaps we should
;   let the user edit the form, but we have no standard editor in the
;   system.
                 
                   (RETURN NIL)))

;   Recover the actual value from the CONS produced by the ERROR1-SET
;   protection

            (SETQ ANS (CAR ANS))
                 
;   Print the answer to PROVE-FILE and, if PROVE-FILE is not the terminal,
;   print a comma or the failure message, as appropriate, to the terminal
;   to indicate completion of the event.

            (ITERPRI PROVE-FILE)
            (IPRINC ANS PROVE-FILE)
            (COND ((NOT (EQ PROVE-FILE NIL))
                   (COND ((EQ ANS NIL)
                          (ITERPRI NIL)
                          (IPRINC FAILURE-MSG NIL)
                          (ITERPRI NIL))
                         (T (IPRINC (QUOTE |,|) NIL)
                            (COND ((< (OUR-LINEL NIL NIL)
                                      (IPOSITION NIL NIL NIL))
                                   (ITERPRI NIL)))))))

;   Exit if the command failed.

            (COND ((EQ ANS NIL) (RETURN NIL)))

;   Otherwise, continue looping.

            (GO LOOP)))))

(DEFUN DO?-EVENTS (EVENTS)

  ;; same as do-events, except that maybe-do-event is applied to each form

  (LET (ANS FORM)
    (PROG NIL
          (SETQ UNDONE-EVENTS EVENTS)
          LOOP
          (COND ((NULL UNDONE-EVENTS)
                 (RETURN T)))
          (SETQ FORM (maybe-do-event (CAR UNDONE-EVENTS)))

;   Print out the event form to PROVE-FILE and, if PROVE-FILE is not the
;   terminal, print the name to the terminal

          (ITERPRIN 1 PROVE-FILE)
          (IPRINC EVENT-SEPARATOR-STRING PROVE-FILE)
          (ITERPRIN 2 PROVE-FILE)
          (PPRIND FORM 0 0 PROVE-FILE)
          (ITERPRI PROVE-FILE)
          (COND ((NOT (EQ PROVE-FILE NIL))
                 (IPRINC (safe-CADR FORM) NIL)))

;   Evaluate the event form

          (SETQ ANS (ERROR1-SET (EVAL FORM)))
          (COND ((NULL ANS)
                 
;   A SOFT ERROR1 occurred during the evaluation.  Perhaps we should
;   let the user edit the form, but we have no standard editor in the
;   system.
                 
                 (RETURN NIL)))

;   Recover the actual value from the CONS produced by the ERROR1-SET
;   protection

          (SETQ ANS (CAR ANS))
                 
;   Print the answer to PROVE-FILE and, if PROVE-FILE is not the terminal,
;   print a comma or the failure message, as appropriate, to the terminal
;   to indicate completion of the event.

          (ITERPRI PROVE-FILE)
          (IPRINC ANS PROVE-FILE)
          (COND ((NOT (EQ PROVE-FILE NIL))
                 (COND ((EQ ANS NIL)
                        (ITERPRI NIL)
                        (IPRINC FAILURE-MSG NIL)
                        (ITERPRI NIL))
                       (T (IPRINC (QUOTE |,|) NIL)
                          (COND ((< (OUR-LINEL NIL NIL)
                                    (IPOSITION NIL NIL NIL))
                                 (ITERPRI NIL)))))))

;   Exit if the command failed.

          (COND ((EQ ANS NIL) (RETURN NIL)))

;   Otherwise, continue looping.

          (SETQ UNDONE-EVENTS (CDR UNDONE-EVENTS))
          (GO LOOP))))

(defmacro cheat-on-proofs (form)
  `(let ((form ',form)
	 (saved-prove
	  (symbol-function 'prove-fn))
	 (saved-bash1
	  (if (fboundp 'bash1)
	      (symbol-function 'bash1))))
     (unwind-protect
	 (progn (setf (symbol-function 'prove-fn)
		      (symbol-function 'prove-cheat))
		(setf (symbol-function 'bash1)
		      (symbol-function 'bash1-cheat))
		,form)
       (setf (symbol-function 'prove-fn)
	     saved-prove)
       (if (fboundp 'bash1)
	   (setf (symbol-function 'bash1)
		 saved-bash1)))))

(defun prove-cheat (term)
  (declare (ignore term))
  (er warning nil |Prover| |has| |been| |told| |to| |believe|
      |everything| EXCL)
  t)

(defun bash1-cheat (term)
  ;; Gives the result of BASH and thus sets PROCESS-CLAUSES (but not
  ;; PROCESS-HIST) appropriately.
  (declare (ignore term))
  (er warning nil BASH |has| |been| |told| |to| |believe|
      |everything| EXCL)
  (setq process-clauses nil)
  t)

(defun skim-events (eventlist)
  (cheat-on-proofs (do-events eventlist)))