;;; -*-  Mode: Lisp; Package: USER; Syntax: Common-Lisp; Base: 10 -*- ;;;;

;  Copyright (C) 1994 by Matt Kaufmann and Computational Logic, Inc.
;  All Rights Reserved.

;  Copying of this file is authorized to those who have read and
;  agreed with the terms in the "Pc-Nqthm-1992 GENERAL PUBLIC SOFTWARE
;  LICENSE" at the beginning of the Pc-Nqthm file "basis-pc.lisp".

;  Pc-Nqthm Version 1992 (for use with Nqthm-1992)

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

; This file contains functions which are redefined from NQTHM.  All
; but the first three or so have to do with definitions with
; interactive termination proofs.

(DEFUN ADD-SUB-FACT (ATM PROP VAL TUPLE INIT)

;   Here is the spec for ADD-SUB-FACT.  It takes 5 args ATM PROP VAL TUPLE and
;   INIT but only a few of these make sense in combination.  To store a new
;   fact you call ADD-SUB-FACT using ATM PROP and VAL.  If PROP is a variable
;   declared below we either CONS VAL to the front of PROPs top level value or
;   set PROP to VAL depending on whether PROP is ADDITIVE or SINGLE.  SET is
;   used in both cases.  If PROP is DCELL it means PUTD1 ATM to VAL.
;   Otherwise, PROP had better be an additive or single property declared below
;   and if so the appropriate ADDITIVE or SINGLE PUT1 is done.  If you want to
;   delete a previously added fact you call ADD-SUB-FACT with all args but
;   TUPLE NIL.  TUPLE should be the undo tuple produced by the adding of the
;   fact in question.  Before you begin to add or sub facts you must first
;   initialize the library file.  You do that by calling ADD-SUB-FACT with INIT
;   T and all other args NIL.  The initialization sets LIB-PROPS to the list of
;   properties declared below in the order of declaration -- making the first
;   property declared the one of highest priority.  Because the list of
;   declarations is used to generated LIB-PROPS you must include in it all of
;   the properties used by the event level abstraction itself, even those these
;   properties aren't technically yours.  These properties are IDATE SATELLITES
;   MAIN-EVENT EVENT, SEXPR and LOCAL-UNDO-TUPLES.  They should all be declared
;   with type HIDDEN rather than ADDITIVE or SINGLE.  The code will cause an
;   error if you leave out any built-in prop or use HIDDEN on any nonbuilt-in
;   one -- the whole purpose of your knowing about these properties and the
;   token HIDDEN is just to allow you to specify where in the list of
;   priorities they should be kept.  The other thing that initialization does
;   is set all variables declared below to NIL.  The HIDDEN variable CHRONOLOGY
;   should be declared explicitly.  We force you to do that so you'll always
;   remember we've claimed that variable name.

;   No property or variable name may contain lower case letters or be NIL.  If
;   this convention is violated the code produced for ADD-SUB-FACT is garbage
;   because we generate the code with SUBST's that hit lower case names and we
;   sometimes generate SELECTQs with NIL first elements of clauses.

;   For ADDITIVE data you must supply a form, which may involve VAL as a free
;   var, for computing from VAL some datum to be stored in the undo tuple.
;   This datum must be sufficient for distinguishing that VAL from all others
;   in that ADDITIVE pot.  In particular, to find the VAL in question the
;   undoing mechanism scans the pot and evaluates the form again for each
;   entry, with VAL bound to the entry, and removes from the pot the first
;   entry for which that form computes an EQUAL datum.  The form in question
;   must not contain any free variables other than VAL and must not cause any
;   side-effects.

  (ADD-SUB-FACT-BODY (TYPE-PRESCRIPTION-LST ADDITIVE PROPERTY (CAR VAL))
                     (LEMMAS ADDITIVE PROPERTY (ACCESS REWRITE-RULE NAME VAL))
                     (TOTALP-LST ADDITIVE PROPERTY (CAR VAL))
                     (LINEAR-LEMMAS ADDITIVE PROPERTY
                                    (ACCESS LINEAR-LEMMA NAME VAL))
                     (QUICK-BLOCK-INFO SINGLE PROPERTY)
                     (SDEFN SINGLE PROPERTY)
                     (LISP-CODE SINGLE PROPERTY)
                     (SUBRP SINGLE PROPERTY)
                     (BODY SINGLE PROPERTY)
                     (ERRATICP SINGLE PROPERTY)
                     (TYPE-RESTRICTIONS SINGLE PROPERTY)
                     (INDUCTION-MACHINE SINGLE PROPERTY)
                     (LEVEL-NO SINGLE PROPERTY)
                     (JUSTIFICATIONS SINGLE PROPERTY)
                     (IDATE HIDDEN PROPERTY)
                     (ELIMINATE-DESTRUCTORS-SEQ SINGLE PROPERTY)
                     (ELIMINATE-DESTRUCTORS-DESTS SINGLE PROPERTY)
                     (CONTROLLER-POCKETS SINGLE PROPERTY)
                     (SATELLITES HIDDEN PROPERTY)
                     (MAIN-EVENT HIDDEN PROPERTY)
                     (IMMEDIATE-DEPENDENTS0 ADDITIVE PROPERTY VAL)
                     (EVENT HIDDEN PROPERTY)
                     (LOCAL-UNDO-TUPLES HIDDEN PROPERTY)
                     (SEXPR HIDDEN PROPERTY)

;   SEXPR is used to remember the defn of 1fns and is managed by PUTD1 and
;   friends.  It is never hit with ADD-FACT and the 1fn never finds its way
;   onto LIB-ATOMS-WITH-PROPS.  We include it here to remind ourselves that
;   this property name is used already.

                     (NONCONSTRUCTIVE-AXIOM-NAMES ADDITIVE VARIABLE VAL)
                     (*1*BTM-OBJECTS ADDITIVE VARIABLE VAL)
                     (RECOGNIZER-ALIST ADDITIVE VARIABLE VAL)
                     (TRUE-COMPOUND-RECOGNIZER-ALIST ADDITIVE VARIABLE
                                                     (CDDR VAL))
                     (FALSE-COMPOUND-RECOGNIZER-ALIST ADDITIVE VARIABLE
                                                      (CDDR VAL))
                     (SHELL-ALIST ADDITIVE VARIABLE VAL)
                     (SINGLETON-TYPE-SETS ADDITIVE VARIABLE VAL)
                     (GENERALIZE-LEMMAS ADDITIVE
                                        VARIABLE
                                        (ACCESS GENERALIZE-LEMMA NAME VAL))
                     (SHELL-POCKETS ADDITIVE VARIABLE VAL)
                     (DEFINED-FUNCTIONS-TOGGLED ADDITIVE VARIABLE VAL)
                     (DISABLED-LEMMAS ADDITIVE VARIABLE VAL)
                     (*1*THM-MODE$ SINGLE VARIABLE VAL)
                     (FREE-VARS-IN-REWRITE-RHSS ADDITIVE VARIABLE VAL)
		     ;;; here's the new one
		     (STATE-STACK-LIST ADDITIVE VARIABLE (CAR VAL))
                     (CHRONOLOGY HIDDEN VARIABLE)))

(DEFEVENT PROVE-LEMMA (NAME TYPES TERM &OPTIONAL HINTS)
  (LET ((NAME NAME)
	(TYPES TYPES)
	(TERM TERM)
	(HINTS HINTS))
       
       ;; Here are the cases applicable for interactive hints.
       ;; 1. No PREVIOUS-EVENT (i.e., fresh session) -- the following are not
       ;;    mutually exclusive, but they are exhaustive (i.e., if there is no
       ;;    PREVIOUS-EVENT hint and the lemma is T then there must be a
       ;;    START-GOAL hint).
       ;;    a. Lemma to "prove" is not simply T
       ;;    b. There is a START-GOAL
       ;; 2. There is a PREVIOUS-EVENT, and no START-GOAL hint is present.
       
       ;; The following version of PROVE-LEMMA only adds an entry to the STATE-STACK-LIST when
       ;; the goal MAIN is not proved.
  
       (EVENT-COMMAND
	(LIST (QUOTE PROVE-LEMMA) NAME TYPES TERM HINTS)
	(LET ((IN-PROVE-LEMMA-FLG T) PROVE-ANS FINISH-FLG)
	     (MATCH! (CHK-ACCEPTABLE-LEMMA NAME TYPES TERM)
		     (LIST NAME TYPES TERM))
	     (COND
	      ((INTERACTIVE-HINTS-P HINTS)
	       (SETQ FINISH-FLG (NOT (EQUAL TERM TRUE)))
	       (MATCH! (CHK-ACCEPTABLE-INTERACTIVE-HINTS HINTS FINISH-FLG)
		       (LIST HINTS))
	       (COND
		((and REPLAY-FLG
                      (null petitio-principii))
		 ;; REPLAY-FLG is set globally to T, but it is locally set to NIL upon
		 ;; invocation of the EXIT command of the proof checker, or more
		 ;; precisely, by the code in FINISH-INTERACTIVELY-REWRITE which deals
		 ;; with the user's response to the query generated by the EXIT command.
		 (REPLAY-FOR-PROVE-LEMMA TERM HINTS)))
	       (COND
		((OR petitio-principii
                     (AND (NULL FINISH-FLG)
			  (INT= (LENGTH (NEW-PART-OF-STATE-STACK STATE-STACK))
				(LENGTH (ASSOC-EQ (QUOTE INSTRUCTIONS) HINTS))))
		     (ITERATE FOR GL IN (ALL-GOALS (CAR STATE-STACK))
			      ALWAYS (EQUAL (ACCESS GOAL CONC GL) TRUE)))
		 (MAKE-EVENT NAME 
			     (LIST (QUOTE PROVE-LEMMA)
				   NAME TYPES TERM HINTS))
		 (ADD-LEMMA0 NAME TYPES TERM)
		 (COND ((and (null petitio-principii)
                             (NULL FINISH-FLG))
			(ADD-FACT NIL (QUOTE STATE-STACK-LIST)
				  (CONS NAME
					(CONS
					 (MAKE-NEW-RECORD
					  (QUOTE STATE)
					  (LIST (CONS (QUOTE INSTRUCTION) `(START ,NAME)))
					  (CAR STATE-STACK))
					 STATE-STACK)))))
		 (DEPEND NAME
			 (UNION-EQ
			  (IF (ASSOC-EQ (QUOTE PREVIOUS-EVENT) HINTS)
			      (LIST (CADR (ASSOC-EQ (QUOTE PREVIOUS-EVENT) HINTS)))
			      NIL)
			  (UNION-EQ (and (null petitio-principii)
                                         (ACCESS STATE CUMULATIVE-LEMMAS-USED
                                                 (CAR STATE-STACK)))
				    (ALL-FNNAMES TERM))))
		 NAME)
		(T
		 (PRINT-INTERACTIVE-PROOF-FAILURE NAME))))
	      (T
	       (MATCH! (CHK-ACCEPTABLE-HINTS HINTS)
		       (LIST HINTS))
	       (UNWIND-PROTECT
		;; the rest is literally nqthm code
         (PROGN
            
;   Before calling PROVE we call APPLY-HINTS.  APPLY-HINTS sets some global
;   variables that affect the theorem-prover.  We enter an UNWIND-PROTECT here
;   so that we can set those variables to their standard default values no
;   matter how we exit PROVE.
            
           (SETQ PROVE-ANS
                 (PROVE-FN (APPLY-HINTS HINTS TERM)))
           (COND (PROVE-ANS
                  (MAKE-EVENT NAME (COND (HINTS
                                          (LIST (QUOTE PROVE-LEMMA)
                                                NAME TYPES TERM HINTS))
                                         (T (LIST (QUOTE PROVE-LEMMA)
                                                  NAME TYPES TERM))))
                  (ADD-LEMMA0 NAME TYPES TERM)
                  (DEPEND NAME
                          (UNION-EQ ALL-LEMMAS-USED
                                    (UNION-EQ
                                     (EXTRACT-DEPENDENCIES-FROM-HINTS
                                      HINTS)
                                     (ALL-FNNAMES TERM))))))
           (COND (PROVE-ANS NAME)
                 (T NIL)))
       (PROGN (ITERATE FOR X IN HINT-VARIABLE-ALIST
                       DO (SET (CADR X) (CADDDR X)))
              (SETQ LOCAL-DISABLEDP-HASH-FLAG NIL)))))))))

(DEFUN MAKE-LIB (FILE &OPTIONAL COMPILE-FLG)
; patched to make STATE-STACK-LIST benign in space consumption -- thus, the
; only change is the surrounding LET form below
  (CHK-COMMAND-STATUS NIL)
  (COND (PETITIO-PRINCIPII
         (RETURN-FROM MAKE-LIB T)))
 (let ((state-stack-list (saved-state-stack-list state-stack-list)))
  (LET ((*READ-BASE* 10)
        (*PRINT-BASE* 10)
        (*READTABLE* (COPY-READTABLE NIL))
        *PRINT-RADIX*
        *PRINT-PRETTY*
        *PRINT-LEVEL*
        *PRINT-LENGTH*
        (*PRINT-CASE* :UPCASE)
        TEMP PROP-FILE FN-FILE PROP-FILE-NAME FN-FILE-NAME
        (DATE (IDATE)))
    (CHK-INIT)
    (PRINC (FORMAT NIL "~%Making the lib for ~s.~%" FILE) PROVE-FILE)
    (ITERPRI PROVE-FILE)
    (SETQ PROP-FILE (OPEN (EXTEND-FILE-NAME FILE FILE-EXTENSION-LIB)
                          :DIRECTION :OUTPUT
                          :IF-EXISTS :RENAME-AND-DELETE))
    (FORMAT PROP-FILE
            ";;; -*-  Mode: Lisp; Package: USER; Syntax: Common-Lisp; ~
            Base: 10 -*- ;;;; ~%(IN-PACKAGE \"USER\")~2%")

    (SETQ PROP-FILE-NAME
          (NAMESTRING (TRUENAME PROP-FILE)))
    (TERPRI PROP-FILE)
    (PRINC (QUOTE |;   Creation date:  |) PROP-FILE)
    (PRINT-IDATE DATE PROP-FILE)
    (TERPRI PROP-FILE)
    (PRINC (QUOTE |;   Created as:  |) PROP-FILE)
    (PRINC PROP-FILE-NAME PROP-FILE)
    (TERPRI PROP-FILE)
    (TERPRI PROP-FILE)
    (PRINC "(EVAL-WHEN (LOAD EVAL COMPILE)
           (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))"
           PROP-FILE)
    (TERPRI PROP-FILE)
    (PRINT (LIST (QUOTE INIT-LIB) (KWOTE LIB-PROPS) (KWOTE LIB-VARS))
           PROP-FILE)
    (PRINT (LIST (QUOTE SETQ) (QUOTE LIB-DATE) DATE) PROP-FILE)
    (PRINT (LIST (QUOTE SETQ) (QUOTE LIB-VERSION)
                 (KWOTE THEOREM-PROVER-LIBRARY-VERSION))
           PROP-FILE)
    (ITERATE FOR VAR IN LIB-VARS
             DO (PRINT
                 (LIST (QUOTE SETQ) VAR (KWOTE (SYMBOL-VALUE VAR)))
                 PROP-FILE))
    (PRINT (LIST (QUOTE SETQ)
                 (QUOTE LIB-ATOMS-WITH-PROPS)
                 (KWOTE LIB-ATOMS-WITH-PROPS))
           PROP-FILE)
    (PRINT (LIST (QUOTE SETQ)
                 (QUOTE LIB-ATOMS-WITH-DEFS)
                 (KWOTE LIB-ATOMS-WITH-DEFS))
           PROP-FILE)
    (ITERATE FOR ATM IN LIB-ATOMS-WITH-PROPS DO
             (PRINT
              (LIST
               (QUOTE PUT1-LST)
               (KWOTE ATM)
               (KWOTE
                (ITERATE FOR PROP IN LIB-PROPS
                         NCONC
                         (PROGN
                           (SETQ TEMP (GET ATM PROP A-VERY-RARE-CONS))  
                           (COND ((NOT (EQ A-VERY-RARE-CONS TEMP))
                                  (LIST PROP TEMP)))))))
              PROP-FILE))
    (ITERATE FOR ATM IN (REVERSE LIB-ATOMS-WITH-DEFS) WITH TEMP DO
             (PROGN 
               (SETQ TEMP
                     (GET ATM
                          (QUOTE SEXPR)))
               (PRINT
                (LIST (QUOTE PUT1-LST)
                      (KWOTE ATM)
                      (KWOTE (LIST (QUOTE SEXPR)
                                   (LIST (QUOTE LAMBDA)
                                         (CADR TEMP)
                                         (CADDR TEMP)))))
                PROP-FILE)))
    (CLOSE PROP-FILE)
    (SETQ FN-FILE (OPEN (EXTEND-FILE-NAME FILE FILE-EXTENSION-LISP)
                        :DIRECTION :OUTPUT
                        :IF-EXISTS :RENAME-AND-DELETE))
    (FORMAT FN-FILE
            ";;; -*-  Mode: Lisp; Package: USER; Syntax: Common-Lisp; Base: ~
             10 -*- ;;;; ~%(IN-PACKAGE \"USER\")~2%")
    (SETQ FN-FILE-NAME
          (NAMESTRING (TRUENAME FN-FILE)))
    (TERPRI FN-FILE)
    (PRINC (QUOTE |;   Creation date:  |) FN-FILE)
    (PRINT-IDATE DATE FN-FILE)
    (TERPRI FN-FILE)
    (PRINC (QUOTE |;   Created as:  |) FN-FILE)
    (PRINC FN-FILE-NAME FN-FILE)
    (TERPRI FN-FILE)
    (PRINC (QUOTE |;   Associated file:  |) FN-FILE)
    (PRINC PROP-FILE-NAME FN-FILE)
    (TERPRI FN-FILE)
    (PRINC "(EVAL-WHEN (LOAD EVAL COMPILE)
           (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))"
           FN-FILE)
    (TERPRI FN-FILE)
    (PRINT `(COND ((NOT (EQUAL LIB-DATE ,DATE))
                   (ERROR1 (QUOTE (PROGN |Attempt| |to| |load| |the|
                                         |wrong| |.lisp| |or| |compiled|
                                         |file| EXCL |the| |variable|
                                         |LIB-DATE| |has| |the| |value|
                                         (!PPR LIB-DATE NIL) |,|
                                         |but| |the| |date| |in| |the|
                                         |.lisp| |or| |compiled|
                                         |file| |is| (!PPR DATE NIL) |.|
                                         ))
                           (LIST (CONS 'LIB-DATE LIB-DATE)
                                 (CONS 'DATE ,DATE))
                           (QUOTE HARD))))
           FN-FILE)
    (ITERATE FOR ATM IN (REVERSE LIB-ATOMS-WITH-DEFS) WITH TEMP DO
             (PROGN (SETQ TEMP (GET ATM (QUOTE SEXPR)))
                    (PRINT (LIST (QUOTE DEFUN)
                                 ATM
                                 (CADR TEMP)
                                 (CADDR TEMP))
                           FN-FILE)))
;  Once upon a time we did this, but changed it for Release 7.  See
;  NOTE-LIB1.

;   (PRINT (LIST (QUOTE COMPILE-IF-APPROPRIATE-AND-POSSIBLE)
;                    (QUOTE LIB-ATOMS-WITH-DEFS))
;              FN-FILE)

    (CLOSE FN-FILE)
    (SETQ LIB-DATE DATE)
    (COND (COMPILE-FLG
           (PRINC
            (FORMAT NIL "~%Compiling the lib for ~s.~%" FILE)
            PROVE-FILE)
           (ITERPRI PROVE-FILE) 
           (COMPILE-FILE FN-FILE-NAME)
           (LOAD (EXTEND-FILE-NAME FILE FILE-EXTENSION-BIN))
           (PRINC
            (FORMAT NIL "~%Finished compiling the lib for ~s.~%" FILE)
            PROVE-FILE)
           (ITERPRI PROVE-FILE)))
    (SETQ LIB-FILE (EXTEND-FILE-NAME FILE FILE-EXTENSION-LIB))
    (PRINC (FORMAT NIL "~%Finished making the lib for ~s.~%" FILE)
           PROVE-FILE)
    (ITERPRI PROVE-FILE)
    (LIST PROP-FILE-NAME FN-FILE-NAME))))

;;;; Everything below has to do with the extension to handle DEFN events whose termination
;;;; proofs are constructed interactively.

(DEFUN PROVE-TERMINATION (FORMALS RM MACHINE)
  (LET ((PROOF-START-TIME-IN-60THS (TIME-IN-60THS))
        ALL-PROCESS-CLAUSES)
    (SETQ PROVE-TERMINATION-LEMMAS-USED NIL)
    (PROG1
     (COND
      (INTERACTIVE-DEFN-HINT
       (COND
        (REPLAY-FLG
         (REPLAY-FOR-DEFN
          (START-TERM-FOR-INTERACTIVE-PROVE-TERMINATION FORMALS RM MACHINE)
          (CDR INTERACTIVE-DEFN-HINT))))
       (SETQ PROVE-TERMINATION-LEMMAS-USED (ACCESS STATE CUMULATIVE-LEMMAS-USED (CAR STATE-STACK)))
       (ITERATE FOR GL IN (ALL-GOALS (CAR STATE-STACK))
                ALWAYS (EQUAL (ACCESS GOAL CONC GL) TRUE)))
      (T
        (AND (COND ((AND (EQ *1*THM-MODE$ (QUOTE NQTHM))
                         (EQ (CAR RM) (QUOTE ORD-LESSP)))
                    (COND
                     ((AND (SIMPLIFY-CLAUSE-MAXIMALLY
                            (LIST (CONS-TERM (QUOTE ORDINALP)
                                             (LIST (CADR RM)))))
                           (NULL PROCESS-CLAUSES))
                      (SETQ PROVE-TERMINATION-LEMMAS-USED
                            (UNION-EQUAL PROCESS-HIST
                                         PROVE-TERMINATION-LEMMAS-USED))
                      T)
                     (T (SETQ ALL-PROCESS-CLAUSES
                              (UNION-EQUAL PROCESS-CLAUSES
                                           ALL-PROCESS-CLAUSES))
                        (ER WARNING ((X (CONS-TERM (QUOTE ORDINALP)
                                                   (LIST (CADR RM))))
                                     (Y (COND ((AND (NVARIABLEP (CADR RM))
                                                    (NOT (FQUOTEP (CADR RM)))
                                                    (EQ (FFN-SYMB (CADR RM))
                                                        'CONS))
                                               (FARGN (CADR RM) 1))
                                              (T NIL))))
                            |The| |formula| (!TERM X NIL) |has| |not| |been|
                            |proved| |.|
                            (COND (Y |This| |often| |happens| |because| |the|
                                     |first| |component| |,| (!TERM Y '|,|)
                                     |of| |the| |alleged| |ordinal| |is| |not|
                                     |strictly| |positive| |.|)))
                        NIL)))
                   (T T))
             (ITERATE FOR X IN MACHINE
                      ALWAYS
                      (COND
                       ((AND (SIMPLIFY-CLAUSE-MAXIMALLY
                              (NCONC1 (ITERATE
                                       FOR H IN (ACCESS TESTS-AND-CASE TESTS X)
                                       COLLECT (NEGATE-LIT H))
                                      (CONS-TERM (CAR RM)
                                                 (LIST (SUB-PAIR-VAR
                                                        FORMALS
                                                        (ACCESS
                                                         TESTS-AND-CASE
                                                         CASE X)
                                                        (CADR RM))
                                                       (CADR RM)))))
                             (NULL PROCESS-CLAUSES))
                        (SETQ PROVE-TERMINATION-LEMMAS-USED
                              (UNION-EQUAL PROCESS-HIST
                                           PROVE-TERMINATION-LEMMAS-USED))
                        T)
                       (T (SETQ ALL-PROCESS-CLAUSES
                                (UNION-EQUAL PROCESS-CLAUSES
                                             ALL-PROCESS-CLAUSES))
                          NIL))))))
     (SETQ PROVE-TERMINATION-CASES-TRIED
           ;;Note:  we don't care about all-process-clauses in the interactive case
           (CONS (CONS RM ALL-PROCESS-CLAUSES)
                 PROVE-TERMINATION-CASES-TRIED))
     (COND
      ((NOT INTERACTIVE-DEFN-HINT) 
       (SETQ PROVE-TIME (+ PROVE-TIME
                           (- (TIME-IN-60THS) PROOF-START-TIME-IN-60THS))))))))

(DEFUN PUT-INDUCTION-INFO (FNNAME FORMALS BODY RELATION-MEASURE-LST)

;   If FNNAME is recursive we store JUSTIFICATIONS, INDUCTION-MACHINE, and
;   QUICK-BLOCK-INFO properties.  If only one JUSTIFICATION is stored and in it
;   the RELATION is NIL then we did not establish termination.  ALL-LEMMAS-USED
;   is side-effected to contain lemma names used to clean up the
;   INDUCTION-MACHINE.

  (PROG (T-MACHINE I-MACHINE JUSTIFICATIONS)
        (SETQ PROVE-TERMINATION-CASES-TRIED NIL)
        (SETQ T-MACHINE (TERMINATION-MACHINE FNNAME BODY NIL))
        (COND ((NULL T-MACHINE)
               (SETQ ALL-LEMMAS-USED NIL)
               (RETURN NIL)))
        (OR RELATION-MEASURE-LST
            (SETQ RELATION-MEASURE-LST
                  (GUESS-RELATION-MEASURE-LST FORMALS T-MACHINE)))
	;; My patch.  For simplicity, interactive definitions will never consider more than
	;; one suggested relation-measure list.
	(COND
	 ((AND (CDR RELATION-MEASURE-LST) INTERACTIVE-DEFN-HINT)
	  (ER WARNING NIL |Omitting| |all| |extra| |relation-measure| |pairs| |for| |the|
	      |purpose| |of| |an| |interactively| |verified| |definition| |.|)
	  (SETQ RELATION-MEASURE-LST (LIST (CAR RELATION-MEASURE-LST)))))
        (SETQ JUSTIFICATIONS
              (COND ((AND PETITIO-PRINCIPII
                          (EQUAL (LENGTH RELATION-MEASURE-LST) 1))
                     (LET ((RM (CAR RELATION-MEASURE-LST))
                           (PROVE-TERMINATION-LEMMAS-USED NIL))
                       (LIST (MAKE JUSTIFICATION
                                   (ALL-VARS (CADR RM))
                                   (CADR RM)
                                   (CAR RM)
                                   PROVE-TERMINATION-LEMMAS-USED))))
                    (T
                     (ITERATE FOR RM IN RELATION-MEASURE-LST
                              WHEN (PROVE-TERMINATION FORMALS RM
                                                      T-MACHINE)
                              COLLECT (MAKE JUSTIFICATION
                                            (ALL-VARS (CADR RM))
                                            (CADR RM)
                                            (CAR RM)
                                            PROVE-TERMINATION-LEMMAS-USED)))))
        (COND (JUSTIFICATIONS
               (ADD-FACT FNNAME (QUOTE JUSTIFICATIONS)
                         JUSTIFICATIONS)))
        (SETQ ALL-LEMMAS-USED NIL)

;   We set ALL-LEMMAS-USED to NIL to forget the lemmas put there by PROVE so we
;   can now accumulate the lemmas used by REMOVE-REDUNDANT-TESTS in
;   INDUCTION-MACHINE.

        (SETQ I-MACHINE
              (INDUCTION-MACHINE FNNAME BODY NIL))

;   We store the INDUCTION-MACHINE even for inadmissible BOOT-STRAP
;   fns like V&C$.  Otherwise, we would be fooled into thinking such
;   functions were non-recursive.  It is illegitimate to make use of
;   the INDUCTION-MACHINE as an induction schema unless the
;   JUSTIFICATIONS property is non-NIL.

        (ADD-FACT FNNAME (QUOTE INDUCTION-MACHINE) I-MACHINE)
        (COND (FORMALS
               (ADD-FACT FNNAME (QUOTE QUICK-BLOCK-INFO)
                         (QUICK-BLOCK-INFO FORMALS I-MACHINE))))
        (RETURN NIL)))

(DEFUN DEFN-WRAPUP (WON-FLG)
  (SETQ WON-FLG (COND (WON-FLG (QUOTE DEFN-OK))
                      (T NIL)))
  (COND ((NOT (EQ LEMMA-STACK ORIG-LEMMA-STACK))
         (ITERPRI T)
         (ER WARNING NIL DEFN-WRAPUP |found| |a| |non-trivial| LEMMA-STACK
             EXCL)))
  (COND ((NOT (EQ LINEARIZE-ASSUMPTIONS-STACK
                  ORIG-LINEARIZE-ASSUMPTIONS-STACK))
         (ITERPRI T)
         (ER WARNING NIL DEFN-WRAPUP |found| |a| |non-trivial|
             LINEARIZE-ASSUMPTIONS-STACK EXCL)))
  (COND ((NOT WON-FLG)
         (COND ((NULL PROVE-TERMINATION-CASES-TRIED)
                (ER SOFT () |The| |admissibility| |of| |this| |definition|
                    |has| |not| |been| |established| |.|  |The| |theorem|
                    |prover's|
                    |heuristics| |found| |no| |plausible| |measure| |to|
                    |justify|
                    |the| |recursion| |.|  |In| |particular| |,| |no| |single|
                    |argument| |of| |the| |function| |is| |both| |tested|
                    |in| |each| |branch| |and| |changed| |in| |each|
                    |recursive|
                    |call| |.| |The| |definition| |is| |rejected| |.|))
	       ;; Here's the patch
	       (interactive-defn-hint
		(ER SOFT
                    ((WHY (ITERATE FOR X IN PROVE-TERMINATION-CASES-TRIED
                                   COLLECT
                                   (LIST (QUOTE PROGN)
                                         (PQUOTE |Relation: |)
                                         (LIST (QUOTE !PPR)
                                               (KWOTE (CAAR X)) NIL)
                                         (QUOTE CR)
                                         (PQUOTE |Measure:  |)
                                         (LIST (QUOTE !PPR)
                                               (KWOTE (CADAR X)) NIL)
                                         (QUOTE CR)))))
		    |Interactive| |proof| |not| |completed| excl CR
		    |The| |definition| |is| |rejected| |.|
                    |Below| |are| |the| |relations| |and|
                    |measures| |tried| |.| CR CR
		    (!LIST1 WHY
			    (PQUOTE (PROGN CR CR))
			    (PQUOTE (PROGN CR CR)))
                    CR CR))
               (T

                (ER SOFT
                    ((WHY (ITERATE FOR X IN PROVE-TERMINATION-CASES-TRIED
                                   COLLECT
                                   (LIST (QUOTE PROGN)
                                         (PQUOTE |Relation: |)
                                         (LIST (QUOTE !PPR)
                                               (KWOTE (CAAR X)) NIL)
                                         (QUOTE CR)
                                         (PQUOTE |Measure:  |)
                                         (LIST (QUOTE !PPR)
                                               (KWOTE (CADAR X)) NIL)
                                         (QUOTE CR)
                                         (PQUOTE |Unproved goals:|)
                                         (QUOTE CR)
                                         (LIST (QUOTE !CLAUSE-SET)
                                               (KWOTE (CDR X)) 5)))))
                    |The| |admissibility| |of| |this| |definition|
                    |has| |not| |been| |established| |.|  |The| |simplifier|
                    |could| |not| |prove| |that| |the| |measure(s)|
                    |tried| |decrease| |in| |each| |recursive|
                    |call| |.| |The| |definition| |is| |rejected| |.|
                    |Below| |are| |listed| |the| |relations| |and|
                    |measures| |tried| |and| |some| |of| |the|
                    |unproved| |goals| |for| |each| |.| CR CR
                    (!LIST1 WHY
                            (PQUOTE (PROGN CR CR))
                            (PQUOTE (PROGN CR CR)))
                    CR CR)))))
  (OR PETITIO-PRINCIPII
      (IO (QUOTE FINISHED)
          NIL NIL NIL (LIST WON-FLG))))

(DEFEVENT DEFN (NAME ARGS BODY &OPTIONAL RELATION-MEASURE-LST interactive-defn-hint)
  ;; only the final argument, interactive-defn-hint, is new, together with the check for
  ;; acceptability of that argument
  (EVENT-COMMAND
   (LIST (QUOTE DEFN) NAME ARGS BODY RELATION-MEASURE-LST)
   (match! (chk-acceptable-interactive-defn-hint interactive-defn-hint)
	   (list interactive-defn-hint))
   (let ((start-marker `(START (DEFN ,NAME ,ARGS ,BODY ,RELATION-MEASURE-LST)))
	 (IN-DEFN-FLG T))
     (MATCH! (CHK-ACCEPTABLE-DEFN NAME ARGS BODY RELATION-MEASURE-LST)
	     (LIST NAME ARGS BODY RELATION-MEASURE-LST))
     (MAKE-EVENT NAME (COND (INTERACTIVE-DEFN-HINT
			     (LIST (QUOTE DEFN) NAME ARGS BODY
				   RELATION-MEASURE-LST INTERACTIVE-DEFN-HINT))
			    (RELATION-MEASURE-LST
			     (LIST (QUOTE DEFN) NAME ARGS BODY
				   RELATION-MEASURE-LST))
			    (T (LIST (QUOTE DEFN) NAME ARGS BODY))))
     (DEFN0 NAME ARGS BODY RELATION-MEASURE-LST NIL NIL)
     (DEPEND
      NAME
      (REMOVE
       NAME
       (UNION-EQ
	(ALL-FNNAMES BODY)
	(UNION-EQ
	 ALL-LEMMAS-USED
	 (ITERATE FOR TEMP IN (GET NAME (QUOTE JUSTIFICATIONS))
		  WITH ITERATE-ANS
		  DO (SETQ
		      ITERATE-ANS
		      (UNION-EQ (COND
				 ((NULL (ACCESS JUSTIFICATION
						RELATION TEMP))
				  NIL)
				 (T (UNION-EQ
				     (ALL-FNNAMES (ACCESS
						   JUSTIFICATION
						   MEASURE-TERM
						   TEMP))
				     (ADD-TO-SET (ACCESS JUSTIFICATION
							 RELATION TEMP)
						 (ACCESS JUSTIFICATION
							 LEMMAS TEMP)))))
				ITERATE-ANS))
		  FINALLY (RETURN ITERATE-ANS))))))
    
     (COND ((NOT (ADMITTED-FUNCTIONP NAME))
	    (UNDO-BACK-THROUGH-FN NAME)
	    (DEFN-WRAPUP NIL)
	    NIL)
	   (T (PRINT-DEFN-MSG NAME ARGS)
	      (DEFN-WRAPUP (ADMITTED-FUNCTIONP NAME))
	      NAME)))))

(DEFUN NOTE-LIB (FILE &OPTIONAL LOAD-COMPILED-CODE-FLG)
;; patched to allow for noting pure Boyer-Moore libraries without unbinding STATE-STACK-LIST
  (LET ((*READ-BASE* 10)
        (*READTABLE* (COPY-READTABLE NIL))
        FILE1 FILE2)
    (CHK-COMMAND-STATUS T)
    (PRINT-NQTHM-DISCLAIMER)
    (SETQ FILE1 (EXTEND-FILE-NAME FILE FILE-EXTENSION-LIB))
    (COND ((PROBE-FILE FILE1) NIL)
          (PETITIO-PRINCIPII
           (PRINEVAL (PQUOTE (PROGN CR CR |file| |not| |found| |:|
                                    (!PPR FILE '|.|)
                                    |skipping| |this| NOTE-LIB |.|))
                     (LIST (CONS 'FILE FILE1))
                     0 PROVE-FILE)
           (RETURN-FROM NOTE-LIB T))
          (T (ER SOFT (FILE1) |file| |not| |found:| (!PPR FILE1 '|.|))))
    (KILL-LIB)
    (LOAD FILE1)
    
    (SETQ LIB-FILE FILE1)
    (SETQ FILE2
          (EXTEND-FILE-NAME FILE
                            (COND (LOAD-COMPILED-CODE-FLG FILE-EXTENSION-BIN)
                                  (T FILE-EXTENSION-LISP))))
    (PROCLAIM-NQTHM-FILE FILE)
    (IF LOAD-COMPILED-CODE-FLG
        (OR (LOAD FILE2 :IF-DOES-NOT-EXIST NIL)
            (ER SOFT (FILE2) |You| |have| |supplied| |a| |non-NIL| |second|
                |argument| |to| NOTE-LIB |,| |indicating| |a| |desire| |to|
                |load| |a| |compiled| |file| |.| |However| |,| |an| |attempt|
                |to| |load| |the| |file| (!PPR FILE2 NIL) |has| |failed| |.|))
        (LOAD FILE2))
    (COND ((OR (NOT (BOUNDP (QUOTE LIB-VERSION)))
               (NOT (EQUAL LIB-VERSION THEOREM-PROVER-LIBRARY-VERSION)))
           (SETQ LIB-VERSION-TROUBLE T)
           (ERROR1 (QUOTE (PROGN |The| |library| |being| |loaded| |was|
                                 |created| |with| |a| |format| |different|
                                 |from| |the| |one| |appropriate| |for| |this|
                                 |version| |of| |the| |theorem-prover| |.|
                                 |you| |are| |advised| |to| |remake| |all| |of|
                                 |your| |libraries| |from| |scratch| |.|
                                 |it| |is| |possible| |that| |you| |will|
                                 |meet| |with| |nonsensical| |results| |by|
                                 |using| |the| |results| |of| |this| |call|
                                 |to| NOTE-LIB |.|))
                   NIL
                   (QUOTE WARNING))))
    ;; here's the patch
    (cond ((not (boundp 'state-stack-list))
           (setq state-stack-list nil)))
    (cond ((not (member-eq 'state-stack-list lib-vars))
           (let* ((y (our-last lib-vars))
                  (old-y (copy-list y)))
             (setf (car y) 'state-stack-list)
             (setf (cdr y) old-y))))
    (LIST FILE1 FILE2)))

;;; We conclude with changes to STORE-SENT and FERTILIZE-CLAUSE so
;;; that they ignore IN-PROVE-LEMMA-FLG.  Also, STORE-SENT has been
;;; modified to recognize OTF-FLG so that BASH can create new subgoals
;;; rather than revert to "induct" on the original goal.


;;************************************************************;;
;;                    Change to STORE-SENT                    ;;
;;************************************************************;;

(DEFUN STORE-SENT (CL HIST)

  ;; modified from nqthm code so that otf-flg can suppress reverting
  ;; to induct on original goal (idea taken from Acl2)

;   This function is FUNCALLed and therefore may not be made a MACRO.

  (LET (CL-SET)
    (COND ((NULL CL)
           (IO (QUOTE STORE-SENT) CL HIST NIL (LIST (GET-STACK-NAME STACK)))
           (WRAPUP NIL))
          (DO-NOT-USE-INDUCTION-FLG
           (IO (QUOTE STORE-SENT)
               CL HIST NIL (LIST (GET-STACK-NAME STACK) (QUOTE QUIT)))
           (WRAPUP NIL))
          ((AND
            (NOT (AND IN-PROVE-LEMMA-FLG (ASSOC-EQ (QUOTE INDUCT) HINTS)))
	    ;; I don't want to do the start-over thing if otf-flg is set; this is the only patch.
	    (not otf-flg)
            (OR
             (AND
              (NULL STACK)
              (ITERATE FOR X IN HIST THEREIS
                       (NOT
                        (MEMBER-EQ (CAR X)
                                   (QUOTE (SETTLED-DOWN-CLAUSE
                                           SIMPLIFY-CLAUSE
                                           SETUP))))))
             (AND STACK (NOT (ASSOC-EQ (QUOTE BEING-PROVED) STACK)))))

;   Abort and push the input clause to work on if (a) this is the first time
;   we've ever pushed anything and we've done anything to the input other than
;   simplify it, or (b) we have not yet gone into the first induction for the
;   original conjecture but have already pushed one simplified subgoal.

           (SETQ STACK NIL)
           (SETQ CL-SET (CNF-DNF THM (QUOTE C)))

;   Once upon a time we backed up to the output of PREPROCESS in PROVE.
;   However, PREPROCESS -- and CLAUSIFY-INPUT -- applies unconditional rewrite
;   rules and we want the ability as users to type in exactly what the system
;   inducts on.  The theorem that PREPROCESS screwed us on was HACK1 when it
;   distributed TIMES and GCD.

           (IO (QUOTE STORE-SENT)
               CL NIL NIL (LIST (GET-STACK-NAME STACK)
                                CL-SET))
           (PUSH-CLAUSE-SET CL-SET)
           (THROW (QUOTE SIMPLIFY-LOOP) NIL))
          (T (SETQ CL-SET (LIST CL))
             (IO (QUOTE STORE-SENT) CL HIST NIL (LIST (GET-STACK-NAME STACK)))
             (PUSH-CLAUSE-SET CL-SET)))))

;; The definition below occurs here rather than in nqthm-patches.lisp
;; because it refers to the special variable *UNTRANSLATE-EVENT-TYPES*,
;; which is defined in basis-pc.lisp.

(DEFUN UNTRANSLATE-EVENT (EVENT)
  (LET (NAME TYPES TERM ARGS BODY HINTS INTERACTIVE-HINT OLD-NAME-OR-LIST FS)
    (CASE (CAR EVENT)
          (ADD-AXIOM (MATCH! EVENT (ADD-AXIOM NAME TYPES TERM))
                     (LIST (QUOTE ADD-AXIOM)
                           NAME TYPES (UNTRANSLATE TERM)))
          (DEFN (COND ((INT= (LENGTH EVENT) 4)
                       (MATCH! EVENT (DEFN NAME ARGS BODY))
                       (LIST (QUOTE DEFN) NAME ARGS (UNTRANSLATE BODY)))
		      ;; here is support for interactive DEFN hints
		      ((INT= (LENGTH EVENT) 6)
                       (MATCH! EVENT (DEFN NAME ARGS BODY HINTS interactive-hint))
                       (LIST (QUOTE DEFN) NAME ARGS (UNTRANSLATE BODY)
                             (ITERATE FOR X IN HINTS COLLECT
                                      (LIST (CAR X)
                                            (UNTRANSLATE (CADR X))))
                             interactive-hint))
                      (T (MATCH! EVENT (DEFN NAME ARGS BODY HINTS))
                         (LIST (QUOTE DEFN) NAME ARGS (UNTRANSLATE BODY)
                               (ITERATE FOR X IN HINTS COLLECT
                                        (LIST (CAR X)
                                              (UNTRANSLATE (CADR X))))))))
          (PROVE-LEMMA (COND ((INT= (LENGTH EVENT) 4)
                              (MATCH! EVENT (PROVE-LEMMA NAME TYPES TERM))
                              (LIST (QUOTE PROVE-LEMMA)
                                    NAME TYPES (UNTRANSLATE TERM)))
                             (T (MATCH! EVENT
                                        (PROVE-LEMMA NAME TYPES TERM HINTS))
                                (LIST (QUOTE PROVE-LEMMA)
                                      NAME TYPES (UNTRANSLATE TERM)
                                      (UNTRANSLATE-HINTS HINTS)))))
          (FUNCTIONALLY-INSTANTIATE
           (MATCH! EVENT
                   (FUNCTIONALLY-INSTANTIATE NAME TYPES TERM
                                             OLD-NAME-OR-LIST FS HINTS))
           (LIST (QUOTE FUNCTIONALLY-INSTANTIATE) NAME TYPES
                 (UNTRANSLATE TERM)
                 OLD-NAME-OR-LIST
                 (UNTRANSLATE-FS FS)
                 (UNTRANSLATE-HINTS HINTS)))
          (CONSTRAIN
           (MATCH! EVENT (CONSTRAIN NAME TYPES TERM FS HINTS))
           (LIST (QUOTE CONSTRAIN) NAME TYPES (UNTRANSLATE TERM)
                 (UNTRANSLATE-FS FS)
                 (UNTRANSLATE-HINTS HINTS)))
          (OTHERWISE
	   (IF (MEMBER-EQ (CAR EVENT) *UNTRANSLATE-EVENT-TYPES*)
	       EVENT
	       (ER HARD (EVENT) |The| |following| |event| |was| |not|
		   |recognized| |by|
		   UNTRANSLATE-EVENT |,| (!PPR EVENT (QUOTE |.|))))))))

